#!/usr/bin/perl
package SmartClient::DeviceStatusCache;
use strict;
use warnings;
our $VERSION = "0.1";

my $MARKER_STRING = 'to_be_changed';

use Cache::FastMmap;
use Log::Log4perl;
use SmartClient::Config;

sub new
{
    my ( $class, @params ) = @_;
    my $logger = Log::Log4perl->get_logger();

    if ( ref $class ) {
        $logger->warn('Constructor must not be used to clone an instance');
        return;
    }

    my $self = {
        'filename'  => undef,
        'readonly'  => undef,
        'num_pages' => undef,
        'page_size' => undef,
        'config'    => undef,
        @params
    };

    $self->{config} ||= SmartClient::Config->new( '/etc/smartclient/base',
        '/etc/smartclient/ldap2dnsdhcpd' );
    $self->{num_pages} ||=
      $self->{config}->get( 'DEVICE_STATUS_CACHE_MMAP_NUM_PAGES', 1 )
      || '137';
    $self->{page_size} ||=
      $self->{config}->get( 'DEVICE_STATUS_CACHE_MMAP_PAGE_SIZE', 1 )
      || '64k';

    my $filename = $self->{'filename'};
    if ( !defined $filename ) {
        $logger->warn('No filename given as parameter');
        return;
    }

    $logger->info( "Open device status cache $filename with "
          . $self->{num_pages}
          . ' pages of '
          . $self->{page_size}
          . ' size' );

    my $init_file = 1;
    if ( $self->{'readonly'} ) {
        $init_file = 0;
        $logger->info("cache will be opened readonly");
    }

    my $status_cache = Cache::FastMmap->new(
        'share_file' => $filename,
        'init_file'  => $init_file,
        ## page_size default ist 64k, this is ok for device-map
        ## for 10384 Entries 127 pages of this size are needed
        ## should be a prime number
        'num_pages' => $self->{num_pages},
        'page_size' => $self->{page_size},
        ##'context'    => $sc_cache,
        ##'read_cb'    => sub { $_[0]->getDeviceStatus($_[1]) },
        ##'expire_time'   => 1m,
    );

    if ( !$status_cache ) {
        $logger->error(
            "failed to initialize device status FastMmap file $filename");
    }

    $self->{'status_cache'} = $status_cache;
    $self->{'update_list'}  = [];

    return bless $self, $class;
}

=head2 setDeviceStatus

 my $dn         = 'cn=test,dc=one,dc=two';
 my $status_ref = {'boot': 'local', 'version': 20}; 
 if ( $status_cache->setDeviceStatus( $dn, $status_ref ) ) {
    print "Set Status for $dn"         
 }

=over 4

=item Parameter

=over 4

=item dn-Name of the item for which status should be set

=item hash reference with values to be set

=back

=item Return

boolean true, if status could be set successfully

=back

=cut

sub setDeviceStatus
{
    my ( $self, $dn, $status_href ) = @_;
    my $logger = Log::Log4perl->get_logger();

    if ( !ref $self ) {
        $logger->error("method can't be used as a function");
        return;
    }

    if ( $self->{'readonly'} ) {
        $logger->warn(
            'cache was initialized as readonly. Will not change the entry');
        return;
    }
    my $old_status_href = $self->{'status_cache'}->get($dn);

    if ( defined $old_status_href ) {
        if ( isSame( $old_status_href, $status_href ) ) {
            $logger->debug("no changes for $dn, so will not mark it");
            return 1;
        }
        $logger->debug("got reference for $dn and will mark it");
        $old_status_href->{$MARKER_STRING} = 1;
    } else {
        $logger->debug(
            "got no reference for $dn, so will create new one and mark it");
        $old_status_href = { $MARKER_STRING => 1 };
    }

    $self->addChange(
        'dn'          => $dn,
        'status_href' => $status_href,
        'type'        => 'UPDATE'
    );

    if ( !$self->{'status_cache'}->set( $dn, $old_status_href ) ) {
        $logger->warn(
            "could not set status info for $dn into $self->as_string()");
        return;
    }

    return 1;
}

=head2 getDeviceStatus

  my $status_href = $status_cache->get($dn);
  if (!defined $status_href) {
      $logger->info("no status for $dn could be found");
  }

=over 4

=item Parameter

 dn name of the device to get the status for

=item Return

 hash ref to status information

=back

=cut

sub getDeviceStatus
{
    my ( $self, $dn ) = @_;
    my $logger = Log::Log4perl->get_logger();

    if ( !defined $dn ) {
        $logger->info('no dn was given, so we will do nothing');
        return;
    }

    return $self->{'status_cache'}->get($dn);
}

=head2 removeDeviceStatus

  if ( $status_cache->removeDeviceStatus($dn) ) {
      print "OK status for $dn is marked for delete\n";
  }

=over 4

=item Parameter

 dn name of entry for which the status should be removed 

=item Return

 boolean true, if device status can be removed

=back

=cut

sub removeDeviceStatus
{
    my ( $self, $dn ) = @_;
    my $logger = Log::Log4perl->get_logger();

    if ( my $status_href = $self->{'status_cache'}->get($dn) ) {
        $status_href->{$MARKER_STRING} = 1;
        $self->{'status_cache'}->set( $dn, $status_href );
        $self->addChange( 'dn' => $dn, 'type' => 'REMOVE' );
        return 1;
    }

    $logger->info(
        "status for $dn was not in the status cache, so it will not be removed"
    );
    return;
}

=head2 addChange

=over 4

=item Parameter

=item Return

=back

=cut

sub addChange
{
    my $self   = shift;
    my %params = (
        'dn'          => undef,
        'status_href' => undef,
        'type'        => undef,
        @_
    );

    my $dn          = $params{'dn'};
    my $status_href = $params{'status_href'};
    my $change_type = uc( $params{'type'} );

    my $logger = Log::Log4perl->get_logger();
    if ( !defined $dn ) {
        $logger->warn("no dn was given, so no change can be recorded");
        return;
    }

    if ( !( $change_type eq 'UPDATE' or $change_type eq 'REMOVE' ) ) {
        $logger->warn("change has to be on of UPDATE or REMOVE");
        return;
    }

    push @{ $self->{'update_list'} }, [ $dn, $change_type, $status_href ];

    return 1;

}

=head2 commitUpdates

  $status_cache->commitUpdates();

=over 4

=item Parameter

nothing

=item Return

nothing

=back

=cut

sub commitUpdates
{
    my $self   = shift;
    my $logger = Log::Log4perl->get_logger();

    if ( !ref $self ) {
        $logger->warn(
            'was not called as a method. commitUpdates will not work');
        return;
    }

    my $status_cache = $self->{'status_cache'};

    if ( !ref $status_cache ) {
        $logger->warn(
            'status cache was not found, so no Updates will be commited');
        return;
    }

    my @scheduled_updates = @{ $self->{'update_list'} || [] };
    my $nr_of_updates = @scheduled_updates;
    $logger->debug("committing $nr_of_updates changes");
    for my $update_element_aref (@scheduled_updates) {
        my ( $dn, $change_type, $status_href ) = @{$update_element_aref};
        my $result;
        if ( $change_type eq 'UPDATE' ) {
            $result = $status_cache->set( $dn, $status_href );
        } elsif ( $change_type eq 'REMOVE' ) {
            $result = $status_cache->remove($dn);
        }
        if ( !$result ) {
            $logger->warn("Could not update $dn to new status");
        }
    }
    $self->{'update_list'} = [];
    return;
}

=head2 isDirty

=head3 Example

 $status_cache->setDeviceStatus($dn, $status_href);
 ok($status_cache->isDirty($dn), 'entry marked as changed');
 $status_cache->rollbackUpdates();
 ok(!$status_cache->isDirty($dn), 'update was canceled');

=over 4

=item Parameter

 dn name of the entry

=item Return

 true if dn is in the status cache and marked as to_be_changed

=back

=cut

sub isDirty
{
    my ( $self, $dn ) = @_;

    my $logger = Log::Log4perl->get_logger();

    if ( !ref $self ) {
        $logger->warn('was not called as a method, so we can not do our work.');
        return;
    }

    my $status_cache = $self->{'status_cache'};

    if ( !ref $status_cache ) {
        $logger->warn(
            'status cache was not found, so no Updates will be commited');
        return;
    }

    my $status_href = $status_cache->get($dn);
    if ( ref $status_href and defined $status_href->{$MARKER_STRING} ) {
        return 1;
    }

    return;
}

=head2 rollbackUpdates

=head3 Example

  $status_cache->setDeviceStatus($dn, $status_href);
  ok($status_cache->isDirty($dn), 'entry marked as changed');
  $status_cache->rollbackUpdates();
  ok(!$status_cache->isDirty($dn), 'update was canceled');
 
=over 4

=item Parameter

 none
 
=item Return

 none

=back

=cut

sub rollbackUpdates
{
    my $self = shift;

    my $logger = Log::Log4perl->get_logger();

    if ( !ref $self ) {
        $logger->warn(
            'we were not called as a method, so we will not do anything.');
        return;
    }
    my $status_cache = $self->{'status_cache'};

    if ( !ref $status_cache ) {
        $logger->warn(
            'status cache was not found, so no Updates will be rolled back');
    }

    $logger->debug('rollbackUpdates');
    my @update_list = @{ $self->{'update_list'} };
    for my $update_element_aref (@update_list) {
        my ($dn) = @{$update_element_aref};
        my $status_href = $status_cache->get($dn);
        if ( ref $status_href ) {
            delete $status_href->{$MARKER_STRING};
            $status_cache->set( $dn, $status_href )
              or $logger->warn("could not set status_cache info for $dn");
        } else {
            $logger->info(
                "no status_cache info for $dn could be found, strange...I wrote that info before..."
            );
        }
    }

    $self->{'update_list'} = [];
    return;
}

# internally used function
# checks if two hashes have the same values
sub isSame
{
    my ( $old_href, $new_href ) = @_;
    return (  isSameOneWay( $old_href, $new_href )
          and isSameOneWay( $new_href, $old_href ) );
}

# internally used function
# checks if values from one hash, are present in another one and have the same values
sub isSameOneWay
{
    my ( $dit_href, $dat_href ) = @_;
    while ( my ( $dit_key, $dit_value ) = each %{$dit_href} ) {
        if (
            not( defined $dat_href->{$dit_key}
                and $dat_href->{$dit_key} eq $dit_value )
          )
        {
            return;
        }
    }
    return 1;
}

# Package must return a true value
1;
