#!/usr/bin/perl -w
package SmartClient::DhcpEntries;
use strict;
use warnings;

=head1 NAME

=begin comment

Name des Moduls,Grobkonzept

=end comment

B<SmartClient::DhcpEntries>

=head1 SYNOPSIS

=begin comment

  kurze Beschreibung, was das Programm macht, Feinkonzept

=end comment

Objekt und Funktion zum Auslesen der dhcpd.leases Datei.

 SmartClient::DhcpEntries->new(lease    => "IP-Addresse",
				 ethernet => "MAC-Adresse",
				 starts   => "Start-Datum",
				 ends     => "Ends-Datum",
				 hostname => "Host-Name");

 read_dhcp_leases();

=head1 DESCRIPTION

=begin comment

  muss fuer das Grobkonzept ausgefuellt werden

=end comment

=head2 Allg. Beschreibung und Aufgaben

Objekt und Funktion zum Auslesen der dhcpd.lease-Datei eines DHCP-Servers.


=head2 Funktionsweise / Design / Architektur / Sprache

=over 4

=item Funktionsweise

B<SmartClient::DhcpEntries> ermöglicht die Speicherung eines Eintrags
 der dhcpd.leases -Datei. Die Datei dhcpd.leases hat folgenden Aufbau:

 lease 10.1.254.88 {
        starts 4 2001/04/19 06:14:34;
        ends 5 2001/04/20 06:14:34;
        hardware ethernet 00:d0:59:2b:c6:b4;
        uid 01:00:d0:59:2b:c6:b4;
	client-hostname "dhcppc7"; 
 }

Unbedingt notwendig ist das Vorhandensein einer IP-Addresse und einer MAC-Adresse. Ohne diese Angaben wird keine gültige Instanz des Objektes gebildet.

Ein Eintrag muß unbedingt eine öffenende und schließende geschweifte Klammer enthalten, um zu verhindern, daß bei einem gleichzeitigen Zugriff auf die Datei, ein unvollständiger Datensatz ausgelesen wird.
   
Die Funktion B<read_dhcp_leases> liest die Datei dhcpd.leases vollstänig aus und liefert einen Hash aus MAC-Adresse und Referenz zu einem DhcpEntries-Objekt zurück.

B<Erweiterung für Token-Ring>

Eine Token-Ring Hardware Adressse hat das gleiche Format wie eine Ethernet-MAC-Adresse.
Jedoch wird bei DHCP unterschieden, ein lease file Eintrag für Token-Ring:


 lease 10.1.254.89 {
        starts 4 2001/04/19 06:14:34;
        ends 5 2001/04/20 06:14:34;
        hardware token-ring 00:d0:59:2b:c6:b4;
        uid 01:00:d0:59:2b:c6:b4;
	client-hostname "dhcppc8"; 
 }

Auch in die dhcpd.conf muß bei den fixed-address Einträgen hardware token-ring
statt hardware ethernet eingetragen werden.

Um die Sourcecode Änderungen zu minimieren, wird die interne Struktur um
den boolean Wert is_tokenring erweitert.


=item Architektur

Linux, i386

=item Sprache

  Perl

=back

=head2 Sicherheitskonzept

keine Angabe

=head2 Einordnung / Abhängigkeiten / Voraussetzungen

=begin comment

Abhaenigkeiten von anderen Perlmodulen werden durch
require Statements sichergestellt.

=end comment

require L<DB::Tools::IP>

require L<DB::Tools::MAC>

  

=head2 Interfaces(Programm/Benutzer)

Der Standort der dhcpd.leases-Datei kann angeben werden.

=head2 Status (Fehler)

 Bei einem Fehler wird ein nicht definiertes Objekt zurückgeben.
 Kann die Datei dhcpd.leases nicht geöffnet werden, wird das Programm
 mit entsprechender Fehlermeldung verlassen.

=head2 Version

=begin comment

  Versionsnummer fuer das Modul,  Feinkonzept,

=end comment

0.1

=cut

our $VERSION = 0.1;

=head2 CVS Version

  $Id: DhcpEntries.pm 11063 2015-08-12 11:48:48Z FSchumacher $

=cut

=head2 Installation und sonstige Hinweise

=begin comment

Feinkonzept

=end comment

Modul muß im @INC -Pfad stehen.

=head1 AUTHOR

=begin comment

name and e-mail

=end comment

Frank Balzer mailto:frank.balzer@suse.de

SuSE Linux Solutions AG mailto: rheinruhr@suse.de

=head1 SEE ALSO

L<SmartClient::IP>

L<SmartClient::MAC>

=cut

=head1 Exportierte Module, Methoden, Variablen und Funktionen

=over 4

=item Variablen

 lease:    IP-Adresse
 starts:   Datum und Uhrzeit des Eintrags
 ends:     Verfalls-Datum und -Uhrzeit
 ethernet: MAC-Adresse
 hostanme: Hostname


=back

=head1 Optional exportierte Module, Methoden, Variablen und Funktionen

=over 4

=item Funktionen

 read_dhcp_leases:  Funktion zum Auslesen der dhcpd.leases-Datei

=back

=cut

use Data::Dumper;
use Time::Local;
use SmartClient::Constant;
use Date::Calc qw(Add_Delta_DHMS);
use base 'Exporter';

our @EXPORT = qw(lease starts ends ethernet hostname );

our @EXPORT_OK = qw(read_dhcp_leases clear_dhcp_leases);

## fallback-filename for dhcpd.leases path

my $DEFAULT_DHCP_LOG = "/var/dhcp/dhcpd.leases";

=head1 Modul

=begin comment

Beschreibung des Moduls

=end comment

 SmartClient::DhcpEntries( lease    =>  "IP-Adresse",
			     ethernet =>  "MAC-Adresse",
			     starts   =>  "start-date",
			     ends     =>  "end-date",
			     hostname =>  "hostname");

 # Beispiel
 # 
 use SmartClient::DhcpEntries;
 $dhcp_entry = SmartClient::DhcpEntries->new(ethernet => "10:ab:e1:b1:11:22",
					       lease    => "10.1.1.1.");

=cut

# constructor
sub new
{
    my $package = shift;
    my $data    = {
        'ipAddress'  => "",
        'macAddress' => "",
        'starts'     => "",
        'ends'       => "",
        'hostname'   => "",
        'media'      => 0,
        'status'     => STATUS_UNDEF,
        @_,
    };
    my $entry = undef;

    if ( !testIP( $data->{'ipAddress'} ) || !testMAC( $data->{'macAddress'} ) )
    {
        return;
    }
    return bless( $data, $package );
}

=head2 getStarts

gibt die Startzeit als unixtime zurueck

=item Return

die Startzeit als unixtime

=cut

sub getStarts
{
    my $self       = shift;
    my @start_date = @{ $self->{'starts'}->{'date'} };
    my @start_time = @{ $self->{'starts'}->{'time'} };
    return timegm(
        reverse(@start_time), $start_date[2],
        $start_date[1] - 1,   $start_date[0]
    );
}

=head2 getEnds

gibt die Endezeit als unixtime zurueck

=item Return

die Endezeit als unixtime

=cut

sub getEnds
{
    my $self     = shift;
    my @end_date = @{ $self->{'ends'}->{'date'} };
    my @end_time = @{ $self->{'ends'}->{'time'} };
    return timegm( reverse(@end_time), $end_date[2], $end_date[1] - 1,
        $end_date[0] );
}

=head2 getIPAddr

gibt die IP-Adresse zurueck

=item Return

die IP-Adresse

=cut

sub getIPAddr
{
    my $self = shift;
    return $self->{'ipAddress'};
}

=head2 getMACAddr

gibt die MAC-Adresse zurueck

=item Return

die MAC-Adresse

=cut

sub getMACAddr
{
    my $self = shift;
    return $self->{'macAddress'};
}

=head2 getHostname

gibt den Hostnamen zurueck

=item Return

den Hostnamen

=cut

sub getHostname
{
    my $self = shift;
    return $self->{'hostname'};
}

=head2 getStatus

gibt den Status zurueck

=item Return

der Status

=cut

sub getStatus
{
    my $self = shift;
    return $self->{'status'};
}

=head2 getMedia

gibt das Medium zurueck

=item Return

das Medium

=cut

sub getMedia
{
    my $self = shift;
    return $self->{'media'};
}

=head2 isTokenring

gibt zurueck, ob das Medium Token-Ring ist 

=item Return

wahr, wenn das Medium Token-Ring ist, sonst falsch

=cut

sub isTokenring
{
    my $self = shift;
    if ( $self->{'media'} eq 'token-ring' ) {
        return 1;
    } else {
        return 0;
    }
}

=head2 isExpired

Prueft, ob ein Eintrag abgelaufen ist

=over 4

=item Parameter

Eine Array-Referenz mit der Anzahl der Tage, Stunden, Minuten und
Sekunden (in dieser Reihenfolge) für den Ablauf des Eintrags

=item Return

Wahr, wenn der Eintrag abgelaufen ist, sonst falsch

=back

=cut

sub isExpired
{

    my ( $self, $expiration_delta ) = @_;

    my $logger = Log::Log4perl->get_logger();

    if (   !defined($expiration_delta)
        || ref($expiration_delta) ne "ARRAY"
        || scalar( @{$expiration_delta} != 4 )
        || scalar( grep { $_ =~ /^\d+$/ } @{$expiration_delta} ) != 4 )
    {
        $logger->warn(
            "Call without or wrong expiration delta -> values set to 0");
        $expiration_delta = [ 0, 0, 0, 0 ];
    }

    my $expire_time_stamp = timegm( gmtime() );
    my @expire_date_time  = Add_Delta_DHMS(
        @{ $self->{'ends'}->{'date'} },
        @{ $self->{'ends'}->{'time'} },
        @{$expiration_delta}
    );
    my $expire_date_time_gm = timegm(
        $expire_date_time[5],     $expire_date_time[4],
        $expire_date_time[3],     $expire_date_time[2],
        $expire_date_time[1] - 1, $expire_date_time[0]
    );

    if ( $expire_time_stamp > $expire_date_time_gm ) {
        return 1;
    } else {
        return 0;
    }

}

=head2 testIP()

S<  >B<testIP($ip)> Methoden zum Testen der IP-Adresse.

Liefert Bei Erfolg 1 zurück. Ansonsten 0;

=cut

sub testIP
{
    my $ip_addr = shift;
    require SmartClient::IP;
    my $ip = SmartClient::IP->new($ip_addr);
    if ( !$ip ) {
        Log::Log4perl->get_logger()
          ->error("[DHCP] Invalid parameter for IP-address");
        return 0;
    }
    return 1;
}

=head2 testMAC()

S<  >B<testMAC($ip)> Methoden zum Testen der MAC-Adresse.

Liefert bei Erfolg 1 zurück. Ansonsten 0;

=cut

sub testMAC
{
    my $mac_addr = shift;
    require SmartClient::MAC;
    my $mac = SmartClient::MAC->new($mac_addr);
    if ( !$mac ) {
        Log::Log4perl->get_logger()
          ->error("[DHCP] Invalid parameter for MAC-address");
        return 0;
    }
    return 1;
}

=head1 Funktionen

=head2 read_dhcp_leases

liest die dhcpd.leases Datei.

S<  >B<read_dhcp_leases()>

S<  >B<read_dhcp_leases($filename)>

Liefert bei Erfolg einen Hash zurück.

 #Beispiel
  use SmartClient::DhcpEntries(read_dhcp_leases);
  my %hash = read_dhcp_leases($ARGV[0]);
  
  foreach $i (keys %hash) {
      print $i . "has entry ";
      print $hash{$i}{'lease'}  . ",";
      print $hash{$i}{'starts'}  . ",";
      print $hash{$i}{'hostname'}  . "\n";
   }


=cut

sub read_dhcp_leases
{
    my $filename = shift;
    if ( !$filename ) {
        $filename = $DEFAULT_DHCP_LOG;
    }
    my %entry_list;
    my $logger = Log::Log4perl->get_logger();

    if ( $logger->is_debug() ) {
        $logger->debug("Start read dhcp lease file");
    }
    open( my $fh, '<', $filename ) or return %entry_list;
    my $ip       = undef;
    my $mac      = undef;
    my $media    = undef;
    my $starts   = undef;
    my $ends     = undef;
    my $status   = STATUS_UNDEF;
    my $hostname = undef;
    my $begin    = undef;

    while (<$fh>) {
        ## excluding comments
        if ( $_ =~ /^[#]/ ) {
            next;
        }
        if ( !$begin ) {
            if (/\{/) {
                require SmartClient::IP;
                my @list = split( /\s+/, $_ );
                foreach my $i (@list) {
                    next if $i eq 'lease';
                    $ip = SmartClient::IP->new($i);
                    if ($ip) {
                        $begin = 1;
                        last;
                    }
                }
            } else {
                next;
            }
        } elsif (/starts/) {
            my @list = split( /[\s;]+/, $_ );
            if ( @list >= 3 ) {
                $starts = parseDateTimeString( $list[-2] . " " . $list[-1] );
            }
        } elsif (/ends/) {
            my @list = split( /[\s;]+/, $_ );
            if ( @list >= 3 ) {
                $ends = parseDateTimeString( $list[-2] . " " . $list[-1] );
            }
        } elsif (/^\s+binding state active/) {
            if ( $status == STATUS_UNDEF ) {
                $status = STATUS_ACTIV;
            }
        } elsif (/^\s+binding state free/) {
            if ( $status == STATUS_UNDEF || $status == STATUS_ACTIV ) {
                $status = STATUS_INACTIV;
            }
        } elsif (/ethernet/) {
            $media = 'ethernet';
            require SmartClient::MAC;
            my @list = split( /[\s;]/x, $_ );
            foreach my $i (@list) {
                $mac = SmartClient::MAC->new($i);
                if ($mac) {
                    last;
                }
            }
        } elsif (/token-ring/) {
            $media = 'token-ring';
            require SmartClient::MAC;
            my @list = split( /[\s;]/x, $_ );
            foreach my $i (@list) {
                $mac = SmartClient::MAC->new($i);
                if ($mac) {
                    last;
                }
            }
        } elsif (/hostname/) {
            my @list = split( /[\s;\"]+/, $_ );
            if ( @list >= 2 ) {
                $hostname = $list[-1];
            }
        } elsif (/^\s*\}\s*$/)
        {    # a leases entry ends with an single } on a line.
             # if an uid field has a } in it, it would match the old single } rule
            if (   $mac
                && $ip
                && $ip->returnAddress()
                && $mac->returnAddress() )
            {
                if ( $logger->is_debug() ) {
                    $logger->debug( "read of lease for "
                          . $mac->returnAddress()
                          . " complete" );
                }
                my $entry = SmartClient::DhcpEntries->new(
                    ipAddress  => $ip->returnAddress(),
                    macAddress => $mac->returnAddress(),
                    starts     => $starts,
                    ends       => $ends,
                    hostname   => $hostname,
                    media      => $media,
                    status     => $status,
                );
                if ($entry) {
                    my $mac_address = $mac->returnAddress();
                    if ( exists $entry_list{$mac_address} ) {
                        my $old_entry = $entry_list{$mac_address}->{'act'};
                        if ( $entry->getStarts() gt $old_entry->getStarts() ) {
                            if ( $logger->is_debug() ) {
                                $logger->debug( "use newer entry: "
                                      . $entry->getStarts()
                                      . " greater than "
                                      . $old_entry->getStarts() );
                            }
                            $entry_list{$mac_address}->{'act'} = $entry;
                        }
                    } else {
                        $entry_list{$mac_address} = {
                            act => $entry,
                            all => []
                        };
                    }
                    push @{ $entry_list{$mac_address}->{'all'} }, $entry;
                }
            }
            $ip       = undef;
            $mac      = undef;
            $media    = undef;
            $starts   = undef;
            $ends     = undef;
            $status   = STATUS_UNDEF;
            $hostname = undef;
            $begin    = undef;
        }
    }
    close($fh);
    if ( $logger->is_debug() ) {
        $logger->debug("End read dhcp lease file");
    }
    return %entry_list;
}

=head2 isLeaseToBeCleared

Prueft, ob ein Eintrag geloescht werden soll

=item Parameter

hash consisting of:

=over 4

=item C<ip>

Die IP-Adresse des Eintrags

=item C<mac>

Die MAC-Adresse des Eintrags

=item C<macs_to_clear_from_leases>

Ein Hash in dem die zu loeschenden Eintraege enthalten sind

=item C<cleared_macs>

Ein Hash in dem die Eintraege vermerkt sind, zu denen bereits eine
Protokollierung der Loeschung erfolgt ist. 

=back

=item Return

Wahr, wenn der Eintrag geloescht werden soll, sonst falsch

=back

=cut

sub isLeaseToBeCleared
{
    my %params = (
        'ip'                        => undef,
        'mac'                       => undef,
        'macs_to_clear_from_leases' => undef,
        'cleared_macs'              => undef,
        @_
    );

    my $lease_ip                  = $params{'ip'};
    my $lease_mac                 = $params{'mac'};
    my $macs_to_clear_from_leases = $params{'macs_to_clear_from_leases'};
    my $cleared_macs              = $params{'cleared_macs'}->{'on_request'};

    my $logger = Log::Log4perl->get_logger();

    if (    $lease_ip
        and $lease_mac
        and ref($macs_to_clear_from_leases) eq "HASH"
        and ref($cleared_macs)              eq "HASH"
        and exists $macs_to_clear_from_leases->{$lease_mac}
        and $macs_to_clear_from_leases->{$lease_mac} <= 1 )
    {
        if (   !exists $cleared_macs->{$lease_mac}
            || !exists $cleared_macs->{$lease_mac}->{$lease_ip} )
        {
            $logger->info( "removing entry with MAC "
                  . $lease_mac
                  . " and IP "
                  . $lease_ip
                  . " from leases" );
            $cleared_macs->{$lease_mac}->{$lease_ip} = 1;
        } else {
            ++$cleared_macs->{$lease_mac}->{$lease_ip};
        }
        return 1;
    } else {
        return 0;
    }

}

=head2 isLeaseExpired

Prueft, ob ein Eintrag abgelaufen ist

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<ip>

Die IP-Adresse des Eintrags

=item C<mac>

Die MAC-Adresse des Eintrags

=item C<expire_ts>

Die Karenz-Zeit in Unix-Zeit zum Ende des Eintrags

=item C<expiration_delta>

Eine Array-Referenz mit der Anzahl der Tage, Stunden, Minuten und
Sekunden (in dieser Reihenfolge) für den Ablauf des Eintrags

=item C<cleared_macs>

Ein Hash in dem die Eintraege vermerkt sind, zu denen bereits eine
Protokollierung der Loeschung erfolgt ist. 

=back

=item Return

Wahr, wenn der Eintrag abgelaufen ist, sonst falsch

=back

=cut

sub isLeaseExpired
{

    my %params = (
        'ip'               => undef,
        'mac'              => undef,
        'ends'             => undef,
        'status'           => undef,
        'expire_ts'        => undef,
        'expiration_delta' => undef,
        'cleared_macs'     => undef,
        @_
    );

    my $lease_ip          = $params{'ip'};
    my $lease_mac         = $params{'mac'};
    my $lease_ends        = $params{'ends'};
    my $lease_status      = $params{'status'};
    my $expire_time_stamp = $params{'expire_ts'};
    my $expiration_delta  = $params{'expiration_delta'};
    my $cleared_macs      = $params{'cleared_macs'}->{'expired'};

    my $logger = Log::Log4perl->get_logger();

    my @expire_date_time = Add_Delta_DHMS(
        @{ $lease_ends->{'date'} },
        @{ $lease_ends->{'time'} },
        @{$expiration_delta}
    );
    my $expire_date_time_gm = timegm(
        $expire_date_time[5],     $expire_date_time[4],
        $expire_date_time[3],     $expire_date_time[2],
        $expire_date_time[1] - 1, $expire_date_time[0]
    );

    if (    $lease_ip
        and $lease_mac
        and $lease_ends
        and $lease_status
        and ref($cleared_macs) eq "HASH"
        and $expire_time_stamp > $expire_date_time_gm )
    {
        if (   !exists $cleared_macs->{$lease_mac}
            || !exists $cleared_macs->{$lease_mac}->{$lease_ip} )
        {
            $logger->info(
                (
                    ( $lease_status == STATUS_INACTIV )
                    ? "Inactive "
                    : "Active "
                )
                . "lease entry with MAC "
                  . $lease_mac
                  . " and IP "
                  . $lease_ip
                  . " is expired as configured by LEASES2LDAP_LEASES_EXPIRE_AFTER, remove it."
            );
            $cleared_macs->{$lease_mac}->{$lease_ip} = 1;
        } else {
            ++$cleared_macs->{$lease_mac}->{$lease_ip};
        }
        return 1;
    } else {
        return 0;
    }

}

=head2 clear_dhcp_lease_entry

test wether a mac is to be removed from leases and
only write the entry if that is not the case.

=over 4

=item parameter

hash consisting of:

=over 4

=item C<cleared_leases_file_handle>

File-handle for writing the C<leases_entry>
if this entry is not to be cleared.

=item C<macs_to_clear_from_leases>

A hash-reference with MAC-Addresses which will
be cleared from leases file.

=item C<expire_time_stamp>

The timestamp for the expiration of the entry

=item C<expiration_delta>

An array reference with the number of days, hours,
minutes and seconds (in that order) for the expiration
of the entry

=item C<cleared_macs>

A hash-reference with MAC-Addresses which were really 
removed from leases file.

=item C<lease_entry>

An array with the entry from the leases file which
will be tested against the Parameter
C<macs_to_clear_from_leases>. If the MAC-Address
does not occur, then the entry must be written to
the C<cleared_leases_file_handle>.

=back

=item Return

none

=back

=cut

sub clear_dhcp_lease_entry
{

    my %params = (
        'cleared_leases_file_handle' => undef,
        'macs_to_clear_from_leases'  => undef,
        'expire_time_stamp'          => undef,
        'expiration_delta'           => undef,
        'cleared_macs'               => undef,
        'lease_entry'                => undef,
        @_
    );

    my $clfh                      = $params{'cleared_leases_file_handle'};
    my $macs_to_clear_from_leases = $params{'macs_to_clear_from_leases'};
    my $expire_time_stamp         = $params{'expire_time_stamp'};
    my $expiration_delta          = $params{'expiration_delta'};
    my $cleared_macs              = $params{'cleared_macs'};
    my $lease_entry               = $params{'lease_entry'};

    my $logger = Log::Log4perl->get_logger();

    if ( @{$lease_entry} ) {
        my $lease_ip;
        my $lease_mac;
        my $lease_ends;
        my $lease_status = STATUS_UNDEF;
        foreach my $lease_entry_line ( @{$lease_entry} ) {
            $lease_ip = $1 if ( $lease_entry_line =~ /^lease\s+([\d.]+)\s+\{/ );
            $lease_mac = $1
              if ( $lease_entry_line =~
                /^\s+hardware ethernet ((?:[0-9a-f]{2}:){5}[0-9a-f]{2});/ );
            if ( $lease_entry_line =~ /ends/ ) {
                my @lease_ends_list = split( /[\s;]+/, $lease_entry_line );
                if ( @lease_ends_list >= 3 ) {
                    $lease_ends =
                      parseDateTimeString(
                        $lease_ends_list[-2] . " " . $lease_ends_list[-1] );
                }
            }
            if ( $lease_entry_line =~ /^\s+binding state active/ ) {
                if ( $lease_status == STATUS_UNDEF ) {
                    $lease_status = STATUS_ACTIV;
                }
            }
            if ( $lease_entry_line =~ /^\s+binding state free/ ) {
                if ( $lease_status == STATUS_UNDEF ) {
                    $lease_status = STATUS_INACTIV;
                }
            }
        }
        if ( !$lease_ip || !$lease_mac ) {
            $logger->info("no lease found in lease_entry: [" . join("[CR]", @{$lease_entry}) . "]");
            return;
        }
        if (
            !isLeaseToBeCleared(
                'ip'                        => $lease_ip,
                'mac'                       => $lease_mac,
                'macs_to_clear_from_leases' => $macs_to_clear_from_leases,
                'cleared_macs'              => $cleared_macs
            )
            && !isLeaseExpired(
                'ip'               => $lease_ip,
                'mac'              => $lease_mac,
                'ends'             => $lease_ends,
                'status'           => $lease_status,
                'expire_ts'        => $expire_time_stamp,
                'expiration_delta' => $expiration_delta,
                'cleared_macs'     => $cleared_macs
            )
          )
        {
            print $clfh "$_\n" foreach ( @{$lease_entry} );
        }
    }
    return;
}

=head2 clear_dhcp_leases

clear the leases file of dhcpd

=over 4

=item parameter

hash consisting of:

=over 4

=item C<leases_file_name>

The name of the leases file of DHCP-Daemon.

=item C<cleared_leases_file_name>

The name of the new leases file, into which the
entries are written, which not to be removed.

=item C<macs_to_clear_from_leases>

A hash-reference with MAC-Addresses which will
be cleared from leases file.

=item C<expiration_delta>

An array reference with the number of days, hours,
minutes and seconds (in that order) for the expiration
of the entry

=back

=item return

ref to hash with cleared macs (maybe empty) or undef in case of an error

=back

=cut

sub clear_dhcp_leases
{

    my %params = (
        'leases_file_name'          => $DEFAULT_DHCP_LOG,
        'cleared_leases_file_name'  => $DEFAULT_DHCP_LOG . ".cleared",
        'macs_to_clear_from_leases' => undef,
        'expiration_delta'          => undef,
        @_
    );

    my $leases_file               = $params{'leases_file_name'};
    my $leases_file_cleared       = $params{'cleared_leases_file_name'};
    my $macs_to_clear_from_leases = $params{'macs_to_clear_from_leases'};
    my $expiration_delta          = $params{'expiration_delta'};

    my $logger = Log::Log4perl->get_logger();

    if (   !defined($expiration_delta)
        || ref($expiration_delta) ne "ARRAY"
        || scalar( @{$expiration_delta} != 4 )
        || scalar( grep { $_ =~ /^\d+$/ } @{$expiration_delta} ) != 4 )
    {
        $logger->warn(
            "Call without or wrong expiration delta -> values set to 0");
        $expiration_delta = [ 0, 0, 0, 0 ];
    }

    open( my $leases_fh, '<', $leases_file )
      or do {
        $logger->error(
            "Error opening $leases_file for reading! Leases are NOT cleared");
        return;
      };
    open( my $cleared_fh, '>', $leases_file_cleared )
      or do {
        $logger->error(
            "Error opening $leases_file_cleared for writing! Leases are NOT cleared"
        );
        return;
      };

    my $expire_time_stamp = timegm( gmtime() );
    my $cleared_macs      = { 'expired' => {}, 'on_request' => {} };
    my $lease_entry       = [];
    my $leases_line;
    while ( defined( $leases_line = <$leases_fh> ) ) {
        if ( $leases_line =~ /^\s*(?:#.*|)$/ ) {
            print $cleared_fh "$leases_line";
            next;
        }
        chomp $leases_line;
        if ( $leases_line =~ /^lease/ ) {
            clear_dhcp_lease_entry(
                'cleared_leases_file_handle' => $cleared_fh,
                'macs_to_clear_from_leases'  => $macs_to_clear_from_leases,
                'expire_time_stamp'          => $expire_time_stamp,
                'expiration_delta'           => $expiration_delta,
                'cleared_macs'               => $cleared_macs,
                'lease_entry'                => $lease_entry
            );
            $lease_entry = [$leases_line];
        } else {
            push @{$lease_entry}, $leases_line;
        }

    }
    clear_dhcp_lease_entry(
        'cleared_leases_file_handle' => $cleared_fh,
        'macs_to_clear_from_leases'  => $macs_to_clear_from_leases,
        'expire_time_stamp'          => $expire_time_stamp,
        'expiration_delta'           => $expiration_delta,
        'cleared_macs'               => $cleared_macs,
        'lease_entry'                => $lease_entry
    );

    close($cleared_fh);
    close($leases_fh);

    return $cleared_macs;
}

=head2 parseDateTimeString

parses a string of format "2006/01/01 10:12:30" into its time components and returns it as timelocal(...)

=over 4

=item parameter

string of format "YYYY/mm/dd HH:MM:SS"

=item return

integer seconds since epoch

=back

=cut

sub parseDateTimeString
{
    my $date_time = shift;

    my ( $day_part, $time_part ) = split( /\s/, $date_time );
    my @parsed_date = split( /\//, $day_part );
    my @parsed_time = split( /:/,  $time_part );
    return { 'date' => [@parsed_date], 'time' => [@parsed_time] };
}

1;
