#!/usr/bin/perl -w

use strict;

#no strict "refs";

#
# $Id: DislocatedDevicePermission.pm 7570 2011-11-07 11:05:56Z FSchumacher $
#

#
# package name
#
package SmartClient::DislocatedDevicePermission;

#
# version  of the package
#
our $VERSION = 0.1;

=encoding utf8

=head1 NAME

C<SmartClient::DislocatedDevicePermission>

=head2 SVN Version

 $Id: DislocatedDevicePermission.pm 7570 2011-11-07 11:05:56Z FSchumacher $

=head1 SYNOPSIS

determine what permission a dislocated device should have



=head1 DESCRIPTION

=head2 Purpose

determine what permission a dislocated device should have.
This is company specific, therefore seperated in an own module

=head2 Realisation

=head2 Error handling

=head2 Context and Dependencies

=over 4

=item Context

=item Dependencies


=back



=head2 Interfaces

=begin comment

    Command line parameter are already descripied 
    at SYNOPSIS and usage

=end comment

=over 4

=item Programm

=item User

=back



=head2 Status

=begin comment

    descripes the TODOs/Errors for this programm

=end comment


=head2 Installation and other Notes



=head1 AUTHOR

    smartclient@dass-it.de




=head1 SEE ALSO



=cut

#
# module definition
#

use Exporter;
our @ISA = qw( Exporter );

# =head1 symbols exported by default
#
# - variable1
# - function1
# - function2
#
# =cut
#
our @EXPORT = qw( refreshDislocatedDevicePermission );

# =head1 symbols exported on request
#
# =cut
#
# @EXPORT_OK = qw( );

#
# used modules
#
use SmartClient::Config;
use SmartClient::Constant;
use SmartClient::ScCache;

#
# declarations
#

=head1 Constants

=head1 Global Variables


=head1 Usage

  # get an instance of this class
  my $permission_decider = SmartClient::DislocatedDevicePermission->new(
     'sc_cache' => $sc_cache,
     'config' => $config
  );
  
  # now use it to decide wether a dislocated device may be moved or added
  foreach my $dislocated_device (@dislocated_devices) {
     if ($permission_decider->refreshPermission($dislocated_device) == PERM_GRANTED) {
        # it is allowed to move/add this device
        ... 
     }
  }
  
=head1 Functions

=head2 new

constructor for this class

=head3 Example

  # get an instance of this class
  my $permission_decider = SmartClient::DislocatedDevicePermission->new(
     'sc_cache' => $sc_cache,
     'config' => $config
  );

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<sc_cache>

instance of C<SmartClient::ScCache>

=item C<config>

instance of C<SmartClient::Config> (defaults to C<config> instance of C<sc_cache>)

=back

=item Return

class instance

=back

=cut

sub new
{
    my $class = shift;
    if ( ref $class ) {
        $class = ref $class;
    }
    my $data = {
        'sc_cache' => undef,
        'config'   => undef,
        @_
    };

    my $logger = Log::Log4perl->get_logger();

    #configure fallback to default permission function
    my $permissionFunction = \&permissionDefault;

    #try to get permission function defined by config
    my $config = $data->{'config'} || $data->{'sc_cache'}->{'config'};

    if ( ref $config ) {
        my $configuredFunctionName =
          $config->get('LEASES2LDAP_DISLOCATION_DEVICE_PERMISSION_FUNCTION')
          || 'permissionDefault';
        if ( exists( &{$configuredFunctionName} ) ) {
            $permissionFunction = \&{$configuredFunctionName};
        } elsif ($configuredFunctionName) {
            $logger->warn(
                'configured permission function LEASES2LDAP_DISLOCATION_DEVICE_PERMISSION_FUNCTION="'
                  . $configuredFunctionName
                  . '" could not be found.' );
        }
    } else {
        $logger->info('No config was given in constructor.');
    }
    $data->{'permission_function'} = $permissionFunction;
    return bless $data, $class;
}

=head2 refreshPermission

method to refresh the decision for automatic move/add of dislocated devices 
evaluates and sets the permission for a dislocated device

=over 4

=item Parameter

reference to hash of dislocated device as built by SmartClient::Leases2ldap. 
The data of this hash will be modified by the method!

=item Return

integer: C<PERM_...>
Also modifies C<permission_level> and C<permission> of dislocated device!

=back

=cut

sub refreshPermission
{
    my $self                   = shift;
    my $href_dislocated_device = shift;
    my $logger                 = Log::Log4perl->get_logger();

    #default action is PERM_UNDEF to deny move or add;
    my $permission = PERM_UNDEF;

    my $permissionFunction = $self->{'permission_function'};
    if ( ref $permissionFunction ) {

        # emulate a method invocation for reference to function
        # TODO: is there a better (more readable) way to do this?
        # maybe remember only the name of the method and do something like:
        # if ($self->can($method_name)) {
        #     $permission = $self->$method_name($href_dislocated_device);
        # }
        $permission = $permissionFunction->( $self, $href_dislocated_device );
    } else {
        $logger->error('no permission function defined.');
    }
    return $permission;
}

=head2 permissionDefault

this is just a wrapper to L<rpermissionAddAutoMoveAdmin|SmartClient::DislocatedDevice::permissionAddAutoMoveAdmin>
This method will be called by L<refreshPermission|SmartClient::DislocatedDevice::refreshPermission>

=cut

sub permissionDefault
{
    my $self                   = shift;
    my $href_dislocated_device = shift;
    return $self->permissionAddAutoMoveAdmin($href_dislocated_device);
}

=head2 permissionAddAutoMoveAdmin

Default method to allow auto move/add of workstations/printers.

It will autoallow adding of new workstations and will require admin
intervention for moving of devices. 

This method will be called by L<refresPermission|SmartClient::DislocatedDevice::refreshPermission>

=over 4

=item Parameter

reference to dislocated device hash consisting of

=over 4

=item ...

=back

=item Return

integer: constant C<PERM_...> permission to move/add device

=back

=cut

sub permissionAddAutoMoveAdmin
{
    my $self                   = shift;
    my $href_dislocated_device = shift;

    if ( $href_dislocated_device->{'action'} == ACTION_ADD ) {

        $href_dislocated_device->{'permission_level'} = PERM_LEVEL_AUTO;

        # GRANTED can only be set if PERM_LEVEL_AUTO
        $href_dislocated_device->{'permission'} = PERM_GRANTED;

    } elsif ( $href_dislocated_device->{'action'} == ACTION_MOVE ) {
        $href_dislocated_device->{'permission_level'} = PERM_LEVEL_ADMIN;
    }

    return $href_dislocated_device->{'permission'};
}

=head2 dbkPermissionAddAutoMoveAdmin

special method to allow auto move/add of workstations/printers for Debeka.

It will allow automatic move out of and into specially marked locations for 
all workstations.

This method will be called by L<refreshPermission|SmartClient::DislocatedDevice::refreshPermission>

=over 4

=item Parameter

reference to dislocated device hash consisting of

=over 4

=item ...

=back

=item Return

integer: constant C<PERM_...> permission to move/add device

=back

=cut

sub dbkPermissionAddAutoMoveAdmin
{
    my $self                   = shift;
    my $href_dislocated_device = shift;

    my $logger = Log::Log4perl->get_logger();
    $logger->debug('starting debeka permission decider');

    my $sc_cache = $self->{'sc_cache'};

    if ( $href_dislocated_device->{'action'} == ACTION_ADD ) {

        $href_dislocated_device->{'permission_level'} = PERM_LEVEL_AUTO;

        # GRANTED can only be set if PERM_LEVEL_AUTO
        $href_dislocated_device->{'permission'} = PERM_GRANTED;

    } elsif ( $href_dislocated_device->{'action'} == ACTION_MOVE ) {
        my $destination_location = $href_dislocated_device->{'location_dn_new'};
        my $source_location =
          $sc_cache->getLocationForDn(
            $href_dislocated_device->{'current_ldap_dn'} );
        my $compiled_permission_regex = $self->permissionRegex();
        my $device_type_dn =
          $href_dislocated_device->{'current_ldap_entry'}
          ->get_value('dbkDeviceTypeDn');
        if ($device_type_dn) {
            my @auto_move_device_type_dns = split(
                '\s+',
                $self->{'config'}->get(
                    'LEASES2LDAP_DISLOCATION_DEVICE_AUTO_MOVE_DEVICE_TYPE_DNS')
                  || ''
            );
            if (@auto_move_device_type_dns) {
                for my $auto_move_device_type_dn (@auto_move_device_type_dns) {
                    if ( lc($auto_move_device_type_dn) eq lc($device_type_dn) )
                    {
                        $href_dislocated_device->{'permission_level'} =
                          PERM_LEVEL_AUTO;
                        $href_dislocated_device->{'permission'} = PERM_GRANTED;
                        $logger->debug( 'Auto move LDAP entry '
                              . $href_dislocated_device->{'current_ldap_dn'}
                              . ' with device-type '
                              . $device_type_dn );
                        last;
                    }
                }
            }
            my @auto_move_equal_device_type_dns = split(
                '\s+',
                $self->{'config'}->get(
                    'LEASES2LDAP_DISLOCATION_DEVICE_AUTO_MOVE_EQUAL_DEVICE_TYPE_DNS'
                  )
                  || ''
            );
            if (@auto_move_equal_device_type_dns) {
                my ($dest_dev_type) = $sc_cache->getReferenceAttribute(
                    entry_dn                 => $destination_location,
                    reference_attribute_name => 'dbkDeviceTypeDn'
                );
                for my $auto_move_equal_device_type_dn (
                    @auto_move_equal_device_type_dns)
                {
                    if (
                        lc($auto_move_equal_device_type_dn) eq
                        lc($device_type_dn)
                        && lc($auto_move_equal_device_type_dn) eq
                        lc($dest_dev_type) )
                    {
                        $href_dislocated_device->{'permission_level'} =
                          PERM_LEVEL_AUTO;
                        $href_dislocated_device->{'permission'} = PERM_GRANTED;
                        $logger->debug( 'Auto move LDAP entry '
                              . $href_dislocated_device->{'current_ldap_dn'}
                              . ' with device-type '
                              . $device_type_dn
                              . ' to location with same device-type' );
                        last;
                    }
                }
            }
        }
        if (   $destination_location =~ $compiled_permission_regex
            or $source_location =~ $compiled_permission_regex )
        {
            $href_dislocated_device->{'permission_level'} = PERM_LEVEL_AUTO;
            $href_dislocated_device->{'permission'}       = PERM_GRANTED;
            $logger->debug( 'Auto move LDAP entry '
                  . $href_dislocated_device->{'current_ldap_dn'}
                  . ' from location '
                  . $source_location
                  . ' to location '
                  . $destination_location );
        } else {
            $href_dislocated_device->{'permission_level'} = PERM_LEVEL_ADMIN;
        }
    }

    return $href_dislocated_device->{'permission'};
}

sub permissionRegex
{
    my ($self) = @_;
    if ( $self->{compiled_permission_regex} ) {
        return $self->{compiled_permission_regex};
    }
    my $permission_regex =
      $self->{'config'}->getWithDefault(
        'LEASES2LDAP_DISLOCATION_DEVICE_AUTO_MOVE_PERMISSION_REGEX', '^$' );

    return $self->{compiled_permission_regex} = qr{$permission_regex}im;
}

1;
