#!/usr/bin/perl -w

package SmartClient::LDAP;

=head1 NAME

SmartClient::LDAP

=head1 SYNOPSIS

=begin comment

  kurze Beschreibung, was das Programm macht, Feinkonzept

=end comment

Das Modul SmartClient::Ldap ist ein wrapper für Net::LDAP mit erweiterten features

=head1 DESCRIPTION

=head2 Allg. Beschreibung und Aufgaben

 - Trennung Lese/Schreiboperationen
 - mehrere LDAP Server
 - Lastverteilung
 - Priorisierung
 - autom. Fallback auf andere LDAP Server im Fehlerfall
   (timeout, retries)
 - Methoden: add, search, modify, unbind (Net::LDAP), update (Net::LDAP::Entry)

TODO: abfangen von Fehlern, Fallback

=head2 Funktionsweise / Design / Architektur / Sprache

=over 4

=item Funktionsweise

 - TODO: genaue Beschreibung

In der Konfigurationsdatei muß eine Zeile mit der folgenden Struktur
eingetragen werden:

 ALL_LDAPSERVERS="<ldaphost1>,<ro|rw>,<priority>[;<ldaphost2>,<ro|rw>,<priority>][;...]"

Bei zwei LDAP Servern, die Master-Master Replikation unterstützen, macht z.B.
folgende Zeile Sinn:

 ALL_LDAPSERVERS="ldap1,rw,1;ldap2,rw,2;ldap1,ro,2;ldap2,ro,1"

Damit werden LDAP Verbindung, die Schreibzugriffe benötigen primär zu ldap1
und nur-Lese Zugriffe primär auf ldap2.

=back 4

=cut

use strict;
use Data::Dumper;
use Log::Log4perl;
use Net::LDAP;
use List::Util qw (max);
use SmartClient::LdapUtil;

# version  of the package

my $VERSION = 0.1;

=head1 AUTHOR

=begin comment

Name und e-mail

=end comment

Stephan Dühr
mailto: stephan.duehr@dass-it.de

=cut

# exportieren der Module
use Exporter;

# use Data::Dumper; #debug

my @ISA = qw( Exporter );

# Festlegen der zu exportierenden Module, Methoden, Funktionen, Variablen, Konstanten
#@EXPORT = qw(init_branches get_regions);

=head1 exportierte Module

S<  >B<Modul>

=cut

# Festlegen der optional zu exportierenden Module, Methoden, Funktionen, Variablen, Kostanten

## @EXPORT_OK;

=head1 optional exportierte Module

S<  >B<Modul>

S<  >B<Funktion>

S<  >B<Variable>

=cut

=head1 Modul

=begin comment

Beschreibung des Moduls

=end comment

S<  >B<Modul> { I<parameter> =E<gt> I<value>, ... };

S<  >B<Modul> I<\@param1>, I<\@param2>[, I<$option1> ] [, I<$option2> ] ];   

=cut

#Programmcode: constructor

sub new
{
    my $package = shift;    ## packagename

    # Definition of defaults
    my $data = {
        'CONFIG' => undef,    ## reference to SmartClient::Config object
        ##   has to contain ALL_LDAPSERVER, ALL_LDAPBASE, ALL_LDAPTIMEOUT
        'LDAPSERVERS' =>
          undef,    ## reference to a hash containing info of LDAP Servers
        ##   key: servername, value: array of priority and
        ##   write capability ("ro": readonly, "rw": readwrite)
        'base' => undef,    ## default LDAP base
        ## (can be overwritten at every LDAP operation)
        'TIMEOUT'    => undef,    ## timeout for LDAP operations
        'LDAPBINDDN' => undef,    ## LDAP bind DN for modifying operations
        'LDAPBINDPW' => undef,    ## LDAP bind password for modifying operations
        'MAXRETRIES' => 2,        ## max. number of retries
        'LDAPSTATUS' =>
          undef,    ## reference to a hash containing LDAP status such
        ##   as failcount, retry, etc. (to be detailed)
        ##   key: Servername
        'LDAPCONN' => undef,    ## LDAP connection (Net::LDAP object)
        'NEEDCAP'  => undef,    ## needed capability ("ro"|"rw"|"ri")
        'LDAPHOST' => undef,    ## actual ldap host
        'SRVLIST'  => undef,    ## list of ldap server ordered by priority
        'DEBUG'    => 0         ## print debug msgs if set to true
    };

    # get constructor parameters
    while (@_) {
        my $field = shift;
        if (@_) {
            $data->{$field} = shift;
            if ( !$data->{$field} ) {
                $data->{$field} = "";
            }
        }
    }
    if ( init($data) ) {

        # return the object-hash
        return bless($data, $package);
    } else {
        return;
    }
}

sub init
{
    my $data = shift;

    #init log4perl if it is not already initialized
    Log::Log4perl::init_once('/etc/smartclient/log4perl.noninteractive');
    my $logger = Log::Log4perl->get_logger();
    ## TODO: check parameters:
    # - SmartClient::Config object containing ALL_LDAPSERVERS and ALL_LDAPTIMEOUT
    # - correct information in ALL_LDAPSERVERS
    if ( !exists $data->{'CONFIG'}->{'ALL_LDAPSERVERS'} ) {
        $logger->fatal('No entry ALL_LDAPSERVERS defined in Config file');
        return;
    }

    # parse ALL_LDAPSERVERS, a config file line has to look like this:
    # ALL_LDAPSERVERS="ldap.server.debeka.de,rw,4;ldap2.server.debeka.de,ro,5;"
    $logger->debug( "parsing ALL_LDAPSERVERS setting: "
          . $data->{'CONFIG'}->{'ALL_LDAPSERVERS'} );
    $data->{'LDAPSERVERS'} = {};
    my @ldap_servers = split( /;/, $data->{'CONFIG'}->{'ALL_LDAPSERVERS'} );
    foreach my $ldap_server (@ldap_servers) {

        # split each LDAP Server Entry into:
        # [0]: name
        # [1]:  ro: read only capability
        #       rw: read write
        #       ri: read internal attributes (e.g. modification time)
        # [2]: priority
        my @ldap_srv = split( /,/, $ldap_server );
        if ( $ldap_srv[1] eq $data->{'NEEDCAP'} ) {
            $data->{'LDAPSERVERS'}->{ $ldap_srv[0] } =
              [ $ldap_srv[1], $ldap_srv[2] ];
        }
    }

    # test
    if ( $logger->is_debug() ) {
        foreach my $key ( keys %{ $data->{'LDAPSERVERS'} } ) {
            $logger->debug( "LDAP Servers from Config file: " 
                  . $key . ","
                  . $data->{'LDAPSERVERS'}->{$key}->[0] . ","
                  . $data->{'LDAPSERVERS'}->{$key}->[1] );
        }
    }

    # choose appropriate ldap server according to needed capability (ro|rw|ri)
    # and priority by building a hash (key: priority, value: array of servers
    # for that prio. Only server of the needed capability (ro|rw) are put
    # into the hash.
    $data->{'SRVLIST'} = [];
    my %srv_prio;
    foreach my $key ( keys %{ $data->{'LDAPSERVERS'} } ) {

        #print $data->{'LDAPSERVERS'}->{$key}->[0], ",",
        #      $data->{'LDAPSERVERS'}->{$key}->[1],"\n"; #debug
        if ( $data->{'LDAPSERVERS'}->{$key}->[0] eq $data->{'NEEDCAP'} ) {
            my $prio = $data->{'LDAPSERVERS'}->{$key}->[1];
            if ( !$srv_prio{$prio} ) {
                $srv_prio{$prio} = [];
            }
            push( @{ $srv_prio{$prio} }, $key );
        }
    }
    my @prio_list = keys %srv_prio;
    my $max_prio  = max(@prio_list);
    foreach my $prio ( sort @prio_list ) {
        foreach my $srv ( @{ $srv_prio{$prio} } ) {
            push( @{ $data->{'SRVLIST'} }, $srv );
        }
    }

    $logger->debug( sub { Data::Dumper->Dump( [$data] ) } );

    # make LDAP connection
    $data->{'LDAPCONN'} = undef;
    while ( !$data->{'LDAPCONN'} ) {
        if ( @{ $data->{'SRVLIST'} } == 0 ) {
            $logger->fatal('Could not find appropriate LDAP Server');
            return;
        }
        $data->{'LDAPHOST'} = shift( @{ $data->{'SRVLIST'} } );

        $logger->debug(
            "trying to connect to LDAP Server " . $data->{'LDAPHOST'} );
        $data->{'LDAPCONN'} =
          Net::LDAP->new( $data->{'LDAPHOST'}, version => 3 );

        #$data->{'LDAPCONN'} = Net::LDAP->new($data->{'LDAPHOST'});
        if ( $data->{'LDAPCONN'} ) {
            my $ldap = $data->{'LDAPCONN'};
            if ( $data->{'LDAPBINDDN'} ) {

                #print STDERR "trying to bind to LDAP Server ".$data->{'LDAPHOST'}."\n"; #debug
                #print STDERR "DN: ", $data->{'LDAPBINDDN'}, "\n"; #debug
                #print STDERR "PW: ", $data->{'LDAPBINDPW'}, "\n"; #debug
                my $mesg =
                  $ldap->bind( $data->{'LDAPBINDDN'},
                    password => $data->{'LDAPBINDPW'} );

                if ( $logger->is_debug() ) {
                    $logger->debug( "ldap bind gave: "
                          . $mesg->code . " : "
                          . Data::Dumper->Dump( [$mesg] ) );
                }

                if ( $mesg->code ) {
                    $logger->warn( "error binding to ldap server using "
                          . $data->{'LDAPBINDDN'} . ' and server '
                          . $data->{'LDAPHOST'} . " : "
                          . $mesg->error );
                    $data->{'LDAPCONN'} = undef;
                    next;
                }
            }
        } else {
            $logger->warn(
                "error connecting to LDAP Server " . $data->{'LDAPHOST'} );
        }
    }

    if ( !$data->{'base'} ) {
        $data->{'base'} = $data->{'CONFIG'}->{'ALL_LDAPBASE'};
        if ( !$data->{'base'} ) {
            $logger->warn("ALL_LDAPBASE is undefined");
        }
    }

    #print STDERR "using LDAP Server ", $data->{'LDAPHOST'}, "\n";
    if ( !$data->{'LDAPCONN'} ) {
        return;
    }
    return $data;
}

=head2 search

LDAP Suche mit Net::LDAP::search()

=over 4

=item Parameter

 siehe man Net::LDAP und man Net::LDAP::Search

=item Rckgabewert

 siehe man Net::LDAP

=back

=cut

sub search
{
    my $self        = shift;
    my @search_args = @_;
    my $ldap        = $self->{'LDAPCONN'};

    #print STDERR Data::Dumper->Dump([@search_args]); #debug
    my $result = $ldap->search( base => $self->{'base'}, @search_args );
    return $result;
}

=head2 getEntry

get the entry to a specific DN
(uses LDAP::search)

=over 4

=item Parameter

    dn
    [ attributes (default: '*')]

=item Rckgabewert

 Net::LDAP::Entry or undef

=back

=cut

sub getEntry
{
    my $self  = shift;
    my $dn    = shift;
    my $attrs = shift || ['*'];

    my $ldap = $self->{'LDAPCONN'};

    return ldap_get_dn_entry(
        LDAP  => $self->{'LDAPCONN'},
        DN    => $dn,
        ATTRS => $attrs
    );
}

=head2 add

LDAP add mit Net::LDAP::add()

=over 4

=item Parameter

 siehe man Net::LDAP

=item Rckgabewert

 siehe man Net::LDAP

=back

=cut

sub add
{
    my $self   = shift;
    my @args   = @_;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $ldap->add(@args);
    return $result;
}

=head2 modify

LDAP modify mit Net::LDAP::modify()

=over 4

=item Parameter

 siehe man Net::LDAP

=item Rckgabewert

 siehe man Net::LDAP

=back

=cut

sub modify
{
    my $self   = shift;
    my @args   = @_;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $ldap->modify(@args);
    return $result;
}

=head2 moddn

LDAP moddn mit Net::LDAP::moddn()

=over 4

=item Parameter

=item Return

=back

=cut

sub moddn
{
    my $self   = shift;
    my @args   = @_;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $ldap->moddn(@args);
    return $result;
}

=head2 inner

undocumented method of Net::LDAP as passthrough

=over 4

=item Parameter

=item Return

=back

=cut

sub inner
{
    my $self   = shift;
    my @args   = @_;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $ldap->moddn(@args);
    return $result;
}

=head2 update

update an LDAP Entry using Net::LDAP::Entry::update()

=over 4

=item Parameter

 Net::LDAP::Entry Objekt

=item Rckgabewert

 siehe man Net::LDAP::Message

=back

=cut

sub update
{
    my $self   = shift;
    my $entry  = shift;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $entry->update($ldap);
    return $result;
}

=head2 unbind

calls Net::LDAP::unbind(), this method is somewhat more compatibility with Net::LDAP,
and may be changed using a destructor that closes the LDAP connection.

=over 4

=item Parameter

 keine

=item Rckgabewert

 siehe man Net::LDAP::Message

=back

=cut

sub unbind
{
    my $self   = shift;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $ldap->unbind;
    return $result;
}

=head2 check_connection

Check if the LDAP connection still works,
otherwise reconnect

=over 4

=item Parameter

 keine

=item Rckgabewert

 true if connection ok or reconnect succeeded
 otherwise false

=back

=cut

sub check_connection
{
    my $self   = shift;
    my @args   = @_;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = 0;
    my $msg;
    my $logger = Log::Log4perl->get_logger();

    if ($ldap) {
        $msg = $ldap->search(
            base      => "",
            scope     => 'base',
            filter    => '(objectClass=*)',
            attrs     => ['supportedLDAPVersion'],
            timelimit => $self->{'TIMEOUT'}
        );

        #my $rootDSE = $ldap->root_dse();
    }

    if ( !ref($msg) || $msg->code ) {

        #print "type of msg: ", ref($msg), "\n"; #debug
        #print Data::Dumper->Dump([$msg]); #debug
        my $error_msg = "";
        if ( ref($msg) ) {
            $error_msg = $msg->error;
        } else {
            $error_msg = "general LDAP error";
        }
        $logger->warn( "'connection to LDAP Server "
              . $self->{'LDAPHOST'}
              . " failed: "
              . $error_msg
              . ", reconnecting'" );
        if ( !init($self) ) {
            return 0;
        }
    }
    return 1;
}

sub set_values
{
    my $self = shift;

    my %params = (
        'DN'     => undef,
        'VALUES' => undef,
        'ENTRY'  => undef,
        @_,
    );

    return ldap_set_values(
        LDAP => $self->{'LDAPCONN'},
        %params
    );
}

sub getNetLdap
{
    my $self = shift;
    return $self->{'LDAPCONN'};
}

sub getBase
{
    my $self = shift;
    return $self->{'base'};
}

1;
