#!/usr/bin/perl -w
use strict;

package SmartClient::LdapCache;

=head1 NAME

SmartClient::LdapCache

=head2 SVN Version

 $Id: LdapCache.pm 8040 2012-07-24 13:39:15Z FSchumacher $

=head1 SYNOPSIS

=begin comment

Class for cached use of Net::LDAP::Entry objects

=end comment

Class for cached use of Net::LDAP::Entry objects

=head1 DESCRIPTION

=head2 Allg. Beschreibung und Aufgaben

This Module provides methods cached use of Net::LDAP::Entry objects
for the purpose of reducing LDAP server load by requesting only one
search operation and storing the resulting Net::LDAP::Entry objects
in memory.

For refreshing the cached entries, a modifyTimestamp based search
operation is used which only finds and transfers the entries that
changed in the LDAP server since cache creation or last refresh.

It is important to consider that entries which are deleted in LDAP
remain in the cache even after refreshing because a modifyTimestamp
based search only finds changed or added entries.

=head2 Funktionsweise / Design / Architektur / Sprache

=over 4

=item Funktionsweise

 - TODO: genaue Beschreibung

=back 4

=cut

use Data::Dumper;
use Log::Log4perl;
use Net::LDAP qw(LDAP_NO_SUCH_OBJECT);
use Net::LDAP::Entry;
use SmartClient::LDAP;
use SmartClient::LdapUtil;
use SmartClient::Util;
use SmartClient::IP;
use SmartClient::Constant;

# version  of the package

our $VERSION = 0.1;

=head1 AUTHOR

=begin comment

Name und e-mail

=end comment

Stephan Dühr
mailto: stephan.duehr@dass-it.de

=cut

# exportieren der Module

# Festlegen der zu exportierenden Module, Methoden, Funktionen, Variablen, Konstanten

=head1 exportierte Module

S<  >B<Modul>

=cut

# Festlegen der optional zu exportierenden Module, Methoden, Funktionen, Variablen, Kostanten

=head1 optional exportierte Module

S<  >B<Modul>

S<  >B<Funktion>

S<  >B<Variable>

=cut

=head1 Modul

=begin comment

Beschreibung des Moduls

=end comment

S<  >B<Modul> { I<parameter> =E<gt> I<value>, ... };

S<  >B<Modul> I<\@param1>, I<\@param2>[, I<$option1> ] [, I<$option2> ] ];   

=cut

my $LOGGER = Log::Log4perl->get_logger();

#Programmcode: constructor

sub new
{
    my $package = shift;

    # Definition of defaults
    my $data = {

        # reference to SmartClient::Config object
        'config' => undef,

        # time (in seconds) that a operation may needs
        # to be visible at the LDAP server.
        # required, to find all new entries,
        # because eDirectory eventually changes object
        # with older modifyTimestamp
        # 5 seconds are evaluated for eDirectory
        # From config: ALL_LDAP_SYNC_INTERVAL
        'modifyTimestamp_offset' => -1,

        # domain component for standards.
        # From config. Used at treeSearch
        'standard_dc' => undef,

        # LDAP entry cache
        'entrycache' => undef,

        # LDAP entry cache base
        'entrycache_base' => undef,

        # LDAP entry cache filter
        'entrycache_filter' => undef,

        # LDAP entry cache attribute list
        'entrycache_attrs' => undef,

        # LDAP entry cache lowercase setting
        'entrycache_lowercase' => undef,

        # timestamp of last cache refresh
        'last_check_timestamp' => undef,
        @_
    };

    # internal instance variables
    # this is used by assignLocations and getAssignedLocations
    $data->{'unassignable_entries_found_flag'} = 0;
    $data->{'delete_notify_dn_hash'}           = {};
    $data->{'delete_notify_locations_hash'}    = {};

    $data->{'added_location'}   = {};
    $data->{'removed_location'} = {};

    # the following entries are use by the hierarchical cache
    $data->{'hierarchical_cache'}                   = {};
    $data->{'hierarchical_cache_objectclass_names'} = {};

    # hierarchical LDAP cache for every leaf
    #   which has entries in scope one
    $data->{'hierarchical_cache_dirty'} = {};
    $data->{'index_for_attr'} = { 'macaddress' => {}, 'iphostnumber' => {} };
    if ( init($data) ) {

        # return the object-hash
        return bless( $data, $package );
    } else {
        return;
    }
}

sub init
{
    my $logger = $LOGGER;

    my $data = shift;

    $data->{'standard_dc'} ||= $data->{'config'}->get('ALL_LDAP_STANDARD_DC');

    # use negative LDAP Sync Interval as modifyTimestamp offset
    # and check, that offset it as least -1
    my $ldap_sync_interval = $data->{'config'}->get('ALL_LDAP_SYNC_INTERVAL');
    if ($ldap_sync_interval) {
        $data->{'modifyTimestamp_offset'} = -$ldap_sync_interval;
    }
    $logger->debug(
        "modifyTimestamp_offset: " . $data->{'modifyTimestamp_offset'} );

    # check ldap attributes
    if ( !defined( $data->{'entrycache_attrs'} ) ) {
        my @attrs =
          split( ",", $data->{'config'}->get('ALL_LDAP_ATTRIBUTES_SC_CACHE') );
        if (@attrs) {
            $data->{'entrycache_attrs'} = \@attrs;
        } else {
            $data->{'entrycache_attrs'} = [ '*', 'modifyTimestamp' ];
            $logger->warn(
                "no list of attributes defined. Falling back to all attributes and modifyTimestamp"
            );
        }
    }
    $logger->debug(
        "attrs: [ " . join( ", ", @{ $data->{'entrycache_attrs'} } ) . " ]" );

    # check if hierarchical cache should be initialized for some objectClasses
    if ( $data->{'config'}->get( 'ALL_LDAP_CACHE_OBJECTCLASS_NAMES', 1 ) ) {
        my @objectclass_names = split /,/,
          $data->{'config'}->get( 'ALL_LDAP_CACHE_OBJECTCLASS_NAMES', 1 );
        foreach my $objectclass_name (@objectclass_names) {
            $data->{'hierarchical_cache_objectclass_names'}
              ->{$objectclass_name} = 1;
        }
    }

    return $data;
}

=head2 normalize

gets an dn and return the same dn in normalized form.
Should be used, before accessing the entry cache.

=over 4

=item Parameter

    dn

=item Return

    dn

=back

=cut

sub normalize
{
    my $self = shift;
    my $dn   = shift;

    return ldap_normalize( $dn, ( !$self->{'entrycache_lowercase'} ) );
}

=head2 generateEntryCache

Generates a hash consisting of Net::LDAP::Entry Objects received
by an LDAP search operation, the hash key is the DN. 

This method obsoletes ldap_gen_entrycache from SmartClient::LdapUtil,
but generateEntryCache uses a LDAP filter expression in contrast to
objectClass name in ldap_gen_entrycache.


=over 4

=item Parameter

Hash consisting of:

 sc_ldap:     SmartClient::LDAP object (mandatory)
 filter:      LDAP search filter expression
              default: objectClass=*
 attrs:       List of attribute names to include in the cache
              default: ['*']
              Normaly specified in constructor
 base:        Base DN for the search
              default: ALL_LDAPBASE from the config object given
			           the the constructor
 lowercase:   by default the hash keys (DN) are used in lowercase,
              whereas the LDAP results can be mixed case. This must
			  be considered when trying to find a given DN in the
			  cache.
              If storing the hash keys in lowercase is not wanted,
			  use lowercase => 0
 ldif:        used for testing this class. Usage will disable ldap
              search and use correspondig ldif file instead.

=item Return

 SmartClient::LdapCache object or undef in case of error

=back

=cut

sub generateEntryCache
{
    my $self  = shift;
    my %param = (
        'base'      => $self->{'config'}->{'ALL_LDAPBASE'},
        'filter'    => $self->{'config'}->{'ALL_LDAP_FILTER_GET_ALL'},
        'attrs'     => undef,
        'lowercase' => 1,
        'ldif'      => undef,
        @_,
    );

    my $logger = $LOGGER;
    $logger->debug("generateEntryCache was called");

    # TODO: check parameters:
    # - make sure that modifyTimestamp is contained in attrs
    #   if attrs given as parameter

    # check if sc_ldap parameter was passed
    # this will be disabled by passing a ldif filename
    my $ldif    = $param{'ldif'};
    my $sc_ldap = $param{'sc_ldap'};
    if ( !$sc_ldap and !$ldif ) {
        $logger->error("passing parameter sc_ldap is mandatory");
        return;
    }
    if ( ref($sc_ldap) ne 'SmartClient::LDAP' and !$ldif ) {
        $logger->error("parameter sc_ldap must be a SmartClient::LDAP object");
        return;
    }

    # check if ldap search filter is defined
    if ( !$param{'filter'} ) {
        $param{'filter'} = '(objectClass=*)';
        $logger->warn(
            "no filter defined. Falling back to " . $param{'filter'} );
    }

    # save entry cache parameters in object
    $self->{'entrycache_base'}   = $param{'base'};
    $self->{'entrycache_filter'} = $param{'filter'};
    ##  entrycache_attrs are already set by constructor
    ## (neccessary to check reference_attrs from ScCache)
    ## $self->{'entrycache_attrs'}     = $param{'attrs'};
    $self->{'entrycache_lowercase'} = $param{'lowercase'};

    my $attrs = $self->{'entrycache_attrs'};

    my $result;
    if ($ldif) {

        #generate entries from ldif
        use SmartClient::FakeLdapResult;
        $result = SmartClient::FakeLdapResult->new( 'ldif' => $ldif );
    } else {

        #read entries from ldap
        $result = $sc_ldap->search(
            base   => $param{'base'},
            scope  => 'sub',
            filter => $param{'filter'},
            attrs  => $attrs,
            ## needed?
            timelimit => $sc_ldap->{'TIMEOUT'}
        );
    }
    if ( ldap_check_error( $result, 0 ) ) {
        return;
    }

    foreach my $entry ( $result->entries ) {
        my %entry_hash = entryToHash($entry);
        my $message    = $self->checkEntryAttributes(%entry_hash);
        if ($message) {
            $logger->warn( '[LDAP-DATA] attributes check for '
                  . $entry->dn
                  . ": $message" );
        }
        $self->setInternalEntryHash( \%entry_hash );
    }

    # save maximum modifyTimestamp in object
    $self->{'last_check_timestamp'} = ldap_get_max_tstamp($result);

    #    $self->createObjectClassIndex();
    $self->assignLocations();
    return $self->{'entrycache'};
}

=head2 refreshEntryCache

refreshes the entry cache

By combining the filter used for creating the entry cache with
a modifyTimestamp filter for a LDAP search to find all changed
entries and update them in the entry cache.

=over 4

=item Parameter

Hash consisting of:

 sc_ldap:     SmartClient::LDAP object (mandatory)

=item Return

A list of DN which were changed relative to the last check
(empty list if nothing is found) or
undef in case of error.

=back

=cut

sub refreshEntryCache
{
    my $self  = shift;
    my %param = (
        sc_ldap => undef,
        @_,
    );
    my @changed_dn = ();
    my @deleted_dn = ();

    my $logger = $LOGGER;

    # get SmartClient::LDAP connection object
    # check if sc_ldap parameter was passed
    my $sc_ldap = $param{'sc_ldap'};
    if ( !$sc_ldap ) {
        $logger->error("passing parameter sc_ldap is mandatory");
        return;
    }
    if ( ref($sc_ldap) ne 'SmartClient::LDAP' ) {
        $logger->error("parameter sc_ldap must be a SmartClient::LDAP object");
        return;
    }

    # add modifyTimestamp comparison to filter
    # use a negative offset,
    # because
    my $timestamp_filter = ldap_filter_add_timestamp(
        filter    => $self->{'entrycache_filter'},
        timestamp => calcTimestampWithOffset(
            'timestamp' => $self->{'last_check_timestamp'},
            'offset'    => $self->{'modifyTimestamp_offset'}
        )
    );
    $logger->debug( 'timestamp filter: ' . $timestamp_filter );

    # LDAP search for changed data
    my $ldap_result = $sc_ldap->search(
        base   => $self->{'entrycache_base'},
        filter => $timestamp_filter,
        attrs  => $self->{'entrycache_attrs'}
    );
    if ( ref $ldap_result ) {
        if ( $ldap_result->code
            and ( $ldap_result->code != LDAP_NO_SUCH_OBJECT ) )
        {

            # LDAP error different from NO_SUCH_OBJECT
            # (normal if nothing changed):
            # in this case it can't be determined if
            # anything changed by timestamp
            $logger->warn( 'LDAP error: ' . $ldap_result->error );
            return;
        }

        # return empty list if nothing changed

        # update changed entries and build list of
        # the changed entries DN for returning
        foreach my $entry ( $ldap_result->entries ) {
            if ( $self->hasEntryChanged($entry) ) {
                $self->setEntry($entry);
                push( @changed_dn, $entry->dn() );
            }
        }
    } else {
        $logger->warn("ldap search result was not a reference");
        if ( !$sc_ldap->check_connection() ) {
            $logger->warn("ldap connection seems to be down.");
        }
    }

    # process DNs which got a delete notification
    foreach
      my $notified_delete_dn ( keys %{ $self->{'delete_notify_dn_hash'} } )
    {
        $logger->debug("removing $notified_delete_dn from cache");
        if ( !$self->removeEntry($notified_delete_dn) ) {

            #TODO: these warnings should probably be in removeEntry
            $logger->warn("could not remove entry $notified_delete_dn");
            if ( $logger->is_info() ) {
                my $local_base =
                  ldap_normalize( ldap_shiftbase($notified_delete_dn) );
                $logger->info( "hierarchical_cache for $local_base:"
                      . Dumper( $self->{'hierarchical_cache'}->{$local_base} )
                );
            }
        }
        push( @deleted_dn, $notified_delete_dn );
    }

    # look if there are any locations marked as changed. Those will be added to changed_dn
    foreach my $changed_location ( keys %{ $self->{'changed_locations_hash'} } )
    {
        $logger->info(
            "adding location $changed_location to changed_dn since it was in changed_locations_hash"
        );
        push @changed_dn, $changed_location;
    }
    $self->{'changed_locations_hash'} = {};

    if ( @changed_dn || @deleted_dn ) {

        # refresh location assignment
        $self->assignLocations( refreshed_dn_aref => \@changed_dn );
    }

    # save maximum timestamp of result
    if (@changed_dn) {
        $self->{'last_check_timestamp'} = ldap_get_max_tstamp($ldap_result);
    }

    return ( @changed_dn, @deleted_dn );
}

=head2 objectClassIntersection

Builds intersection of objectclass indexed DNs of entries
in the cache for a given list of objectClasses using the
data structure created by createObjectClassIndex.

Important: It must be considered that this does only work similar
to LDAP search filters, because only objectClass is considered
and objectClass inheritance is not applied.

=head3 Example

 my @person_accounts = $Ldap_Cache->objectClassIntersection(
    'base' => 'o=company,c=de',
    'scope' => 'base',
    'objectclasses' => [ 
        'person',
        'posixAccount' 
    ]
 );

=over 4

=item Parameter

hash consisting of:

C<objectclasses> reference to list of objectClass names

C<base> base of search (defaults to entrycache_base)

C<scope> searchscope one of C<base>|C<one>|C<sub> (defaults to sub)

=item Return

List of DN for entries in the cache having all the given objectclasses

=back

=cut

#TODO this function should be handled inside of getObjectClassIndex
sub objectClassIntersection
{
    my $self   = shift;
    my %params = (
        'base'          => $self->{'entrycache_base'},
        'scope'         => 'sub',
        'objectclasses' => undef,
        @_
    );
    my $base  = $params{'base'};
    my $scope = $params{'scope'};
    if ( !ref $params{'objectclasses'} ) {
        return;
    }
    my @objectclasses = @{ $params{'objectclasses'} };

    my $objectclass_count = @objectclasses;
    my %dn_count;
    my @intersection = ();

    foreach my $objectclass (@objectclasses) {
        my $counter = 0;
        foreach my $dn (
            $self->getObjectClassIndex(
                'base'             => $base,
                'scope'            => $scope,
                'objectclass_name' => $objectclass
            )
          )
        {
            $dn_count{$dn}++;
            $counter++;
        }
    }
    foreach my $dn ( keys %dn_count ) {
        if ( $dn_count{$dn} == $objectclass_count ) {
            push( @intersection, $dn );
        }
    }
    return @intersection;
}

=head2 objectClassUnion

Builds union of objectclass indexed DNs of entries
in the cache for a given list of objectClasses using the
data structure created by createObjectClassIndex.

Important: It must be considered that this does only work similar
to LDAP search filters, because only objectClass is considered
and objectClass inheritance is not applied.

=head3 Example

 my @person_accounts = $Ldap_Cache->objectClassUnion(
    'base' => 'o=company,c=de',
    'scope' => 'sub',
    'objectclasses' => [ 
        'dbkWorkstation',
        'dbkPrinter' 
    ]
 );

=over 4

=item Parameter

hash consisting of:

C<objectclasses> reference to list of objectClass names

C<base> ldap search base (defaults to entrycache_base)

C<scope> searchscope one of C<one>|C<base>|C<sub> (defaults to C<sub>)

=item Return

List of DN for entries in the cache having any of the given objectclasses

=back

=cut

#TODO this function should be handled inside of getObjectClassIndex
sub objectClassUnion
{
    my $self   = shift;
    my %params = (
        'base'          => $self->{'entrycache_base'},
        'scope'         => 'sub',
        'objectclasses' => undef,
        @_
    );
    if ( !ref $params{'objectclasses'} ) {
        return;
    }
    my @objectclasses = @{ $params{'objectclasses'} };
    my $base          = $params{'base'};
    my $scope         = $params{'scope'};
    my @union         = ();

    my $objectclass_count = @objectclasses;
    my %dn_union;

    foreach my $objectclass (@objectclasses) {
        foreach my $dn (
            $self->getObjectClassIndex(
                'base'             => $base,
                'scope'            => $scope,
                'objectclass_name' => $objectclass
            )
          )
        {
            $dn_union{$dn}++;
        }
    }
    @union = keys %dn_union;
    return @union;
}

=head2 findLocation

find the location for a given DN

=head3 Example

 my $location_dn = $Ldap_Cache->findLocation('cn=pc00268,ou=gs0010,ou=Koeln,o=smartclient,c=de');

=over 4

=item Parameter

DN

=item Return

DN of the scLocation container for the given DN

=back

=cut

sub findLocation
{
    my $self = shift;
    my $dn   = shift;

    my %locations_hash =
      map { $self->normalize($_) => $_ } $self->getAllLocationDns;

    my $location_dn = $self->normalize($dn);
    while ( !$locations_hash{$location_dn} ) {
        $location_dn = ldap_shiftbase($location_dn);
        if ( ( !$location_dn )
            || $location_dn eq $self->{'entrycache_base'} )
        {
            return;
        }
    }
    return $location_dn;
}

=head2 getAllLocations

=over 4

=item Parameter

=item Return

=back

=cut

sub getAllLocationDns
{
    my $self = shift;
    my $base = shift;
    if ( !$base or ( length($base) < length( $self->{'entrycache_base'} ) ) ) {
        $base = $self->{'entrycache_base'};
    }

    return $self->getObjectClassIndex(
        'objectclass_name' => 'scLocation',
        'base'             => $base
    );
}

=head2 assignLocations

Assigns Location to all objects in the entry cache if applicable. This
is done by using the hierarchical cache (see getObjectClassIndex)
for scLocation which yields a list of all Locations. To assign an entry
its DN is cut off by one DN component from the left until it matches
a location DN.

The assigned location DN is stored in the entry cache for each assignable
entry in the 'location' field. This will be empty for unassignable entries.

This method is implicitly called by generateEntryCache.

=head3 Examples

Example for getting the location DN for a given entry DN:

 $entry_location_dn = $Ldap_Cache->getLocationForDn($entry_dn);

Example for getting all entry DNs assigned to a given location DN:

 @location_entry_dn_list = map { $_->dn } $Ldap_Cache->search('base' => $location_dn, 'attr' => '*', 'value' => '*');

or better retrieve special objectClasses 

 @location_ws_dn_list = $Ldap_Cache->getObjectClassIndex('base' => $location_dn, 'objectclass_name' => 'dbkWorkstation');

=over 4

=item Parameter

Hash consisting of:

Optional parameter refreshed_dn_aref: if given a reference to a list
of DNs which changed as reported by refreshEntryCache this means only
to renew the assignment of the given DNs, otherwise the assignment is
built for the whole cache content based on the hierarchical cache for
scLocation.

=item Return

Reference to the locations hash structure

=back

=cut

sub assignLocations
{
    my $self  = shift;
    my %param = (
        'refreshed_dn_aref' => undef,
        @_,
    );

    my $locations_hash = {};
    my @unassignable;
    my $entrycache      = $self->{'entrycache'};
    my $process_dn_list = [];
    my $logger          = $LOGGER;

    $self->{'unassignable_entries_found_flag'} = 0;

    # check if a list of refreshed DNs was passed
    if ( $param{'refreshed_dn_aref'} ) {
        $process_dn_list = $param{'refreshed_dn_aref'};
    } else {
        $process_dn_list = [ keys %{$entrycache} ];
        $logger->debug("no changed dn list passed, assigning all");

        # build new location hash of all location DNs
        my @sclocation_list =
          $self->getObjectClassIndex( 'objectclass_name' => 'scLocation' );
    }

    # walk through entry cache and find location
    # by successively cutting off a part of the DN
    foreach my $entry_dn ( @{$process_dn_list} ) {
        $entry_dn = $self->normalize($entry_dn);
        if ( $entry_dn =~ m/ou=global/io ) {

            #entry in global affects every other entry
            $self->{'unassignable_entries_found_flag'} = 1;
        }
        my $location_dn = $entry_dn;
        while ( !$self->isLocation( 'dn' => $location_dn )
            and !( $location_dn eq $self->{'entrycache_base'} ) )
        {
            $location_dn = ldap_shiftbase($location_dn);
            if ( ( !$location_dn )
                || $location_dn eq $self->{'entrycache_base'} )
            {
                $logger->debug("could not assign $entry_dn");

                #this is a unassingable entry, which has no global effect.
                #It will be handled by getAssignedLocation correctly.
                last;
            }
        }
        if ( $self->isLocation( 'dn' => $location_dn ) ) {
            $self->setLocationForDn( $entry_dn, $location_dn );
        }
    }

    # process delete notification
    if ( keys %{ $self->{'delete_notify_dn_hash'} } ) {
        $logger->debug("processing delete notifications");
        foreach my $deleted_dn ( keys %{ $self->{'delete_notify_dn_hash'} } ) {
            if ( $deleted_dn =~ m/ou=global/io ) {

                #entry in global affects every other entry
                $self->{'unassignable_entries_found_flag'} = 1;
            }
            my $assigned_location_dn = $self->findLocation($deleted_dn);
            if ( $self->isLocation( 'dn' => $deleted_dn ) ) {

                #this dn is a location, so mark the upper region as changed.
                #TODO: we are lazy and mark it as a global change
                $self->{'unassignable_entries_found_flag'} = 1;
                $logger->info(
                    "deleted a location, so we will mark a global change");
            }
            if ($assigned_location_dn) {
                $self->{'delete_notify_locations_hash'}
                  ->{$assigned_location_dn} = 1;
                $logger->debug(
                    "assigned deleted entry $deleted_dn to $assigned_location_dn"
                );
            } else {

                # this entry can be a standards object for a container for locations,
                # so get all locations for the parent and mark them as dirty
                #TODO: $superior_dn = ldap_shiftbase($dn) if ( $dn =~ m/^cn=/io ) ?
                $logger->debug("found unassignable deleted entry $deleted_dn");
                my $superior_dn = $deleted_dn;
                if ( $deleted_dn =~ m/^cn=/io ) {
                    $superior_dn = ldap_shiftbase($deleted_dn);
                }
                foreach
                  my $subordinate_dn ( $self->getAllLocationDns($superior_dn) )
                {
                    $self->{'delete_notify_locations_hash'}->{$subordinate_dn} =
                      1;
                }
            }
        }

        # clear the the list of DNs which got a delete notfication
        $self->{'delete_notify_dn_hash'} = {};
    }
    my $unassignable_count = @unassignable;
    if ( $unassignable_count > 0 ) {
        $logger->info("found $unassignable_count unassignable entries:");
        foreach my $unassigned_dn (@unassignable) {
            $logger->info("unassigned: $unassigned_dn");
        }

        # this means that reference image entries or superordinate
        # standards entries were changed. In that case
        # we set the unassignable_entries_found_flag so that
        # getAssignedLocations returns the list of all locations.
        # TODO: this is at least necessary for refImage, scRefBootOptions
        # and superordinate standards entries
        $self->{'unassignable_entries_found_flag'} = 1;
    }
    return 1;
}

=head2 isLocation

check if an entry is a location

=head3 Example

usage with a dn as parameter

  if ($ldap_cache->isLocation('dn' => 'ou=location,o=company,c=de')) {
    #ok dn is a location
  }

another example with an entry as parameter

  my $entrry = $ldap_cache->getEntry($location_dn);
  if ($ldap_cache->isLocation('entry' => $entry)) {
    # ok entry is a location
  }

=over 4

=item Parameter

hash consisting of

C<dn> string representation of dn which should be checked (optional, will be used if no entry is given)

C<entry> Net::LDAP::Entry object (optional, if not given C<dn> has to be passed)

C<with_network> boolean. Decides if location should additionally be checked for a network-object

=item Return

boolean true if dn, or entry is a location (possibly with network-object)

=back

=cut

sub isLocation
{
    my $self   = shift;
    my %params = (
        'dn'           => undef,
        'entry'        => undef,
        'with_network' => undef,
        @_
    );
    my $logger = $LOGGER;
    my $entry  = $params{'entry'};
    if ( !$entry ) {
        my $dn = $params{'dn'};
        if ($dn) {
            $entry = $self->getEntry($dn);
        } else {
            $logger->warn(
                "isLocation has to be called with either entry or dn");
            return;
        }
        if ( !$entry ) {
            $logger->info("Could not find entry for location dn: $dn");
            return;
        }
    }
    if ( $params{'with_network'} ) {

        #test, if this location has valid network information
        if ( !$self->getLocationIpObject( $entry->dn ) ) {
            return;
        }
    }
    return has_objectclass( $entry, 'scLocation' );
}

=head2 setLocationForDn

sets a location_dn for a given entry_dn. This has to be done, so that
the location for a dn can be looked up easily.

=head3 Example

   if ( $ldap_cache->setLocationForDn('cn=pc,ou=gs,o=company,c=de', 'ou=gs,o=company,c=de') ) {
     print "ok, set location";
   };

=over 4

=item Parameter

=over 4

=item entry_dn

dn of the entry

=item location_dn

dn of the location the entry is in

=back

=item Return

Boolean: true on success

=back

=cut

sub setLocationForDn
{
    my $self = shift;
    my ( $entry_dn, $location_dn ) = @_;
    $entry_dn = $self->normalize($entry_dn);
    if ( $self->{'entrycache'}->{$entry_dn} ) {
        $self->{'entrycache'}->{$entry_dn}->{'location'} = $location_dn;
        return 1;
    }
    return;
}

=head2 getLocationForDn

gets the location dn the entry is located in

=head3 Example

  my $location_dn = $ldap_cache->getLocationForDn('cn=pc,ou=gs,o=company,c=de');

=over 4

=item Parameter

C<dn> of the entry for which the location should be looked up

=item Return

string dn

=back

=cut

sub getLocationForDn
{
    my $self = shift;
    my $dn   = shift;
    if ( defined $dn and $dn and $self->hasEntry($dn) ) {
        $dn = $self->normalize($dn);
        if ( $self->{'entrycache'}->{$dn}->{'location'} ) {

            #ok, the entry has already been assigned to a location
            return $self->{'entrycache'}->{$dn}->{'location'};
        } else {

            #no location was assigned yet, so find it and bind it
            my $location_dn = $self->findLocation($dn);
            if ($location_dn) {

                #found a location
                $self->{'entrycache'}->{$dn}->{'location'} = $location_dn;
                return $location_dn;
            }
        }

    }

    #no location found
    return;
}

=head2 setEntry

add a Net::LDAP::Entry object to the cache

=over 4

=item Parameter

    - Net::LDAP::Entry object

=item Return

    NONE

=back

=cut

sub setEntry
{
    my $self  = shift;
    my $entry = shift;

    my $logger = $LOGGER;

    my $dn = $entry->dn;
    $dn = $self->normalize($dn);

    # remove old values from index
    #TODO: if attr_value in attr_name is an empty hash after delete, it should be deleted aswell
    #      this should be changed in removeEntry and setInternalEntryHash too
    if ( $self->hasEntry($dn) ) {
        my $old_entry = $self->getEntry($dn);
        foreach my $attr_name ( $self->listIndices() ) {
            my @attr_values = $old_entry->get_value($attr_name);
            foreach my $attr_value (@attr_values) {
                delete $self->{'index_for_attr'}->{ lc($attr_name) }
                  ->{$attr_value}->{$dn};
            }
        }
    } else {

        #this entry seems to be a new entry
        if ( $self->isLocation( 'entry' => $entry ) ) {

            #and it is a location, so mark it as global
            $self->{'unassignable_entries_found_flag'} = 1;
            $self->{'added_location'}->{$dn} = 1;
        }
    }
    my %entry_hash  = entryToHash($entry);
    my $entry_cache = $self->{'entrycache'};
    $entry_cache->{$dn}->{'entry'}     = \%entry_hash;
    $entry_cache->{$dn}->{'ldapentry'} = $entry;

    my $message = $self->checkEntryAttributes(%entry_hash);
    if ($message) {
        $logger->warn("[LDAP-DATA] attributes check for $dn: $message");
    }

    # delete ip_object if it is there
    if ( exists $entry_cache->{$dn}->{'ip_object'} ) {
        delete $entry_cache->{$dn}->{'ip_object'};
    }

    # update the hierarchical cache too
    my $local_base_dn = ldap_shiftbase($dn);
    $self->{'hierarchical_cache'}->{$local_base_dn}->{'entries'}->{$dn} = 1;
    $self->{'hierarchical_cache_dirty'}->{$local_base_dn} = 1;

    # keep index up to date
    foreach my $attr_name ( $self->listIndices() ) {
        my @attr_values = $entry->get_value($attr_name);
        foreach my $attr_value (@attr_values) {
            $self->{'index_for_attr'}->{ lc($attr_name) }->{$attr_value}
              ->{$dn} = 1;
        }
    }

    return 1;
}

=head2 setEntries

Sets all entries in one mass operation

=head3 Example

  $ldap_cache->setEntries({a => $entryA, b => $entryB, c => $entryC });

=over 4

=item Parameter

reference to hash with the entries as values

=item Return

nothing

=back

=cut

sub setEntries
{
    my ( $self, $href_entries ) = @_;
    for my $entry ( values %{$href_entries} ) {
        $self->setEntry($entry);
    }
}

=head2 getEntry

Returns the Net::LDAP::Entry object from the cache for a given DN

=head3 Example

 my $entry = $ldap_cache->getEntry('ou=test,o=smartclient,c=de');

=over 4

=item Parameter

DN of the cache entry object

=item Return

Net::LDAP::Entry object or undef if the given DN does not exist
in the cache

=back

=cut

sub getEntry
{
    my $self = shift;
    my $dn   = shift;
    if ( $dn and $self->{'entrycache'} ) {
        $dn = $self->normalize($dn);
        if ( $self->{'entrycache'}->{$dn} ) {
            return $self->{'entrycache'}->{$dn}->{'ldapentry'}
              if defined $self->{'entrycache'}->{$dn}->{'ldapentry'};
            my $entry_hash = $self->{'entrycache'}->{$dn}->{'entry'};
            my $entry      = hashToEntry( %{$entry_hash} );
            $self->{'entrycache'}->{$dn}->{'ldapentry'} = $entry;
            return $entry;
        }
    }
    return;
}

=head2 hasEntry

checks if entry given by dn is in entrycache

=over 4

=item Parameter

    - dn string representing dn of entry

=item Return

    - boolean true if entry exists in entrycache

=back

=cut

sub hasEntry
{
    my $self = shift;
    my $dn   = shift;

    if ( $dn and $self->{'entrycache'} ) {
        $dn = $self->normalize($dn);
        return defined $self->{'entrycache'}->{$dn};
    }
    return;
}

=head2 getInternalEntryHash

will return full entry hash from entrycache. This can be used to change information
which is stored with entry in entrycache.

USE THIS FUNCTION WITH CARE.

=head3 Example

    my %full_entry_hash = $ldap_cache->getInternalEntryHash($dn);
    $full_entry_hash{'location'} = ldap_shift_base($dn);
    $ldap_cache->setInternalEntryHash(%full_entry_hash);

=over 4

=item Parameter

dn of the entry to get (mandatory)

=item Return

hash containing ldap entry as hash and other internal stuff

=back

=cut

sub getInternalEntryHash
{
    my $self = shift;
    my $dn   = shift;

    if ( $dn and $self->{'entrycache'} ) {
        $dn = $self->normalize($dn);
        return $self->{'entrycache'}->{$dn};
    }
    return;
}

=head2 setInternalEntryHash

stores a hash into entrycache. 

USE THIS FUNCTION WITH UTMOST CARE.

=over 4

=item Parameter

reference to hash representing all needed stuff for the entrycache. If you don't know
what stuff is needed for the entrycache you MUSTN'T USE THIS METHOD

=item Return

boolean: true if entry could be added successfully

=back

=cut

sub setInternalEntryHash
{
    my $self                 = shift;
    my $href_full_entry_hash = shift;
    if ( $href_full_entry_hash->{'DN'} ) {
        my $dn = $self->normalize( $href_full_entry_hash->{'DN'} );
        my %attr_name_map =
          map { lc($_) => $_ } keys %{$href_full_entry_hash};

        delete $self->{'entrycache'}->{$dn}->{'ldapentry'};

        # remove old values from indices for cache
        if ( exists $self->{'entrycache'}->{$dn}->{'entry'} ) {
            my $href_old_entry = $self->{'entrycache'}->{$dn}->{'entry'};
            foreach my $attr_name ( $self->listIndices() ) {
                next
                  if ( !exists $attr_name_map{$attr_name} );
                my $attr_values =
                  $href_old_entry->{ $attr_name_map{$attr_name} };
                next
                  if ( !$attr_values );
                foreach my $attr_value ( @{$attr_values} ) {
                    if ($attr_value) {
                        delete $self->{'index_for_attr'}->{ lc($attr_name) }
                          ->{$attr_value}->{$dn};
                    }
                }
            }
        }
        $self->{'entrycache'}->{$dn}->{'entry'} = $href_full_entry_hash;

        #delete ip_object if it is there
        if ( exists $self->{'entrycache'}->{$dn}->{'ip_object'} ) {
            delete $self->{'entrycache'}->{$dn}->{'ip_object'};
        }

        # prepare entries for hierarchical cache
        my $local_base_dn = ldap_shiftbase($dn);
        if ( !defined $self->{'hierarchical_cache'}->{$local_base_dn} ) {
            $self->{'hierarchical_cache'}->{$local_base_dn} = {};
            $self->{'hierarchical_cache'}->{$local_base_dn}->{'entries'} = {};
            $self->{'hierarchical_cache'}->{$local_base_dn}->{'objectclasses'} =
              {};
        }

        # update indices for cache
        foreach my $attr_name ( $self->listIndices() ) {
            next if ( !exists $attr_name_map{$attr_name} );
            my $attr_values =
              $href_full_entry_hash->{ $attr_name_map{$attr_name} };
            next if ( !$attr_values );
            foreach my $attr_value ( @{$attr_values} ) {
                if ($attr_value) {
                    $self->{'index_for_attr'}->{ lc($attr_name) }->{$attr_value}
                      ->{$dn} = 1;
                }
            }
        }

        # mark entry cache dirty for objectClass index usage
        $self->{'hierarchical_cache'}->{$local_base_dn}->{'entries'}->{$dn} = 1;
        $self->{'hierarchical_cache_dirty'}->{$local_base_dn} = 1;
        return 1;
    }
    return;
}

=head2 getBase

returns the base for the entrycache

=over 4

=item Parameter

	- none
	
=item Return

	string base of entrycache

=back

=cut

sub getBase
{
    my $self = shift;
    return $self->{'entrycache_base'};
}

=head2 getFilter

returns the filter which was used to generate the entrycache

=over 4

=item Parameter

	- none
	
=item Return

	string filter of entrycache

=back

=cut

sub getFilter
{
    my $self = shift;
    return $self->{'entrycache_filter'};
}

=head2 removeEntry

reomves an entry from cache

=head3 Example

    if ($ldap_cache->removeEntry('pc=00123,ou=test,o=smartclient,c=de')) {
        print "dn removed";
    };

=over 4

=item Parameter

DN of the cache entry object

=item Return

    boolean:
        true,   if entry could be deleted from cache,
        false,  otherwise

=back

=cut

sub removeEntry
{
    my $self   = shift;
    my $dn     = shift;
    my $logger = $LOGGER;
    $dn = $self->normalize($dn);

    if ( $self->hasEntry($dn) ) {

        if ( $logger->is_debug() ) {
            $logger->debug("remove entry $dn from LdapCache");
        }

        # delete values from cache indices
        my $old_entry = $self->getEntry($dn);
        foreach my $attr_name ( $self->listIndices() ) {
            my @attr_values = $old_entry->get_value($attr_name);
            foreach my $attr_value (@attr_values) {
                delete $self->{'index_for_attr'}->{ lc($attr_name) }
                  ->{$attr_value}->{$dn};
            }
        }

        #mark location as changed
        my $location_dn = $self->getLocationForDn($dn);
        if ($location_dn) {
            $logger->info("mark location $location_dn as changed");
            $self->{'changed_locations_hash'}->{$location_dn} = 1;
        }

        #if it is a location, mark a global change
        if ( $self->isLocation( 'dn' => $dn ) ) {
            $self->{'unassignable_entries_found_flag'} = 1;
            $self->{'removed_location'}->{$dn} = 1;
        }

        #remove the entry
        delete( $self->{'entrycache'}->{$dn} );

        # modify hierarchical cache and mark it as dirty
        my $local_base_dn = ldap_shiftbase($dn);
        if ( $self->{'hierarchical_cache'}->{$local_base_dn}->{'entries'}
            ->{$dn} )
        {
            delete $self->{'hierarchical_cache'}->{$local_base_dn}->{'entries'}
              ->{$dn};
            $self->{'hierarchical_cache_dirty'}->{$local_base_dn} = 1;
        }
        return 1;
    }
    return;
}

=head2 getAssignedLocations

For a given list of DNs this method returns the List of Locations
to which the given DNs were assigned by assignLocations.

=head3 Example

 my @locations = $ldap_cache->getAssignedLocations(
     'cn=pc00123,ou=client,ou=test1,o=smartclient,c=de',
	 'cn=pc00124,ou=client,ou=test1,o=smartclient,c=de',
	 'cn=pc00035,ou=client,ou=test2,o=smartclient,c=de',
	 'cn=pc00038,ou=client,ou=test2,o=smartclient,c=de'
 );
 
would return the following list:

 ('ou=test1,o=smartclient,c=de', 'ou=test2,o=smartclient,c=de')

=over 4

=item Parameter

List of DNs in the LDAP cache

=item Return

List of Locations to which the given DNs were assigned by
assignLocations. If only unassignable DNs are given, an
empty List will be returned.

=back

=cut

sub getAssignedLocations
{
    my $self    = shift;
    my @dn_list = @_;

    my %locations_hash;
    my @locations_list = ();

    my $logger = $LOGGER;

    # if unassignable_entries_found_flag was set by assignLocations
    # this means that reference image entries or superordinate
    # standards entries were changed. In that case
    # reset the flag and return the list of all locations
    if ( $self->{'unassignable_entries_found_flag'} ) {
        $logger->debug("unassignable flag set, returning all locations");
        $logger->debug( "value of unassignable flag is ",
            $self->{'unassignable_entries_found_flag'} );
        $self->{'unassignable_entries_found_flag'} = 0;
        $self->{'delete_notify_locations_hash'}    = {};
        $self->{'changed_locations_hash'}          = {};
        return $self->getAllLocationDns();
    }

    if (@dn_list) {
        foreach my $dn (@dn_list) {
            if ( $dn =~ m/ou=global/io ) {

                #global change detected.
                @locations_list = $self->getAllLocationDns();
                last;
            }
            $logger->debug($dn);
            my $location_dn = $self->getLocationForDn($dn);
            if ($location_dn) {
                $logger->debug("belongs to $location_dn");
                $locations_hash{$location_dn} = 1;
            } else {

                # maybe this is a region, or standards object for a region or other higher order container
                # so get all locations which could be affected by this change
                #TODO: $superior_dn = ldap_shiftbase($dn) if ( $dn =~ m/^cn=/io ) ?
                my $superior_dn = $dn;
                if ( $dn =~ m/^cn=/io ) {
                    $superior_dn = ldap_shiftbase($dn);
                }
                foreach
                  my $subordinate_dn ( $self->getAllLocationDns($superior_dn) )
                {
                    $locations_hash{$subordinate_dn} = 1;
                }
            }
        }
        if ( !@locations_list ) {
            @locations_list = keys %locations_hash;
        }
    }

    # add locations affected by DNs which got a delete notification
    foreach
      my $location_dn ( keys %{ $self->{'delete_notify_locations_hash'} } )
    {
        $logger->debug("adding $location_dn affected by delete notification");
        push( @locations_list, $location_dn );
    }
    $self->{'delete_notify_locations_hash'} = {};

    foreach my $location_dn ( keys %{ $self->{'changed_locations_hash'} } ) {
        $logger->debug("adding $location_dn which was marked as changed");
        push( @locations_list, $location_dn );
    }
    $self->{'changed_locations_hash'} = {};

    return @locations_list;
}

=head2 entryToHash

Converts the content of an Net::LDAP::Entry object into a hash.
To convert back into an Net::LDAP::Entry object, use hashToEntry
which must be called with a reference to a hash returned by
entryToHash.

To preserve the DN of the LDAP entry, it is stored in the hash
element with the key 'DN'. The other keys of the hash are the
attribute names, the values are references to arrays holding
the values.

=head3 Example

 my %ldap_data = entryToHash($ldap_entry);

Example for converting a hash returned by entryToHash into
a Net::LDAP::Entry object:

 my $entry = hashToEntry(%ldap_data);

=over 4

=item Parameter

Net::LDAP::Entry object

=item Return

hash which consists of the attributes used in the entry as key
and the values as arrays. The DN of the LDAP entry is stored
in the hash key 'DN'.

=back

=cut

sub entryToHash
{
    my $entry = shift;
    my %ldap_data;
    $ldap_data{'DN'} = $entry->dn();
    foreach my $attr ( $entry->attributes ) {

        # make sure to allocate mem for the values
        # by using a new array variable to avoid
        # problems with shared mem usage
        my @values = $entry->get_value($attr);
        $ldap_data{$attr} = \@values;
    }
    return %ldap_data;
}

=head2 hashToEntry

Converts a hash returned by entryToHash into a Net::LDAP::Entry
object. See entryToHash for more details.

=head3 Example

 my $entry = hashToEntry(%ldap_data);

=over 4

=item Parameter

hash as returned by entryToHash

=item Return

Net::LDAP::Entry object

=back

=cut

sub hashToEntry
{
    my %ldap_data = @_;
    my $entry     = Net::LDAP::Entry->new();
    $entry->dn( $ldap_data{'DN'} );
    delete $ldap_data{'DN'};
    $entry->add(%ldap_data);
    return $entry;
}

=head2 treeSearch

searches for a "standards" attribute for a given dn in the entrycache.
Search will start at C<start_dn> and walk up the tree like a frog.

Pseudocode to walk like a frog:

 test dn
 loop until attr found:
    move dn one level up
    test dn
    test standards_name_for_dn 
 
If a C<devicetype_dn> is given or extracted via C<smartdevicetype>, C<standard_dc> will be disabled.

=head3 Example

  my ($entry) = $ldap_cache->treeSearch(
    'start_dn' => 'pc00257,ou=gs,o=debeka,c=de',
    'attr' => 'dbkRefPcDn'
  );

=over 4

=item Parameter

Hash consisting of:

  start_dn:        string where to start the search (mandatory)
  root_dn:         base of the tree (default $self->{'entrycache_base'})
  standard_dc:     prefix for standards (default $self->{'standard_dc'})
  devicetype_dn:   devicetype for which standards should be searched 
                   (default: undef)
  smartdevicetype: boolean if devicetype_dn should be guessed from entry
                   specified by start_dn (default: false)
  attr:            attribute which has to be set in entry
  
=item Return

array with one element, which is the entry

=back

=cut

sub treeSearch
{
    my $self  = shift;
    my %param = (
        'start_dn'        => undef,
        'root_dn'         => $self->{'entrycache_base'},
        'standard_dc'     => $self->{'standard_dc'},
        'devicetype_dn'   => undef,
        'smartdevicetype' => undef,
        'attr'            => undef,
        @_,
    );

    my $start_dn        = $param{'start_dn'};
    my $root_dn         = $param{'root_dn'};
    my $standard_dc     = $param{'standard_dc'};
    my $devicetype_dn   = $param{'devicetype_dn'};
    my $smartdevicetype = $param{'smartdevicetype'};
    my $attr            = $param{'attr'};

    my $entry_cache = $self->{'entrycache'};

    my $logger = $LOGGER;

    my $entry;
    my $dn = $start_dn;
    $dn = $self->normalize($dn);

    my $treesearch_cache_key;

    # if no devicetype was given and smartdevicetype is enabled, try to get it from the dn
    if ( !$devicetype_dn and $smartdevicetype ) {
        if ( $self->hasEntry($dn) ) {
            my $entry_for_devicetype = $self->getEntry($dn);
            if ( $entry_for_devicetype->exists('dbkDeviceTypeDn') ) {
                $devicetype_dn =
                  $entry_for_devicetype->get_value('dbkDeviceTypeDn');
                $logger->debug(
                    "undef standards prefix since dbkDeviceTypeDn was found by smartsearch"
                );
                $standard_dc = undef;
            }
        }
    } elsif ($devicetype_dn) {
        $logger->debug(
            "undef standards prefix since dbkDeviceTypeDn was given explicitly"
        );
        $standard_dc = undef;
    }

    if ( $logger->is_debug() ) {
        if ($standard_dc) {
            $logger->debug( "Using standards prefix "
                  . $standard_dc
                  . " objects for treeSearch" );
        } elsif ($devicetype_dn) {
            $logger->debug( "Using standards objects by devicetype_dn "
                  . $devicetype_dn
                  . " for treeSearch" );
        } else {
            $logger->debug("Ignoring standards objects for treeSearch");
        }
    }

    while ( length($dn) >= length($root_dn) ) {
        $logger->debug( "Test dn: " . $dn . " for attribute" );
        if ( $self->hasEntry($dn) ) {
            $entry = $self->getEntry($dn);
            if ( $entry->exists($attr) ) {
                last;
            } else {
                $entry = undef;
            }
        }

        if ( $dn !~ /^cn=/io and exists $self->{'treesearch_cache'} ) {
            $treesearch_cache_key =
                ( $dn              || '' )
              . ( $standard_dc     || '' )
              . ( $devicetype_dn   || '' )
              . ( $smartdevicetype || '' )
              . ( $attr            || '' );
            if ( exists $self->{'treesearch_cache'}->{$treesearch_cache_key} ) {
                return $self->getEntry(
                    $self->{'treesearch_cache'}->{$treesearch_cache_key} );
            }
        }
        if ( !$standard_dc and !$devicetype_dn ) {

            # removing first domain component
            $dn = ldap_shiftbase($dn);
        } elsif (
            $self->isStandardsObject(
                'dn'            => $dn,
                'devicetype_dn' => $devicetype_dn,
                'standard_dc'   => $standard_dc
            )
          )
        {

            # Seems this was the standard object, so strip first component and domain
            $dn = ldap_shiftbase( ldap_shiftbase($dn) );
        } else {
            my $new_dn = $self->constructStandardsName(
                'dn'            => $dn,
                'devicetype_dn' => $devicetype_dn,
                'standard_dc'   => $standard_dc
            );
            if ( !$new_dn ) {
                $dn = ldap_shiftbase($dn);
            } else {
                $dn = $new_dn;
            }
        }
    }

    if ( $treesearch_cache_key and ref $entry ) {
        $self->{'treesearch_cache'}->{$treesearch_cache_key} = $entry->dn;
    }

    # returns undef if nothing was found
    return $entry;
}

=head2 search

searches entrycache for objects with attribute equals a specified value

=head3 Example

  my @entries = $ldap_cache->search(
      'base'  => 'ou=testgs,o=company,c=de', 
      'scope' => 'sub', 
      'attr'  => 'macAddress', 
      'value' => '00:11:22:33:44:55'
  );

=over 4

=item Parameter

hash consisting off:

C<base> base of ldap tree where the search should start (default: entrycache_base)

C<scope> one of C<base>|C<one>|C<sub> (default: C<sub>)

C<attr> name of ldap attribute which should be compared (default: undef)

C<value> value of ldap attribute (default: undef)

C<casesensitive> boolean if compare of ldap attribute value should be made 
casesensitive (uses compare function) (default: 0)

=item Return

array consisting of all entry objects matching the search criteria

=back

=cut

sub search
{
    my $self = shift;

    my $logger = $LOGGER;

    my $attr_case   = $self->{'entrycache_lowercase'};
    my $entry_cache = $self->{'entrycache'};

    my %param = (

        'base'  => $self->{'entrycache_base'},
        'scope' => 'sub',                        # 'base' | 'one' | 'sub'
        ##'filter'      => undef,
        'attr'            => undef,
        'value'           => undef,
        'casesensitive'   => 0,
        'recursive_level' => 0,
        @_,
    );

    my $value_case = $param{'casesensitive'};
    my $attr       = convertCase( $param{'attr'}, $attr_case );
    my $value      = convertCase( $param{'value'}, $value_case );
    my $base       = $self->normalize( $param{'base'} );
    my $scope      = lc( $param{'scope'} );

    if (    $scope eq 'sub'
        and $base eq $self->normalize( $self->{'entrycache_base'} ) )
    {
        foreach my $attr_in_index ( $self->listIndices() ) {
            if ( lc($attr_in_index) eq lc($attr) ) {

                #this search will be done by index search
                return $self->searchByIndex(
                    'attr'  => $attr,
                    'value' => $value
                );
            }
        }
        $logger->info(
            "Index would be used for search if index for attr was enabled, which is not the case: $base scope $scope for $attr = $value"
        );
    }
    my @result = ();

    if ( defined $param{'recursive_level'}
        and $param{'recursive_level'} > 10000 )
    {
        my $logger = $LOGGER;
        $logger->warn(
            "reached max recursive level in search (current search base in recursion is $base searching for $attr = $value)"
        );
        return @result;
    }

    #for the scopes one and sub we have to look at our children
    if ( $scope eq 'one' or $scope eq 'sub' ) {
        foreach my $dn (
            keys %{ $self->{'hierarchical_cache'}->{$base}->{'entries'} } )
        {

            my $local_scope = $scope;

            #scope one search is a base scope for all its children
            if ( $local_scope eq 'one' ) {
                $local_scope = 'base';
            }

            # now do the recursive search
            push @result,
              $self->search(
                'base'            => $dn,
                'scope'           => $local_scope,
                'attr'            => $attr,
                'value'           => $value,
                'casesensitive'   => $value_case,
                'recursive_level' => ( $param{'recursive_level'} + 1 )
              );
        }
    }

    # scope one doesn't want to look in direct base
    if ( $scope eq 'base' or $scope eq 'sub' ) {
        my $entry = $self->getEntry($base);
        my $ldap_value;
        if ( ref $entry ) {
            ($ldap_value) = $entry->get_value($attr);
        }
        if ( $ldap_value or $attr eq '*' ) {
            if ( compare( $ldap_value, $value, $value_case )
                or $value eq '*' )
            {
                push( @result, $entry );
            }
        }
    }

    return @result;
}

=head2 searchByIndex

searches for values by using the index. Only attributes in $self->{'index_attribute_names'} can be searched.

=head3 Example

  my @entries = $ldap_cache->searchByIndex('attr'=>'macAddress', 'value'=> '00:30:05:aa:bb:cc');

=over 4

=item Parameter

hash consisting of:

C<attr> Name of the attribute to search after

C<value> of the above mentioned attribute

=item Return

array composed of all matching entries

=back

=cut

sub searchByIndex
{
    my $self   = shift;
    my %params = (
        'attr'  => undef,
        'value' => undef,
        @_
    );
    my $logger = $LOGGER;
    my $attr   = lc( $params{'attr'} );
    my $value  = $params{'value'};

    #check if attr is in index
    if ( !ref $self->{'index_for_attr'}->{$attr} ) {
        $logger->warn( 'no entries found in index cache for attr ' . $attr );
    }
    my @result        = ();
    my $index_entries = $self->{'index_for_attr'}->{$attr}->{$value};
    if ($index_entries) {
        foreach my $index_entry ( keys %{$index_entries} ) {
            push @result, $self->getEntry($index_entry);
        }
        if (@result) {
            return @result;
        }
    }
    return;
}

=head2 isStandardsObject

Tests if a given cache object is a standards object for the optionally given
devicetype.
If no devicetype is given, it will be tested for the prefix based standards object.

=head3 Example

	my $isStandard = $ldap_cache->isStandardsObject(
						dn => 'cn=standards,o=debeka,c=de',
						standard_dc => 'cn=standards'
						);

	my $isDTStandard = $ldap_cache->isStandardsObject( 
						dn => 'cn=knutStandard,o=debeka,c=de', 
						devicetype_dn => 'cn=deviceKnuTPC,ou=global,o=debeka,c=de'
						);

=over 4

=item Parameter

Hash consisting of:

	dn:	name of object to test (mandatory)
	devicetype_dn: name of devicetype object (default: undef)
	standard_dc: prefix for standards object (default: $self->{'standard_dc'})

=item Return

Boolean

=back

=cut

sub isStandardsObject
{
    my $self   = shift;
    my %params = (
        'dn'            => undef,
        'devicetype_dn' => undef,
        'standard_dc'   => $self->{'standard_dc'},
        @_
    );

    my $devicetype_dn = $params{'devicetype_dn'};
    my $dn            = $params{'dn'};
    my $standard_dc   = $params{'standard_dc'};
    my $logger        = $LOGGER;

    if ( !$dn ) {
        $logger->warn(
            'isStandardsObject() should be called with dn, but none given');
        return 0;
    }

    if ( !$devicetype_dn ) {

        # we were given no special devicetype, so try the standard_dc
        if ($standard_dc) {
            my ($actual_dc) = ldap_splitdn($dn);
            if ( $actual_dc eq $standard_dc ) {
                return 1;
            }
        }

        # since we have no devicetype, we have to give up here
        return 0;
    }

    my $entry = $self->getEntry($dn);
    if (
            ( ref $entry )
        and ( has_objectclass( $entry, 'dbkDeviceTypeStandards' ) )
        and (
            ldap_entry_has_value(
                ENTRY     => $entry,
                ATTR_NAME => 'dbkDeviceTypeDn',
                VALUE     => $devicetype_dn,
                LOWERCASE => 1,
            )
        )
      )
    {
        $logger->debug(
            'dn: ' . $dn . " is standards object for " . $devicetype_dn );
        return 1;
    } else {
        $logger->debug(
            'dn: ' . $dn . " is not a standards object for " . $devicetype_dn );
    }

    return 0;
}

=head2 constructStandardsName

Builds a string of the C<dn> of a standards object for a given location. 
Either a devicetype or standards prefix must be given.

If no standards object could be located for a given devicetype, the dn
will be stripped by one component and the remaining string will be given
as result.

=head3 Example

	my $standardsDn = $ldap_cache->constructStandardsName( 
						dn => 'o=debeka,c=de',
						standard_dc => 'cn=standard'
						);

	my $standardsDn = $ldap_cache->constructStandardsName( 
						dn => 'cn=knutStandard,o=debeka,c=de', 
						devicetype_dn => 'ou=koblenz,o=debeka,c=de'
						);

=over 4

=item Parameter

Hash consisting of:

 - dn:	name of location for which standards object name should be built
        (mandatory)
 - devicetype_dn: name of devicetype object (default: undef)
 - standard_dc: prefix for standards object (default: $self->{'standard_dc'})

=item Return

String dn if a standards object could be found, else undef

=back

=cut

sub constructStandardsName
{
    my $self   = shift;
    my %params = (
        'dn'            => undef,
        'devicetype_dn' => undef,
        'standard_dc'   => $self->{'standard_dc'},
        @_
    );

    my $dn            = $params{'dn'};
    my $devicetype_dn = $params{'devicetype_dn'};
    my $standard_dc   = $params{'standard_dc'};
    my $logger        = $LOGGER;

    #TODO: should test with !defined $standard_dc
    if ( !$devicetype_dn and !$standard_dc ) {
        $logger->error(
            'Can\'t construct a standards name without either devicetype_dn or standard_dc. Returning undef.'
        );

        # no standards object could be found, so return undef
        return;
    }

    if ($devicetype_dn) {
        $devicetype_dn = lc($devicetype_dn);
        if ( $logger->is_debug() ) {
            $logger->debug( "Test " 
                  . $dn
                  . " for devicetype_dn standard "
                  . $devicetype_dn );
        }

        #get all potential objects with oc dbkDeviceTypeStandards
        my @potentialStandards = $self->getObjectClassIndex(
            'base'             => $dn,
            'scope'            => 'one',
            'objectclass_name' => 'dbkDeviceTypeStandards'
        );
        my @standard_object_dns = ();
        foreach my $potentialStandardDn (@potentialStandards) {
            my $entry = $self->getEntry($potentialStandardDn);
            if ( !ref $entry ) {
                $logger->warn("no entry found for $potentialStandardDn");
                next;
            }
            my @values = $entry->get_value('dbkDeviceTypeDn');
            foreach my $v (@values) {
                $logger->debug( 'dbkDeviceTypeDn:' . $v );
                if ( lc($v) eq $devicetype_dn ) {
                    $logger->debug( "found standards object "
                          . $potentialStandardDn
                          . "by devicetype_dn "
                          . $devicetype_dn );

                    # found standard for this devicetype in location, so set it as value for next round
                    push @standard_object_dns, $potentialStandardDn;
                }
            }
        }
        if ( $#standard_object_dns > 0 ) {
            $logger->debug(
                "found more than one potential standards object, removing cn=standards"
            );
            my $std_prefix = $self->{'config'}->get('ALL_LDAP_STANDARD_DC');
            @standard_object_dns =
              grep { $_ !~ /^$std_prefix,/io } @standard_object_dns;
            if ( $#standard_object_dns > 0 ) {
                $logger->warn(
                    "[LDAP] there are more than two standards object for devicetype $devicetype_dn in $dn"
                );
            }
        }
        my $found = shift @standard_object_dns;
        if ($found) {
            $logger->debug(
                "using standards object $found for devicetype $devicetype_dn");
            return $found;
        }
        $logger->debug( 'no standards object for devicetype '
              . $devicetype_dn
              . ' found in '
              . $dn );
        return;
    } else {
        $logger->debug( 'constructing dn for standard_dc: '
              . $standard_dc . ' dn: '
              . $dn );
        if ( $standard_dc ne "" ) {
            $dn = $standard_dc . "," . $dn;
        }
    }
    return $dn;
}

=head2 getStandardsNameForLocation

Searches the standards object for a given location or workstation and
returns its C<dn>.
Either prefix based or devicetype based search will be used. Default
is prefix based searching with prefix C<'cn=standards'>.

=head3 Example

   # Prefix based searching 
   $location_dn = 'ou=gs0001,ou=test,o=company,c=de';
   $standard_dn = $ldap_cache->getStandardsNameForLocation(
           'dn' => $location_dn,
           'standard_dc' => 'cn=standards',
       );
       
   # devicetype based searching
   $standard_dn = $ldap_cache->getStandardsNameForLocation(
           'dn' => $location_dn,
           'devicetype_dn' => 'cn=deviceKnuTPC,ou=global,o=company,c=de',
       );
       
=over 4

=item Parameter

Hash consisting of:

=over 4

=item dn

name of location for which standards object should found (mandatory)

=item devicetype_dn

name of devicetype object (default: undef)

=item standard_dc

prefix for standards object (default: $self->{'standard_dc'})

=back

=item Return

string dn of standards object or undef if no standards object could be found

=back 

=cut

sub getStandardsNameForLocation
{
    my $self   = shift;
    my %params = {
        'dn'            => undef,
        'devicetype_dn' => undef,
        'standard_dc'   => $self->{'standard_dc'},
        @_
    };
    my $dn            = $params{'dn'};
    my $devicetype_dn = $params{'devicetype_dn'};
    my $standard_dc   = $params{'standard_dc'};

    my $logger = $LOGGER;

    my $standard_dn = undef;

    if ($devicetype_dn) {

        # we won't search prefix based if a devicetype_dn is given
        $standard_dc = undef;
    }

    my $entry = $self->treeSearch(
        'start_dn'      => $dn,
        'devicetype_dn' => $devicetype_dn,
        'standard_dc'   => $standard_dc,
        'attr'          => 'scRefBootOptionsDn',
    );

    if ($entry) {
        $standard_dn = $entry->dn();
        $logger->debug(
            "found standard dn: " . $standard_dn . " for dn: " . $dn );
    } else {
        $logger->warn( "no standard dn for dn: " . $dn . " found." );
    }

    return $standard_dn;
}

=head2 getLocationIpObject

get a SmartClient::IP object of a location

=over 4

=item Parameter

string: DN of a location

=item Return

C<SmartClient::IP> object ref or undef

=back

=cut

sub getLocationIpObject
{
    my $logger = $LOGGER;

    my $self        = shift;
    my $location_dn = shift;

    my $href_internal_location = $self->getInternalEntryHash($location_dn);
    if ( $href_internal_location->{'ip_object'} ) {
        return $href_internal_location->{'ip_object'};
    }

    my $location_entry = $self->getEntry($location_dn);

    if ( !$location_entry ) {

        # TODO: error message if not ignore location
        return;
    }

    my $location_ip_network = $location_entry->get_value('ipNetworkNumber');

    my $location_ip_netmask = $location_entry->get_value('ipNetmaskNumber');

    if ( !$location_ip_netmask or !$location_ip_network ) {

        #TODO: info message
        return;
    }

    $logger->debug( $location_dn . ": "
          . $location_ip_network . "/"
          . $location_ip_netmask );

    my $ip_object =
      SmartClient::IP->new( $location_ip_network, $location_ip_netmask );
    $href_internal_location->{'ip_object'} = $ip_object;
    return $ip_object;
}

=head2 getLocation( % )

Detemines the location to a value, e.g. IP address

=over 4

=item Parameter

Hash consisting of:

 - ip_address (string)
 - dn (string)

=item Return

DN of a location of undef

=back

=cut

sub getLocation
{
    my $self  = shift;
    my %param = (
        'dn'         => undef,
        'ip_address' => undef,
        @_,
    );

    my $dn         = $self->normalize( $param{'dn'} );
    my $ip_address = $param{'ip_address'};
    my $logger     = $LOGGER;

    my $location_dn = undef;

    if ($dn) {
        $location_dn = $self->getLocationForDn($dn);
    } elsif ($ip_address) {
        $location_dn = $self->getLocationForIp($ip_address);
    } else {
        $logger->warn( "can't determine location for " . Dumper(%param) );
    }

    return $location_dn;
}

=head2 getAllEntryDnIterator

get an iterator over all dn in entrycache

=head3 Example

    my $iterator = $ldap_cache->getAllEntryDnIterator();
    while (my $dn = $iterator->()) {
        print "DN: $dn";
    }

=over 4

=item Parameter

base: ldap base 

=item Return

iterator over all dn in entrycache

=back

=cut

sub getAllEntryDnIterator
{
    my $self     = shift;
    my $base     = shift;
    my @all_keys = keys %{ $self->{'entrycache'} };
    my $counter  = 0;

    return sub {
        if ( $counter > $#all_keys ) {

            #end of iterator
            return;
        }
        return $all_keys[ $counter++ ];
      }
}

=head2 getAllEntryDn

get an array containing all dn in entrycache

=over 4

=item Parameter

NONE
    
=item Return

array containing all dn in entrycache

=back

=cut

sub getAllEntryDn
{
    my $self   = shift;
    my %params = (@_);
    return keys %{ $self->{'entrycache'} };
}

=head2 notifyDelete

notify the deletion of a list of DNs
This method will scan recursively all subentries of the given DNs
and notify their delete respectively.

=over 4

=item Parameter

array: list of deleted DNs

=item Return

Current list of DNs notified for delete 

=back

=cut

sub notifyDelete
{
    my $logger = $LOGGER;

    my $self           = shift;
    my @delete_dn_list = @_;

    my $delete_notify_dn_hash = $self->{'delete_notify_dn_hash'};

    foreach my $delete_dn (@delete_dn_list) {
        foreach my $delete_entry (
            $self->search(
                'base'  => $delete_dn,
                'attr'  => '*',
                'value' => '*'
            )
          )
        {
            if ( ref $delete_entry ) {
                my $dn = $self->normalize( $delete_entry->dn );
                $delete_notify_dn_hash->{$dn} = 1;
            }
        }
    }
    return keys %{$delete_notify_dn_hash};
}

=head2 getObjectClassIndex

gets all dn's which have a specific objectclass

=head3 Example

 my @dn_list = $ldap_cache->getObjectClassIndex(
     'scope' => 'base', 
     'objectclass_name' => 'scWorkstation', 
     'base' => 'ou=testgs,o=compay,c=de'
 );

=over 4

=item Parameter

hash consisting of:

C<base> ldap base (default entrycache_ldapbase)

C<objectclass_name> name of objectclass (default: undef)

C<scope> one of C<sub>|C<one>|C<base> (default C<sub>)

C<subs_exclude> undef or name of objectclass 
                (eg. scLocation, if sub search should exclude sub locations)
                only subtree are excludes. This object in base is ignored. 

=item Return

array consisting of dn's with corresponding objectclass

=back

=cut

sub getObjectClassIndex
{
    my $self   = shift;
    my %params = (
        'base'             => $self->{'entrycache_base'},
        'objectclass_name' => undef,
        'scope'            => 'sub',
        'subs_exclude'     => undef,
        'recursive_level' => 0,    #used for internal householding
        @_
    );
    my $logger = $LOGGER;

    if (   !defined $params{'objectclass_name'}
        or !$params{'objectclass_name'} )
    {
        $logger->error('method must be called with parameter objectclass_name');
        return;
    }
    if ( !$params{'base'} or !$params{'scope'} ) {
        $logger->error(
            'method must be called with sensible parameter base and/or scope');
        return;
    }
    my $base  = $self->normalize( $params{'base'} );
    my $scope = lc( $params{'scope'} );

    # normalize object classes doesn't work,
    # so object classes must be given correctly
    #my $objectclass_name = $self->normalize( $params{'objectclass_name'} );
    my $objectclass_name = $params{'objectclass_name'};

    #my $subs_exclude     = $self->normalize( $params{'subs_exclude'} );
    my $subs_exclude    = $params{'subs_exclude'};
    my $recursive_level = $params{'recursive_level'};

    if ( $recursive_level > 1000 ) {
        $logger->warn(
            "deep recursion has happened. Force exit for current parameter: base->$base scope->$scope oc->$objectclass_name"
        );
    }

    # tell cache to build an objectclass index for this objectclass
    if ( !$self->{'hierarchical_cache_objectclass_names'}->{$objectclass_name} )
    {
        $self->{'hierarchical_cache_objectclass_names'}->{$objectclass_name} =
          LDAPCACHE_CACHE_OBJECTCLASS_NEW;
        $self->checkForNewObjectClassNames();
    }
    if ( scalar( keys %{ $self->{'hierarchical_cache_dirty'} } ) > 0 ) {

        #some entries were changed
        $self->cleanHierarchicalCache();
    }

    # now get all keys
    my @result             = ();
    my $href_local_base_dn = $self->{'hierarchical_cache'}->{$base};

    # scope base has to look one step up.
    # sub is base + recursive sub with defined $recursive_level (>0)
    if ( !$recursive_level ) {

        # first search
        if ( $scope eq 'base' or $scope eq 'sub' ) {
            my $base_of_base = ldap_shiftbase($base);
            if ( $self->{'hierarchical_cache'}->{$base_of_base}
                ->{'objectclasses'}->{$objectclass_name}->{$base} )
            {
                push @result, $base;
            }
        }
    }

    # scope sub has to search recursively
    if ( $scope eq 'sub' ) {
        foreach my $leaf_dn ( keys %{ $href_local_base_dn->{'entries'} } ) {

            # only to recursiv search,
            # if not excluded by subs_exclude
            if (
                not(   $subs_exclude
                    && $href_local_base_dn->{'objectclasses'}->{$subs_exclude}
                    && $href_local_base_dn->{'objectclasses'}->{$subs_exclude}
                    ->{$leaf_dn} )
              )
            {
                my @recursive_result = $self->getObjectClassIndex(
                    'objectclass_name' => $objectclass_name,
                    'base'             => $leaf_dn,
                    'scope'            => $scope,
                    'subs_exclude'     => $subs_exclude,
                    'recursive_level'  => $recursive_level + 1,
                );
                push @result, @recursive_result;
            }
        }
    }

    # scope one and sub are interested in the objectClasses of current node
    if ( $scope eq 'one' or $scope eq 'sub' ) {
        foreach my $key (
            keys
            %{ $href_local_base_dn->{'objectclasses'}->{$objectclass_name} } )
        {
            push @result, $key;
        }
    }

    return @result;
}

=head2 cleanHierarchicalCache

iterates over all leafs in hierarchical cache wich are marked as dirty and rebuilds their objectclass cache.
This method will be called internally if getObjectClassIndex is used.

=over 4

=item Parameter

None

=item Return

Always true

=back

=cut

sub cleanHierarchicalCache
{
    my $self = shift;

    my $logger = $LOGGER;
    $logger->debug("clean hierachical cache");
    $self->checkForNewObjectClassNames();

    #iterate over all marked leafs
    foreach my $local_base_dn ( keys %{ $self->{'hierarchical_cache_dirty'} } )
    {
        $logger->debug("clean hierachical cache for leaf $local_base_dn;");

        # we will build the local cache entirely new
        $self->{'hierarchical_cache'}->{$local_base_dn}->{'objectclasses'} = {};
        my $href_objectclass_names =
          $self->{'hierarchical_cache_objectclass_names'};

        my $href_local_base_dn_entries =
          $self->{'hierarchical_cache'}->{$local_base_dn}->{'entries'};
        my $href_local_base_dn_objectclasses =
          $self->{'hierarchical_cache'}->{$local_base_dn}->{'objectclasses'};

        # iterate over all entries which can be found in scope one
        foreach my $dn ( keys %{$href_local_base_dn_entries} ) {

            #search for all objectclasses in entries and store them in objectclasses cache
            $dn = $self->normalize($dn);
            my $entry = $self->getEntry($dn);
            if ( ref $entry ) {
                my @objectclasses =
                  $self->getEntry($dn)->get_value('objectClass');
                foreach my $objectclass (@objectclasses) {
                    next
                      if ( !$href_objectclass_names->{$objectclass} );
                    if ( !defined
                        $href_local_base_dn_objectclasses->{$objectclass} )
                    {
                        $href_local_base_dn_objectclasses->{$objectclass} = {};
                    }
                    $href_local_base_dn_objectclasses->{$objectclass}->{$dn} =
                      1;
                }
            }
        }

        #wash away the dirty mark
        delete $self->{'hierarchical_cache_dirty'}->{$local_base_dn};
    }
    return 1;
}

=head2 checkForNewObjectClassNames

checks if new objectclasses should be used in the hierarchical objectclass index. 
If new objectclass names are found, the whole hierarchical cache will be marked as dirty.
This method will be called internally and automatically by getObjectClassIndex.

=over 4

=item Parameter

None

=item Return

Always true

=back

=cut

sub checkForNewObjectClassNames
{
    my $self   = shift;
    my $logger = $LOGGER;

    my $mark_as_dirty;
    my $href_objectclass_names =
      $self->{'hierarchical_cache_objectclass_names'};
    foreach my $objectclass_name ( keys %{$href_objectclass_names} ) {
        if ( $href_objectclass_names->{$objectclass_name} ==
            LDAPCACHE_CACHE_OBJECTCLASS_NEW )
        {
            $mark_as_dirty = 1;
            $href_objectclass_names->{$objectclass_name} =
              LDAPCACHE_CACHE_OBJECTCLASS_OK;
            $logger->info(
                "Use $objectclass_name in hierarchical objectclass index");
        }
    }
    if ($mark_as_dirty) {
        foreach my $leaf_dn ( keys %{ $self->{'hierarchical_cache'} } ) {
            $self->{'hierarchical_cache_dirty'}->{$leaf_dn} = 1;
            $logger->debug("mark $leaf_dn in hierarchical cache as dirty");
        }
    }
    return 1;
}

=head2 hasEntryChanged

=over 4

=item Parameter

=item Return

=back

=cut

sub hasEntryChanged
{
    my $self      = shift;
    my $new_entry = shift;
    if ($new_entry) {

        my $dn        = $new_entry->dn();
        my $old_entry = $self->getEntry($dn);
        if ( ref $old_entry ) {
            my @new_attributes = $new_entry->attributes();
            my @old_attributes = $old_entry->attributes();

            #check if attribute names are the same
            if ( !isEqualArray( \@new_attributes, \@old_attributes ) ) {
                return 1;
            }

            #now check each attributes values
            foreach my $attr_name (@old_attributes) {
                next
                  if ( lc($attr_name) eq 'modifytimestamp' );
                my @old_attrs = $old_entry->get_value($attr_name);
                my @new_attrs = $new_entry->get_value($attr_name);
                if ( !isEqualArray( \@old_attrs, \@new_attrs ) ) {
                    return 1;
                }
            }

            #no differences could be found
            #so update modifyTimestamp to newer one
            if ( $new_entry->exists('modifyTimestamp') ) {
                my $modify_timestamp = $new_entry->get_value('modifyTimestamp');
                $self->updateModifyTimestamp(
                    'dn'               => $dn,
                    'modify_timestamp' => $modify_timestamp
                );
            }
            return;
        }
    }
    return 1;
}

=head2 updateModifyTimestamp

updates the modification timestamp of an entry in the cache without poisoning the hierarchical cache.

=head3 Example

  $ldap_cache->updateModifyTimestamp(
      'dn'=>'cn=pc01,ou=testgs,o=test,c=de',
      'modify_timestamp' => '20060701123000Z'
  );

=over 4

=item Parameter

hash consisting of:

C<dn> string representing the entry in the cache

C<modify_timestamp> string representation of timestamp in ldap format

=item Return

boolean: true if update of modification timestamp could be done

=back

=cut

sub updateModifyTimestamp
{
    my $self   = shift;
    my %params = (
        'dn'               => undef,
        'modify_timestamp' => undef,
        @_
    );

    my $dn               = $params{'dn'};
    my $modify_timestamp = $params{'modify_timestamp'};
    if ( !$dn or !$modify_timestamp ) {
        ##don't know what to change
        return;
    }
    if ( !( $modify_timestamp =~ m/^\d{14}Z$/ ) ) {
        ##doesn't look like a timestamp in ldap format
        return;
    }

    #TODO: should timestamp only be updated if it is newer than original one?
    $dn = $self->normalize($dn);
    my $entry = $self->getEntry($dn);
    if ( !$entry ) {
        ##entry seems to be not in cache
        return;
    }

    # FIXME: don't change the entry. We are modifying internal structures and should call $self->setEntry after change.
    $entry->replace( 'modifyTimestamp', $modify_timestamp );
    my %entry_hash = entryToHash($entry);
    $self->{'entrycache'}->{$dn}->{'entry'} = \%entry_hash;
    return 1;
}

=head2 createIndices

create indices for all attributes named in listIndices

=over 4

=item Parameter

none

=item Return

true

=back

=cut

sub createIndices
{
    my $self = shift;

    #clear all indices
    foreach my $index_attr ( $self->listIndices() ) {
        $self->{'index_for_attr'}->{$index_attr} = {};
    }

    #now iterate over all entries in cache
    foreach my $dn ( keys %{ $self->{'entrycache'} } ) {
        my $href_entry = $self->{'entrycache'}->{$dn}->{'entry'};
        my %attr_name_map = map { lc($_) => $_ } keys %{$href_entry};

        # update indices for cache
        foreach my $attr_name ( $self->listIndices() ) {
            my $attr_values = $href_entry->{ $attr_name_map{$attr_name} };
            next if ( !$attr_values );
            foreach my $attr_value ( @{$attr_values} ) {
                if ($attr_value) {
                    $self->{'index_for_attr'}->{ lc($attr_name) }->{$attr_value}
                      ->{$dn} = 1;
                }
            }
        }
    }
    return 1;
}

=head2 listIndices

list all attribute names which can be searched by index

=over 4

=item Parameter

none

=item Return

array containing all attributes for which an index will be kept

=back

=cut

sub listIndices
{
    my $self   = shift;
    my %params = (@_);
    if ( $self->{'index_for_attr'} ) {
        return keys %{ $self->{'index_for_attr'} };
    }
    return;
}

=head2 addIndexForAttribute

adds an index for a given attribute name

=head3

  $ldap_cache->addIndexForAttribute('ipHostnumber');

=over 4

=item Parameter

string representing attribute name

=item Return

=back

=cut

sub addIndexForAttribute
{
    my $self = shift;
    my $attr = lc(shift);

    if ( !exists $self->{'index_for_attr'}->{$attr} ) {
        $self->{'index_for_attr'}->{$attr} = {};
        $self->createIndices();
    }

    return 1;
}

=head2 addObjectClassIndex

explicitly add a objectClass to the objectclass index.

=head3 Example

  $ldap_cache->addObjectClassIndex('dbkWorkstation');

=over 4

=item Parameter

name of the objectClass

=item Return

C<true> on success, C<false> else

=back

=cut

sub addObjectClassIndex
{
    my $self = shift;
    my $name = shift;

    my $logger = $LOGGER;
    if ( !$name ) {
        $logger->warn(
            "Could not add objectClassName to index, since no name was given.");
        return;
    }
    if ( !exists $self->{'hierarchical_cache_objectclass_names'}->{$name}
        or $self->{'hierarchical_cache_objectclass_names'}->{$name} ne
        LDAPCACHE_CACHE_OBJECTCLASS_OK )
    {
        $self->{'hierarchical_cache_objectclass_names'}->{$name} =
          LDAPCACHE_CACHE_OBJECTCLASS_NEW;
        my $href_hierarchical_cache_dirty = $self->{'hierarchical_cache_dirty'};
        foreach my $leaf_dn ( keys %{ $self->{'hierarchical_cache'} } ) {
            $href_hierarchical_cache_dirty->{$leaf_dn} = 1;
            $logger->debug("mark $leaf_dn in hierarchical cache as dirty");
        }
    }
    return 1;
}

=head2 deleteIndexForAttribute

removes an index for a given attribute name

=head3 Example

  $ldap_cache->deleteIndexForAttribute('ipHostnumber');

=over 4

=item Parameter

string representing attribute name

=item Return

true

=back

=cut

sub deleteIndexForAttribute
{
    my $self = shift;
    my $attr = lc(shift);

    if ( !exists $self->{'index_for_attr'}->{$attr} ) {
        delete $self->{'index_for_attr'}->{$attr};
        $self->createIndices();
    }

    return 1;
}

=head2 sortByCommaCount

Sort by the number of commata.

  is_deeply ['a,b,c', 'b,c', 'a'], sortByCommaCount('a', 'a,b,c', 'b,c')

=over 4

=item Parameter

list with dns to sort

=item Returns

sorted list, where entries with most commata are first

=back

=cut

sub sortByCommaCount
{

    # sort the dns by the numbers of commata in a dn
    return map { $_->[1] }
      sort     { $b->[0] <=> $a->[0] }
      map { [ scalar( () = $_ =~ /,/g ) => $_ ] } @_;
}

=head2 notifyOfModification

This method notifies the cache of changes, which happened outside the scope of the cache.
For example deletions made with an administration tool. 

It expects an array reference where changes have happened. The method will scan those areas
and resync the cache to it. As the resync happens all valid entries will be cummulated in an 
array. A reference to that array will be returned, if it holds at least one entry.

=over 4

=item Parameter

reference to array of dn of modified entries

=item Return

reference to array containing dn

=back

=cut

sub notifyOfModification
{
    my $self              = shift;
    my $aref_modified_dns = shift;
    my $logger            = $LOGGER;
    my %new_dns           = ();

    foreach my $dn ( sortByCommaCount( @{$aref_modified_dns} ) ) {
        $logger->info("notified of change in $dn");

        # hash of entry dns which are to delete from the cache below this one (including itself)
        my %dns_to_delete = ();
        foreach my $current_entry (
            $self->search(
                'base'  => $dn,
                'scope' => 'sub',
                'attr'  => '*',
                'value' => '*'
            )
          )
        {
            next if ( !ref $current_entry );
            $dns_to_delete{ $current_entry->dn } = 1;
        }

        #search for new entries with dn as base and re-insert them
        my $sc_ldap = $self->getLdapConnection();
        my $result  = $sc_ldap->search(
            base   => $dn,
            scope  => 'sub',
            filter => $self->{'entrycache_filter'},
            attrs  => $self->{'entrycache_attrs'},
            ## needed?
            timelimit => $sc_ldap->{'TIMEOUT'}
        );
        if ( ldap_check_error( $result, 0 ) ) {
            $logger->error("Could not reload entries below $dn.");
            return;
        }

        foreach my $entry ( $result->entries ) {
            next if ( !ref $entry );
            $new_dns{ $entry->dn } = $entry;
            delete $dns_to_delete{ $entry->dn };
        }
        $self->setEntries( \%new_dns );
        foreach my $old_dn ( keys %dns_to_delete ) {
            $self->removeEntry($old_dn);
        }
    }

    # update changed_locations_hash
    foreach my $dn ( keys %new_dns ) {
        if ( $self->isLocation( 'dn' => $dn ) ) {
            $logger->info("mark location $dn as changed");
            $self->{'changed_locations_hash'}->{ lc($dn) } = 1;
        }
    }

    if ( keys %new_dns ) {
        my @new_dns = keys %new_dns;
        return \@new_dns;
    }

    #no entries were changed
    return;

}

sub getLocationForIp
{
    my $self        = shift;
    my $ip_address  = shift;
    my $logger      = $LOGGER;
    my $location_dn = undef;

    $logger->debug("searching location for IP address $ip_address");

    # go through all location and check if IP address
    # lies inside the location network
    my @sclocation_list =
      $self->getObjectClassIndex( 'objectclass_name' => 'scLocation' );
    foreach my $dn (@sclocation_list) {
        my $location_ip_object = $self->getLocationIpObject($dn);
        if (   $location_ip_object
            && $location_ip_object->contains($ip_address) )
        {
            $location_dn = $dn;
            last;
        }
    }
    return $location_dn;
}

=head2 getLdapConnection

creates and returns a new ldap connection based on the config of this instance

=head3 Example

  my $ldap_con = $ldap_cache->getLdapConnection('needcap' => 'rw');

=over 4

=item Parameter

hash consisting of

=over 4

=item needcap

one of C<ri>, C<ro> or C<rw>

=back

=item Return

=back

=cut

sub getLdapConnection
{
    my $self   = shift;
    my %params = (
        'needcap' => 'ri',
        @_
    );
    my $sc_ldap;
    my $needcap = $params{'needcap'};
    my $counter = 0;
    my $logger  = $LOGGER;
    while ( !$sc_ldap ) {
        $sc_ldap = SmartClient::LDAP->new(
            'CONFIG' => $self->{'config'},
            'LDAPBINDDN' =>
              $self->{'config'}->get('LDAP2DNSDHCPD_LDAP_ADMIN_DN'),
            'LDAPBINDPW' =>
              $self->{'config'}->get('LDAP2DNSDHCPD_LDAP_ADMIN_PW'),
            'NEEDCAP' => $needcap
        );
        if ( !$sc_ldap ) {
            sleep( $self->{'config'}->get( '', 1 ) || 10 );
            if ( $counter++ % 5 ) {
                $logger->warn(
                    "[LDAP] no ldap connection could be established after $counter tries"
                );
            }
        }

    }

    return $sc_ldap;
}

=head2 checkEntryAttributes

check each attribute in a given hash against all attribute checks from
C<%checkSyntax>. This hash has the format:

 attribute_name => [ regex, fref_check ]

for example for a macAddress check:

 'macAddress' => [ qr/^([a-f0-9]{2}:){5}[a-f0-9]{2}$/io, undef ]
 
for ipHostNumber this test could be:

 'ipHostNumber' => [
   qr/^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/o,
   sub {
       my $ip = shift;
       my @numbers = split( /\./, $ip );
       grep { $_ > 255 or $_ < 0 } @numbers;
   }
 ]

The anonymous C<fref_check> will be given the scalar attribute value as first parameter.

=head3 Example

 my %entry_hash = $ldap_cache->entryToHash($entry);
 my $message = $ldap_cache->checkEntryAttributes(%entry_hash);
 if ($message) {
   # something rotten
 }

=over 4

=item Parameter

hash of entry as returned by C<SmartClient::LdapCache::entryToHash>

=item Return

string containing error message with all attributes which didn't pass the configured tests

=back

=cut

sub checkEntryAttributes
{
    my $self       = shift;
    my %entry_hash = (@_);

    my $logger = $LOGGER;
    my $dn     = $entry_hash{'dn'};
    my @message;
    my %checkSyntax = (
        'macAddress'   => [ qr/^([a-f0-9]{2}:){5}[a-f0-9]{2}$/io, undef ],
        'ipHostNumber' => [
            qr/^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/o,
            sub {
                my $ip = shift;
                my @numbers = split( /\./, $ip );
                grep { $_ > 255 or $_ < 0 } @numbers;
              }
        ],
        'dbkDhcpRange'      => [ qr{^([0-9.]+(?:\s*[,]\s*[0-9.]+)?)$}o, undef ],
        'dbkDhcpFixedRange' => [ qr{^([0-9.]+(?:\s*[,]\s*[0-9.]+)?)$}o, undef ],
    );

    # check all attributes for which checks are available
    while ( my ( $attribute, $checks ) = each %checkSyntax ) {
        my ( $regex, $fref_check ) = @{$checks};
        if ( exists $entry_hash{$attribute} ) {
            my @values = @{ $entry_hash{$attribute} };
            foreach my $value (@values) {
                if ( $value !~ $regex ) {
                    push @message, "wrong $attribute: \'$value\'";
                } elsif ( $fref_check and $fref_check->($value) ) {
                    push @message, "wrong $attribute: \'$value\'";
                }
            }
        }
    }

    return join( ", ", @message );
}

1;
