#!/usr/bin/perl -w
use strict;

package SmartClient::LdapUtil;

=head1 NAME

SmartClient::LdapUtil

=head2 SVN Version

	$Id: LdapUtil.pm 8038 2012-07-24 11:38:32Z FSchumacher $

=cut

# exportieren der Module
use Exporter;
our @ISA = qw(Exporter);

require Net::LDAP;
use Net::LDAP qw(LDAP_NO_SUCH_OBJECT);
use SmartClient::Util;
use SmartClient::CheckUtil;
use Net::Netmask;
use Net::LDAP::Filter;
use Time::gmtime;
use Data::Dumper;
use Log::Log4perl;

# Festlegen der zu exportierenden Module, Methoden, Funktionen, Variablen, Konstanten
our @EXPORT =
  qw(ldap_treesearch ldap_splitdn ldap_shiftbase ldap_cut_base ldap_normalize ldap_check_error ldap_entry_has_value has_objectclass ldap_get_dn_entry ldap_get_ref_entry ldap_get_cn_from_dn ldap_dn2dnsname ldap_cn2dnsname get_zonename ldap_getLocationNetmask ldap_get_location_entry ldap_cn2addr ldap_gen_entrycache ldap_get_refpc_entry_cached ldap_gen_entrycache_successive ldap_set_values ldap_get_next_available_number ldap_get_next_available_workstation_ip ldap_get_next_available_printer_ip ldap_get_next_available_ip ldap_get_next_available_workstation_dn ldap_get_next_available_printer_dn ldap_get_next_available_dn get_next_available_ip ldap_ip2offset ldap_offset2ip ldap_entry2ip get_location_baselist getContainer getContainerIPs getLocationIPs getWorkstationDefaultAttributes getPrinterDefaultAttributes ldap_chk_tstamp ldap_get_min_tstamp ldap_get_max_tstamp calcTimestampWithOffset ldap_get_attribute_syntax ldap_attribute_is_dn ldap_filter_add_timestamp ldap_construct_standards_name ldap_is_standards_object ldap_getListOfDNsThatMatchesFilter isInSubtree);

=head1 exportierte Module, Methoden, Variablen und Funktionen

=over 4

=item Funktionen

=over 4

=item ldap_treesearch

Suche nach bestimmten LDAP-Objekten von einen bestimmten Knoten bis zur Wurzel

=item ldap_splitdn

Trennen eines DN in seine erste Komponente und den Rest.

=item ldap_check_error

Überprfen des Ergebnisses einer LDAP-Anfrage auf aufgetretene Fehler,
ggf. per Syslog und auf STDERR Fehlermeldung ausgeben.

=item ldap_entry_has_value

Überprfen eines LDAP-Eintrags vom Typ Net::LDAP::Entry auf vorkommen
eines bestimmten Attributwertes.

=item ldap_get_dn_entry

Liefert den LDAP-Eintrag zu einem gegebenen DN

=item ldap_get_cn_from_dn

Liefert die cn-Komponente aus einem DN

=item ldap_dn2dnsname

Liefert einen DNS-Namen zu einem DN, Beispiel:

 cn=pc00263,ou=gs201004,ou=test,o=organisation,c=de
 => pc00263,gs201004.test.organisation.de

=back

=back

=cut

=head1 Funktionen

=head2 Funktion ldap_treesearch

LDAP-Suche, die bei einem spezifizierten Knoten beginnt und sukzessive
bis zur Wurzel nach einem bestimmten Objekt sucht bis eines gefunden wird.
Es kann entweder per LDAP und FILTER gesucht werden
oder falls vorhanden innerhalb eines vorher mit 
ldap_gen_entrycache generierten Hash aus Net::LDAP::Entry Objekten.
Dieses kann für wiederholte und haeufige Aufrufen die Last auf dem LDAP Server erheblich reduzieren.
Durch die derzeitige Beschraenkung von ldap_gen_entrycache auf Objektklassen kann ldap_treesearch mit ENTRYCACHE auch nur für solche Suchanfragen eingesetzt werden.

=over 4

=item Parameter

 - ENTRYCACHE: Referenz auf den von ldap_gen_entrycache generierten Hash
 - LDAP: Objekt zum Zugriff auf LDAP-Server (Typ Net::LDAP::Message)
 - FILTER: LDAP-Filter (siehe man ldap_search, RFC 2254)
 - STARTDN: Suche beginnt bei dieser DN
 - ROOTDN: bis zu dieser DN (einschl.) wird gesucht.
 		if not defined, search stops at ''
 - STANDARDDN: Erste Komponente der DN fr nachgeordnete Suche
	(default: cn=standards)
 - DEVICETYPEDN: DN des Devicetype Objektes, für das die Standards Objekte gesucht werden sollen
    (default: undef)
 - LOWERCASE: 1: DN wird lowercase als hash key verwendet
              0: DN wird wie von LDAP geliefert als hash key verwendet
        (default: 0)

Beispiel:

 my $entry = ldap_treesearch(
			LDAP => $ldap,
			FILTER => '(objectClass=dbkDsap)',
			STARTDN => 'cn=pc00261,ou=gs201004,ou=test,o=organisation,c=de',
			ROOTDN => 'o=organisation,c=de' );
 oder
 my $entry = ldap_treesearch(
		ENTRYCACHE => $entrycache_ref,
		STARTDN => 'cn=pc00261,ou=gs201004,ou=test,o=organisation,c=de',
		ROOTDN => 'o=organisation,c=de');
 

=item Rueckgabe

Falls gefunden, Objekt vom Typ Net::LDAP::Entry, sonst undef.

=back

=cut

sub ldap_treesearch
{
    my %param = (
        ENTRYCACHE   => undef,
        LDAP         => undef,
        ATTR         => undef,
        FILTER       => undef,
        STARTDN      => undef,
        ROOTDN       => "",
        STANDARDCN   => 'cn=standards',
        DEVICETYPEDN => undef,
        SCOPE        => 'base',
        LOWERCASE    => 0,
        DEBUG        => 0,
        @_,
    );

    my $dn            = $param{STARTDN};
    my $entrycache    = $param{ENTRYCACHE};
    my $ldap          = $param{LDAP};
    my $attr          = $param{ATTR};
    my $filter        = $param{FILTER};
    my $std_cn        = $param{STANDARDCN};
    my $devicetype_dn = $param{DEVICETYPEDN};
    my $scope         = $param{SCOPE};
    my $lc            = $param{LOWERCASE};
    my $rootdn        = $param{ROOTDN};
    my ( $cn, $actualbase ) = ldap_splitdn($dn);

    my $result;
    my $entry;

    my $logger = Log::Log4perl->get_logger();

    if ( !$dn ) {
        $logger->error("ldap_treesearch: called without dn");
        return;
    }

    if ($ldap) {
        if ($entrycache) {
            $logger->error(
                "ldap_treesearch: LDAP and ENTRYCACHE can not be used together"
            );
            return;
        }
        if ($attr) {
            my ( $c_package, $c_filename, $c_line ) = caller;
            $logger->error(
                "ldap_treesearch called from $c_package in line $c_filename in line $c_line: LDAP should not be used with ATTR. Redeclaring as ldap_treesearch with filter"
            );
            return ldap_treesearch(
                LDAP         => $ldap,
                FILTER       => $attr . '=*',
                STARTDN      => $dn,
                ROOTDN       => $param{ROOTDN},
                STANDARDCN   => $std_cn,
                DEVICETYPEDN => $devicetype_dn,
                LOWERCASE    => $lc,
                DEBUG        => $param{DEBUG}
            );
        }
        if ( !$filter ) {
            $logger->error(
                "ldap_treesearch: both LDAP and FILTER must be defined");
            return;
        }
    } elsif ($entrycache) {
        if ( !$attr ) {
            $logger->error(
                "ldap_treesearch: both ENTRYCACHE and ATTR must be defined");
            return;
        }
        if ($filter) {
            $logger->error(
                "ldap_treesearch: ENTRYCACHE cannot be used with FILTER");
            return;
        }
        if ( $scope ne 'base' ) {
            $logger->error(
                "ldap_treesearch: ENTRYCACHE cannot be used with other scopes than base"
            );
            return;
        }

    } else {
        $logger->error(
            "ldap_treesearch: either ENTRYCACHE or LDAP must be defined");
        return;
    }

    while ( length($dn) >= length( $param{ROOTDN} ) ) {

        # uses entrycache if available
        if ($entrycache) {
            $dn = lc($dn) if $lc;

            #print "treesearch: probing  <$dn>\n" if $param{DEBUG}; #debug
            if ( $entrycache->{$dn} ) {
                my $tmp_entry = $entrycache->{$dn};

                #print "treesearch: checking <$dn>\n" if $param{DEBUG}; #debug
                if ( $tmp_entry->exists($attr) ) {
                    $entry = $tmp_entry;
                    last;
                }
            }
        } else {    # query LDAP
                    # search LDAP for the actual dn
            $result = $ldap->search(
                base   => $dn,
                scope  => $scope,
                filter => $filter
            );

            # when searching for a DN with base => $dn,
            # and nothing was found,
            # LDAP-Servers yield a specific error,
            # which is not really an error
            # in this context. Here, the loop must go on.
            if ( $result->code and ( $result->code != LDAP_NO_SUCH_OBJECT ) ) {
                ldap_check_error( $result, 0 );    # don't die
                return;
            }
            if ( $result->count > 0 ) {
                if ( $param{DEBUG} ) {
                    $logger->error( " LDAP_TREESEARCH found "
                          . $result->count
                          . " entries for "
                          . $dn );
                }

                # there should be no or one entry
                # this will ignore
                # possibly existing further
                # entries found
                ($entry) = $result->all_entries;
                last;
            }
        }

        # example for a search sequence (dn = dc1,dc2,dc3)
        # - dc1,dc2,dc3
        # - std_cn,dc1,dc2,dc3
        # - dc2,dc3
        # - std_cn,dc2,dc3
        # and so on.
        # if std_cn is not defined, its just
        # - dc1,dc2,dc3
        # - dc2,dc3
        # - dc3
        if ( !$std_cn ) {

            # removing first domain component
            $dn = ldap_shiftbase($dn);
        } elsif (
            ldap_is_standards_object(
                'DN'           => $dn,
                'STANDARDDN'   => $std_cn,
                'DEVICETYPEDN' => $devicetype_dn,
                'LDAP'         => $ldap,
                'ROOTDN'       => $rootdn,
                'ENTRYCACHE'   => $entrycache,
                'LOWERCASE'    => $lc,
            )
          )
        {

            # removing std_cn
            $dn = ldap_shiftbase($dn);

            # removing next domain component
            $dn = ldap_shiftbase($dn);
        } else {
            my $next_dn = ldap_construct_standards_name(
                'DN'           => $dn,
                'STANDARDDN'   => $std_cn,
                'DEVICETYPEDN' => $devicetype_dn,
                'LDAP'         => $ldap,
                'ENTRYCACHE'   => $entrycache,
                'LOWERCASE'    => $lc,
            );
            if (  !$next_dn
                or $next_dn eq ""
                or $next_dn eq $dn )
            {

                # no standards object name could be constructed, so take next component
                $dn = ldap_shiftbase($dn);
            } else {
                $dn = $next_dn;
            }
        }
    }

    return $entry;
}

=head2 Funktion ldap_construct_standards_name

Baut für einen übergebenen devicetype_dn oder Standards Prefix den dn des
nächsten Standards Objektes im Baum. Wenn in dem übergebenen Baumabschnitt 
kein Standards Objekt vorhanden ist, so wird die erste Komponente des 
übergebenen dn's abgeschnitten.

=over 4

=item Parameter

 - ENTRYCACHE: Referenz auf den von ldap_gen_entrycache generierten Hash
 - LDAP: Objekt zum Zugriff auf LDAP-Server (Typ Net::LDAP::Message)
 - FILTER: LDAP-Filter (siehe man ldap_search, RFC 2254)
 - DN: Suche beginnt bei dieser DN
 - STANDARDDN: Erste Komponente der DN fr nachgeordnete Suche
	(default: undef)
 - DEVICETYPEDN: DN des Devicetypes, für das ein Standards Objekt gesucht wird
 - LOWERCASE: 1: DN wird lowercase als hash key verwendet
              0: DN wird wie von LDAP geliefert als hash key verwendet
        (default: 0)

Beispiel:

 my $dn = ldap_construct_standards_name(
		ENTRYCACHE => $entrycache_ref,
		DN => 'ou=gs201004,ou=test,o=organisation,c=de',
		STANDARDDN => 'cn=standards');
 

=item Rueckgabe

String dn des nächsten Standardobjektes

=back

=cut

sub ldap_construct_standards_name
{
    my %params = (
        'DN' => undef,

        'DEVICETYPEDN' => undef,
        ## or
        'STANDARDDN' => undef,

        'LDAP' => undef,
        ## or
        'ENTRYCACHE' => undef,
        'LOWERCASE'  => 0,

        @_
    );
    my $devicetype_dn = $params{'DEVICETYPEDN'};
    my $location_dn   = $params{'DN'};
    my $standard_dc   = $params{'STANDARDDN'};
    my $ldap          = $params{'LDAP'};
    my $entrycache    = $params{'ENTRYCACHE'};
    my $lc            = $params{'LOWERCASE'};

    my $logger = Log::Log4perl->get_logger();

    if ($devicetype_dn) {
        if ($ldap) {
            my $result = $ldap->search(
                'base'  => $location_dn,
                'scope' => 'one',
                'filter' =>
                  '(&(objectClass=dbkDeviceTypeStandards)(dbkDeviceTypeDn='
                  . $devicetype_dn . '))',
                'attrs' => ['cn']
            );
            if ( $result->code || $result->count == 0 ) {

                #something horrible has happened so return nothing
                return;
            }

            # we have found a dbkStandard object for this devicetype
            return $result->entry(0)->dn();
        } elsif ($entrycache) {

            #TODO test function
            if ($lc) {
                $location_dn = lc($location_dn);
            }

            #TODO implement something nicer
            $logger->info('Warning. very expensive search ahead');
            foreach my $potential_standard_dn ( keys %{$entrycache} ) {
                if ($lc) {
                    $potential_standard_dn = lc($potential_standard_dn);
                }
                my $potential_standard_location =
                  ldap_shiftbase($potential_standard_dn);
                if ( $potential_standard_location eq $location_dn ) {
                    my $entry = $entrycache->{$potential_standard_dn};
                    if ( has_objectclass( $entry, 'dbkDeviceTypeStandards' ) ) {
                        if (
                            ldap_entry_has_value(
                                ENTRY     => $entry,
                                ATTR_NAME => 'dbkDeviceTypeDn',
                                VALUE     => $devicetype_dn,
                                LOWERCASE => 1,
                            )
                          )
                        {
                            return $potential_standard_dn;
                        }
                    }
                }
            }
        }
        return;
    } elsif ($standard_dc) {
        if ( $location_dn =~ m/^cn=/i ) {

            # cn= can't be a container, so skip it
            return;
        } else {
            my $new_standard_dn = undef;
            if ( $standard_dc ne "" ) {
                $new_standard_dn = $standard_dc . "," . $location_dn;
            } else {
                $new_standard_dn = $location_dn;
            }
            if ($entrycache) {
                if ( !$entrycache->{$new_standard_dn} ) {
                    $new_standard_dn = undef;
                }
            }
            return $new_standard_dn;
        }
    }

    $logger->error(
        'ldap_construct_standards_name was called with no standards prefix nor devicetype_dn'
    );
    return;
}

=head2 Funktion ldap_is_standards_object

Baut für einen übergebenen devicetype_dn oder Standards Prefix den dn des
nächsten Standards Objektes im Baum. Wenn in dem übergebenen Baumabschnitt 
kein Standards Objekt vorhanden ist, so wird die erste Komponente des 
übergebenen dn's abgeschnitten.

=over 4

=item Parameter

 - ENTRYCACHE: Referenz auf den von ldap_gen_entrycache generierten Hash
 - LDAP: Objekt zum Zugriff auf LDAP-Server (Typ Net::LDAP::Message)
 - DN: Suche beginnt bei dieser DN
 - STANDARDDN: Erste Komponente der DN fr nachgeordnete Suche
	(default: undef)
 - DEVICETYPEDN: DN des Devicetype Objektes
    (default: undef)
 - LOWERCASE: 1: DN wird lowercase als hash key verwendet
              0: DN wird wie von LDAP geliefert als hash key verwendet
        (default: 0)

Beispiel:

 my $isStandard = ldap_is_standards_object(
		ENTRYCACHE => $entrycache_ref,
		DN => 'cn=standards,ou=gs201004,ou=test,o=organisation,c=de',
		STANDARDDN => 'cn=standards');

oder

 my $isStandard = ldap_is_standards_object(
		ENTRYCACHE => $entrycache_ref,
		DN => 'cn=windowsStandards,ou=gs201004,ou=test,o=organisation,c=de',
		DEVICETYPEDN => 'cn=deviceWindowsPC,ou=global,o=organisation,c=de');
 

=item Rueckgabe

Boolean

=back

=cut

sub ldap_is_standards_object
{
    my %params = (
        'DN'           => undef,
        'STANDARDDN'   => undef,
        'DEVICETYPEDN' => undef,
        'LDAP'         => undef,
        'ENTRYCACHE'   => undef,
        'LOWERCASE'    => 0,
        @_
    );

    my $dn            = $params{'DN'};
    my $standard_dc   = $params{'STANDARDDN'};
    my $devicetype_dn = $params{'DEVICETYPEDN'};
    my $entrycache    = $params{'ENTRYCACHE'};
    my $lc            = $params{'LOWERCASE'};
    my $ldap          = $params{'LDAP'};

    my $logger = Log::Log4perl->get_logger();

    if ( !$dn ) {

        # no dn given
        return 0;
    }

    if ($devicetype_dn) {
        if ($ldap) {
            my $result = $ldap->search(
                base  => $dn,
                scope => 'base',
                filter =>
                  '(&(objectClass=dbkDeviceTypeStandards)(dbkDeviceTypeDn='
                  . $devicetype_dn . '))',
                attrs => ['cn'],
            );
            if ( $result->code || ( $result->count == 0 ) ) {

                # either this was no standards object or ldap is down
                return 0;
            } elsif ( $result->count > 0 ) {

                # this is a standards object for given devicetype
                return 1;
            }

            #TODO more descriptive error message
            $logger->error('shouldnt be here');
            return 0;
        } elsif ($entrycache) {

            #TODO test me
            if ($lc) {
                $dn = lc($dn);
            }
            if ( $entrycache->{$dn} ) {
                my $entry = $entrycache->{$dn};
                if ( has_objectclass( $entry, 'dbkDeviceTypeStandards' ) ) {
                    return ldap_entry_has_value(
                        ENTRY     => $entry,
                        ATTR_NAME => 'dbkDeviceTypeDn',
                        VALUE     => $devicetype_dn,
                        LOWERCASE => 1,
                    );
                }
            }
        }
        $logger->error(
            "devicetype given, but no ldap nor entrycache, so cant decide, if "
              . $dn
              . " is a standards object." );
        return 0;
    } elsif ($standard_dc) {
        if ( $standard_dc ne "" ) {
            my ($actual_cn) = ldap_splitdn($dn);
            return ( $actual_cn eq $standard_dc );
        } else {

            #TODO: standard_dc equals "". should object be always a standards object?
            return 1;
        }
    }
    $logger->error(
        'ldap_is_standards_object was called with no standards prefix nor devicetype_dn'
    );
    return 0;
}

=head1 Funktionen

=head2 Funktion ldap_treesearch_cached

obsolete: use ldap_treesearch instead

=cut

sub ldap_treesearch_cached
{
    my %param = (
        ENTRYCACHE => undef,
        STARTDN    => undef,
        ROOTDN     => undef,
        STANDARDCN => 'cn=standards',
        @_,
    );
    return ldap_treesearch(%param);
}

=head2 Funktion ldap_shiftbase

entfernt die erste Komponente einer LDAP-Base bzw. dn

=over 4

=item Parameter

String, der eine Base oder DN enthält

=item Rckgabewert

Base ohne die erste Komponente

Beispiel: ldap_shiftbase('ou=gs234,ou=berlin,o=organisation,c=de') liefert
'ou=berlin,o=organisation,c=de'

=back

=cut

sub ldap_shiftbase
{

    # given a dn, this function returns the appropriate base by removing
    # the first component of the dn string
    # more generic, from a comma separated string, this function removes
    # the first part
    my ($dn) = @_;
    $dn =~ s/^[^,]+,?//o;
    return $dn;
}

=head2 Funktion ldap_check_error

Überprfung des Ergebnisses eines Net::LDAP::Search auf Fehler

=over 4

=item Parameter

 - LDAP-Result (Typ Net::LDAP::Search)
 - Abbruch bei Fehler (Typ Boolean): 0: kein Abbruch, 1: Abbruch

=item Rckgabewert

LDAP-Fehlermeldung, außerdem, wird per Syslog eine Meldung
ausgegeben und das Programm wird abgebrochen.

=back

=cut

sub ldap_check_error
{

    # check LDAP result for error, send error message to syslog
    # and die with error message
    my $result = "";
    my $logger = Log::Log4perl->get_logger();
    my ( $ldap_result, $die_on_error ) = @_;
    if ( $ldap_result->code != 0 ) {
        $result = $ldap_result->error;
        $logger->fatal( "LDAP-Error: " . $ldap_result->error );
        if ($die_on_error) {
            show_callstack();
            die $ldap_result->error;
        }
    }
    return $result;
}

=head2 Funktion ldap_cut_base

    entfernt Parameter $2 vom hinteren Ende von Parameter $1
    Wird verwendet, um die LDAP Base von einem DN zu entfernen.
    Dient der besseren Darstellung.

=over 4

=item Parameter

 - 1: Array reference of DN strings
 - 2: LDAP Base String

=item Rckgabewert

array of DNs without there bases

=back

=cut

sub ldap_cut_base
{
    my $dn_array = shift;
    my $base     = shift;

    my @result;

    # short dn: strip ldap base
    #print Dumper( $dn_array );
    foreach my $i ( @{$dn_array} ) {
        my $short = $i;
        $short =~ s/$base$//g;
        push( @result, $short );
    }
    return @result;
}

=head2 Funktion ldap_splitdn

Teilt einen gegebenen DN in den ersten Teil (z.B. cn=...) und den
Rest.

Beispiel:

 ldap_splitdn('cn=pc00234,ou=gs234,ou=berlin,o=organisation,c=de')
 Ergebnis: ['cn=pc00234', 'ou=gs234,ou=berlin,o=organisation,c=de']

=over 4

=item Parameter

 - DN (String)

=item Rckgabewert

Liste aus zwei Elementen:
 
 1. erste Komponente des DN
 2. restliche Komponenten des DN

=back

=cut

sub ldap_splitdn
{

    # split a given dn into its first part (eg. cn=...) and the remaining parts
    # and return it as a list of two strings
    # so that it can be used as filter and base in an LDAP search
    # Example:
    # "cn=refimagetest01,ou=global,o=organisation,c=DE"
    # => ("cn=refimagetest01", "ou=global,o=organisation,c=DE")
    # Parameter: dn
    my ($dn) = @_;

    my @result = ();

    if ( defined $dn ) {
        my @splitdn = split( /,/, $dn );
        push( @result, shift(@splitdn) );
        push( @result, join( ",", @splitdn ) );
    }

    return @result;
}

=head2 ldap_normalize( $;$ )

    converts DN into standard formate (lowercase, whitespace removed).
    Required, if working with a cache

=over 4

=item Parameter

 - string: DN
 - [boolean: case sensitive (default: false)]

=item Returnd

    normalized DN

=back

=cut

sub ldap_normalize( $;$ )
{
    my $dn            = shift;
    my $casesensitive = shift;

    $dn = convertCase( $dn, $casesensitive );

    # TODO: remove whitespaces

    return $dn;
}

=head2 Funktion has_objectclass

Überprüft, ob ein LDAP-Eintrag eine bestimmte Objektklasse besitzt.

=over 4

=item Parameter

 - LDAP-Eintrag gemäß Net::LDAP::Entry
 - Objektklassenname

=item Rückgabewert

boolean

=back

=cut

sub has_objectclass
{
    my ( $ldap_entry, $objectclass_name ) = @_;
    return ldap_entry_has_value(
        ENTRY     => $ldap_entry,
        ATTR_NAME => 'objectclass',
        VALUE     => $objectclass_name,
        LOWERCASE => 1,
    );
}

=head2 Funktion ldap_entry_has_value

Überprft, ob ein LDAP-Eintrag ein bestimmtes Attribut besitzt.
Wird benötigt, wenn z.B. bei Multivalue Attributen abgefragt werden
soll, ob ein bestimmter Inhalt in diesem Attribut vorkommt.

Beispiel:

 if ( ldap_entry_has_value(
			ENTRY => $ldap_entry,
			ATTR_NAME => 'objectclass',
			VALUE => 'dbkUpdate') {
	...

Überprft, ob der Eintrag die Objektklasse dbkUpdate besitzt.

=over 4

=item Parameter

 - ENTRY: ein LDAP-Eintrag gemäß Net::LDAP::Entry
 - ATTR_NAME: Attributname
 - VALUE: Inhalt des Attributs, nach dem gesucht werden soll

=item Rckgabewert

1 (true), wenn der Wert im Attribut vorkommt, sonst 0 (false)

=back

=cut

sub ldap_entry_has_value
{
    my %param = (
        ENTRY     => undef,
        ATTR_NAME => undef,
        VALUE     => undef,
        LOWERCASE => 0,
        @_,
    );

    my $result = 0;

    my $ldap_entry = $param{ENTRY};
    my $lc         = $param{LOWERCASE};
    my $value      = $param{VALUE};

    if ($lc) {
        $value = lc($value);
    }

    if ( ref $ldap_entry ) {
        my @values = $ldap_entry->get_value( $param{ATTR_NAME} );
        my ($grep_result) =
          grep { $_ eq $value } map { $lc ? lc($_) : $_ } @values;
        if ( defined $grep_result && $grep_result eq $value ) {
            $result = 1;
        }
    }
    return $result;
}

=head2 Funktion ldap_get_dn_entry

Liefert einen LDAP-Eintrag zu einem gegebenem DN.

=over 4

=item Parameter

 - LDAP: Objekt zum Zugriff auf LDAP-Server (Typ Net::LDAP::Message)
 - DN: DN, nach dem gesucht werden soll
 - FILTER: optional, default: (objectClass=*)

=item Rckgabewert

LDAP-Eintrag (Type Net::LDAP::Entry) des gesuchten DN, oder undef, wenn
es keinen Eintrag unter dem gegebenem DN gibt, oder ein Fehler aufgetreten
ist.

=back

=cut

sub ldap_get_dn_entry
{
    my %param = (
        LDAP   => undef,
        DN     => undef,
        FILTER => '(objectclass=*)',
        ATTRS  => ['*'],
        @_,
    );

    my $ldap = $param{LDAP};
    my $entry;
    my $result = $ldap->search(
        base   => $param{DN},
        scope  => 'base',
        filter => $param{FILTER},
        attrs  => $param{ATTRS}
    );

    # looking up DNs directly as base yields an error-code,
    # if not found.
    if ( $result->code and ( $result->code != LDAP_NO_SUCH_OBJECT ) ) {
        ldap_check_error( $result, 0 );    # don't die
        return;
    }

    # we should have found an entry now
    if ( $result->count > 0 ) {

        # there should be no or one entry
        # this will ignore possibly existing further entries found
        ($entry) = $result->all_entries;
    }
    return $entry;
}

=head2 Funktion ldap_get_ref_entry

Liefert einen LDAP-Eintrag, auf den in einem gegebenen Eintrag gezeigt wird,
also den Eintrag, dessen DN in einem bestimmten Attribut eines gegebenen Eintrags steht.

Beispiel: gesucht wird der RefPC-Eintrag zu einer bestimmten Workstation.
Gegeben ist der DN des Workstation-Eintrags sowie Objektklasse und
Attributname des ReferenzDN-Eintrags. Falls im Workstation-Eintrag
ein RefernzDN-Eintrag vorhanden ist, so wird der Eintrag geliefert, auf den dieser zeigt, ansonsten wird im Baum abw�ts nach einem entsprechenden standards-Eintrag gesucht.

=over 4

=item Parameter

 - LDAP: Objekt zum Zugriff auf LDAP-Server (Typ Net::LDAP::Message)
 - DN: DN des Eintrags, zu dem ein Referenz-Eintrag gesucht wird
 - ROOTDN: bis zu dieser DN (einschl.) wird gesucht
 - FILTER: der Filter sollte die Suche auf geeignete Objektklassen einschränken
	die das ReferenzDN-Attribut enthalten
 - REFATTR: Attributname des ReferenzDN-Attributs

=item Rckgabewert

LDAP-Eintrag (Type Net::LDAP::Entry) des gesuchten Referenzeintrags, oder
undef, wenn es keinen Eintrag unter dem gegebenem DN gibt.

=back

=cut

sub ldap_get_ref_entry
{
    my %param = (
        LDAP    => undef,
        DN      => undef,
        ROOTDN  => undef,
        FILTER  => '(objectclass=*)',
        REFATTR => undef,
        @_,
    );

    my $ldap = $param{LDAP};

    my $ref_dn;

    my $ref_entry = ldap_treesearch(
        LDAP    => $ldap,
        STARTDN => $param{DN},
        ROOTDN  => $param{ROOTDN},
        FILTER  => $param{FILTER}
    );

    if ( ref $ref_entry ) {
        $ref_dn = $ref_entry->get_value( $param{REFATTR} );
    }

    # we should now have a valid reference-DN
    my $entry = undef;
    $entry = ldap_get_dn_entry(
        LDAP => $ldap,
        DN   => $ref_dn
    );

    return $entry;
}

=head2 Funktion ldap_get_refpc_entry_cached

Liefert den RefPC Eintrag, der zu einer Workstation gehört, wobei hier
keine LDAP Anfragen durchgefhrt werden, stattdessen mssen Hashes mit
allen benötigten LDAP Einträgen bergeben werden.
Liefert einen LDAP-Eintrag, auf den in einem gegebenen Eintrag gezeigt wird,
also den Eintrag, dessen DN in einem bestimmten Attribut eines gegebenen
Eintrags steht.

Beispiel: gesucht wird der RefPC-Eintrag zu einer bestimmten Workstation.
Gegeben ist der DN des Workstation-Eintrags sowie Objektklasse und
Attributname des ReferenzDN-Eintrags. Falls im Workstation-Eintrag
ein RefernzDN-Eintrag vorhanden ist, so wird der Eintrag geliefert, auf
den dieser zeigt, ansonsten wird im Baum abwärts nach einem entsprechenden 
standards-Eintrag gesucht.

=over 4

=item Parameter

 - WORKSTATION_ENTRYCACHE
 - DSAP_ENTRYCACHE
 - REFPC_ENTRYCACHE
 - DN: DN des Eintrags, zu dem ein Referenz-Eintrag gesucht wird
 - ROOTDN: bis zu dieser DN (einschl.) wird gesucht
 - REFATTR: Attributname des ReferenzDN-Attributs

=item Rckgabewert

LDAP-Eintrag (Type Net::LDAP::Entry) des gesuchten Referenzeintrags, oder
undef, wenn es keinen Eintrag unter dem gegebenem DN gibt.

=back

=cut

sub ldap_get_refpc_entry_cached
{
    my %param = (
        WORKSTATION_ENTRYCACHE => undef,
        DSAP_ENTRYCACHE        => undef,
        REFPC_ENTRYCACHE       => undef,
        WORKSTATIONDN          => undef,
        ROOTDN                 => undef,
        REFATTR                => undef,
        @_,
    );

    my $ws_cache    = $param{WORKSTATION_ENTRYCACHE};
    my $dsap_cache  = $param{DSAP_ENTRYCACHE};
    my $refpc_cache = $param{REFPC_ENTRYCACHE};
    my $entry;
    my $ref_entry;
    my $ref_dn;

    # get Entry of given DN
    if ( !exists $ws_cache->{ $param{WORKSTATIONDN} } ) {
        return;
    }
    $entry = $ws_cache->{ $param{WORKSTATIONDN} };

    # check if the referenceDN-attribute exists and is not empty
    if ( $entry->exists( $param{REFATTR} ) ) {
        $ref_dn = $entry->get_value( $param{REFATTR} );
    }

    # if no referenceDN was found in the given entry, do a treesearch
    # downwards
    if ( !defined $ref_dn ) {

        # begin search with standards in actual container
        my ( $cn, $base ) = ldap_splitdn( $param{WORKSTATIONDN} );

        #$ref_entry = ldap_treesearch(
        #		LDAP => $ldap,
        #		STARTDN => 'cn=standards,' . $base,
        #		ROOTDN => $param{ROOTDN},
        #		FILTER => $param{FILTER}
        #);
        $ref_entry = ldap_treesearch(
            ENTRYCACHE => $dsap_cache,
            STARTDN    => 'cn=standards,' . $base,
            ROOTDN     => $param{ROOTDN},
            ATTR       => $param{REFATTR},
            LOWERCASE  => 1
        );

        if ( defined $ref_entry ) {
            $ref_dn = $ref_entry->get_value( $param{REFATTR} );
        }
    }

    # we should now have a valid reference-DN
    $entry = undef;
    $entry = $refpc_cache->{$ref_dn};

    return $entry;
}

=head2 Funktion ldap_get_cn_from_dn
 
Liefert die cn-Komponente aus einem DN, ohne den Teil cn=

 ldap_get_cn_from_dn("cn=pc00234,ou=gs345,ou=bonn,o=organisation,c=de")
 => pc00234
 
=over 4
 
=item Parameter

ein DN, dessen erste Komponente cn=... enthält
 
=item Rckgabewert

Wert der cn-Komponente, oder undef, wenn keine cn-Komponente im DN vorhanden
 
=back
 
=cut

sub ldap_get_cn_from_dn
{
    my ($dn) = @_;

    my $cn;

    my ($cnpart) = ldap_splitdn($dn);
    if ( defined $cnpart ) {
        $cnpart = lc($cnpart);
        if ( $cnpart =~ /^cn=(.+)$/ ) {
            $cn = $1;
        }
    }
    return $cn;
}

=head2 Funktion ldap_dn2dnsname
 
Liefert zu einem gebeben dn den entsprechenden DNS-Namen, Beispiel:

 cn=pc00263,ou=gs201004,ou=test,o=organisation,c=de
 => pc00263.gs201004.test.organisation.de

LDAP erlaubt Blanks im DN. Sollte dies der Fall sein, so werden
Blanks durch das Zeichen "-" ersetzt, damit sie DNS konform sind.
 
=over 4
 
=item Parameter

DN als String
 
=item Rckgabewert

DNS-Name
 
=back
 
=cut

sub ldap_dn2dnsname
{
    my ($dn) = @_;
    my $dns_name = "";

    if ($dn) {
        my @dn_parts = split( /,/, $dn );
        my @dns_name_parts;
        foreach my $dn_part (@dn_parts) {
            $dn_part =~ /.+=(.+)/;
            push( @dns_name_parts, $1 );
        }
        $dns_name = lc( join( '.', @dns_name_parts ) );

        # replace whitespace by "-"
        # and German Umlauts by 2 letter code
        # for DNS conformity
        $dns_name = convertToValidDnsName($dns_name);
    }

    return $dns_name;
}

=head2 Funktion ldap_cn2dnsname

Liefert zu einem gebeben cn den entsprechenden DNS-Namen, Beispiel:

 pc-respc_1
 => pc-respc-1

LDAP erlaubt Blanks und "_" (underscore) im cn. Sollte dies der Fall sein,
so werden Blanks und "_" durch das Zeichen "-" ersetzt, damit sie
DNS konform sind.

=over 4

=item Parameter

cn als String

=item Rckgabewert

DNS-Name

=back

=cut

sub ldap_cn2dnsname
{
    my ($cn) = @_;
    my $dns_name = "";

    if ($cn) {
        $dns_name = lc($cn);

        # replace whitespace by "-"
        # and German Umlauts by 2 letter code
        # for DNS conformity
        $dns_name = convertToValidDnsName($dns_name);
    }

    return $dns_name;
}

=head2 Funktion get_zonename

Ermitteln des DNS Zonennamens aus dem dn

=over 4

=item Parameter

der dn eines LDAP Eintrags

=item Rückgabewert

Zonenname

=back

=cut

sub get_zonename
{
    my $dn           = shift;
    my $do_shiftbase = shift;

    # if optional 2nd parameter is not given,
    # default is to remove first component of DN
    if ( !defined($do_shiftbase) ) {
        $do_shiftbase = 1;
    }

    my $dn_base;
    if ($do_shiftbase) {
        $dn_base = ldap_shiftbase($dn);
    } else {
        $dn_base = $dn;
    }

    my $zonename = ldap_dn2dnsname($dn_base);

    # it was determined, that all the server entries found in o=debeka,c=de
    # are put in the zone server.debeka.de
    # doing it this way is not good. Better: putting the in Entries
    # in the appropriate LDAP Tree ou=server,o=debeka,c=de
    #FIXME: $config sollte doch irgendwoher kommen.
    # 20060711 stephand@dass-it.de:
    # this is obsolete, it is now desired to generate both zones
    #my $config        = SmartClient::Config->new();
    #my $base_zonename = lc( ldap_dn2dnsname( $config->get('ALL_LDAPBASE') ) );
    #$zonename =
    #  $zonename eq $base_zonename ? "server." . $base_zonename : $zonename;

    return $zonename;
}

=head2 Funktion ldap_cn2addr
 
Berechnung der IP-Adresse aus cn und Netzwerkadresse

=over 4
 
=item Parameter

    - cn: contains the workstations IP relative to the fixed range. Form aannnnn (e.g.: pr00071 or pc00257)
    - firstipaddr: first IP of fixed Range

=item Rckgabewert

 IP-Adresse als String in der Form nnn.nnn.nnn.nnn (z.B. 10.82.17.1)
 Bei ungltigen Parametern wird undef geliefert.
 
=back
 
=cut

sub ldap_cn2addr
{

    # a workstations IP-address is defined in its cn entry,
    # consisting of "pc" + and a number padded with zeroes to 5 digits
    # this must be added to the beginning of the fixed range for this subnet
    # to get the IP.

    my ( $cn, $firstipaddr ) = @_;
    my $newipaddr;

    my $logger = Log::Log4perl->get_logger();

    if ( !defined($cn) ) {
        $logger->error("parameter cn not defined");
        return;
    }

    if ( !defined($firstipaddr) ) {
        $logger->error("parameter firstipaddr not defined for $cn");
        return;
    }

    # convert cn to number by stripping leading non-digit characters (usually "pc")
    my $cni = $cn;

    #TODO: $cni =~ s/^\D+//o; should be faster.
    $cni =~ s/^\D+(\d+)/$1/;
    ## check if $cni really contains a number now
    if ( !( $cni =~ /^\d+$/ ) ) {
        $logger->warn("cn $cn is invalid");
        return;
    }

    # calculate the workstations ip by adding $cni to $firstip using Net::IP
    # this is done by converting the first ip of the range to integer, adding the
    # workstation number and converting the result back to an ip address string.

    # TODO: this is inefficient, check the IP more efficient!
    if ( !chk_ip($firstipaddr) ) {
        $logger->error("$firstipaddr is not a valid IP address");
        return;
    }

    ## this is not very efficient:
    #$newipaddr = Net::IP::ip_bintoip( Net::IP::ip_inttobin( ($firstip->intip())+$cni ,4),4);
    #return $newipaddr;

    ## this is much faster (thanks to holgerm):
    return join(
        ".",    # Result is seperated with dots
        unpack(
            "C4",    # a networkaddress becomes a quadruple of unsigned chars
            pack(
                "N",    # an integer becomes a network address

                unpack(
                    "N",    # a network address becomes an integer
                    pack( "C4", split( /\./, $firstipaddr ) )
                  )         # ip becomes a network address
                  + $cni
            )
        )
    );

}

=head2 Funktion ldap_cn2addr_ws_stgt
 
Berechnung der IP-Adresse aus cn und Netzwerkadresse fr Workstations mit
dem Namenschema fr die Stuttgarter Versicherung.

=over 4
 
=item Parameter

 CN: cn als String in der Form ws0hnn (z.B: ws0e23)
 NETWORK: Netzwerkadresse als String in der Form nnn.nnn.nnn.nnn,
          z.B. 10.82.16.0
 
=item Rckgabewert

 IP-Adresse als String in der Form nnn.nnn.nnn.nnn (z.B. 10.82.17.1)
 Bei ungltigen Parametern wird undef geliefert.
 
=back
 
=cut

sub ldap_cn2addr_ws_stgt
{

    # a workstations IP-address is defined in its cn entry,
    # consisting of "ws0" + one hex digit + 2 decimal digits (zero padded)
    # where the hex digit is the 3rd octet and the 2 decimal digits are
    # the 4th octet of the workstations IP address. The first two octets
    # are taken from the network address.

    # Parameters:
    # - cn: contains the workstations IP relative to the fixed range
    # - network_addr: Network Address

    my ( $cn, $network_addr ) = @_;
    my $newipaddr;
    my $logger = Log::Log4perl->get_logger();

    #defined($cn) || die "cn2addr(): Parameter cn not defined";

    if ( !defined($cn) ) {
        $logger->fatal("ldap_cn2addr: Parameter cn not defined");
        return;
    }

    if ( !defined($network_addr) ) {
        $logger->fatal("ldap_cn2addr: Parameter network_addr not defined");
        return;
    }

    if ( $cn =~ /(ws0)([0-9A-Fa-f])(\d{2})/ ) {
        my @network_split = split( /\./, $network_addr );
        $newipaddr =
          join( ".", $network_split[0], $network_split[1], hex($2), $3 + 0 );
    } else {
        $logger->fatal(
            "ldap_cn2addr_ws_stgt: $cn in network $network_addr is invalid");
        return;
    }

    return $newipaddr;
}

=head2 getLocationNetmask( $ldap, $location_dn, [$filter] )

checks if a specific value is part of a LDAP entry attribute (multivalue)


=over 4

=item Parameter

 - ldap connection
 - location_dn
 - [ldap filter]

=item Return

  string: IP netmask

=back

=cut

sub ldap_getLocationNetmask( $$;$ )
{
    my $ldap        = shift;
    my $location_dn = shift;
    my $filter      = shift || "(objectclass=scLocation)";

    my $attrs = ['ipNetmaskNumber'];

    # TODO: still needed? use Config?
    #my $dn    = 'cn=standards,' . $location_dn;
    my $dn = $location_dn;

    my $entry = ldap_get_dn_entry(
        LDAP   => $ldap,
        DN     => $dn,
        FILTER => $filter,
        ATTRS  => $attrs,
    );

    my $netmask;
    if ($entry) {
        $netmask = $entry->get_value('ipNetmaskNumber');
    }

    return $netmask;
}

=head2 ldap_get_location_entry()
 
Liefert das dkbLocation-Objekt zu einem gegebenen DN, beispielsweise
wird zu einem dbkPrinter oder dbkWorkstation-Eintrag der zugeh�ige
dbkLocation-Eintrag (cn=standards und objectClass=dbkLocation) ben�igt:

 my $pc_dn = 'cn=pc00257,ou=gs0030,ou=koblenz,o=organisation,c=de';
 my $location_entry = ldap_get_location_entry($pc_dn);
 
=over 4
 
=item Parameter

 LDAP: Verbindung zum LDAP-Server (Type Net::LDAP::Message)
 DN: eines dbkPrinter oder dbkWorkstation Eintrags
 
=item Rckgabewert

dbkLocation bzw. scLocation Eintrag oder undef,
wenn kein Eintrag gefunden wurde.
 
=back
 
=cut

sub ldap_get_location_entry
{
    my %param = (
        'LDAP'    => undef,
        'STARTDN' => undef,
        @_,
    );

    my $ldap    = $param{'LDAP'};
    my $startdn = $param{'STARTDN'};

    my $config = SmartClient::Config->new();

    my $location_entry = ldap_treesearch(
        LDAP    => $ldap,
        STARTDN => $startdn,
        ROOTDN  => $config->get('ALL_LDAPBASE'),
        FILTER  => $config->get('ALL_LDAP_FILTER_LOCATION')
    );

    return $location_entry;
}

=head2 Funktion ldap_gen_entrycache
 
Generiert einen Hash bestehend aus Net::LDAP::Entry Objekten, key: dn
 
=over 4
 
=item Parameter

 LDAP:        Net::LDAP Object 
 BASE:        Base DN for the search
 OBJECTCLASS: Name der Objektklasse, zu der der Cache aufgebaut werden soll
 ATTRLIST:    Liste der Attribute, die in den Cache aufgenommen werden
 
=item Rückgabewert

Referenz auf den Entry-Cache Hash
 
=back
 
=cut

sub ldap_gen_entrycache
{
    my %param = (
        'LDAP'        => undef,
        'BASE'        => undef,
        'OBJECTCLASS' => undef,
        'ATTRLIST'    => undef,
        'LOWERCASE'   => 0,
        @_,
    );

    my %entrycache;
    my $ldap   = $param{'LDAP'};
    my $result = $ldap->search(
        base   => $param{'BASE'},
        scope  => 'sub',
        filter => 'objectClass=' . $param{'OBJECTCLASS'},
        attrs  => $param{'ATTRLIST'}
    );
    ldap_check_error( $result, 1 );

    foreach my $entry ( $result->entries ) {
        my $dn = $entry->dn;
        $dn = lc($dn) if $param{LOWERCASE};
        $entrycache{$dn} = $entry;
    }
    return \%entrycache;
}

=head2 Funktion ldap_gen_entrycache_successive
 
Generiert einen Hash bestehend aus Net::LDAP::Entry Objekten, key: dn
Dabei wird sukzessive von einer gegebenen Base DN aus mit Scope one gesucht
und jeweils eine Komponente des DN entfernt bis zum erreichen der Root Base.

Anwendungsbeispiel: Ermitteln aller dbkDsap Objekte, zu einer Lokation.
 
=over 4
 
=item Parameter

 LDAP:        Net::LDAP o. SmartClient::LDAP Object 
 STARTBASE:   Start Base DN for the search
 ROOTBASE:    Root Base DN for the search
 FILTER:      LDAP Filterausdruck, i.d.R. objectClass=...
 ATTRLIST:    Liste der Attribute, die in den Cache aufgenommen werden
 
=item Rckgabewert

Referenz auf den Entry-Cache Hash
 
=back
 
=cut

sub ldap_gen_entrycache_successive
{
    my %param = (
        'LDAP'      => undef,
        'STARTBASE' => undef,
        'ROOTBASE'  => undef,
        'FILTER'    => undef,
        'ATTRLIST'  => undef,
        @_,
    );

    my %entrycache;
    my $ldap = $param{'LDAP'};
    my $base = $param{'STARTBASE'};

    # inside the loop, all bases including the Root Base must be used.
    # CAUTION: this assummes that the Start Base and Root Base end
    # in the same string, this is not checked here.
    while ( length($base) >= length( $param{'ROOTBASE'} ) ) {
        my $result = $ldap->search(
            base   => $base,
            scope  => 'one',
            filter => $param{'FILTER'},
            attrs  => $param{'ATTRLIST'}
        );
        ldap_check_error( $result, 1 );

        foreach my $entry ( $result->entries ) {
            $entrycache{ $entry->dn } = $entry;
        }

        # remove first component from base
        $base = ldap_shiftbase($base);
    }
    return \%entrycache;
}

=head2 Funktion ldap_set_values
 
Schreiben von Attributen in einen existierende LDAP Eintrag. Die Attribute
werden zum Eintrag hinzugefgt (add), falls noch nicht vorhanden oder
andernfalls ersetzt (replace).

=over 4
 
=item Parameter

 LDAP:        Net::LDAP o. SmartClient::LDAP Object, muß mit Schreibrecht
              auf den zu ändernden Eintrag gebunden sein.
 DN:          DN des zu modifizierenden Eintrags
 VALUES:      Hash-Referenz gemäß Net::LDAP->modify, falls darin keine
              objectClass gegeben ist, wird grundsätzlich replace
			  durchgefhrt, was fehlschlagen wird, wenn die entsprechenden
			  Werte noch nicht im Eintrag vorhanden sind.
			  Funktioniert zunächst nur, wenn genau eine objectClass gegeben
			  ist.
 ENTRY:       optional kann ein Net::LDAP::Entry Objekt bergeben werden.
              Diese Funktion muß dann keine LDAP Suche durchfhren, um
			  den Eintrag zu analysieren.
 
=item Rckgabewert

Ergebnis der LDAP Operation (Net::LDAP::Message->code)
 
=back
 
=cut

sub ldap_set_values
{
    my %param = (
        'LDAP'   => undef,
        'DN'     => undef,
        'VALUES' => undef,
        'ENTRY'  => undef,
        @_,
    );

    my $ldap   = $param{'LDAP'};
    my $entry  = $param{'ENTRY'};
    my $values = $param{'VALUES'};
    my $dn     = $param{'DN'};

    my $objectclass;
    my $do_replace = 0;

    my $logger = Log::Log4perl->get_logger();

    # check if Values Hash specifies an objectClass
    if ( $values->{'objectClass'} ) {
        $objectclass = $values->{'objectClass'};
    } else {
        $do_replace = 1;
    }

    # if no Entry was given, we must do a search to determine
    # the objectClasses of the Entry
    if ($objectclass) {
        if ( !defined $entry ) {
            my $res = $ldap->search(
                base   => $dn,
                scope  => 'base',
                filter => 'objectClass=*',
                attrs  => ['objectClass']
            );
            if ( $res->code ) {
                $logger->warn("ldap_set_values(): $dn not found in LDAP");
                return 0;
            }
            ($entry) = $res->entries;
        }

        # check if the entry already has the required objectClass
        if (
            ldap_entry_has_value(
                ENTRY     => $entry,
                ATTR_NAME => 'objectClass',
                VALUE     => $objectclass
            )
          )
        {
            $do_replace = 1;
        }

    }    # if ( $objectclass )

    # now do the LDAP operation
    my $result;
    if ($do_replace) {

        # if we do a replace, possibly existing objectclass must be
        # removed from the values hash
        if ( $values->{'objectClass'} ) {
            delete $values->{'objectClass'};
        }
        $result = $ldap->modify( $dn, replace => $values );
    } else {

        # add must be done
        $result = $ldap->modify( $dn, add => $values );
    }
    if ( $result->is_error() ) {

        $logger->warn( "error modifying $dn: " . $result->error() );
    }
    return ( $result->code, $result->error() );
}

sub show_callstack
{
    my $i = 0;
    print "\nCallstack:\n";
    while (
        my (
            $package,   $filename, $line,       $subroutine, $hasargs,
            $wantarray, $evaltext, $is_require, $hints,      $bitmask
        )
        = caller( $i++ )
      )
    {
        print "$package $filename $line $subroutine\n";
    }
}

=head2 ldap_get_next_available_number

searches next unused number and returns it

=over 4
 
=item Parameter

	LDAP:			LDAP connection
	SEARCHBASE		DN within the location container
        'START'         => 10000,
        'END'           => 65535,
        'ATTRIBUTE'     => 'uidNumber',

 
=item return

	number 
	or undef on error
 
=back

=cut

sub ldap_get_next_available_number
{
    my %param = (
        'LDAP'       => undef,
        'SEARCHBASE' => undef,
        'START'      => 10000,
        'END'        => 65535,
        'ATTRIBUTE'  => 'uidNumber',
        @_,
    );

    my $ldap      = $param{'LDAP'};
    my $attribute = $param{'ATTRIBUTE'};

    my $number;

    # work around: searches "=*",
    # because ">=" doesn't work always.
    # eg. it fails for uidNumber>=1000 (return nothing)
    my $search = $ldap->search(
        base   => $param{'SEARCHBASE'},
        filter => "(&(objectClass=posixAccount)($attribute=*))",
        scope  => 'sub'
    );
    if ( $search->code == 0 ) {
        my %numbers = ();
        foreach my $entry ( $search->entries ) {
            $numbers{ $entry->get_value($attribute) } = 1;
        }

        $number = getFirstAvailable(
            'USED_NUMBERS_HREF' => \%numbers,
            'START'             => $param{'START'},
            'END'               => $param{'END'}
        );

    }

    return $number;
}

=head2 ldap_get_next_available_workstation_ip

searches next unused IP address in that location
and returns it

=over 4
 
=item Parameter

	LDAP:			LDAP connection
	SEARCHBASE		DN within the location container
	ROOTDN			base DN (searches only up to this DN)
 
=item return

	IP address for a new workstation
	or undef on error
 
=back

=cut

sub ldap_get_next_available_workstation_ip
{
    my %param = (
        LDAP                  => undef,
        SEARCHBASE            => undef,
        ROOTDN                => undef,
        CONTAINER_OBJECTCLASS => "scWorkstationContainer",
        IP_RANGE_ATTRIBUTE    => "dbkDhcpFixedRange",
        @_,
    );

    return ldap_get_next_available_ip(%param);
}

=head2 ldap_get_next_available_printer_ip

searches next unused IP address in that location
and returns it

=over 4
 
=item Parameter

	LDAP:			LDAP connection
	SEARCHBASE		DN within the location container
	ROOTDN			base DN (searches only up to this DN)
 
=item return

	IP address for a new printer
	or undef on error
 
=back

=cut

sub ldap_get_next_available_printer_ip
{
    my %param = (
        LDAP                  => undef,
        SEARCHBASE            => undef,
        ROOTDN                => undef,
        CONTAINER_OBJECTCLASS => "scPrinterContainer",
        IP_RANGE_ATTRIBUTE    => "scPrinterIpRange",
        @_,
    );

    return ldap_get_next_available_ip(%param);
}

=head2 ldap_get_next_available_ip

searches next unused IP address in that location
and return it

=over 4
 
=item Parameter

	LDAP:			LDAP connection
	SEARCHBASE		DN within the location container
	ROOTDN			base DN (searches only up to this DN)
	CONTAINER_OBJECTCLASS 	LDAP container
	IP_RANGE_ATTRIBUTE 	LDAP Attribute that holds the IP range

=item return

	IP address
 
=back

=cut

sub ldap_get_next_available_ip
{
    my %param = (
        LDAP                  => undef,
        SEARCHBASE            => undef,
        ROOTDN                => undef,
        CONTAINER_OBJECTCLASS => "scWorkstationContainer",
        IP_RANGE_ATTRIBUTE    => "dbkDhcpFixedRange",
        @_,
    );

    my $ldap                  = $param{'LDAP'};
    my $searchbase            = $param{'SEARCHBASE'};
    my $rootdn                = $param{'ROOTDN'};
    my $container_objectclass = $param{'CONTAINER_OBJECTCLASS'};
    my $ip_range_attribute    = $param{'IP_RANGE_ATTRIBUTE'};

    my $logger = Log::Log4perl->get_logger();

    if ( !$ldap or !$searchbase ) {
        $logger->error("can't determine next available dn. missing parameter");
        return;
    }

    #
    # get dhcp fixed range
    #
    my $locationEntry = ldap_get_location_entry(
        LDAP    => $ldap,
        STARTDN => $searchbase
    );
    my $network        = $locationEntry->get_value('ipNetworkNumber');
    my $netmask        = $locationEntry->get_value('ipNetmaskNumber');
    my $dhcpFixedRange = $locationEntry->get_value($ip_range_attribute);
    my ( $fixed_from, $fixed_to ) = split( /,/, $dhcpFixedRange );

    #
    # get all IPs from this location
    #
    my $href_ip = getLocationIPs(
        LDAP       => $ldap,
        SEARCHBASE => $searchbase,
        NETWORK    => $network,
        NETMASK    => $netmask
    );

    my $ip = get_next_available_ip(
        USED_IP_HREF => $href_ip,
        START_IP     => $fixed_from,
        END_IP       => $fixed_to,
        NETWORK      => $network,
        NETMASK      => $netmask
    );

    return $ip;
}

=head2 ldap_get_next_available_workstation_dn

searches next unused IP address in that location
and return the corresponding DN to that IP

=over 4
 
=item Parameter

	LDAP:			LDAP connection
	SEARCHBASE		DN within the location container
	ROOTDN			base DN (searches only up to this DN)
        GENERATOR_FORMATSTRING  Formatstring for naming of new workstation (default 'pc${gsip}')
        GENERATOR_IDS           Generators to be filled in formatstring (default { 'gsip' => { type => 'IP', format => '%05d' } })
 
=item return

	DN for a new workstation
	or undef on error
 
=back

=cut

#TODO CONTAINER_OBJECTCLASS als konfigurierbaren Suchfilter ummodeln
#TODO dbkDeviceTypeDn und ldap entry soll übergeben werden können
sub ldap_get_next_available_workstation_dn
{
    my %param = (
        LDAP         => undef,
        SEARCHBASE   => undef,
        ROOTDN       => undef,
        DEVICETYPEDN => undef,
        ENTRY        => undef,

        #        CONTAINER_OBJECTCLASS => "scWorkstationContainer",
        CONTAINER_OBJECTCLASS  => "*",
        IP_RANGE_ATTRIBUTE     => "dbkDhcpFixedRange",
        ATTR                   => "dbkDeviceTypeDn",
        GENERATOR_FORMATSTRING => 'pc${gsip}',
        GENERATOR_IDS => { 'gsip' => { 'type' => 'IP', 'format' => '%05d' } },
        @_,
    );

    my $logger = Log::Log4perl->get_logger();

    # if a devicetype_dn was given, assume it is same as generator_dn
    # TODO change GENERATORDN to DEVICETYPEDN since they are the same
    if ( $param{DEVICETYPEDN} ) {
        $param{GENERATORDN} = $param{DEVICETYPEDN};
    }

    # Hole dbkDeviceTypeDN aus den Standards und ermittle so
    # den Namensgenerator
    my %generator = ldap_get_generator(%param);
    if ( $generator{GENERATOR_FORMATSTRING} && $generator{GENERATOR_IDS} ) {
        $logger->info( " Using generator: "
              . $generator{GENERATOR_FORMATSTRING} . " : "
              . $generator{GENERATOR_IDS} );
        $param{GENERATOR_FORMATSTRING} = $generator{GENERATOR_FORMATSTRING};
        $param{GENERATOR_IDS}          = $generator{GENERATOR_IDS};
        $param{DEVICE_TYPE_DN}         = $generator{DEVICE_TYPE_DN};
    }
    return ldap_get_next_available_dn(%param);
}

=head2 Funktion ldap_get_generator

Funktion holt für eine Lokation (C<cn=standards>) die zuständigen Generatoren
und den Formatstring. Diese Informationen werden für die Namensgenerierung
von neuen LDAP Objekten benutzt.
Kann auf zwei Weisen benutzt werden. Entweder wird ein generatorDn übergeben, oder
es wird aus der LDAP Suchbasis der generatorDn bestimmt.

=over 4

=item Parameter

	LDAP:		ldap connection
	ATTR:		attribute name for generatorDn reference
	SEARCHBASE:	starting point for recursive treesaerch
	ROOTDN:		point to stop  treesearch
	GENERATORDN:	string containing dn of the generator

=item return
	
	hash holding GENERATOR_FORMAT_STRING, reference to hash of GENERATOR_IDS 
		
=back

=cut 

sub ldap_get_generator
{
    my %param = (
        LDAP        => undef,
        ATTR        => undef,
        SEARCHBASE  => undef,
        ROOTDN      => undef,
        GENERATORDN => undef,
        @_
    );
    my $ldap                = $param{LDAP};
    my $attr                = $param{ATTR};
    my $searchbase          = $param{SEARCHBASE};
    my $defaultFormatstring = 'pc${gsip}';
    my $defaultGeneratorIds = {
        'gsip' => {
            'type'   => 'IP',
            'format' => '%05d'
        }
    };
    my $generatorDn = undef;

    my $logger = Log::Log4perl->get_logger();

    if ( !$param{GENERATORDN} ) {
        $logger->info( "get " 
              . $attr . " for "
              . $searchbase
              . " with rootdn="
              . $param{ROOTDN}
              . " in ldap: "
              . $ldap );
        my $generatorDnEntry = ldap_treesearch(
            LDAP    => $ldap,
            ATTR    => $attr,
            STARTDN => $searchbase,
            ROOTDN  => $param{ROOTDN},
            DEBUG   => 1
        );
        if ( !$generatorDnEntry ) {
            $logger->error( "no generator found for startdn " . $searchbase );
            return (
                GENERATOR_FORMATSTRING => $defaultFormatstring,
                GENERATOR_IDS          => $defaultGeneratorIds
            );
        }
        $generatorDn = $generatorDnEntry->get_value($attr);
    } else {
        $generatorDn = $param{GENERATORDN};
    }
    $logger->info( "Searching for " . $generatorDn );
    my $result = $ldap->search(
        base   => $generatorDn,
        scope  => 'base',
        filter => 'objectClass=*'
    );
    $logger->info( "result(" . $result->code . "): " . $result );

    my $formatstring;
    my %generatorIds;
    if ( $result->count > 0 ) {
        foreach my $entry ( $result->all_entries ) {
            if (    $entry->exists('dbkObjectIdFormatString')
                and $entry->exists('dbkObjectIdGenerator') )
            {
                $formatstring = $entry->get_value('dbkObjectIdFormatstring');

                # copy attribute values in our own list
                my $generatorsRef =
                  [ @{ $entry->get_value( 'dbkObjectIdGenerator', asref => 1 ) } ];
                foreach my $gen ( @{$generatorsRef} ) {

                    # Jeder Generator hat eine eindeutige ID.
                    my $genId = $gen;
                    $genId =~ s/.*id=([^,]*).*/$1/;
                    if ( $genId eq "" ) {
                        $logger->error( "Generator "
                              . $generatorDn
                              . " has no or malformed id. Therefore we skip it."
                        );
                        next;
                    }
                    if ( $generatorIds{$genId} ) {
                        $logger->error( "Generator "
                              . $generatorDn
                              . " has duplicate id "
                              . $genId
                              . " Therefore we skip it." );
                        next;
                    }

                    # Alle anderen Attribute sind nicht zwingend vorhanden und von Generator
                    # zu Generator eventuell unterschiedlich
                    my %genParams;
                    $gen =~ s/id=[^,]*,?//g;
                    while ( $gen =~ m/=/ ) {
                        my $genParamKey   = $gen;
                        my $genParamValue = $gen;
                        $genParamKey   =~ s/^([^=]*).*/$1/;
                        $genParamValue =~ s/^[^=]*=([^,]*).*/$1/;
                        $genParams{$genParamKey} = $genParamValue;
                        $gen =~ s/^[^,]*,?//;
                    }
                    $generatorIds{$genId} = \%genParams;
                }
            } else {
                $logger->error( "defect generator "
                      . $generatorDn
                      . " for searchbase "
                      . $searchbase );
                $formatstring = $defaultFormatstring;
                %generatorIds = %{$defaultGeneratorIds};
            }
        }
    } else {
        $logger->error( "no generator "
              . $generatorDn
              . " found for startdn  "
              . $searchbase );
        $formatstring = $defaultFormatstring;
        %generatorIds = %{$defaultGeneratorIds};

    }
    $logger->info( " Found generator: " . $formatstring );
    return (
        DEVICE_TYPE_DN         => $generatorDn,
        GENERATOR_FORMATSTRING => $formatstring,
        GENERATOR_IDS          => \%generatorIds
    );
}

=head2 ldap_get_next_available_printer_dn

searches next unused IP address in that location
and return the corresponding DN to that IP

=over 4
 
=item Parameter

	LDAP:			LDAP connection
	SEARCHBASE		DN within the location container
	ROOTDN			base DN (searches only up to this DN)
        GENERATOR_FORMATSTRING  Formatstring for naming of new printer (default 'pr${gsip}')
        GENERATOR_IDS           Generators to be filled in formatstring (default { 'gsip' => { type => 'IP', format => '%05d' } })
 
=item return

	DN for a new printer
	or undef on error
 
=back

=cut

sub ldap_get_next_available_printer_dn
{
    my %param = (
        LDAP                   => undef,
        SEARCHBASE             => undef,
        ROOTDN                 => undef,
        CONTAINER_OBJECTCLASS  => "scPrinterContainer",
        IP_RANGE_ATTRIBUTE     => "scPrinterIpRange",
        GENERATOR_FORMATSTRING => 'pr${gsip}',
        GENERATOR_IDS          => {
            'gsip' => {
                'type'   => 'IP',
                'format' => '%05d'
            }
        },
        @_,
    );

    return ldap_get_next_available_dn(%param);
}

=head2 ldap_get_next_available_dn

searches next unused IP address in that location
and return the corresponding DN to that IP

=over 4
 
=item Parameter

	LDAP:			LDAP connection
	SEARCHBASE		DN within the location container
	ROOTDN			base DN (searches only up to this DN)
	CONTAINER_OBJECTCLASS 	LDAP container
	IP_RANGE_ATTRIBUTE 	LDAP Attribute that holds the IP range
        GENERATOR_FORMATSTRING  Formatstring for naming of new object (default 'pc${gsip}')
        GENERATOR_IDS           Generators to be filled in formatstring (default { 'gsip' => { type => 'IP', format => '%05d' } })

=item return

	DN for a new printer
	or undef on error
 
=back

=cut

#TODO CONTAINER_OBJECTCLASS als Suchfilter ummodeln und überprüfen
#ob CONTAINER_OBJECTCLASS genutzt wird
sub ldap_get_next_available_dn
{
    my %param = (
        LDAP                   => undef,
        SEARCHBASE             => undef,
        ROOTDN                 => undef,
        CONTAINER_OBJECTCLASS  => "scWorkstationContainer",
        IP_RANGE_ATTRIBUTE     => "dbkDhcpFixedRange",
        DEVICETYPE_DN          => undef,
        ENTRY                  => undef,
        GENERATOR_FORMATSTRING => 'pc${gsip}',
        GENERATOR_IDS          => {
            'gsip' => {
                'type'   => 'IP',
                'format' => '%05d'
            }
        },
        @_,
    );

    my $ldap                   = $param{'LDAP'};
    my $searchbase             = $param{'SEARCHBASE'};
    my $rootdn                 = $param{'ROOTDN'};
    my $container_objectclass  = $param{'CONTAINER_OBJECTCLASS'};
    my $ip_range_attribute     = $param{'IP_RANGE_ATTRIBUTE'};
    my $devicetype_dn          = $param{'DEVICETYPE_DN'};
    my $entry                  = $param{'ENTRY'};
    my $generator_formatstring = $param{'GENERATOR_FORMATSTRING'};
    my %generator_ids          = %{ $param{'GENERATOR_IDS'} };

    my $logger = Log::Log4perl->get_logger();

    if ( !$ldap or !$searchbase ) {
        $logger->error("can't determine next available dn. missing parameter");
        return;
    }

    my $ip = ldap_get_next_available_ip(%param);
    if ( !defined($ip) ) {
        return;
    }

    #
    # get dhcp fixed range
    #
    my $locationEntry = ldap_get_location_entry(
        LDAP    => $ldap,
        STARTDN => $searchbase
    );
    my $network = $locationEntry->get_value('ipNetworkNumber');
    my $netmask = $locationEntry->get_value('ipNetmaskNumber');

    #my $cn = generatorHash{$devicetype_dn}->generateName(...);
    my $cn = ldap_generate_name(
        IP_ADDR                => $ip,
        NETWORK                => $network,
        NETMASK                => $netmask,
        GENERATOR_IDS          => \%generator_ids,
        GENERATOR_FORMATSTRING => $generator_formatstring,
        LDAP                   => $ldap,
        SEARCHBASE             => $searchbase,
        ENTRY                  => $entry,
        DEVICETYPE_DN          => $devicetype_dn,
    );
    $logger->info("Got $cn for $ip in $network with $generator_formatstring\n");

    # get relevant subcontainer and use first one
    my ($container) = getContainer(
        LDAP       => $ldap,
        SEARCHBASE => $searchbase,
        FILTER =>
          "(&(objectClass=$container_objectclass)(!(objectClass=scLocation)))",
        RESULT_FILTER => "(objectClass=$container_objectclass)"
    );

    return ( "cn=" . $cn . "," . $container, $ip, $param{DEVICE_TYPE_DN} );
}

=head2 Funktion ldap_generate_name


Generates a name to be used as C<cn> with the given generator

=over 4

=item parameter

	IP_ADDR assigned IP address for the LDAP object
	NETWORK assigned network for the LDAP object
	NETMASK assigned netmask for the LDAP object
	GENERATOR_IDS unparsed array  of generators to be used for name creation
	LDAP Connection
	SEARCHBASE base of network-ou where newly created object will be placed

=item return

	readily composed string for the creation of cn
	
=back

=cut

sub ldap_generate_name
{
    my %param = (
        IP_ADDR                => undef,
        NETWORK                => undef,
        NETMASK                => undef,
        ENTRY                  => undef,
        DEVICETYPE_DN          => undef,
        GENERATOR_IDS          => undef,
        GENERATOR_FORMATSTRING => undef,
        LDAP                   => undef,
        SEARCHBASE             => undef,
        @_,
    );

    #TODO use entry and devicetype to decide wether we should keep our old cn

    #
    # translate IP address to name
    #
    my $number;
    my $generatorFormatstring = $param{GENERATOR_FORMATSTRING};
    my %generatorIds          = %{ $param{GENERATOR_IDS} };
    my $uniqueGeneratorId;

    my $logger = Log::Log4perl->get_logger();

    # iterate over all given generators.
    foreach my $generatorId ( keys %generatorIds ) {
        my $generatorIdType   = uc( $generatorIds{$generatorId}{'type'} );
        my $generatorIdFormat = $generatorIds{$generatorId}{'format'};

        if ( $generatorIdType eq "IP" ) {
            $generatorIds{$generatorId}{'id'} = $generatorId;
            $generatorFormatstring = generate_name_IP(
                GENERATOR     => $generatorIds{$generatorId},
                FORMAT_STRING => $generatorFormatstring,
                IP_ADDR       => $param{IP_ADDR},
                NETWORK       => $param{NETWORK},
                NETMASK       => $param{NETMASK}
            );
            if ( !$generatorFormatstring ) {
                return;
            }
        } elsif ( $generatorIdType eq "LDAP" ) {
            $generatorIds{$generatorId}{'id'} = $generatorId;
            $generatorFormatstring = generate_name_LDAP(
                GENERATOR     => $generatorIds{$generatorId},
                FORMAT_STRING => $generatorFormatstring,
                LDAP          => $param{LDAP},
                SEARCHBASE    => $param{SEARCHBASE}
            );
            if ( !$generatorFormatstring ) {
                return;
            }
        } elsif ( $generatorIdType eq "XMLRPC" ) {
            $generatorIds{$generatorId}{'id'} = $generatorId;
            $generatorFormatstring = generate_name_XMLRPC(
                GENERATOR     => $generatorIds{$generatorId},
                FORMAT_STRING => $generatorFormatstring,
                SEARCHBASE    => $param{SEARCHBASE},
                IP_ADDR       => $param{IP_ADDR},
                NETWORK       => $param{NETWORK},
                NETMASK       => $param{NETMASK}
            );
            if ( !$generatorFormatstring ) {
                return;
            }
        } elsif ( $generatorIdType eq "UNIQUE" ) {

            # Save UNIQUE generator for the last step
            if ($uniqueGeneratorId) {
                $logger->error( "More than one UNIQUE Generator in "
                      . $generatorId
                      . " . This will most certainly not work.\n" );
            }
            $uniqueGeneratorId = $generatorId;
        }
    }
    if ($uniqueGeneratorId) {
        $logger->info( "Call generate_name_UNIQUE "
              . $uniqueGeneratorId . " => "
              . $generatorIds{$uniqueGeneratorId} );
        $generatorIds{$uniqueGeneratorId}{'id'} = $uniqueGeneratorId;
        return generate_name_UNIQUE(
            GENERATOR     => $generatorIds{$uniqueGeneratorId},
            FORMAT_STRING => $generatorFormatstring,
            SEARCHBASE    => $param{SEARCHBASE},
            LDAP          => $param{LDAP}
        );
    }

    return $generatorFormatstring;
}

sub generate_name_XMLRPC
{
    my %param = (
        GENERATOR     => undef,
        FORMAT_STRING => undef,
        @_
    );
    my $generatorFormatstring = $param{FORMAT_STRING};
    my $generatorId           = $param{GENERATOR}{'id'};
    my $generatorIdFormat     = $param{GENERATOR}{'format'};
    if ( !$generatorIdFormat ) {
        $generatorIdFormat = "%s";
    }

    my $logger = Log::Log4perl->get_logger();

    # Dieser Typ soll auf einem XML-RPC Server (wie etwa Zope) eine beliebige
    # Methode aufrufen. Als Parameter werden die Id, der Formatstring und ein
    # eventuell vorhandener rpcgeneratortype übergeben.
    # Vom XML-RPC Server muss die
    require RPC::XML;
    require RPC::XML::Client;
    my $client = RPC::XML::Client->new( $param{GENERATOR}{'url'} );
    if (   $param{GENERATOR}{'realm'}
        && $param{GENERATOR}{'user'}
        && $param{GENERATOR}{'pass'} )
    {
        $client->credentials(
            $param{GENERATOR}{'realm'},
            $param{GENERATOR}{'user'},
            $param{GENERATOR}{'pass'}
        );
    }

    # Standardwerte in die zu uebergebene Struktur schreiben
    my %struct;
    for my $myKey ( keys %{ $param{GENERATOR} } ) {
        $logger->info( 'XMLRPC Generator: ' 
              . $myKey . ' => '
              . $param{GENERATOR}{$myKey} );
        $struct{$myKey} = RPC::XML::string->new( $param{GENERATOR}{$myKey} );
    }

    # Alle anderen Werte in die uebergebenen Struktur
    $struct{'generatorId'}       = RPC::XML::string->new($generatorId);
    $struct{'generatorIdFormat'} = RPC::XML::string->new($generatorIdFormat);
    $struct{'generatorIdFormatstring'} =
      RPC::XML::string->new($generatorFormatstring);
    $struct{'ldapSearchBase'} = RPC::XML::string->new( $param{SEARCHBASE} );
    $struct{'ipAddress'}      = RPC::XML::string->new( $param{IP_ADDR} );
    $struct{'network'}        = RPC::XML::string->new( $param{NETWORK} );
    $struct{'netmask'}        = RPC::XML::string->new( $param{NETMASK} );

    $logger->info( "Try to ask XMLRPC Server for: "
          . $struct{'generatorId'}->as_string() . "/"
          . $struct{'generatorIdFormatstring'}->as_string() );

    my $response = $client->send_request( $param{GENERATOR}{'method'},
        RPC::XML::struct->new(%struct) );

    if ($response) {
        if ( !${ $response->value }{'name'} ) {
            my $reason = "No reason given";
            if ( ${ $response->value }{'reason'} ) {
                $reason = ${ $response->value }{'reason'};
            }
            $logger->error( "XMLRPC gave no name for "
                  . $struct{'ipAddress'}->as_string() . " in "
                  . $struct{'ldapSearchBase'}->as_string()
                  . " [ reason="
                  . $reason . ", id="
                  . $struct{'generatorId'}->as_string()
                  . ", format="
                  . $struct{'generatorIdFormat'}->as_string()
                  . ", Formatstring="
                  . $struct{ 'generatorIdFormatstring'->as_string() }
                  . " ]" );
            return;
        } else {
            $generatorFormatstring = ${ $response->value }{'name'};
            $logger->info( "XMLRPC answer: "
                  . $generatorFormatstring . " [ "
                  . %{ $response->value }
                  . " ] " );
        }
    } else {
        $logger->error( "XMLRPC call gave no answer for URL:"
              . $param{GENERATOR}{'url'}
              . " method:"
              . $param{GENERATOR}{'method'} );
        return;
    }
    return $generatorFormatstring;
}

sub generate_name_IP
{
    my %param = (
        GENERATOR     => undef,
        FORMAT_STRING => undef,
        IP_ADDR       => undef,
        NETWORK       => undef,
        NETMASK       => undef,
        @_
    );
    my $generatedFormatstring = $param{FORMAT_STRING};
    my $generatorId           = $param{GENERATOR}{'id'};
    my $generatorIdFormat     = $param{GENERATOR}{'format'};

    # Old smartclient algorithm
    my $number = ldap_ip2offset(
        IP_ADDR => $param{IP_ADDR},
        NETWORK => $param{NETWORK},
        NETMASK => $param{NETMASK}
    );

    if ( !defined($number) ) {
        return;
    }
    if ( !$generatorIdFormat ) {
        $generatorIdFormat = '%05d';
    }
    $generatedFormatstring =~ s/\$\{\Q$generatorId\E\}/$generatorIdFormat/;
    $generatedFormatstring = sprintf $generatedFormatstring, $number;

    return $generatedFormatstring;
}

#TODO Felix meint funktioniert nicht. Und er muß es wissen
sub generate_name_LDAP
{
    my %param = (
        GENERATOR     => undef,
        FORMAT_STRING => undef,
        LDAP          => undef,
        SEARCHBASE    => undef,
        @_
    );
    my $generatedFormatstring    = $param{FORMAT_STRING};
    my $generatorId              = $param{GENERATOR}{'id'};
    my $generatorIdFormat        = $param{GENERATOR}{'format'};
    my $generatorIdLdapAttribute = $param{GENERATOR}{'attribute'};

    #Find matching ldap attribute
    my $searchAttr;
    if ( $param{GENERATOR}{'attribute'} ) {
        $searchAttr = $param{GENERATOR}{'attribute'};
    } else {
        $searchAttr = $generatorId;
    }

    my $ldapEntry = ldap_treesearch_by_attr(
        LDAP    => $param{LDAP},
        ATTR    => $searchAttr,
        STARTDN => $param{SEARCHBASE},
        DEBUG   => 1
    );

    if ( !defined $ldapEntry ) {
        return;
    }
    $generatedFormatstring =~ s/\$\{\Q$generatorId\E\}/$generatorIdFormat/;
    my $number = $ldapEntry->get_value($generatorIdLdapAttribute);
    $generatedFormatstring = sprintf( $generatedFormatstring, $number );
    return $generatedFormatstring;
}

sub generate_name_UNIQUE
{
    my %param = (
        GENERATOR     => undef,
        FORMAT_STRING => undef,
        LDAP          => undef,
        SEARCHBASE    => undef,
        @_
    );

    my $logger = Log::Log4perl->get_logger();

    my $generatorFormatstring = $param{FORMAT_STRING};
    my $config                = SmartClient::Config->new();
    if ( $config->get('ALL_GENERATOR_UNIQUE_NAME_XMLRPC_URL', 1) ) {
        my $xmlrpcUrl = $config->get('ALL_GENERATOR_UNIQUE_NAME_XMLRPC_URL');
        $param{GENERATOR}{'url'}    = $xmlrpcUrl;
        $param{GENERATOR}{'method'} = "generator.uniqueName";
        if ( $config->get('ALL_GENERATOR_UNIQUE_NAME_XMLRPC_METHOD', 1) ) {
            $param{GENERATOR}{'method'} =
              $config->get('ALL_GENERATOR_UNIQUE_NAME_XMLRPC_METHOD');
        }
        $logger->info( "get generated unique name from xmlrpc "
              . $param{GENERATOR}{'url'} . " by "
              . $param{GENERATOR}{'method'} );
        return generate_name_XMLRPC(
            GENERATOR     => $param{GENERATOR},
            SEARCHBASE    => $param{SEARCHBASE},
            FORMAT_STRING => $param{FORMAT_STRING}
        );
    }
    my $generatorIdFormat = $param{GENERATOR}{'format'};
    my $generatorId       = $param{GENERATOR}{'id'};
    if ( !$generatorIdFormat ) {
        $logger->error( "no format for generator set. using \%04d for "
              . $generatorId . " => "
              . $param{GENERATOR} );
        $generatorIdFormat = '%04d';
    }
    my $generatorIdMin = $param{GENERATOR}{'min'};
    my $generatorIdMax = $param{GENERATOR}{'max'};
    $logger->info( "Result is " . $generatorFormatstring );
    if ( !$generatorIdMin ) {
        $generatorIdMin = 1;
    }
    if ( !$generatorIdMax ) {
        $generatorIdMax = 10000;
    }
    my $uniqueId = ldap_get_next_unique_id(
        GENERATOR_ID           => $generatorId,
        GENERATOR_ID_MIN       => $generatorIdMin,
        GENERATOR_ID_MAX       => $generatorIdMax,
        GENERATOR_ID_FORMAT    => $generatorIdFormat,
        GENERATOR_FORMATSTRING => $generatorFormatstring,
        LDAP                   => $param{LDAP}
    );
    if ( !defined($uniqueId) ) {
        return;
    }
    return $uniqueId;
}

=head2 Funktion ldap_get_next_unique_id

searches for the first free number in range given by GENERATOR_MON 
and GENERATOR_MAX with namespace GENERATOR_ID.
To accomplish this, a LDAP search string is generated. For this the GENERATOR_ID in the 
format string is replaced by C<*>. Then all C<cn> are searched for and stored in a hash. 
Lastly the first free unique C<cn> is searched by iterating from C<GENERATOR_MIN> to 
C<GENERATOR_MAX> and inserting this id into the original format string.

Example:
 	format string: rl1234${ziffer}
 	GENERATOR_ID: ziffer
 	generated search filter: cn=rl1234* 

=over 4

=item Parameter

	GENERATOR_ID namespace for generator
	GENERATOR_MIN minimal number for searchstart
	GENERATOR_MAX maximum number. 
	GENERATOR_FORMAT special format string to be used for name generation
	SEARCH_BASE
	FILTER
	LDAP Connection

=item return

	next unused number for generator namespace identified by GENERATOR_ID

=back

=cut

sub ldap_get_next_unique_id
{
    my %param = (
        GENERATOR_ID                => undef,
        GENERATOR_ID_MIN            => undef,
        GENERATOR_ID_MAX            => undef,
        GENERATOR_ID_FORMAT         => undef,
        GENERATOR_ID_ATTRIBUTE_NAME => 'cn',
        GENERATOR_FORMATSTRING      => undef,
        LDAP                        => undef,
        SEARCH_BASE                 => "c=de",
        FILTER                      => "objectClass=*",
        @_,
    );
    my $ldap          = $param{LDAP};
    my $generatorId   = $param{GENERATOR_ID};
    my $attributeName = $param{GENERATOR_ID_ATTRIBUTE_NAME};

    my $logger = Log::Log4perl->get_logger();

    # LDAP Filter bei dem die Generator Id durch * ersetzt wird.
    my $filter = "(&("
      . $param{FILTER} . ")("
      . $attributeName . "="
      . $param{GENERATOR_FORMATSTRING} . "))";
    $filter =~ s/\$\{\Q$generatorId\E\}/*/;

    # printf Format, für die Erzeugung der
    my $generatorPrintFormat = $param{GENERATOR_FORMATSTRING};
    $generatorPrintFormat =~
      s/\$\{\Q$generatorId\E\}/$param{GENERATOR_ID_FORMAT}/;
    $logger->info(
        "Using " . $filter . " for ldap search in " . $param{SEARCH_BASE} );

    my $ldap_result = $ldap->search(
        base   => $param{SEARCH_BASE},
        scope  => 'sub',
        filter => $filter
    );
    my %id_hash;
    if ( $ldap_result->count > 0 ) {
        foreach my $entry ( $ldap_result->all_entries ) {
            if ( $entry->exists($attributeName) ) {
                if ( !$id_hash{ $entry->get_value($attributeName) } ) {
                    $id_hash{ $entry->get_value($attributeName) } = 1;
                } else {

                    # Wenn dieser Code ausgeführt wird, so bedeutet das, dass dieser cn nicht
                    # eindeutig in der Suchbasis ist.
                    ++$id_hash{ $entry->get_value($attributeName) };
                    $logger->error( "CN "
                          . $entry->get_value($attributeName)
                          . " from "
                          . $entry->dn
                          . " is not unique in "
                          . $param{SEARCH_BASIS}
                          . "." );
                }
            }
        }
    }
    my $minId = $param{GENERATOR_ID_MIN};
    my $maxId = $param{GENERATOR_ID_MAX};
    my $uniqueId;
    for ( my $counter = $minId ; $counter < $maxId ; $counter++ ) {
        $uniqueId = sprintf( $generatorPrintFormat, $counter );
        if ( !$id_hash{$uniqueId} ) {
            $logger->info( "Will use " . $uniqueId . " as unique id." );
            return $uniqueId;
        }
    }
    $logger->error( "Couldn't find a unique Id for generator "
          . $param{GENERATOR_ID}
          . " in base "
          . $param{SEARCH_BASE}
          . " for range ["
          . $minId . ","
          . $maxId
          . "]  with formatstring "
          . $generatorPrintFormat );
    return;
}

=head2 get_next_available_ip

searches between START and END for a number 
not already defined in USED_NUMBERS
Used to generate a new entry from leases2ldap or Webmin

=over 4
 
=item Parameter

	USED_IP_HREF 	hash reference of all IPs in use
	START_IP	start of range (IP number)
	END_IP		end of range ((IP number)
	NETWORK		network IP 
			(needed to calculate numbers from IPs)
	NETMASK		network netmask
			(needed to calculate numbers from IPs)
 
=item return

	next unused IP between START_IP and END_IP
	or undef otherwise
 
=back

=cut

sub get_next_available_ip
{
    my %param = (
        USED_IP_HREF => undef,
        START_IP     => undef,
        END_IP       => undef,
        NETWORK      => undef,
        NETMASK      => undef,
        @_,
    );

    my $logger = Log::Log4perl->get_logger();

    foreach my $key ( keys %param ) {
        if ( !defined $param{$key} ) {
            $logger->error(
                "can't determine next available ip. missing parameter $key");
            return;
        }
    }

    my $used_ip_href = $param{'USED_IP_HREF'};
    my $network = new2 Net::Netmask( $param{'NETWORK'}, $param{'NETMASK'} );
    if ( !$network ) {
        $logger->error(
            "get_next_available_ip: invalid network definition. ($param{'NETWORK'}, $param{'NETMASK'})"
        );
        return;
    }

    my $end_number = $network->match( $param{'END_IP'} );
    my $number     = $network->match( $param{'START_IP'} );

    my $available_ip;
    while ( ( $number <= $end_number ) && !$available_ip ) {
        my $ip = $network->nth($number);
        if ( !$used_ip_href->{$ip} ) {
            $available_ip = $ip;
        }
        $number++;
    }
    return $available_ip;
}

=head2 ldap_ip2offset

translates a ip address to offset.
Used for naming LDAP entries (workstations and printers)

obsoletes:
	LdapWorkstation: ip_to_ws_number

=cut

sub ldap_ip2offset
{
    my %param = (
        IP_ADDR => undef,
        NETWORK => undef,
        NETMASK => undef,
        @_,
    );

    my $ip_addr = $param{'IP_ADDR'};
    my $network = $param{'NETWORK'};
    my $netmask = $param{'NETMASK'};

    my $logger = Log::Log4perl->get_logger();

    if ( !$ip_addr || !$network || !$netmask ) {
        $logger->error(
            "ldap_ip2offset: IP address ($ip_addr) or network number ($network) or netmask ($netmask) is not defined"
        );
        return;
    }

    my $ip_net = new2 Net::Netmask( $network, $netmask );
    if ( !$ip_net ) {
        $logger->error(
            "ldap_ip2offset: network number ($network) or netmask ($netmask) is invalid"
        );
        return;
    }

    # match return the position of an IP in a network
    # (0 if not in network)
    return $ip_net->match($ip_addr);
}

=head2 ldap_ip2number

generates IP address from network address and offset
(used for workstations and printers)

=cut

sub ldap_offset2ip
{
    my %param = (
        OFFSET  => undef,
        NETWORK => undef,
        NETMASK => undef,
        @_,
    );

    my $offset  = $param{'OFFSET'};
    my $network = $param{'NETWORK'};
    my $netmask = $param{'NETMASK'};

    my $logger = Log::Log4perl->get_logger();

    if ( !$offset || !$network || !$netmask ) {
        $logger->error(
            "ldap_ip2number: IP offset ($offset) or network number ($network) or netmask ($netmask) is not defined"
        );
        return;
    }

    my $ip_net = new2 Net::Netmask( $network, $netmask );
    if ( !$ip_net ) {
        $logger->error(
            "ldap_ip2number: network number ($network) netmask ($netmask) is invalid"
        );
        return;
    }

    return $ip_net->nth($offset);
}

=head2 ldap_entry2ip
 
Ermittelt die IP-Adresse eines LDAP Entries.
Entweder aus der definierten ipHostNumber oder ansonsten
nach der SmartClient Namenskonvention
 
=over 4
 
=item Parameter

 ENTRY		LDAP entry, dessen IP Adresse ermittelt werden soll
 [NETWORK]	zu ENTRY gehöriges Netzwerk. 
 		Wird benötigt, falls ipHostNumber nicht definiert ist.
		Kann ggf. aus LOC_ENTRY ermittelt werden.
 [NETMASK]	
 [LOC_ENTRY] 	der scLocation LDAP Entry zu dem ENTRY gehört
 [WARNINGS]     wenn gesetzt, wird im Fall 
                das keine IP Adresse ermittelt werden konnte, 
                eine Warnung ins Syslog geschrieben
                default ist an
 
=item Return

 IP-Adresse des Geraetes oder undef, wenn ein Fehler aufgetreten ist.
 
=back
 
=cut

sub ldap_entry2ip
{
    my %param = (
        ENTRY     => undef,
        NETWORK   => undef,
        NETMASK   => undef,
        LOC_ENTRY => undef,
        WARNINGS  => 1,
        @_,
    );

    my $entry          = $param{'ENTRY'};
    my $network        = $param{'NETWORK'};
    my $location_entry = $param{'LOC_ENTRY'};

    my $ipaddr;

    my $logger = Log::Log4perl->get_logger();

    if ( $entry->exists('ipHostNumber') ) {
        $ipaddr = $entry->get_value('ipHostNumber');
    } else {
        if ( !$network and $location_entry ) {
            $network = $location_entry->get_value('ipNetworkNumber');
        }
        if (
            lc( ( scalar $entry->get_value('dbkNoDNSName') ) || 'false' ) eq
            'true' )
        {

            #this device needs no ip address
            return;
        }
        $ipaddr = ldap_cn2addr( $entry->get_value('cn'), $network );
    }
    if ( !$ipaddr && $param{'WARNINGS'} ) {
        $logger->error(
            "ldap_entry2ip: could not get IP address for " . $entry->dn );
    }

    return $ipaddr;
}

=head2 Funktion get_location_baselist

Generieren einer Liste aus bases, die die Objektklasse scLocation 
haben. Die Objektklasse scLocation ist neu und dient dazu, Container
als Lokation zu definieren.

=over 4

=item Parameter

 - Root-Base

=item Rückgabewert

Liste der Bases mit ObjectClass=scLocation

=back

=cut

sub get_location_baselist
{
    my %param = (
        LDAP       => undef,
        SEARCHBASE => undef,
        @_
    );

    my $ldap      = $param{LDAP};
    my $root_base = $param{SEARCHBASE};
    my $config    = SmartClient::Config->new();

    my @baselist = ();

    my $result = $ldap->search(
        timelimit => $config->get('ALL_LDAPTIMEOUT'),
        base      => $root_base,
        scope     => 'sub',
        filter    => '(objectClass=scLocation)'
    );
    $result->code && die $result->error;

    foreach my $entry ( $result->entries ) {
        push( @baselist, $entry->dn );
    }

    return @baselist;
}

sub getContainer
{
    my %param = (
        LDAP       => undef,
        SEARCHBASE => undef,
        FILTER =>
          '(&(|(objectClass=scWorkstationContainer)(objectClass=scPrinterContainer)(objectClass=scServerContainer))(!(objectClass=scLocation)))',
        RESULT_FILTER => 'objectClass=*',
        @_
    );

    my $ldap          = $param{LDAP};
    my $filter        = $param{FILTER};
    my $result_filter = $param{RESULT_FILTER};

    my @container = ( $param{SEARCHBASE} );
    my @result_container;

    my $logger = Log::Log4perl->get_logger();

    # get all sub container,
    # as long as they are no other locations (own network)
    while (@container) {
        my $ldap_result;
        my $searchbase = shift @container;

        # check is this searchbase is also a result
        $ldap_result = $ldap->search(
            base   => $searchbase,
            scope  => 'base',
            filter => $result_filter
        );
        if ( $ldap_result and ( $ldap_result->count > 0 ) ) {
            push( @result_container, $searchbase );
        }

        # check for subcontainer
        $ldap_result = $ldap->search(
            base   => $searchbase,
            scope  => 'one',
            filter => $filter
        );
        if ( !$ldap_result ) {
            $logger->error("no result, while looking for sub container");
            return;
        }
        foreach my $entry ( $ldap_result->all_entries ) {
            push( @container, $entry->dn );
        }
    }

    return @result_container;
}

sub getContainerIPs
{
    my %param = (
        LDAP       => undef,
        SEARCHBASE => undef,
        NETWORK    => undef,
        NETMASK    => undef,
        @_,
    );

    my $logger = Log::Log4perl->get_logger();

    my $ldap   = $param{'LDAP'};
    my $result = $ldap->search(
        base  => $param{'SEARCHBASE'},
        scope => 'one',
        filter =>
          '(|(objectClass=dbkWorkstation)(objectClass=dbkPrinter)(objectClass=dbkServer)(objectClass=dbkService))',
        attrs => [ 'cn', 'ipHostNumber' ]
    );
    if ( !$result ) {
        $logger->error("no result, while looking for IP addresses");
        return;
    }

    my %ip_hash = ();
    foreach my $entry ( $result->all_entries ) {
        my $ip = ldap_entry2ip(
            'ENTRY'   => $entry,
            'NETWORK' => $param{'NETWORK'},
            'NETMASK' => $param{'NETMASK'}
        );
        if ($ip) {
            $ip_hash{$ip} = $entry;
        } else {
            $logger->error( "no ip address defined for " . $entry->dn() );
        }
    }
    return \%ip_hash;
}

sub getLocationIPs
{
    my %param = (
        LDAP       => undef,
        SEARCHBASE => undef,
        NETWORK    => undef,
        NETMASK    => undef,
        @_,
    );

    my $ldap = $param{'LDAP'};

    # get all relevant subcontainer (scWorkstationContainer, ...)
    my @subcontainer = getContainer(
        LDAP       => $ldap,
        SEARCHBASE => $param{'SEARCHBASE'}
    );

    my %ipHash;
    foreach my $container_dn (@subcontainer) {

        # fill ipHash with IPs from each subcontainer
        my $ipHRef = getContainerIPs(
            LDAP       => $ldap,
            SEARCHBASE => $container_dn,
            NETWORK    => $param{'NETWORK'},
            NETMASK    => $param{'NETMASK'}
        );
        %ipHash = ( %ipHash, %$ipHRef );
    }

    return \%ipHash;
}

sub getWorkstationDefaultAttributes
{
    my $dn = shift @_;

    my $config = SmartClient::Config->new();

    my %attributes = (
        'objectClass'     => [ 'dbkCarrierClass', 'dbkWorkstation' ],
        'dbkAuthorName'   => $0,
        'dbkImageVersion' => 0,
        'dbkStatus'       => "0",
        'dbkCachingImageServer' =>
          uc( $config->get('ALL_CACHING_IMAGE_SERVER_DEFAULT_VALUE') )
    );

    if ($dn) {
        $attributes{'cn'} = ldap_get_cn_from_dn($dn);
    }

    return %attributes;
}

sub getPrinterDefaultAttributes
{
    my $dn = shift @_;

    my $config = SmartClient::Config->new();

    my %attributes = (
        'objectClass'   => [ 'dbkCarrierClass', 'dbkPrinter' ],
        'dbkAuthorName' => $0
    );

    if ($dn) {
        $attributes{'cn'} = ldap_get_cn_from_dn($dn);
    }

    return %attributes;
}

sub ldap_chk_tstamp
{
    my %param = (
        LDAP        => undef,
        LAST_TSTAMP => undef,
        FILTER      => undef,
        BASE        => undef,
        DEBUG       => 0,
        @_,
    );

    my $min_tstamp  = ldap_get_min_tstamp();
    my $last_tstamp = $param{'LAST_TSTAMP'};
    my $ldap        = $param{'LDAP'};
    my $filter      = $param{'FILTER'};
    my $base        = $param{'BASE'};

    #print STDERR "param tstamp: ", $param{'LAST_TSTAMP'}, "\n"; #debug

    ## add timestamp condition to given filter
    my $tstamp_filter = ldap_add_timestamp_filter(
        filter    => $filter,
        timestamp => $last_tstamp
    );

    #print STDERR "tstamp_filter: $tstamp_filter\n"; #debug

    ## now do the search
    my $res = $ldap->search(
        base   => $base,
        filter => $tstamp_filter,
        attrs  => ['modifyTimestamp']
    );
    if ( $res->code and ( $res->code != LDAP_NO_SUCH_OBJECT ) ) {
        ## LDAP error: in this case it can't be determined if
        ## anything changed by timestamp
        ## so generate a timestamp from actual system time
        ## and return that.
        ldap_check_error( $res, 0 );
        return ( ldap_get_min_tstamp() );
    }
    if ( $res->count == 0 ) {
        ## no new data
        return ( $param{'LAST_TSTAMP'} );
    }
    return ( ldap_get_max_tstamp($res) );

}

=head2 ldap_filter_add_timestamp

Add a modifyTimestamp expression to a given filter. Because LDAP filter
expressions only support greater or equal as comparison operator, the
given timestamp is increased by one second for the new filter.

LDAP timestamps have the following format:

 YYYYMMDDhhmmssZ

for example:

 20060424135124Z

Example:

 my $timestamp_filter = ldap_filter_add_timestamp(
     filter => '(objectClass=*)',
	 timestamp => $last_timestamp
 );

=over 4

=item Parameter

 - filter: LDAP filter expression
 - timestamp: LDAP modifyTimestamp value

=item Return

LDAP filter expression with modifyTimestamp comparison

=back

=cut

sub ldap_filter_add_timestamp
{
    my %param = (
        filter    => undef,
        timestamp => undef,
        @_,
    );

    my $last_tstamp = $param{'timestamp'};
    my $filter      = $param{'filter'};

    #print STDERR "param tstamp: ", $param{'LAST_TSTAMP'}, "\n"; #debug

    ## add timestamp condition to given filter
    #require Net::LDAP::Filter;
    #use Time::localtime;
    #print STDERR "filter: $filter\n"; #debug
    my $ldap_filter = Net::LDAP::Filter->new($filter);
    my ($firstkey)  = keys %{$ldap_filter};
    my $filter_tmp  = {};
    $filter_tmp->{$firstkey} = $ldap_filter->{$firstkey};
    delete $ldap_filter->{$firstkey};

    #print STDERR "last_tstamp: $last_tstamp\n"; #debug
    push(
        @{ $ldap_filter->{'and'} },
        {
            'not' => {
                'lessOrEqual' => {
                    'assertionValue' => $last_tstamp,
                    'attributeDesc'  => "modifyTimestamp"
                }
            }
        },
        $filter_tmp
    );
    my $tstamp_filter = $ldap_filter->as_string();
    $ldap_filter = undef;

    return ($tstamp_filter);
}

sub gen_tstamp
{
    my $tm     = gmtime;
    my $tstamp = sprintf(
        "%04d%02d%02d%02d%02d%02dZ",
        $tm->year + 1900,
        $tm->mon + 1,
        $tm->mday, $tm->hour, $tm->min, $tm->sec
    );
    return ($tstamp);

}

sub ldap_get_min_tstamp
{

    # minimal timestamp (1970-01-01 00:00:00)
    # older version (OpenLDAP 2.2) had 10010101000000
    # but that doesn't work with OpenLDAP 2.3
    # which return 0 entries in this case

    return ('19700101000000Z');
}

sub ldap_get_max_tstamp
{
    my $ldap_res = shift;
    my @tstamps;
    my $max_tstamp = "";
    foreach my $entry ( $ldap_res->all_entries ) {
        my $act_tstamp = $entry->get_value('modifyTimestamp');
        if ( !$act_tstamp ) {
            $act_tstamp = ldap_get_min_tstamp();
        }
        if ( $act_tstamp gt $max_tstamp ) {
            $max_tstamp = $act_tstamp;
        }
    }
    return ($max_tstamp);
}

=head2 calcTimestampWithOffset

calculates a timestamp for ldap relative to a given one

=head3 Example

  my $new_timestamp = calcTimestampWithOffset('timestamp'=>'20060102100503Z', 'offset' => -2);

=over 4

=item Parameter

hash consisting of:

C<timestamp> starting time in ldap timestamp format (default: null)

C<offset> interval in s which should be added (default: null)

=item Return

string containing new timestamp

=back

=cut

sub calcTimestampWithOffset
{
    my %params = (
        'timestamp' => undef,
        'offset'    => undef,
        @_
    );

    my $logger    = Log::Log4perl->get_logger();
    my $timestamp = $params{'timestamp'};
    my $offset    = $params{'offset'};

    #example timestamp 20060609060358Z
    #                   YYYYMMDDHHMMSSZ
    my ( $year, $month, $day, $hour, $minute, $second ) =
      ( $timestamp =~ m/^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})Z$/ );
    use Time::Local;
    use POSIX qw(strftime);
    my $time_t =
      timegm( $second, $minute, $hour, $day, $month - 1, $year - 1900 );
    my $new_timestamp =
      strftime( "%Y%m%d%H%M%SZ", @{ gmtime( $time_t + $offset ) } );
    return $new_timestamp;
}

=head2 function ldap_get_attribute_syntax

ermittelt den Typ eines LDAP Attributs anhand des LDAP Schemas.
Bei abgeleiteten Typen kann es hierfür notwendig sein, die übergeordneten Typen zu betrachten.

 
=over 4
 
=item Parameter

    LDAP | SCHEMA   LDAP connection 
                    or (if available) schema of the LDAP connection
    ATTR            function return LDAP syntax for this attribut name
    INTERNAL_ATTR   determines if the syntax of internal attributes
                    is also returned. 
                    Normally this is not needed, 
                    because these attributes 
                    can not be modified by the user.
                    default is off
    DEBUG           if true, debug information is written to STDOUT
                    default is off
 
=item Return

syntax of the LDAP attribut, 
undef if syntax can't be determined
 
=back
 
=cut

sub ldap_get_attribute_syntax
{
    my %param = (
        LDAP          => undef,
        SCHEMA        => undef,
        ATTR          => undef,
        INTERNAL_ATTR => 0,
        DEBUG         => undef,
        @_,
    );

    my $ldap   = $param{'LDAP'};
    my $schema = $param{'SCHEMA'};
    my $attr   = $param{'ATTR'};
    my $debug  = $param{'DEBUG'};

    my $logger = Log::Log4perl->get_logger();

    if ( !$schema && $ldap ) {
        $schema = $ldap->schema();
    }

    if ( !$schema ) {
        $logger->error( "ldap_get_attribute_syntax: SCHEMA not defined: "
              . Dumper(%param) );
        return;
    }

    my $attribute_href = $schema->attribute($attr);

    #if($debug) { print Dumper($attribute_href); }
    if ($attribute_href) {
        if ( ( !$param{'INTERNAL_ATTR'} )
            && $attribute_href->{'no-user-modification'} )
        {
            if ($debug) { print " internal attribute. skipped\n"; }
            return;
        }
        if ( $attribute_href->{'syntax'} ) {
            return $attribute_href->{'syntax'};
        } else {
            foreach my $i ( @{ $attribute_href->{'sup'} } ) {
                if ($debug) { print "sup: $i\n"; }
                my $syntax = ldap_get_attribute_syntax(
                    SCHEMA => $schema,
                    ATTR   => $i,
                    DEBUG  => $debug
                );
                if ($syntax) {
                    return $syntax;
                }
            }
        }
    }
}

=head2 function ldap_attribute_is_dn

ermittelt anhand des LDAP Schemas ob ein LDAP Attribute vom Typ "distinguishedName" (DN) ist.
 
=over 4
 
=item Parameter

(see function ldap_get_attribute_syntax)
 
=item Return

true, if attribute has type distinguishedName
undef, otherwise
 
=back
 
=cut

sub ldap_attribute_is_dn
{
    my %param = (
        LDAP          => undef,
        SCHEMA        => undef,
        ATTR          => undef,
        INTERNAL_ATTR => 0,
        DEBUG         => undef,
        @_,
    );

    my $syntax = ldap_get_attribute_syntax(%param);
    if ( $syntax && ( $syntax eq '1.3.6.1.4.1.1466.115.121.1.12' ) ) {
        return 1;
    }
}

=head2 ldap_getListOfDNsThatMatchesFilter( % )
 
=over 4
 
=item Parameter
 
=item Return
 
=back
 
=cut

sub ldap_getListOfDNsThatMatchesFilter( % )
{
    my %params = (
        'ldap'          => undef,
        'filter'        => '(objectclass=*)',
        'aref_base_DNs' => undef,
        @_
    );

    my $ldap = $params{'ldap'};
    if ( !ref($ldap) ) {
        return;
    }

    my $filter = $params{'filter'};

    # use hash to get list of unique elements
    my %result_container;
    foreach my $dn ( @{ $params{'aref_base_DNs'} } ) {
        my $rv = $ldap->search(
            base   => $dn,
            filter => $filter,
            attr   => ['cn'],
        );
        if ( !$rv->code ) {

            # use hash to get list of unique elements
            foreach my $entry ( $rv->all_entries ) {
                $result_container{ $entry->dn() }++;
            }
        }
    }
    return ( keys %result_container );
}

=head2 isInSubtree

checks, if a DN is part of a subtree of given DNs
 
=over 4
 
=item Parameter

    dn                  dn to be checked
    aref_sub_locations  list if DNs, which are checked, 
                            if DN is in a subtree to them
 
=item Return

1, if dn belongs to a location from aref_sub_locations
0, otherwise
 
=back
 
=cut

sub isInSubtree
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'dn'                 => undef,
        'aref_sub_locations' => [],
        @_,
    );

    my $dn            = $params{'dn'};
    my @sub_locations = @{ $params{'aref_sub_locations'} };

    foreach my $location (@sub_locations) {
        if ( $dn =~ m/$location$/ ) {
            return 1;
        }
    }
    return 0;
}

1;

