#!/usr/bin/perl -w
use strict;

#
# package name
#
package SmartClient::Leases2ldap;

#
# version  of the package
#
our $VERSION = 0.1;

=encoding utf8

=head1 NAME

C<SmartClient::Leases2ldap>

=head2 SVN Version

 $Id: Leases2ldap.pm 9817 2014-09-22 11:24:05Z HZuchel $

=head1 SYNOPSIS

class for the SmartClient leases2ldap functionality


=head1 DESCRIPTION

=head2 Purpose

class for the SmartClient leases2ldap functionality,
including 
- read DHCP leases from leases file
- handling list of dislocated devices
- performing action for dislocated devices (add, move, warn)


=head2 Realisation

=head2 Error handling

=head2 Context and Dependencies

=over 4

=item Context

=item Dependencies


=back



=head2 Interfaces

=begin comment

    Command line parameter are already descripied 
    at SYNOPSIS and usage

=end comment

=over 4

=item Programm

=item User

=back



=head2 Status

=begin comment

    describes the TODOs/Errors for this programm

=end comment


=head2 Installation and other Notes



=head1 AUTHOR

    smartclient@dass-it.de




=head1 SEE ALSO



=cut

#
# module definition
#

use Exporter;
our @ISA = qw( Exporter );

# =head1 symbols exported by default
#
# - variable1
# - function1
# - function2
#
# =cut
#
#our @EXPORT = qw(  );

# =head1 symbols exported on request
#
# =cut
#
#our @EXPORT_OK = qw( );

#
# used modules
#
use Data::Dumper;
use Log::Log4perl;
use Cache::FastMmap;
use Net::LDAP::Entry;
use SmartClient::Config;
use SmartClient::Constant;
use SmartClient::Util;
use SmartClient::DhcpEntries qw(read_dhcp_leases clear_dhcp_leases);
use SmartClient::LdapUtil;
use SmartClient::LDAP;
use SmartClient::ScCache;
use SmartClient::DislocatedDevicePermission;

#use Time::Local;
use POSIX qw(strftime);

#
# declarations
#

=head1 Constants


=head1 Functions

=head2 new

constructor
create
    dislocated devices
    ...

    Could die, if requirement are not fullfilled

=over 4

=item Parameter

    - [logger]
    - config
    - sc_ldap
    - sc_cache


=item Return

    object reference or undef on error

=back

=cut

sub new
{
    my $package = shift;

    # Definition of defaults
    my $data = {

        # reference to SmartClient::Config object
        'config' => undef,

        # SmartClient::LDAP reference (LDAP connection)
        'sc_ldap' => undef,

        # SmartClient::ScCache reference
        # (SmartClient LDAP entry cache)
        'sc_cache' => undef,

        # internal
        'leases_file_mtime' => 0,
        @_
    };

    if ( init($data) ) {

        # return the object-hash
        return bless( $data, $package );
    } else {
        return (undef);
    }
}

=head2 init

checks requirements, parameter
and creates dislocated devices FastMmap object

=over 4

=item Parameter

    same as "new"

=item Return

    boolean

=back

=cut

sub init
{
    my $data = shift;

    # check parameters

    $data->{'logger'} = Log::Log4perl->get_logger();

    my @undefined_parameters =
      getUndefinedHashKeys( $data,
        [ 'logger', 'config', 'sc_ldap', 'sc_cache' ] );

    if (@undefined_parameters) {
        $data->{'logger'}
          ->logdie( "SmartClient::leases2ldap constructor requires parameter "
              . join( ", ", @undefined_parameters ) );
    }

    # checks if config variables exists, otherwise dies
    $data->{'config'}->require(
        'ALL_LDAPSERVER_TYPE',
        'LEASES2LDAP_LEASES_FILE',
        'LEASES2LDAP_CLEARED_LEASES_FILE',
        'LEASES2LDAP_STOP_DHCPD',
        'LEASES2LDAP_START_DHCPD',
        'LEASES2LDAP_MMAP_FILE_DISLOCATED_DEVICES',
        'LEASES2LDAP_LEASES_EXPIRE_AFTER',
        'LEASES2LDAP_LEASES_IGNORE_DEVICE_TYPE_DNS',
    );

    if ( $data->{'config'}->{'LEASES2LDAP_IGNORE_NETWORKS'} ) {
        my @scip_ignore_nets;
        my @ignore_nets =
          split( /\s+/, $data->{'config'}->{'LEASES2LDAP_IGNORE_NETWORKS'} );
        foreach my $ignore_net (@ignore_nets) {
            my $scip_ignore_net = SmartClient::IP->new($ignore_net);
            if ( ref $scip_ignore_net ) {
                $data->{'logger'}->info( "ignoring IP-Addresses from "
                      . $scip_ignore_net->{'first_ip'} . " to "
                      . $scip_ignore_net->{'last_ip'} );
                push( @scip_ignore_nets, $scip_ignore_net );
            } else {
                $data->{'logger'}
                  ->warn( "failed to initialize Network from " . $ignore_net );
            }
        }
        if (@scip_ignore_nets) {
            $data->{'ignore_nets'}            = \@scip_ignore_nets;
            $data->{'ignore_nets_logged_ips'} = {};
        }
    }

    $data->{'leases_file_name'} =
      $data->{'config'}->{'LEASES2LDAP_LEASES_FILE'};
    $data->{'cleared_leases_file_name'} =
      $data->{'config'}->{'LEASES2LDAP_CLEARED_LEASES_FILE'};
    $data->{'macs_to_clear_from_leases'} = {};
    $data->{'dhcpd_stop_cmd'}  = $data->{'config'}->{'LEASES2LDAP_STOP_DHCPD'};
    $data->{'dhcpd_start_cmd'} = $data->{'config'}->{'LEASES2LDAP_START_DHCPD'};

    # FastMmap reference
    # used to store information about dislocated devices
    # and make them accessable for other processes
    my $dislocated_devices = Cache::FastMmap->new(
        share_file =>
          $data->{'config'}->{'LEASES2LDAP_MMAP_FILE_DISLOCATED_DEVICES'},
        init_file => 1
    );

    # TODO: find better solution. Function?
    chmod 0666, $data->{'config'}->{'LEASES2LDAP_MMAP_FILE_DISLOCATED_DEVICES'};

    $data->{'dislocated_devices'} = $dislocated_devices;

    $data->{'permission_decider'} =
      SmartClient::DislocatedDevicePermission->new(
        'sc_cache' => $data->{'sc_cache'},
        'config'   => $data->{'config'}
      );

    my @expiration_delta = ( 3, 0, 0, 0 );
    my @exp_dhms =
      split( /,/, $data->{'config'}->{'LEASES2LDAP_LEASES_EXPIRE_AFTER'} );
    for ( my $exp_dhms_cnt = 0 ; $exp_dhms_cnt < 4 ; $exp_dhms_cnt++ ) {
        if ( $exp_dhms[$exp_dhms_cnt] =~ /^\d+$/ ) {
            $expiration_delta[$exp_dhms_cnt] = $exp_dhms[$exp_dhms_cnt];
        } else {
            $data->{'logger'}->warn( ( $exp_dhms_cnt + 1 )
                . ". entry in expiration delta is invalid -> value set to "
                  . $expiration_delta[$exp_dhms_cnt] );
        }
    }
    $data->{'leases_expire_after'}        = \@expiration_delta;
    $data->{'expired_leases_logged_macs'} = {};
    $data->{'ignore_leases_logged_macs'}  = {};
    $data->{'trigger_clear_dhcp_leases'}  = 0;

    return 1;
}

=head2 getAttributesChangedByMove

determines from a dislocated device 
(eg. C<current_ldap_entry>, C<location_dn_new>, ...) the new name 
and attributes after movement

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<href_device>

hash ref to dislocated device

=item C<device_type_dn>

device type dn (optional)

=back

=item Return

array with

=over 4

=item C<rdn>

new relative DN (eg. cn=pc00123)

=item  hash of changed LDAP entry attributes

=back

=back

=cut

sub getAttributesChangedByMove
{
    my $self   = shift;
    my %params = (
        'href_device'    => undef,
        'device_type_dn' => undef,
        @_,
    );

    my $sc_ldap  = $self->{'sc_ldap'};
    my $sc_cache = $self->{'sc_cache'};

    #my $dislocated_devices = $self->{'dislocated_devices'};

    my $href_device    = $params{'href_device'};
    my $device_type_dn = $params{'device_type_dn'};

    my $entry           = $href_device->{'current_ldap_entry'};
    my $location_dn_new = $href_device->{'location_dn_new'};

    my ( $new_dn, $new_ip, $new_device_type_dn ) =
      $sc_cache->getNextAvailableDeviceInfos(
        'location_dn'     => $location_dn_new,
        'old_entry'       => $entry,
        'new_entry_infos' => {
            'devicetype_dn' => $device_type_dn,
            'mac_address'   => lc( $href_device->{'mac_address'} ),
            'dhcp_hostname' => $href_device->{'hostname'},
        }
      );

    # cn is changed automatically by move
    my ( $rdn, $container ) = ldap_splitdn($new_dn);

    my $old_entry_dn    = "NO OLD DN";
    my @new_description = ();
    if ( ref $entry ) {
        $old_entry_dn    = $entry->dn;
        @new_description = $entry->get_value("description");
    }
    push @new_description, "moved from $old_entry_dn to $location_dn_new";
    my $time_now = strftime( "%d.%m.%Y %H:%M:%S", localtime() );
    my %changed_attributes = (
        ##'cn'                => $cn,
        'ipHostNumber'    => [$new_ip],
        'dbkDeviceTypeDn' => [$new_device_type_dn],
        'description'     => \@new_description,
        'dbkAuthorName'   => ['automatic move'],
        'dbkStatus'       => ["0, $time_now, MOVED: $old_entry_dn"],
    );
    if ( ref $entry ) {
        if ( has_objectclass( $entry, 'dbkWorkstation' ) ) {
            %changed_attributes = (
                'dbkStandardPrinterDn' => []
                ,    # we don't want to establish printer faxes
                'dbkPrinterGroupDn' => [],    # same as above
                %changed_attributes
            );
        }
    }

    return $rdn, %changed_attributes;
}

=head2 refreshDislocatedDeviceAction

checks dislocated device and set action value
If action is modified, also rechecks permission

=over 4

=item Parameter

hash ref C<dislocated device>. The data (C<action>) in the hash will be modified! 

=item Return

C<boolean>: true, if action is modified

=back

=cut

sub refreshDislocatedDeviceAction
{
    my ( $self, $href_device ) = @_;

    my $logger             = $self->{'logger'};
    my $config             = $self->{'config'};
    my $sc_cache           = $self->{'sc_cache'};
    my $dislocated_devices = $self->{'dislocated_devices'};

    my $mac            = $href_device->{'mac_address'};
    my $lease_location = $href_device->{'location_dn_new'};
    my $action_old     = $href_device->{'action'};
    my $modified       = 0;

    # get ldap entries with same mac address
    my @result = $sc_cache->search(
        attr          => 'macAddress',
        value         => lc($mac),
        casesensitive => 0
    );
    my $numberLdapEntries = @result;

    if ( $numberLdapEntries == 0 ) {

        # macAddress not in LDAP
        $href_device->{'action'} = ACTION_ADD;
        if ( $action_old != ACTION_ADD ) {
            $logger->info("macAddress $mac not in LDAP. Add it to LDAP");
            $modified = 1;
        }

    } elsif ( $numberLdapEntries == 1 ) {

        # macAddress in LDAP
        my $ldap_entry    = $result[0];
        my $ldap_dn       = $ldap_entry->dn();
        my $ldap_location = $sc_cache->getLocation( dn => $ldap_dn );
        if ( compare( $lease_location, $ldap_location ) ) {

            # macAddress in LDAP, at correct location
            $href_device->{'action'} = ACTION_NOTHING;
            if ( $action_old != ACTION_NOTHING ) {
                $logger->debug(
                    "mac address $mac from DHCP is already in same location as fixed entry from LDAP ("
                      . $ldap_entry->dn
                      . "). No action required" );
                $modified = 1;
            }

        } else {

            # macAddress in LDAP, but at wrong location. Move it.
            $href_device->{'action'} = ACTION_MOVE;
            my $current_ldap_dn = $href_device->{'current_ldap_dn'} || '';
            if (   ( $action_old != ACTION_MOVE )
                || ( $current_ldap_dn ne $ldap_entry->dn() ) )
            {
                $href_device->{'current_ldap_entry'} = $ldap_entry;
                $href_device->{'current_ldap_dn'}    = $ldap_entry->dn();
                $logger->debug(
                    "mac address $mac: physical location and ldap locations differs: $lease_location != $ldap_location"
                );
                $modified = 1;
            }
        }
    } else {

        # macAddress at multiple locations in LDAP
        # don't know what to do, so generate warning and do nothing
        my $dislocated_device_existing = $dislocated_devices->get($mac);
        if ( $dislocated_device_existing
            && ( $dislocated_device_existing->{'action'} == ACTION_MANUAL ) )
        {

            # this entry is already marked as ACTION_MANUAL in dislocated_devices,
            # so don't generate additional log message
        } else {
            $href_device->{'action'} = ACTION_MANUAL;
            if ( $action_old != ACTION_MANUAL ) {
                $logger->warn( "mac address $mac multiple times in LDAP: "
                      . join( ", ", map { $_->dn() } @result )
                      . ". Manual interaction required" );
                $modified = 1;
            }
        }
    }

    if ($modified) {
        my $permission_decider = $self->{'permission_decider'};
        if ( !ref $permission_decider ) {
            $logger->error(
                'no instance of SmartClient::DislocatedDevicePermisssion found in self instance'
            );
        } else {

            # action has changed. Therefore recheck permission
            $permission_decider->refreshPermission($href_device);
        }
    }

    # return if value has changed
    return $modified;
}

=head2 clearDislocatedDevices

clear the dislocated devices from the entries, which
do not occur in leases file.

=over 4

=item Parameter

A hash with all entries from leases.

=item Return

none

=back

=cut

sub clearDislocatedDevices
{

    my ( $self, %leases ) = @_;

    my $logger             = $self->{'logger'};
    my $dislocated_devices = $self->{'dislocated_devices'};

    my @dislocated_devices_to_delete;
    foreach my $dislocated_device_id ( $dislocated_devices->get_keys(0) ) {
        if ( !exists $leases{$dislocated_device_id} ) {
            push @dislocated_devices_to_delete, $dislocated_device_id;
        }
    }

    foreach my $dislocated_device_id (@dislocated_devices_to_delete) {
        $logger->info(
            "Dislocated device $dislocated_device_id was not found in dhcp leases file. Delete it."
        );
        $dislocated_devices->remove($dislocated_device_id);
    }

}

=head2 refreshDislocatedDeviceFromLease

checks a dhcp lease entry
against

=over 4

=item C<dislocated_devices> list 

=item ldap

=back

also evaluates permissions
and extends the dislocated_devices list

=over 4

=item Parameter

dhcp lease (href)

=item Return

C<boolean>

=back

=cut

sub refreshDislocatedDeviceFromLease
{
    my ( $self, $dhcpEntry ) = @_;

    my $logger             = $self->{'logger'};
    my $sc_cache           = $self->{'sc_cache'};
    my $dislocated_devices = $self->{'dislocated_devices'};

    my $mac      = $dhcpEntry->getMACAddr();
    my $lease_ip = $dhcpEntry->getIPAddr();

    foreach my $switch_net ( @{ $self->{'ignore_nets'} } ) {
        if ( ref $switch_net and $switch_net->overlaps($lease_ip) ) {
            if ( !$self->{'ignore_nets_logged_ips'}->{$lease_ip} ) {
                $logger->info(
                    "IP $lease_ip will be ignored as configured by LEASES2LDAP_IGNORE_NETWORKS"
                );
            }
            $self->{'ignore_nets_logged_ips'}->{$lease_ip} = 2;
            return;
        }
    }

    #the getLocationForIp should be faster, since ScCache has notion of reverse zones,
    #which can be searched more efficiently than the locations in LdapCache
    my $lease_location = $sc_cache->getLocationForIp($lease_ip);

    if ( !$lease_location ) {
        $logger->warn(
            "no LDAP location found for dhcp lease entry $lease_ip ($mac)");
        return;
    }

    # initialize dislocated_device hash
    my %dislocated_device = (
        'mac_address'        => $mac,
        'lease_time_start'   => $dhcpEntry->getStarts(),
        'lease_time_end'     => $dhcpEntry->getEnds(),
        'lease_status'       => $dhcpEntry->getStatus(),
        'lease_ip'           => $lease_ip,
        'hostname'           => $dhcpEntry->getHostname(),
        'media'              => $dhcpEntry->getMedia(),
        'tokenring'          => $dhcpEntry->isTokenring(),
        'current_ldap_dn'    => undef,
        'current_ldap_entry' => undef,
        'location_dn_new'    => $lease_location,
        'action'             => ACTION_UNDEF,
        'permission_level'   => PERM_LEVEL_UNDEF,
        'permission'         => PERM_UNDEF,
    );

    my $href_device_existing = $dislocated_devices->get($mac);

    if ( $dhcpEntry->isExpired( $self->{'leases_expire_after'} ) ) {
        if ( !$self->{'expired_leases_logged_macs'}->{$mac} ) {
            if ($href_device_existing) {
                $logger->info(
                    "Dislocated device $mac is expired as configured by LEASES2LDAP_LEASES_EXPIRE_AFTER. Delete it."
                );
            } else {
                $logger->info(
                    "Lease $mac is expired as configured by LEASES2LDAP_LEASES_EXPIRE_AFTER"
                );
            }
        }
        if ($href_device_existing) {
            $dislocated_devices->remove($mac);
        }
        $self->{'expired_leases_logged_macs'}->{$mac} = 2;
        $self->{'trigger_clear_dhcp_leases'} = 1;
        return;
    } else {
        delete $self->{'expired_leases_logged_macs'}->{$mac};
    }

    if (   $href_device_existing
        && $href_device_existing->{'action'} != ACTION_MANUAL
        && $href_device_existing->{'location_dn_new'} eq
        $dislocated_device{'location_dn_new'}
        && $href_device_existing->{'lease_time_start'} eq
        $dislocated_device{'lease_time_start'}
        && $href_device_existing->{'lease_time_end'} eq
        $dislocated_device{'lease_time_end'}
        && $href_device_existing->{'lease_status'} eq
        $dislocated_device{'lease_status'}
        && $href_device_existing->{'hostname'} eq
        $dislocated_device{'hostname'} )
    {
        $logger->debug(
            "Lease $mac is unchanged and already known as dislocated device");
        return;
    }

    $self->refreshDislocatedDeviceAction( \%dislocated_device );

    my $is_ignored = 0;
    if (ref $dislocated_device{'current_ldap_entry'}) {
        my $device_type_dn =
          $dislocated_device{'current_ldap_entry'}
          ->get_value('dbkDeviceTypeDn');
        if ($device_type_dn) {
            my @leases_ignore_device_type_dns = split(
                '\s+',
                $self->{'config'}->get(
                    'LEASES2LDAP_LEASES_IGNORE_DEVICE_TYPE_DNS')
                  || ''
            );
            if (@leases_ignore_device_type_dns) {
                foreach my $leases_ignore_device_type_dn (
                    @leases_ignore_device_type_dns)
                {
                    if (
                        lc($leases_ignore_device_type_dn) eq
                        lc($device_type_dn) )
                    {
                        $is_ignored = 1;
                    }
                }
            }
        }
    }

    if ($is_ignored) {
        if ( !$self->{'ignore_leases_logged_macs'}->{$mac} ) {
            if ($href_device_existing) {
                $logger->info(
                    "Dislocated device $mac is ignored as configured by LEASES2LDAP_LEASES_IGNORE_DEVICE_TYPE_DNS. Delete it."
                );
            } else {
                $logger->info(
                    "Lease $mac is ignored as configured by LEASES2LDAP_LEASES_IGNORE_DEVICE_TYPE_DNS"
                );
            }
        }
        if ($href_device_existing) {
            $dislocated_devices->remove($mac);
        }
        $self->{'ignore_leases_logged_macs'}->{$mac} = 2;
        return;
    } else {
        delete $self->{'ignore_leases_logged_macs'}->{$mac};
    }


    $dislocated_devices->set( $mac, \%dislocated_device );

}

=head2 refreshDislocatedDevicesFromLeases

checks a dhcp lease entry against

=over 4

=item C<dislocated_devices> list

=item ldap.

=back

Also evaluates permissions and extends 
the C<dislocated_devices> list

=over 4

=item Parameter

none

=item Return

integer with

=over 4

=item C<0>

no new lease file

=item >0

new leases files. Number of DHCP lease entries

=back

=back

=cut

sub refreshDislocatedDevicesFromLeases
{
    my $self = shift;

    my $logger           = $self->{'logger'};
    my $leases_file_name = $self->{'leases_file_name'};

    my $counter = 0;

    # reading the dhcpd.leases file
    if ( checkFileMod( $leases_file_name, \$self->{'leases_file_mtime'} ) ) {
        $logger->info("dhcp leases file modified");
        my %leases_hash = read_dhcp_leases($leases_file_name);

        $self->clearDislocatedDevices(%leases_hash);

        foreach my $href_leases_entry ( values %leases_hash ) {
            $self->refreshDislocatedDeviceFromLease($href_leases_entry->{'act'});
            $counter++;
        }

        foreach my $logged_ip ( keys %{ $self->{'ignore_nets_logged_ips'} } ) {
            if ( $self->{'ignore_nets_logged_ips'}->{$logged_ip} == 2 ) {
                $self->{'ignore_nets_logged_ips'}->{$logged_ip} = 1;
            } else {
                delete $self->{'ignore_nets_logged_ips'}->{$logged_ip};
            }
        }

        foreach
          my $logged_mac ( keys %{ $self->{'expired_leases_logged_macs'} } )
        {
            if ( $self->{'expired_leases_logged_macs'}->{$logged_mac} == 2 ) {
                $self->{'expired_leases_logged_macs'}->{$logged_mac} = 1;
            } else {
                delete $self->{'expired_leases_logged_macs'}->{$logged_mac};
            }
        }

        foreach
          my $logged_mac ( keys %{ $self->{'ignore_leases_logged_macs'} } )
        {
            if ( $self->{'ignore_leases_logged_macs'}->{$logged_mac} == 2 ) {
                $self->{'ignore_leases_logged_macs'}->{$logged_mac} = 1;
            } else {
                delete $self->{'ignore_leases_logged_macs'}->{$logged_mac};
            }
        }

    }

    return $counter;
}

=head2 performAction

If permission is granted,
performs the actions

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<id>

for C<$dislocated_devices>

=item C<permission>

grant permission for this action (int) (optional)

=item C<device_type_dn>

new device type (string) (optional) 

=back

=item Return

hash consisting of:

=over 4

=item C<id>

=item C<action>

=item C<code>

=item C<message>

=item C<subcode>

=item C<submessage>

=item C<old_ldap_dn>

=item C<new_ldap_dn>

=back

=back

=cut

sub performAction
{
    my $self   = shift;
    my %params = (
        'id'             => undef,
        'permission'     => PERM_UNDEF,
        'device_type_dn' => undef,
        @_,
    );

    my $logger             = $self->{'logger'};
    my $sc_ldap            = $self->{'sc_ldap'};
    my $sc_cache           = $self->{'sc_cache'};
    my $dislocated_devices = $self->{'dislocated_devices'};

    my $id             = $params{'id'};
    my $permission     = $params{'permission'};
    my $device_type_dn = $params{'device_type_dn'};

    my %result = (
        'id'          => $id,
        'action'      => undef,
        'code'        => undef,
        'message'     => undef,
        'subcode'     => undef,
        'submessage'  => undef,
        'old_ldap_dn' => undef,
        'new_ldap_dn' => undef,
    );

    my $dislocated_device = $dislocated_devices->get($id);
    if ( not $dislocated_device ) {
        $result{'code'}    = XMLRPC_NO_SUCH_ID;
        $result{'message'} = "no dislocated device $id found";
        $logger->error( $result{'message'} );
        return %result;
    }

    # recheck actions and permissions
    # to prevent that an already added entry
    # will be added again
    $self->refreshDislocatedDeviceAction($dislocated_device);

    my $current_ldap_dn = $dislocated_device->{'current_ldap_dn'};
    my $action          = $dislocated_device->{'action'};

    $result{'old_ldap_dn'} = $current_ldap_dn;
    $result{'action'}      = $action;

    $logger->debug( "$id: action: $action, new device type: "
          . ( $device_type_dn || "(none)" ) );

    # check if permission is granted
    # by function parameter
    # or already in dislocatged device
    if (   ( $permission != PERM_GRANTED )
        && ( $dislocated_device->{'permission'} != PERM_GRANTED ) )
    {
        $result{'code'}    = XMLRPC_NO_PERMISSION;
        $result{'message'} = "no permission for dislocated device $id";
        return %result;
    }

    if ( $action == ACTION_NOTHING ) {

        # nothing to do
        $result{'code'}    = XMLRPC_OK;
        $result{'message'} = "no action necessary for dislocated device $id";
    } elsif ( $action == ACTION_ADD ) {

        # adding dhcp lease entry to ldap
        $self->addNewWorkstation(
            result         => \%result,
            id             => $id,
            device_type_dn => $device_type_dn
        );

    } elsif ( $action == ACTION_MOVE ) {

        # move existing ldap entry to other location
        $self->moveDevice(
            result         => \%result,
            id             => $id,
            device_type_dn => $device_type_dn
        );

    } elsif ( $action == ACTION_MANUAL ) {
        $result{'code'} = XMLRPC_ERROR_MANUAL;
        $result{'message'} =
          "errors on dislocated device $id. Manual corrections required";
        $logger->warn( $result{'message'} );
    } else {
        $result{'code'} = XMLRPC_UNKNOWN_ACTION;
        $result{'message'} =
          "unsupported action $action for dislocated_device $id";
        $logger->warn( $result{'message'} );
    }

    return %result;
}

=head2 performActions

walks through dislocated_devices.
If permission is granted,
performs the actions by calling C<performAction> for each C<id> in C<dislocated_devices> list.

=over 4

=item Parameter

none

=item Return

none

=back

=cut

sub performActions
{
    my $self               = shift;
    my $dislocated_devices = $self->{'dislocated_devices'};

    foreach my $id ( $dislocated_devices->get_keys(0) ) {
        $self->performAction( id => $id );
    }
}

=head2 clearLeases

clear all entries in leases file off the
macs which were added/moved before

=over 4

=item Parameter

none

=item Return

none

=back

=cut

sub clearLeases
{
    my $self                      = shift;
    my $macs_to_clear_from_leases = $self->{'macs_to_clear_from_leases'};

    my $run_clear_dhcp_leases = 0;

    if ( ref($macs_to_clear_from_leases) eq "HASH" ) {
        foreach
          my $mac_to_clear_from_leases ( keys %{$macs_to_clear_from_leases} )
        {
            if ( $macs_to_clear_from_leases->{$mac_to_clear_from_leases} <= 1 )
            {
                $run_clear_dhcp_leases = 1;
                last;
            } else {
                --$macs_to_clear_from_leases->{$mac_to_clear_from_leases};
            }
        }
    }

    if ( $self->{'trigger_clear_dhcp_leases'} || $run_clear_dhcp_leases ) {
        system( $self->{'dhcpd_stop_cmd'} );
        my $cleared_macs = clear_dhcp_leases(
            'leases_file_name'          => $self->{'leases_file_name'},
            'cleared_leases_file_name'  => $self->{'cleared_leases_file_name'},
            'macs_to_clear_from_leases' => $macs_to_clear_from_leases,
            'expiration_delta'          => $self->{'leases_expire_after'}
        );
        system( $self->{'dhcpd_start_cmd'} );
        $self->{'trigger_clear_dhcp_leases'} = 0;
        delete $macs_to_clear_from_leases->{$_}
          foreach ( keys %{ $cleared_macs->{'on_request'} } );
    }

}

=head2 getLeasesForMAC

returns all entries from the lease file to the
specified MAC address

=over 4

=item Parameter

the MAC-address

=item Return

hash consisting of:

=over 4

=item C<code>

0 if lease entries found

=item C<message>

empty if lease entries found

=item C<entries>

hash consisting of:

=over 4

=item C<active>

hash ref with active lease entry

=item C<all>

array ref of all lease entries (href)

=back

=back

=back

=cut

sub getLeasesForMAC
{
    my ( $self, $mac_addr ) = @_;

    my $leases_for_mac = {};
    if ( !$mac_addr or $mac_addr !~ /^(?:[0-9a-f]{2}:){5}[0-9a-f]{2}$/ ) {
        $leases_for_mac->{'code'}    = 9;
        $leases_for_mac->{'message'} =
          'parameter does not exist or is not a device ID';
        return $leases_for_mac;
    }
    $leases_for_mac->{'mac_address'} = $mac_addr;
    $leases_for_mac->{'code'}        = 1;
    $leases_for_mac->{'message'}     = 'no leases found';
    my %leases_hash = read_dhcp_leases( $self->{'leases_file_name'} );
    foreach my $lease_mac ( keys %leases_hash ) {
        if ( $mac_addr eq $lease_mac ) {
            $leases_for_mac->{'code'} = 0;
            delete $leases_for_mac->{'message'};
            my $entries = { 'all' => [], 'active' => undef };
            $entries->{'active'} =
              $self->getLeaseEntryForDhcpEntry(
                $leases_hash{$lease_mac}->{'act'} );
            foreach my $dhcp_entry ( @{ $leases_hash{$lease_mac}->{'all'} } ) {
                push @{ $entries->{'all'} },
                  $self->getLeaseEntryForDhcpEntry($dhcp_entry);
            }
            $leases_for_mac->{'entries'} = $entries;
            last;
        }
    }
    return $leases_for_mac;
}

=head2 getLeaseEntryForDhcpEntry

returns lease entry from object of type SmartClient::DhcpEntries

=over 4

=item Parameter

an object of type SmartClient::DhcpEntries

=item Return

hash consisting of:

=over 4

=item C<ipAddress>

of the lease entry

=item C<macAddress>

of the lease entry

=item C<uxtime_starts>

of the lease entry

=item C<uxtime_ends>

of the lease entry

=item C<media>

of the lease entry

=item C<hostname>

of the lease entry

=item C<status>

=back

=back

=cut

sub getLeaseEntryForDhcpEntry
{
    my ( $self, $dhcpEntry ) = @_;

    my %lease_entry = (
        'ipAddress'     => $dhcpEntry->getIPAddr(),
        'macAddress'    => $dhcpEntry->getMACAddr(),
        'uxtime_starts' => $dhcpEntry->getStarts(),
        'uxtime_ends'   => $dhcpEntry->getEnds(),
        'media'         => $dhcpEntry->getMedia(),
        'hostname'      => $dhcpEntry->getHostname(),
        'status'        => $dhcpEntry->getStatus()
    );

    return \%lease_entry;

}

=head2 triggerDeleteLeasesForMAC

triggers a delete of all entries from the lease file
to the specified MAC address

the next run of clearLeases  will delete all entries

=over 4

=item Parameter

the MAC-address

=item Return

hash consisting of:

=over 4

=item C<code>

0 if the delete is triggered

=item C<message>


=back

=back

=cut

sub triggerDeleteLeasesForMAC
{
    my ( $self, $mac_addr ) = @_;

    my $result = {};
    if ( !$mac_addr or $mac_addr !~ /^(?:[0-9a-f]{2}:){5}[0-9a-f]{2}$/ ) {
        $result->{'code'}    = 9;
        $result->{'message'} = 'parameter does not exist or is not a device ID';
        return $result;
    }
    $self->{'macs_to_clear_from_leases'}->{$mac_addr} = 1;
    $result->{'mac_address'}                          = $mac_addr;
    $result->{'code'}                                 = 0;
    $result->{'message'} = 'deleting lease entries was triggered';

    return $result;
}

=head2 addNewWorkstation

add a new workstation to LDAP

Performs 2 LDAP writes

=over 4

=item global

C<($dislocated_devices)>

C<$sc_ldap> (already bound with write access)

C<$sc_cache>

=item Parameter

hash consisting of:

=over 4

=item C<result>

result hash used by C<performAction>. Will be modified!

=item C<id>

for C<$dislocated_devices>

=item C<device_type_dn>

new device type (optional)

=back

=item Return

C<Net::LDAP::Entry> of the added workstation or undef in case of error

Also modifies result hash!

=back

=cut

sub addNewWorkstation
{
    my $self   = shift;
    my %params = (
        'result'         => undef,
        'id'             => undef,
        'device_type_dn' => undef,
        @_,
    );

    my $logger             = $self->{'logger'};
    my $config             = $self->{'config'};
    my $sc_ldap            = $self->{'sc_ldap'};
    my $sc_cache           = $self->{'sc_cache'};
    my $dislocated_devices = $self->{'dislocated_devices'};

    my $result         = $params{'result'};
    my $id             = $params{'id'};
    my $device_type_dn = $params{'device_type_dn'};

    my $dislocated_device = $dislocated_devices->get($id);

    my $mac_address     = lc( $dislocated_device->{'mac_address'} );
    my $tokenring       = $dislocated_device->{'tokenring'};
    my $location_dn_new = $dislocated_device->{'location_dn_new'};

    my @entries_with_mac_address = $sc_cache->search(
        'attr'          => 'macAddress',
        'value'         => $mac_address,
        'casesensitive' => 0
    );
    if (@entries_with_mac_address) {
        $logger->warn(
            "addNewWorkstation was called to add workstation with duplicate macAddress."
        );
        foreach my $entry_with_duplicate_mac_address (@entries_with_mac_address)
        {
            $logger->warn( "duplicate macAddress: "
                  . $entry_with_duplicate_mac_address->get_value('macAddress')
                  . " found in dn: "
                  . $entry_with_duplicate_mac_address->dn );
        }
        return;
    }
    my $ldap      = $sc_ldap->getNetLdap();
    my $ldap_base = $sc_ldap->getBase();
    my ( $new_dn, $new_ip, $new_device_type_dn ) =
      $sc_cache->getNextAvailableDeviceInfos(
        'location_dn'     => $location_dn_new,
        'new_entry_infos' => {
            'devicetype_dn' => $device_type_dn,
            'mac_address'   => $mac_address,
            'dhcp_hostname' => $dislocated_device->{'hostname'}
        }
      );
    if ( !$new_dn || !$new_ip || !$new_device_type_dn ) {
        $result->{'code'} = XMLRPC_LDAP_FAILURE;
        $result->{'message'} =
          "failed to determine required attributes for new workstation ($mac_address):  dn=$new_dn, ip=$new_ip, device_type_dn=$new_device_type_dn";

        $logger->error( $result->{'message'} );
        return;
    }

    $logger->info(
        "adding workstation (MAC: $mac_address) with DN: $new_dn, IP: $new_ip");

    # use LDAP entry.
    # will be added to LDAP and to Cache later on
    my $entry = Net::LDAP::Entry->new();
    $entry->dn($new_dn);
    $result->{'new_ldap_dn'} = $new_dn;

    # get default workstation attributes
    my %attributes = ( getWorkstationDefaultAttributes($new_dn) );

    # set mac address
    $attributes{'macAddress'} = $mac_address;

    if ( $new_ip ne "0.0.0.0" ) {

        # set IP address
        $attributes{'ipHostnumber'} = $new_ip;
    }

    # add tokenring marker is workstation is detected as tokenring
    if ($tokenring) {
        push( @{ $attributes{'objectClass'} }, "dbkTokenRing" );
    }

    if ($new_device_type_dn) {
        $attributes{'dbkDeviceTypeDn'} = $new_device_type_dn;
    }

    # add attributes to LDAP entry
    $entry->add(%attributes);

    # add new workstation to ldap
    my $ldap_result = $entry->update($ldap);
    if ( $ldap_result->code ) {
        $result->{'code'} = XMLRPC_LDAP_FAILURE;
        $result->{'message'} =
          "failed to add new workstation ($mac_address) into LDAP";
        $result->{'subcode'}    = $ldap_result->code();
        $result->{'submessage'} = $ldap_result->error();

        $logger->error( $result->{'message'} . ": " . $result->{'submessage'} );
        return;
    }

    # we have added an entry to ldap. Now we try to get it back.
    # In eDir this can result in false positives when an entry is
    # added in a place, where a pc was moved away shortly before.
    $entry = $self->getLdapEntrySafely(
        'ldap_con'           => $ldap,
        'filter'             => "macAddress=$mac_address",
        'dn'                 => $new_dn,
        'href_xmlrpc_result' => $result
    );
    if ( !ref $entry ) {
        $result->{'code'} = XMLRPC_LDAP_FAILURE;
        $result->{'message'} =
          "failed to change attributes in newly added workstation LDAP entry  $new_dn";
        $logger->warn(
            "[LDAP-DATA] no newly added entry found for dn $new_dn with filter macAddress=$mac_address,"
              . " so no ACL will be set and entry will not be added into cache in this round."
        );
        return;
    }

    # modify ACLs on NDS, obscurely,
    # adding the ACLs when adding the entry
    # does not work on our test NDS.
    if ( uc( $config->get('ALL_LDAPSERVER_TYPE') ) eq "NDS" ) {
        my @myACLs = ();
        push( @myACLs, "1#subtree# $new_dn  #[Entry Rights]" );
        push( @myACLs, "6#entry# $new_dn #[All Attributes Rights]" );

        $entry->add( 'ACL' => [@myACLs] );

        #TODO: this could lead to updating the wrong entry when ws is added into a place, from which a ws was moved recently
        #more info can be found in moveDevice
        $ldap_result = $entry->update($ldap);
        if ( $ldap_result->code ) {
            $result->{'code'} = XMLRPC_LDAP_FAILURE;
            $result->{'message'} =
              "failed to add ACLs to new workstation ($mac_address) into LDAP";
            $result->{'subcode'}    = $ldap_result->code();
            $result->{'submessage'} = $ldap_result->error();

            $logger->error(
                $result->{'message'} . ": " . $result->{'submessage'} );
            return;
        }
    }

    # set result
    $result->{'code'}    = XMLRPC_OK;
    $result->{'message'} = "added dislocated device $id as $new_dn";

    # remove leases entry from dislocated devices
    $dislocated_devices->remove($id);

    #refresh entry from ldap to get an entry suitable for setEntry
    #without this, setEntry will fail. Maybe it would be better to
    #let LdapCache handle all the work (adding to ldap and cache).
    #For explanations of the search filter look comment in moveDevice
    $entry = $self->getLdapEntrySafely(
        'ldap_con'           => $ldap,
        'filter'             => "macAddress=$mac_address",
        'dn'                 => $new_dn,
        'href_xmlrpc_result' => $result
    );
    if ( !ref $entry ) {
        $result->{'code'} = XMLRPC_LDAP_FAILURE;
        $result->{'message'} =
          "failed to fetch entry after adding ACLs to new workstation ($mac_address) into LDAP";

        $logger->error( $result->{'message'} . ": " . $result->{'submessage'} );
        $logger->error( "[LDAP-DATA] No entry was found after updating ACL."
              . " This is very strange. Entry will not be added to Cache this round."
        );
        return;
    }

    # update cache
    $sc_cache->setEntry($entry);
    $sc_cache->assignLocations( 'refreshed_dn_aref' => [$new_dn] );
    $sc_cache->assignIpDevices(
        'refreshed_locations_aref' => [$location_dn_new] );

    # add macAddress to macs_to_clear_from_leases-hash
    $self->{'macs_to_clear_from_leases'}->{$mac_address} = 1;

    return $entry;
}

=head2 moveDevice

move a IP device (eg. workstation) 
from one LDAP location to another.
Also changes LDAP entry attributes according to the rules.

Performs 1 LDAP read (get moved entry) and 1 LDAP write

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<result>

result hash used by performAction. Will be modified!

=item C<id>

for $dislocated_devices

=item C<device_type_dn>

new device type (optional)

=back

=item Return

C<Net::LDAP::Entry> of the moved workstation
or undef in case of error.
Also modifies result hash

=back

=cut

sub moveDevice
{
    my $self   = shift;
    my %params = (
        'result'         => undef,
        'id'             => undef,
        'device_type_dn' => undef,
        @_,
    );

    my $logger             = $self->{'logger'};
    my $sc_ldap            = $self->{'sc_ldap'};
    my $sc_cache           = $self->{'sc_cache'};
    my $dislocated_devices = $self->{'dislocated_devices'};

    my $result         = $params{'result'};
    my $id             = $params{'id'};
    my $device_type_dn = $params{'device_type_dn'};

    my $dislocated_device = $dislocated_devices->get($id);

    my $dn_old          = $dislocated_device->{'current_ldap_dn'};
    my $location_dn_new = $dislocated_device->{'location_dn_new'};

    my $ldap      = $sc_ldap->getNetLdap();
    my $ldap_base = $sc_ldap->getBase();

    my ( $rdn_new, %modified_attributes ) = $self->getAttributesChangedByMove(
        'href_device'    => $dislocated_device,
        'device_type_dn' => $device_type_dn
    );

    my $dn_new = "$rdn_new,$location_dn_new";
    $result->{'new_ldap_dn'} = $dn_new;

    #TODO: ldap_cache should be notified of this move
    my $ldap_result = $ldap->moddn(
        $dn_old,
        'newrdn'       => "$rdn_new",
        'newsuperior'  => "$location_dn_new",
        'deleteoldrdn' => 1
    );
    if ( $ldap_result->code ) {
        $result->{'code'} = XMLRPC_LDAP_FAILURE;
        $result->{'message'} =
          "failed to move workstation LDAP entry from  $dn_old to $dn_new";
        $result->{'subcode'}    = $ldap_result->code();
        $result->{'submessage'} = $ldap_result->error();

        $logger->error( $result->{'message'} . ": " . $result->{'submessage'} );

        return;
    }

    $logger->debug("moved workstation LDAP entry from $dn_old to $dn_new");

    my $mac;
    my $entry = $sc_cache->getEntry($dn_old);
    if ( ref $entry ) {
        if ( $entry->exists('macAddress') ) {
            $mac = $entry->get_value('macAddress');
        }
    }

    # mark new location as changed
    $logger->info("mark location $location_dn_new as changed");
    $sc_cache->{'changed_locations_hash'}->{ lc($location_dn_new) } = 1;

    # remove old entry from cache
    # this is done here,
    # to prevent the attempt
    # to move it again in case of an error
    $sc_cache->removeEntry($dn_old);

    # get moved LDAP entry
    # eDir seems to say OK to moddn before the object is really moved, so we have to try to read it a few times.
    # Another nice behaviour of eDir is that moving objects can result in wrong answers.
    #
    # Example:
    #  move pc1 -> pc2
    #  move pc3 -> pc1
    #  search for pc1 and chances are that you will get pc2
    #
    # Because of this the search filter is more specific to find a special attribute only found in pc3
    my $filter;
    if ($mac) {
        $filter = "macAddress=$mac";
        $logger->debug("will use $mac for filter");
    } else {
        $filter = '(objectClass=*)';
        $logger->debug("we have no mac for filter, so use all objectClasses");
    }
    my $dn_entry = $self->getLdapEntrySafely(
        'ldap_con'           => $ldap,
        'dn'                 => $dn_new,
        'filter'             => $filter,
        'href_xmlrpc_result' => $result
    );
    if ( !ref $dn_entry ) {
        $result->{'code'} = XMLRPC_LDAP_FAILURE;
        $result->{'message'} =
          "failed to get recently moved workstation LDAP entry $dn_new with mac $mac (moved from $dn_old)";
        $logger->debug(
            "no entry found for $dn_new. This should not happen here. The entry should have got following changes: "
              . Dumper( \%modified_attributes )
              . " . ldap told us: "
              . Dumper($ldap_result) );

        $logger->error( $result->{'message'} . ": " . $result->{'submessage'} );
        return;
    }

    foreach my $attr_name ( keys %modified_attributes ) {

        if ( defined( $modified_attributes{$attr_name} ) ) {

            # check, if the LDAP entry must be modified or added
            if ( $dn_entry->exists($attr_name) ) {
                $dn_entry->replace(
                    $attr_name => $modified_attributes{$attr_name} );
            } else {

                # add
                if ( join( "", @{ $modified_attributes{$attr_name} } ) ne "" ) {
                    $dn_entry->add(
                        $attr_name => $modified_attributes{$attr_name} );
                } else {

                    # replace non existing attribute by empty string => do nothing
                    $logger->debug(
                        "$dn_new ($attr_name): replace non existing attribute by empty string => do nothing"
                    );
                }
            }
        } else {

            # empty/undefined value => remove attribute
            if ( $dn_entry->exists($attr_name) ) {
                $dn_entry->delete($attr_name);
            }
        }
    }

    # replace ACLs?
    # Not needed for eDir.

    # add new workstation to ldap
    my $tries = 0;
    $ldap_result = undef;
    do {
        sleep( 2 * $tries++ );
        if ( ref $ldap_result and $tries > 0 ) {
            $logger->info( "retry to update ldap entry "
                  . $dn_entry->dn
                  . " (last update failed because of "
                  . $ldap_result->error
                  . ")" );
        }
        $ldap_result = $dn_entry->update($ldap);
    } until ( !$ldap_result->code or $tries > 10 );
    if ( $tries > 1 ) {
        $logger->info("it took $tries tries to update entry $dn_new");
    }

    # try to get the newly updated entry, to ensure that eDirectory has really changed the values
    $tries = 0;

    #TODO: it should be nicer to take a key and value from modified_attributes instead of hard coded ipHostNumber
    my $new_ip_address = $modified_attributes{'ipHostNumber'}[0];
    $filter   = "(ipHostNumber=$new_ip_address)";
    $dn_entry = $self->getLdapEntrySafely(
        'ldap_con'           => $ldap,
        'dn'                 => $dn_new,
        'filter'             => $filter,
        'href_xmlrpc_result' => $result
    );
    if ( !ref $dn_entry ) {
        $result->{'code'} = XMLRPC_LDAP_FAILURE;
        $result->{'message'} =
          "failed to refetch workstation LDAP entry from  $dn_old to $dn_new";
        $result->{'subcode'}    = $ldap_result->code();
        $result->{'submessage'} = $ldap_result->error();
        $logger->warn(
            "[LDAP-DATA] could not fetch updated entry dn $dn_new with filter $filter. Entry will not be added to cache this round."
        );
        return;
    }

    # add the new entry to sc_cache
    # old entry has already been deleted from cache before
    # even add it to cache, if modify has failed
    $sc_cache->setEntry($dn_entry);
    $sc_cache->assignLocations( 'refreshed_dn_aref' => [$dn_new] );
    $sc_cache->assignIpDevices(
        'refreshed_locations_aref' => [ $sc_cache->getLocationForDn($dn_new) ]
    );

    if ( $ldap_result->code ) {
        $result->{'code'}       = XMLRPC_LDAP_FAILURE;
        $result->{'message'}    = "failed to modify moved workstation $dn_new";
        $result->{'subcode'}    = $ldap_result->code();
        $result->{'submessage'} = $ldap_result->error();

        $logger->error( $result->{'message'} . ": "
              . $result->{'submessage'} . " => "
              . Dumper( \%modified_attributes ) );
        return;
    }

    # set results
    $result->{'code'} = XMLRPC_OK;
    $result->{'message'} =
      "moved dislocated device $id from $dn_old to $dn_new";

    $dislocated_devices->remove($id);

    # add macAddress to macs_to_clear_from_leases-hash
    $self->{'macs_to_clear_from_leases'}->{$mac} = 2;

    return $dn_entry;
}

=head2 getLdapEntrySafely

This method is only to be used internally in C<SmartClient::Leases2ldap>!

It will refetch a newly added/moved or updated entry from ldap. It will retry the
search up to C<LEASES2LDAP_LDAP_MAX_REFETCH_RETRIES> or C<10> times, if the config variable 
could not be found. Between each search try a growing time intervall will be slept. 

=head3 Example

   my $entry = $leases2ldap->getLdapEntrySafely(
      'ldap_con' => $ldap, 
      'dn' => $new_dn, 
      'filter' => "macAddress=$mac",
      'href_xmlrpc_result' => $result);
   if (!ref $entry) {
      $logger->error('could not refetch entry');
   } else {
      #do something with the entry
      ...
   }

=over 4

=item Parameter

hash consisting of

=over 4

=item C<ldap_con> 

ldap connection

=item C<dn>

base of entry to refetch

=item C<filter>

to be used in refetch

=item C<href_xmlrpc_result>

structure which is used to return error codes to calling xmlrpc client

=back

=item Return

C<Net::LDAP::Entry> object which matches filter criteria or C<undef> when no entry could be found

=back

=cut

sub getLdapEntrySafely
{
    my $self   = shift;
    my %params = (
        'ldap_con'           => undef,
        'filter'             => undef,
        'dn'                 => undef,
        'href_xmlrpc_result' => undef,
        @_
    );
    my $logger   = Log::Log4perl->get_logger();
    my $ldap_con = $params{'ldap_con'};
    if ( !ref $ldap_con ) {
        $logger->error("No ldap connection was passed.");
        return;
    }
    my $filter = $params{'filter'};
    if ( !$filter ) {
        $logger->error(
            "No filter was passed, so we can't garantuee to get the right entry. So we will return nothing."
        );
        return;
    }
    my $dn = $params{'dn'};
    if ( !$dn ) {
        $logger->error(
            "No base dn was given. So no search will be performed and no entry returned."
        );
        return;
    }
    my $href_xmlrpc_result = $params{'href_xmlrpc_result'};
    if ( !ref $href_xmlrpc_result ) {
        $logger->info(
            "no reference to xmlrpc result structure was given, so no ldap errors will be recorded."
        );
        $href_xmlrpc_result = {};
    }
    my $entry;
    my $tries = 0;
    my $max_tries =
      $self->{'config'}->get( 'LEASES2LDAP_LDAP_MAX_REFETCH_RETRIES', 1 ) || 10;
    my $got_correct_entry;
    my $ldap_result;
    do {
        $got_correct_entry = 0;
        sleep( $tries * 2 );
        $ldap_result = $ldap_con->search(
            'base'   => $dn,
            'scope'  => 'base',
            'filter' => $filter
        );
        if (    ref $ldap_result
            and !$ldap_result->code
            and ( $ldap_result->count > 0 ) )
        {
            $entry             = $ldap_result->entry(0);
            $got_correct_entry = 1;
        }
    } until ( $got_correct_entry or ( ++$tries > $max_tries ) );
    if ( ref $entry ) {
        if ( lc( $entry->dn() ) ne lc($dn) ) {
            ########################################################
            # Workaround fuer BUG in LDAP-Schnittstelle von Novell #
            ########################################################
            $logger->warn(
                '[LDAP-DATA] dn from result is not equal to given dn for search, will correct it'
            );
            $entry->dn($dn);
        }
        $logger->debug("needed $tries tries to fetch correct ldap entry");
        return $entry;
    }

    # no entry could be found, so save the last error messages from ldap in the xmlrpc result structure
    $href_xmlrpc_result->{'subcode'}    = $ldap_result->code();
    $href_xmlrpc_result->{'submessage'} = $ldap_result->error();

    $logger->warn(
        "No entry could be found for $dn with filter $filter after $tries tries."
    );
    return;
}

1;
