#!/usr/bin/perl -w
use strict;
use warnings;

package SmartClient::NameGenerator;

=head1 NAME

SmartClient::NameGenerator;

=head2 SVN Version

 $Id: NameGenerator.pm 8042 2012-07-25 10:13:36Z FSchumacher $

=head1 Synopsis

=begin comment

Class for generating names for devices based on so called generators, ldap 
information and ip information.

=end comment

Class for generating names for devices based on so called generators, ldap 
information and ip information.

=head2 Usage

	#!/usr/bin/perl
	use strict;
	use SmartClient::ScCache;
	use SmartClient::NameGenerator;
	use Log::Log4perl qw(:easy);
	Log::Log4perl->easy_init($ERROR);

	#construtct a SmartClient::NameGenerator object
	my $name_generator = SmartClient::NameGenerator->new('sc_cache'=>$sc_cache) or warn "couldn\'t init SmartClient::NameGenerator";

	#add a naming generator for a given devicetype_dn
	$name_generator->addGeneratorForDevicetype(
		'format_string' => 'pc${pc_id}',
		'format_ids' => {
			'pc_id' => {
				'type' => 'IP',
				format => '%05d'
			}
		},
		'devicetype_dn' => 'cn=deviceKnuTPC,ou=global,o=company,c=de',
		'keep_old_cn' => undef,
	);

	#or add it by giving an entry object
	$name_generator->addGeneratorByEntry($sc_chache->getEntry('cn=deviceHVPC,ou=global,o=company,c=de'));

	#generate names based on ldap and/or ip information
	my $device_cn = $name_generator->generateCn(
		'destination_location_entry' => $destination_location_entry,
		'ip_addr' => $ip_addr,
		'new_devicetype_dn' => $new_devicetype_dn,
		'old_ldap_entry' => $old_entry,
	);

=head1 Description

Class to manage generators for different devices and their naming standards

=head1 Methods

Description of the various methods

=cut

our $VERSION = 0.1;
use Net::LDAP;
use SmartClient::LdapUtil;
require SmartClient::ScCache;
use SmartClient::IP;
use Log::Log4perl;
use SmartClient::Config;
use Scalar::Util 'blessed';

# which methods should be exported
# we don't export anything, we are a class (a base class that is)
#our @ISA    = qw('');
#our @EXPORT = qw();

=head2 new

Constructor for class SmartClient::NameGenerator.

=over 4

=item Parameter

	- sc_ldap SmartClient::ScLdap object

=item Return

	- blessed reference

=back

=cut

sub new
{
    my $class = shift;
    my $data  = {
        'sc_cache' => undef,
        @_
    };
    my $logger = Log::Log4perl->get_logger();
    $logger->debug('new naming generator construction');
    if ( !$data->{'sc_cache'} ) {
        $logger->error('no sc_cache object given.');
        return;
    }

    $data->{'generators'} = {};

    return bless $data, $class;
}

=head2 addGeneratorForDevicetype

Adds generator for a devicetype to the nameGenerator object. 
If a generator for a devicetype was already registered (added), the old
generator will be replaced silently by the new one.

=over 4

=item Parameter

	- format_string string containing placeholders for ids, e.g. 'pc${pcId}'
	- format_ids hash with id-keys and their values as hash containing, e.g. 'pcId => {"type" => "IP", "format" => "%05d" }'
	- hashref
		-type one of 'IP'|'UNIQUE'|'LDAP'|'XMLRPC'
		-format c-formatstring, e.g. "%03d"
	- devicetype_dn string representation of devicetype
	- keep_old_cn boolean if old cn should be kept while moving devices

=item Return

	- boolean true if generator for devicetype was successfully registered

=back

=cut

sub addGeneratorForDevicetype
{
    my $self   = shift;
    my %params = (
        'format_string' => undef,
        'foramt_ids'    => undef,
        'devicetype_dn' => undef,
        'keep_old_cn'   => undef,
        @_
    );

    my $logger = Log::Log4perl->get_logger();
    my $devicetype_dn = lc( $params{'devicetype_dn'} || '' );
    if ( !$devicetype_dn ) {
        $logger->error('no devicetype_dn given while adding generator');
        return 0;
    }

    # check for all necessary parameters
    my $has_all_params = 1;
    foreach my $check_key (qw(format_string format_ids)) {
        if ( !$params{$check_key} ) {
            $logger->error( 'missing parameter '
                  . $check_key
                  . ' while adding generator for devicetype_dn '
                  . $devicetype_dn );
            $has_all_params = 0;
        }
    }
    if ( !$has_all_params ) {
        return 0;
    }

    # check if all format_ids in format_string can be found in format_ids
    my $tmp_format_string = $params{'format_string'};
    while ( not $tmp_format_string eq '' ) {
        my ($key) = ( $tmp_format_string =~ m/\$\{([^}]*)\}/ );
        $tmp_format_string =~ s/^[^\$]*\${[^}]*\}//;
        if ( not ${ $params{'format_ids'} }{$key} ) {
            $logger->error(
                "missing format_id $key in format_string for $devicetype_dn");
            $has_all_params = 0;
        }
    }
    if ( !$has_all_params ) {
        return 0;
    }

    if ( ${ $self->{'generators'} }{$devicetype_dn} ) {
        $logger->info(
            'will overwrite generator for devicetype_dn ' . $devicetype_dn );
    }

    #add all information about generator
    ${ $self->{'generators'} }{$devicetype_dn} = {
        'format_string' => $params{'format_string'},
        'format_ids'    => $params{'format_ids'},
        'devicetype_dn' => $params{'devicetype_dn'},
        'keep_old_cn'   => $params{'keep_old_cn'}
    };
    return 1;
}

=head2 addGeneratorByEntry

Adds a generator for a given Net::LDAP::Entry object. The entry will be searched for
attributes C<dbkObjectIdFormatString> and C<dbkObjectIdGenerator>. Those will be
used to form C<format_string> and C<format_ids>, which will be passed to C<addGeneratorForDevicetype>.

=over 4

=item Parameter

	- Net::LDAP::Entry object of a dbkDeviceType

=item Return

	- boolean whether adding generator was successful

=back

=cut

sub addGeneratorByEntry
{
    my $self  = shift;
    my $entry = shift;

    my $logger = Log::Log4perl->get_logger();

    if ( ! ( blessed($entry) && $entry->isa('Net::LDAP::Entry') ) ) {
        $logger->error(
            'addGeneratorByEntry has to be called with a Net::LDAP::Entry object'
        );
        return 0;
    }

    my %dynamic_generator = %{ $self->getDynamicGeneratorByDn( $entry->dn ) };
    return $self->addGeneratorForDevicetype(%dynamic_generator);
}

=head2 getDynamicGeneratorByDn

=over 4

=item Parameter

=item Return

=back

=cut

sub getDynamicGeneratorByDn
{
    my $self = shift;
    my $dn   = shift;

    my $logger = Log::Log4perl->get_logger();
    my $entry  = $self->{'sc_cache'}->getEntry($dn);
    if ($entry) {
        my $generator_dn = lc( $entry->dn );

        my $format_string;
        my %format_ids;
        my $keep_old_cn = 0;
        if (    $entry->exists('dbkObjectIdFormatString')
            and $entry->exists('dbkObjectIdGenerator') )
        {
            $format_string = $entry->get_value('dbkObjectIdFormatstring');

            # copy attribute values in our own list
            my $generators_ref =
              [ @{ $entry->get_value( 'dbkObjectIdGenerator', asref => 1 ) } ];
            foreach my $gen ( @{$generators_ref} ) {

                # Jeder Generator hat eine eindeutige ID.
                my $gen_id = $gen;
                $gen_id =~ s/.*id=([^,]*).*/$1/;
                if ( $gen_id eq "" ) {
                    $logger->error( "Generator "
                          . $generator_dn
                          . " has no or malformed id. Therefore we skip it." );
                    next;
                }
                if ( $format_ids{$gen_id} ) {
                    $logger->error( "Generator "
                          . $generator_dn
                          . " has duplicate id "
                          . $gen_id
                          . " Therefore we skip it." );
                    next;
                }

                # Alle anderen Attribute sind nicht zwingend vorhanden und von Generator
                # zu Generator eventuell unterschiedlich
                # If a generator has a key 'keep_old_cn', the generator will try to
                # keep the name when a device object is moved.
                my %gen_params;
                $gen =~ s/id=[^,]*,?//g;
                while ( $gen =~ m/=/ ) {
                    my $gen_param_key   = $gen;
                    my $gen_param_value = $gen;
                    $gen_param_key   =~ s/^([^=]*).*/$1/;
                    $gen_param_value =~ s/^[^=]*=([^,]*).*/$1/;
                    $gen_params{$gen_param_key} = $gen_param_value;
                    $gen =~ s/^[^,]*,?//;
                    if ( $gen_param_key eq 'keep_old_cn' ) {
                        $keep_old_cn = 1;
                    }
                }
                $format_ids{$gen_id} = \%gen_params;
            }
        }
        my %dynamic_generator = (
            'devicetype_dn' => $generator_dn,
            'format_string' => $format_string,
            'format_ids'    => \%format_ids,
            'keep_old_cn'   => $keep_old_cn
        );
        return \%dynamic_generator;
    } else {
        $logger->error("No name generator found for $dn in sc_cache.");
        return;
    }

}

=head getDeviceTypeDn

  my $devicetype_dn = $self->getDeviceTypeDn($old_entry, $new_location_entry);

=over 4

=item Parameter

list with entries to search for dbkDeviceTypeDn

=item Return

dn of the devicetype for given entries, or default devicetype if none could be found

=back

=cut

sub getDeviceTypeDn
{
    my ( $self, @entries ) = @_;
    my $logger = Log::Log4perl->get_logger();

    for my $entry ( @entries ) {
        if ( $entry->exists('dbkDeviceTypeDn') ) {
            my $devicetype_dn =
              lc( $entry->get_value('dbkDeviceTypeDn') );
            return $devicetype_dn if $devicetype_dn;
        }
    }

    my $devicetype_dn = $self->{'sc_cache'}->{'config'}->getWithDefault(
        'LEASES2LDAP_STANDARD_DEVICE_TYPE_DN',
        'cn=deviceknutpc,ou=global,o=debeka,c=de'
    );
    $logger->info(
        'no devicetype_dn could be found while generating name. Using '
          . $devicetype_dn );
    return $devicetype_dn;
}

{
    my %generators = (
        IP   => 'generateNameIP',
        LDAP => 'generateNameLDAP',
        XMLRPC => 'generateNameXMLRPC',
        DHCP => 'generateNameDHCP',
    );

=head2 dispatchGenerator

Dispatches generator to one configured in %generators hash.

  my $format_string = 'pc${pc_id}';
  my $id            = 'pc_id';
  $format_string = $self->dispatchGenerator('IP', 
        {
            'id' => 'pc_id',
            'format' => '%05d',
        },
        format_string => $format_string,
        ip_addr       => '10.1.1.1',
        network       => '10.10.1.0/22',
        netmask       => '255.255.252.0',
  );
  is $format_string, 'pc00257', 'pc name generated';

=over 4

=item Parameters

=item Return

Formatted string filled with newly generated values

=back

=cut

    sub dispatchGenerator
    {
        my ( $self, $format_id_type, %params ) = @_;
        if ( exists $generators{$format_id_type} ) {
            my $method = $self->can($generators{$format_id_type});
            $params{generator_ref}{id} = $params{format_id};
            return $self->$method(
                %params
            );
        }
        return $params{format_string};
    }
}

=head2 generateCn

generates a cn

=over 4

=item Parameter

 - new_location_entry
 - ip
 - new_devicetype_dn
 - old_entry

=item Return

 - string cn 

=back

=cut

sub generateCn
{
    my $self   = shift;
    my %params = (
        'new_location_entry' => undef,
        'old_entry'          => undef,
        'searchbase'         => undef,
        'new_entry_infos'    => undef,
        @_
    );
    my $logger = Log::Log4perl->get_logger();

    # figure out which devicetype should be used.
    # first take the one given explicitly
    my $devicetype_dn =
      lc( $params{'new_entry_infos'}->{'devicetype_dn'} || '' );
    my $old_entry          = $params{'old_entry'};
    my $ip_addr            = $params{'new_entry_infos'}->{'ip_addr'};
    my $new_location_entry = $params{'new_location_entry'};
    my $new_entry_infos    = $params{'new_entry_infos'};
    my %generators         = %{ $self->{'generators'} };
    if ( !%generators ) {
        %generators = ();
    }
    my $searchbase = $params{'searchbase'};
    if ($searchbase) {
        $searchbase = $self->{'sc_cache'}->{'ldap_base'};
    }

    if ( !$new_location_entry ) {
        $logger->error('no location_entry given');
        return;
    }

    $devicetype_dn ||=
      $self->getDeviceTypeDn( $old_entry, $new_location_entry, );
    $logger->debug( 'will use ' . $devicetype_dn . ' for generating name' );

    if ( !exists $generators{$devicetype_dn} ) {
        $generators{$devicetype_dn} =
          $self->getDynamicGeneratorByDn($devicetype_dn);
    }
    if ( $generators{$devicetype_dn}{'keep_old_cn'} ) {

        #if old dbkDeviceTypeDn and new are the same we can keep the cn
        if (    $old_entry
            and $old_entry->exists('dbkDeviceTypeDn')
            and lc( $old_entry->get_value('dbkDeviceTypeDn') ) eq
            lc($devicetype_dn) )
        {
            return $old_entry->get_value('cn');
        }
    }

    my $format_string = $generators{$devicetype_dn}{'format_string'};
    my %format_ids    = %{ $generators{$devicetype_dn}{'format_ids'} };

    my $network = $new_location_entry->get_value('ipNetworkNumber');
    my $netmask = $new_location_entry->get_value('ipNetmaskNumber');

    my $unique_generator_id;
    my $keep_old_cn;

    # iterate over all given generators.
    foreach my $format_id ( keys %format_ids ) {
        my $format_id_type   = uc( $format_ids{$format_id}{'type'} );
        my $format_id_format = $format_ids{$format_id}{'format'};

        $format_string = $self->dispatchGenerator(
            $format_id_type,
            format_id       => $format_id,
            generator_ref   => $format_ids{$format_id},
            format_string   => $format_string,
            ip_addr         => $ip_addr,
            location        => $new_location_entry,
            network         => $network,
            netmask         => $netmask,
            new_entry_infos => $params{'new_entry_infos'},
            searchbase      => $searchbase,
        );
        if ( !$format_string ) {
            return;
        }

        if ( $format_id_type eq "UNIQUE" ) {

            # Save UNIQUE generator for the last step
            if ($unique_generator_id) {
                $logger->error( "More than one UNIQUE Generator in "
                      . $format_id
                      . " . This will most certainly not work.\n" );
            }
            $unique_generator_id = $format_id;
        }
    }
    if ($unique_generator_id) {
        $logger->info( "Call generateNameUNIQUE "
              . $unique_generator_id . " => "
              . $format_ids{$unique_generator_id} );
        $format_ids{$unique_generator_id}{'id'} = $unique_generator_id;
        return $self->generateNameUNIQUE(
            generator_ref => $format_ids{$unique_generator_id},
            format_string => $format_string
        );
    }

    return $format_string;
}

=head2 generateNameIP

Generate a name based on ip address and network. This method can be used
to generate a name which is compatible to the old smartclient naming scheme.

=head3 Example

	my $format_string = $name_generator->generateNameIP(
		'generator_ref' => {
			'id' => 'ip_id',
			'format' => '%05d'
		},
		'format_string' => 'pc${ip_id}',
		'ip_addr' => '10.1.2.3',
		'network' => '10.1.2.0',
		'netmask' => '255.255.255.0'
	);

=over 4

=item Parameter

	- generator_ref
	- format_string
	- ip_addr
	- network
	- netmask

=item Return

	- formatted string or undef if none could be generated

=back

=cut

sub generateNameIP
{
    my $self = shift;

    my %params = (
        'generator_ref' => undef,
        'format_string' => undef,
        'ip_addr'       => undef,
        'network'       => undef,
        'netmask'       => undef,
        @_
    );
    my $generated_formatstring = $params{'format_string'};
    my $generator_ref          = $params{'generator_ref'};
    my $generator_format       = $generator_ref->{'format'};
    my $generator_id           = $generator_ref->{'id'};

    my $logger = Log::Log4perl->get_logger();

    #TODO: check if parameters are ok
    my $network = $params{'network'};
    $network =~ s-/\d+$--;
    my $netmask = $params{'netmask'};
    my $ip_addr = $params{'ip_addr'};

    my $network_block = SmartClient::IP->new( $network, $netmask );
    if ( !$network_block ) {
        $logger->error( 'could not build a SmartClient::IP object with ' 
              . $network . ':'
              . $netmask );
    }
    my $network_offset = $network_block->match( $params{ip_addr} );

    if ( !$network_offset ) {
        $logger->error( 'no matching ip ' 
              . $ip_addr
              . ' could be found in block '
              . $network . ':'
              . $netmask );
        return;
    }

    # SmartClient::IP->match can give back "0 ", so convert it to a number
    $network_offset += 0;

    if ( !$generator_format ) {
        $generator_format = '%05d';
    }

    $generated_formatstring =~ s/\$\{\Q$generator_id\E\}/$generator_format/;
    $generated_formatstring = sprintf $generated_formatstring, $network_offset;

    return $generated_formatstring;
}

=head2 generateNameUNIQUE

Generates a unique name with respect to searchbase. It will use a xmlrpc service 
instead of generating a unique name itself if 
C<ALL_GENERATOR_UNIQUE_NAME_XMLRPC_URL> and C<ALL_GENERATOR_UNIQUE_NAME_XMLRPC_METHOD>
are defined in C</etc/smartclient/base>. 

=head3 example

	my $format_string = $name_generator->generateNameUNIQUE(
		'generator_ref' => {
			'unique_id' => {
				'id' => 'unique_id',
				'format' => '%05d'
				'min' => 200,
				'max' => 300
			}
		},
		'format_string' => 'pc${unique_id}',
		'searchbase' => 'o=company,c=de'
	);

=over 4

=item Parameter

	- generator_ref
	- format_string
	- searchbase

=item Return

	- formatted string

=back

=cut

sub generateNameUNIQUE
{
    my $self   = shift;
    my %params = (
        'generator_ref' => undef,
        'format_string' => undef,
        'searchbase'    => undef,
        @_
    );

    my $logger = Log::Log4perl->get_logger();

    my %generator = %{ $params{'generator_ref'} };

    # if we should ask a central service for unique names, do it
    my $config = SmartClient::Config->new();
    if ( ref $config && $config->get('ALL_GENERATOR_UNIQUE_NAME_XMLRPC_URL', 1) ) {
        my $xmlrpc_url = $config->get('ALL_GENERATOR_UNIQUE_NAME_XMLRPC_URL');
        $generator{'url'} = $xmlrpc_url;
        $generator{'method'} =
          $config->getWithDefault( 'ALL_GENERATOR_UNIQUE_NAME_XMLRPC_METHOD',
            'generator.uniqueName' );
        $logger->info( "get generated unique name from xmlrpc "
              . $generator{'url'} . " by "
              . $generator{'method'} );
        return $self->generateNameXMLRPC(
            'generator_ref' => \%generator,
            'searchbase'    => $params{searchbase},
            'format_string' => $params{format_string}
        );
    }

    # no central service, so generate a unique name on your own
    my $generator_id_format = $generator{'format'};
    my $generator_id        = $generator{'id'};
    if ( !$generator_id_format ) {
        $logger->error( "no format for generator set. using \%04d for "
              . $generator_id . " => "
              . %generator );
        $generator_id_format = '%04d';
    }
    my $generator_id_min = $generator{'min'};
    my $generator_id_max = $generator{'max'};

    if ( !$generator_id_min ) {
        $generator_id_min = 1;
    }
    if ( !$generator_id_max ) {
        $generator_id_max = 10000;
    }
    my $unique_id = $self->getNextUniqueId(
        generator_id        => $generator_id,
        generator_id_min    => $generator_id_min,
        generator_id_max    => $generator_id_max,
        generator_id_format => $generator_id_format,
        format_string       => $params{format_string},
    );
    return $unique_id;
}

=head2 generateNameXMLRPC

generates a name by calling a xmlrpc service

=head3 Example

	my $format_string = $name_generator->generateNameXMLRPC(
		'generator_ref' => {
			'xmlrpc_id' => {
				'id' => 'xmlrpc_id',
				'type' => 'xmlrpc',
				'format' => '%05d',
				'method' => 'namespace.xmlrpcMethodName',
				'url' => 'http://server:9876/RPC2'
			}
		},
		'format_string' => 'pc${xmlrpc_id}',
		'searchbase' => 'o=company,c=de',
		'ip_addr' => '10.1.2.3',
		'network' => '10.1.2.0',
		'netmask' => '255.255.255.0'
	);

=over 4

=item Parameter

	- generator_ref a reference to a generator hash
	- format_string string with format for generator
	- searchbase ldap searchbase in which device will be put
	- ip_addr ip address of device to be named
	- network network in which device will be put
	- netmask netmask of network in which device will be put

=item Return

	- formatted format_string or undef if xmlrpc call failed

=back

=cut

sub generateNameXMLRPC
{
    my $self  = shift;
    my %param = (
        'generator_ref' => undef,
        'format_string' => undef,
        'searchbase'    => undef,
        'ip_addr'       => undef,
        'network'       => undef,
        'netmask'       => undef,
        @_
    );
    my $logger                    = Log::Log4perl->get_logger();
    my $generator_id_formatstring = $param{'format_string'};
    my $generator_ref             = $param{'generator_ref'};
    my $generator_id              = $generator_ref->{'id'};
    my $generator_id_format       = $generator_ref->{'format'};
    if ( !$generator_id_format ) {
        $logger->info('no generator id format was given. Using default');
        $generator_id_format = "%s";
    }

    # Dieser Typ soll auf einem XML-RPC Server (wie etwa Zope) eine beliebige
    # Methode aufrufen. Als Parameter werden die Id, der Formatstring und ein
    # eventuell vorhandener rpcgeneratortype übergeben.
    # Vom XML-RPC Server muss die
    require RPC::XML;
    require RPC::XML::Client;
    my $client = RPC::XML::Client->new( $generator_ref->{'url'} );
    if (   $generator_ref->{'realm'}
        && $generator_ref->{'user'}
        && $generator_ref->{'pass'} )
    {
        $client->credentials(
            $generator_ref->{'realm'},
            $generator_ref->{'user'},
            $generator_ref->{'pass'}
        );
    }

    # Standardwerte in die zu uebergebene Struktur schreiben
    my %struct;
    for my $my_key ( keys %{$generator_ref} ) {
        $logger->info( 'XMLRPC Generator: ' 
              . $my_key . ' => '
              . $generator_ref->{$my_key} );
        $struct{$my_key} = RPC::XML::string->new( $generator_ref->{$my_key} );
    }

    # Alle anderen Werte in die uebergebenen Struktur
    $struct{'generatorId'}       = RPC::XML::string->new($generator_id);
    $struct{'generatorIdFormat'} = RPC::XML::string->new($generator_id_format);
    $struct{'generatorIdFormatstring'} =
      RPC::XML::string->new($generator_id_formatstring);

    # fill in optional params
    my %optional_params = (
        searchbase => 'ldapSearchBase',
        ip_addr    => 'ipAddress',
        network    => 'network',
        netmask    => 'netmask'
    );
    foreach my $option_key ( keys %optional_params ) {
        if ( $param{$option_key} ) {
            $struct{ $optional_params{$option_key} } =
              RPC::XML::string->new( $param{$option_key} );
        }
    }

    $logger->info( "Try to ask XMLRPC Server for: "
          . $struct{'generatorId'}->as_string() . "/"
          . $struct{'generatorIdFormatstring'}->as_string() );

    my $response = $client->send_request( $generator_ref->{'method'},
        RPC::XML::struct->new(%struct) );

    if ($response) {
        if ( !${ $response->value }{'name'} ) {
            my $reason = "No reason given";
            if ( ${ $response->value }{'reason'} ) {
                $reason = ${ $response->value }{'reason'};
            }
            $logger->error( "XMLRPC gave no name for "
                  . $struct{'ipAddress'}->as_string() . " in "
                  . $struct{'ldapSearchBase'}->as_string()
                  . " [ reason="
                  . $reason . ", id="
                  . $struct{'generatorId'}->as_string()
                  . ", format="
                  . $struct{'generatorIdFormat'}->as_string()
                  . ", Formatstring="
                  . $struct{ 'generatorIdFormatstring'->as_string() }
                  . " ]" );
            return;
        } else {
            $generator_id_formatstring = ${ $response->value }{'name'};
            $logger->info( "XMLRPC answer: "
                  . $generator_id_formatstring . " [ "
                  . %{ $response->value }
                  . " ] " );
        }
    } else {
        $logger->error( "XMLRPC call gave no answer for URL:"
              . $generator_ref->{'url'}
              . " method:"
              . $generator_ref->{'method'} );
        return;
    }
    return $generator_id_formatstring;
}

=head2 generateNameLDAP

Generates a name using ldap entries.
Since the entry for which this name will be generated does not exist, the objects
above the location (including the location) will be used as source for ldap values.

=head3 Example

    my $format_string = $name_generator->generateNameLDAP(
		'generator_ref' => {
			'id'        => 'ldap',
			'format'    => '%s',
			'attribute' => 'ou'
		},
		'format_string' => 'rl${ldap}',
		'location'      => $sc_cache->getEntry($location_dn)
	);

=over 4

=item Parameter

hash consisting of

=over 4

=item C<generator_ref>

reference to hash containing the parameters for the generator.

=over 4

=item C<id>

name of the placeholder in the format string

=item C<format>

format used when replacing the placeholder. Has L<sprintf> format.

=item C<attribute>

name of the ldap attribute which will be used as replacement.

=back

=item C<format_string>

string with a placeholder which will be filled by generator using C<generator_ref>

=item C<location>

C<Net::LDAP> object of the new location in which the entry will be created 

=back

=item Return

=back

=cut

sub generateNameLDAP
{
    my $self   = shift;
    my %params = (
        'generator_ref' => undef,
        'format_string' => undef,
        'location'      => undef,
        @_
    );
    my $generator_id_formatstring = $params{'format_string'};
    my $generator_ref             = $params{'generator_ref'};
    my $location                  = $params{'location'};
    my $logger                    = Log::Log4perl->get_logger();

    if ( ref $generator_ref and ref $location and $generator_id_formatstring ) {
        my $generator_id     = $generator_ref->{'id'};
        my $generator_format = $generator_ref->{'format'} || '%s';
        my $generator_attr   = $generator_ref->{'attribute'} || 'ou';
        my $ldap_entry       = $self->{'sc_cache'}->treeSearch(
            'start_dn' => $location->dn(),
            'attr'     => $generator_attr
        );
        my $ldap_value = "";
        if ( ref $ldap_entry ) {
            $ldap_value = $ldap_entry->get_value($generator_attr);
        } else {
            $logger->warn( "No value for attribute ${generator_attr} found for "
                  . $location->dn() );
        }
        $generator_id_formatstring =~
          s/\$\{\Q$generator_id\E\}/$generator_format/;
        $generator_id_formatstring = sprintf $generator_id_formatstring,
          $ldap_value;
        return $generator_id_formatstring;
    }

    # we were called with wrong parameters, so return nothing
    return;
}

my $dhcp_checks = {
    'no_check'      => sub { 1 },
    'ends_with_mac' => sub {
        my ( $cn, $infos ) = @_;
        my $mac = $infos->{'mac_address'};
        if ( !$mac || $mac !~ m/^(?:[a-f0-9]{2}:){5}[a-f0-9]{2}$/ ) {
            return;
        }
        $mac =~ s/[:]//g;
        return $cn =~ m/$mac$/i;
    },
};

sub generateNameDHCP
{
    my $self   = shift;
    my %params = (
        'generator_ref'   => undef,
        'format_string'   => undef,
        'new_entry_infos' => undef,
        @_
    );
    my $generator_id_formatstring = $params{'format_string'};
    my $generator_ref             = $params{'generator_ref'};
    my $new_entry_infos           = $params{'new_entry_infos'};
    my $logger                    = Log::Log4perl->get_logger();

    if ( ref $generator_ref and $generator_id_formatstring ) {
        my $generator_id     = $generator_ref->{'id'};
        my $generator_format = $generator_ref->{'format'} || '%s';
        my $generator_check  = $generator_ref->{'check'} || 'no_check';
        my $generator_attr   = $generator_ref->{'attribute'} || 'dhcp_hostname';
        my $dhcp_value       = $params{'new_entry_infos'}->{$generator_attr};
        if ( !exists $dhcp_checks->{$generator_check} ) {
            $logger->error("[DHCP-DATA] check $generator_check not defined");
            return;
        }
        if ( !$dhcp_value ) {
            $logger->error("[DHCP-DATA] no value for $generator_attr found");
            return;
        }
        $generator_id_formatstring =~
          s/\$\{\Q$generator_id\E\}/$generator_format/;
        $generator_id_formatstring = sprintf $generator_id_formatstring,
          $dhcp_value;
        if ( !$dhcp_checks->{$generator_check}
            ->( $generator_id_formatstring, $new_entry_infos ) )
        {
            $logger->error(
                "[DHCP-DATA] check $generator_check for attribue $generator_attr in $generator_id_formatstring failed"
            );
            return;
        }
        return $generator_id_formatstring;
    }

    # we were called with wrong parameters, so return nothing
    return;
}

=head2 getNextUniqueId

searches for the first free number in range given by C<generator_id_min>
and C<generator_id_max> with namespace C<generator_id>.
To accomplish this, the entire C<entrycache> is searched. To do this the generator_id in the 
format string is replaced by C<"">. Then all C<cn> are searched for and stored in a hash. 
Lastly the first free unique C<cn> is searched by iterating from C<generator_min> to 
C<generator_max> and inserting this id into the original format string.

Warning:
	the unique generator should be the last one in the C<format_string> and no character 
	should follow it. Otherwise you might experience weird results.

Example:
 	format string: rl1234${ziffer}
 	generator_id: ziffer
 	generated search filter: cn=rl1234 

=over 4

=item Parameter

	- generator_id namespace for generator
	- generator_id_min minimal number for searchstart
	- generator_id_max maximum number. 
	- generator_id_format special format string to be used for name generation
	- format_string formatstring
	- searchbase

=item return

	- next unused number for generator namespace identified by generator_id
	or undef is none could be build

=back

=cut

sub getNextUniqueId
{
    my $self   = shift;
    my %params = (
        generator_id                => undef,
        generator_id_min            => undef,
        generator_id_max            => undef,
        generator_id_format         => undef,
        generator_id_attribute_name => 'cn',
        format_string               => undef,
        searchbase                  => undef,
        @_,
    );

    my $sc_cache   = $self->{'sc_cache'};
    my $searchbase = $params{'searchbase'};
    if ( !$searchbase ) {
        $searchbase = $sc_cache->getBase();
    }
    my $generator_id   = $params{generator_id};
    my $attribute_name = $params{generator_id_attribute_name};

    my $logger = Log::Log4perl->get_logger();

    # Filter bei dem die Generator Id durch "" ersetzt wird.
    my $filter = $params{format_string};
    $filter =~ s/\$\{\Q$generator_id\E\}.*//;

    # printf Format, für die Erzeugung der id
    my $generator_print_format = $params{format_string};
    $generator_print_format =~
      s/\$\{\Q$generator_id\E\}/$params{generator_id_format}/;
    $logger->info( "Using " . $filter . " for ldap search in " . $searchbase );

    my %id_hash;
    my $dn_iterator = $sc_cache->getAllEntryDnIterator();
    while ( my $dn = $dn_iterator->() ) {
        my $entry = $sc_cache->getEntry($dn);
        if ( $entry->exists($attribute_name) ) {
            my $value = $entry->get_value($attribute_name);
            if ( $value =~ m/^\Q$filter\E/i ) {
                if ( !$id_hash{$value} ) {
                    $id_hash{$value} = 1;
                } else {

                    # Wenn dieser Code ausgeführt wird, so bedeutet das, dass dieser cn nicht
                    # eindeutig in der Suchbasis ist.
                    ++$id_hash{$value};
                    $logger->error( "Attribute "
                          . $attribute_name . " = "
                          . $value
                          . " from "
                          . $entry->dn
                          . " is not unique in "
                          . $searchbase
                          . "." );
                }
            }
        }
    }
    my $min_id = $params{generator_id_min};
    my $max_id = $params{generator_id_max};
    my $unique_id;
    for ( my $counter = $min_id ; $counter < $max_id ; $counter++ ) {
        $unique_id = sprintf( $generator_print_format, $counter );
        if ( !$id_hash{$unique_id} ) {
            $logger->info( "Will use " . $unique_id . " as unique id." );
            return $unique_id;
        }
    }
    $logger->error( "Couldn't find a unique Id for generator "
          . $params{generator_id}
          . " in base "
          . $searchbase
          . " for range ["
          . $min_id . ","
          . $max_id
          . "]  with formatstring "
          . $generator_print_format );
    return;
}

=head2 getGenerators

get all configured generators

=over 4

=item Parameter

	- none

=item Return

	- reference to hash of all generators or undef if none is configured

=back

=cut

sub getGenerators
{
    my $self = shift;
    return $self->{'generators'};
}

=head2 getGeneratorForDn

get a generator for a given dn

=over 4

=item Parameter

	- string dn of generator

=item Return

	- reference to hash for given generator or undef is none is found

=back

=cut

sub getGeneratorByDn
{
    my $self         = shift;
    my $generator_dn = shift;
    return $self->{'generators'}->{ lc($generator_dn) };
}

=head2 remove

used to notify the naming generator of removal of an entry in the ScCache

=over 4

=item Parameter

string representing the dn of the entry which was deleted in the ScCache

=item Return

always true

=back

=cut

sub remove
{
    my $self = shift;
    my $dn   = shift;

    if ( $dn && exists $self->{'generators'}->{ lc($dn) } ) {
        remove( $self->{'generators'}->{ lc($dn) } );
    }

    return 1;
}
1;
