#!/usr/bin/perl -w

#
# $Id: Queue.pm 8022 2012-07-19 11:36:18Z HZuchel $
#

#
# package name
#
package SmartClient::Queue;
use strict;
use warnings;

#
# version  of the package
#
our $VERSION = 0.5;

=encoding utf8

=head1 NAME

C<SmartClient::Queue> - shared memory queues


=head1 SYNOPSIS

    use SmartClient::Queue;
    my $q = SmartClient::Queue->new( 'shid' );
    $q->enqueue("foo", "bar");
    my $foo = $q->dequeue;    # The "bar" is still in the queue.
    my $foo = $q->dequeue_nb; # returns "bar", or undef if the queue was empty
    my $left = $q->pending;   # returns the number of items still in the queue



=head1 DESCRIPTION

=head2 Purpose

A queue, as implemented by C<SmartClient::Queue> is a shared memory
data structure much like a list.  Any process can safely 
add elements to the end of the list, or remove elements from the head 
of the list. (Queues don't permit adding or removing elements from 
the middle of the list).
Using shared memory different processes can use the same queue.


=head2 Realisation

uses IPC Shared Memory for Inter Process Communication.
The Queue only uses one IPC SHM segment

=head2 Error handling

=head2 Context and Dependencies

=over 4

=item Context

=item Dependencies


=back



=head2 Interfaces

=begin comment

    Command line parameter are already descripied 
    at SYNOPSIS and usage

=end comment

=over 4

=item Programm

=item User

=back



=head2 Status

=begin comment

    descripes the TODOs/Errors for this programm

=end comment

=head2 Installation and other Notes



=head1 AUTHOR

    smartclient@dass-it.de




=head1 SEE ALSO

L<Threads::Queue>
L<IPC::Shareable>

=cut

#
# module definition
#

# =head1 symbols exported by default
#
# - variable1
# - function1
# - function2
#
# =cut
#
# our @EXPORT = qw(  );

# =head1 symbols exported on request
#
# =cut
#
# our @EXPORT_OK = qw( );

#
# used modules
#
use Data::Dumper;
use IPC::Shareable qw( LOCK_EX LOCK_SH LOCK_NB LOCK_UN );
use SmartClient::Util;

#
# declarations
#

=head1 Constants

=head1 Global Variables




=head1 Functions

=head2 hasValue()

checks if a specific value is part of a LDAP entry attribute (multivalue)

Example:

 if ( hasValue(
        entry => $ldap_entry,
        attr_name => 'objectclass',
        value => 'dbkUpdate' ) {

checks, if the LDAP entry has the objectclass dbkUpdate


=cut

=head2 new()

The C<new> function creates a new empty queue.

=cut

sub new
{
    my $class = shift;

    # TODO: warn, if glue has more than 4 characters
    my $shm_id_queue = shift;    # 'data';

    my $busy_waiting_period = shift || 0.1;

    my $data = {
        'shm_id_queue'        => $shm_id_queue,
        'busy_waiting_period' => $busy_waiting_period,
        'q'                   => undef,
    };

    my $result = bless( $data, $class );

    $data->connect;

    return $result;
}

=head2 connect()

The C<connect> function connects to the defined queue.

=cut

sub connect
{
    my $self = shift;
    my $q    = $self->{'q'};

    # check options (master | slave)
    my %options = (
        create    => 'yes',
        exclusive => 0,
        mode      => 0644,
        size      => 128000,

        #destroy   => 'yes',
    );

    if ( !$q ) {
        my @array = ();
        tie @array, 'IPC::Shareable', $self->{'shm_id_queue'}, {%options}
          or return 0;
        $self->{'q'} = \@array;
        $q = $self->{'q'};

        # initialize queue (push data in and out)
        push @$q, 1;
        pop @$q;
    }

    return 1;
}

=head2 pending

The C<pending> method returns the number of items still in the queue.

=over 4

=item Parameter

 - [boolean: no_unlock: default is to unlock, 
    but internaly a version without unlocking is required. Only for internal use]

=item Return

  integer: number of entries in queue

=back

=cut

sub pending
{
    my $self      = shift;
    my $no_unlock = shift;
    my $q         = $self->{'q'};

    ( tied @$q )->shlock(LOCK_SH);
    my $number = scalar(@$q);
    ( tied @$q )->shunlock() if ( !$no_unlock );

    return $number;
}

=head2 enqueue( LIST )

The C<enqueue> method adds a list of scalars on to the end of the queue.
The queue will grow as needed to accommodate the list.

=over 4

=item Parameter

 - LIST

=item Return

  true

=back

=cut

sub enqueue
{
    my ($self, @params) = @_;
    my $q      = $self->{'q'};

    ( tied @$q )->shlock();
    push @$q, @params;
    ( tied @$q )->shunlock();
    return 1;
}

=head2 dequeue

The C<dequeue> method removes a scalar from the head of the queue and
returns it. If the queue is currently empty, C<dequeue> will block the
thread until another thread C<enqueue>s a scalar.

=cut

sub dequeue
{
    # parameter is timeout.
    # if undef, no timeout will be used
    my ($self, $timeout) = @_;

    my $q = $self->{'q'};

    # tell pending, not to unlock
    # this way, it can be gurantied,
    # that dequeue really delivers an entry
    my $NO_UNLOCK = 1;

    # if queue is empty, wait until something comes in
    while ( ( !$self->pending($NO_UNLOCK) )
        && ( !defined($timeout) || ( $timeout > 0 ) ) )
    {

        # nothing in queue. Unlock the queue and wait
        ( tied @$q )->shunlock();
        sleep_log( $self->{'busy_waiting_period'} );
        $timeout -= $self->{'busy_waiting_period'} if defined($timeout);
    }

    my $result = $self->dequeue_nb();

    return $result;
}

=head2 dequeue_nb

The C<dequeue_nb> method, like the C<dequeue> method, removes a scalar from
the head of the queue and returns it. Unlike C<dequeue>, though,
C<dequeue_nb> won't block if the queue is empty, instead returning
C<undef>.

=over 4

=item Parameter

 - none

=item Return

  first entry of the queue or
  undef, if queue is empty

=back

=cut

sub dequeue_nb
{
    my $self = shift;
    my $q    = $self->{'q'};

    ( tied @$q )->shlock();
    my $result = shift @$q;
    ( tied @$q )->shunlock();

    return $result;
}

=head2 remove

Calling remove() on the object underlying a tie()d variable removes the associated shared memory segment. The segment is removed irrespective of whether it has the destroy option set or not and irrespective of whether the calling process created the segment.

=cut

sub remove
{
    my $self = shift;
    my $q    = $self->{'q'};

    ( tied @$q )->remove;
    if ($q) {
        $self->{'q'} = undef;
    }
    return;
}

=head2 cleanup

This is a class method that provokes IPC::Shareable to remove all shared memory segments created by the process. Segments not created by the calling process are not removed.

=cut

sub cleanup
{
    my $self = shift;

    $self->remove();

    return IPC::Shareable->clean_up;

}

1;
