#!/usr/bin/perl -w
use strict;

#
# $Id: RpcXmlUtil.pm 8044 2012-07-25 10:13:38Z FSchumacher $
#

#
# package name
#
package SmartClient::RpcXmlUtil;

#
# version  of the package
#
our $VERSION = 0.1;

=encoding utf8

=head1 NAME

C<RpcXmlUtil>



=head1 SYNOPSIS

=begin comment

    short description of the program 
    and command line parameter

=end comment



=head1 DESCRIPTION

=head2 Purpose

=head2 Realisation

=head2 Error handling

=head2 Context and Dependencies

=over 4

=item Context

=item Dependencies


=back



=head2 Interfaces

=begin comment

    Command line parameter are already descripied 
    at SYNOPSIS and usage

=end comment

=over 4

=item Programm

=item User

=back



=head2 Status

=begin comment

    descripes the TODOs/Errors for this programm

=end comment


=head2 Installation and other Notes



=head1 AUTHOR

    smartclient@dass-it.de




=head1 SEE ALSO



=cut

#
# used modules
#
use Data::Dumper;
use Log::Log4perl;
use Log::Log4perl::Level;
use Time::HiRes qw( gettimeofday tv_interval );
use RPC::XML;
use RPC::XML::Client;
use SmartClient::Constant;
use POSIX;
use Scalar::Util 'blessed';

#
# module definition
#

use base 'Exporter';

=head1 symbols exported by default

- variable1
- function1
- function2

=cut

our @EXPORT =
  qw(createXmlRpcConnection xmlrpc_convertHashStrings xmlrpc_convertHashInts xmlrpc_convertHashBooleans xmlrpc_convertRefToRpcXmlStructure xmlrpc_convertResult getResult notifyOfModification getDeviceStatus waitForDeviceStatus getScheduleUpdateWorkstations setUpdate createSecondaryConfig setKnutStatus);

=head1 symbols exported on request

=cut

our @EXPORT_OK = qw( );

#
# declarations
#

=head1 Constants

=head1 Global Variables




=head1 Functions

=head2 createXmlRpcConnection

initialize a XML-RPC client and tests the connection.
Logs a warning and returns undef on failure.

=over 4

=item Parameter

 - URL of the XML-RPC Server

=item Return

  RPC-XML::Client

=back

=cut

sub createXmlRpcConnection( $ )
{
    my $logger = Log::Log4perl->get_logger();
    my $url    = shift;

    my $conn     = RPC::XML::Client->new($url);
    my $response = $conn->send_request('system.listMethods');
    if ( ref $response ) {
        $logger->debug(
            "$url provides: " . join( ', ', @{ $response->value } ) );
    } else {
        $logger->warn( "failed to connect to $url: " . $response );
    }

    return $conn;
}

=head2 xmlrpc_convertHashStrings( $$$ )

takes the keys from the source hash, 
convert them to XML-RPC strings 
and store them in target hash

=over 4

=item parameter

    - hash: target XML-RPC hash to be filled
    - hash: source
    - list: keys to be converted

=item return

    result is stored in first parameter

=back

=cut

sub xmlrpc_convertHashStrings
{
    my $xmlrpc_hash = shift;
    my $hash        = shift;
    my $aref_keys   = shift;

    foreach my $key ( @{$aref_keys} ) {
        if ( defined( $hash->{$key} ) ) {
            $xmlrpc_hash->{$key} = RPC::XML::string->new( $hash->{$key} );
        }
    }
}

=head2 xmlrpc_convertHashInts( $$$ )

takes the keys from the source hash,
convert them to XML-RPC integers 
and store them in target hash

=over 4

=item parameter

    - hash: target XML-RPC hash to be filled
    - hash: source
    - list: keys to be converted

=item return

    result is stored in first parameter

=back

=cut

sub xmlrpc_convertHashInts
{
    my $xmlrpc_hash = shift;
    my $hash        = shift;
    my $aref_keys   = shift;

    foreach my $key ( @{$aref_keys} ) {
        if ( defined( $hash->{$key} ) ) {
            $xmlrpc_hash->{$key} = RPC::XML::int->new( $hash->{$key} );
        }
    }
}

=head2 xmlrpc_convertHashBooleans( $$$ )

takes the keys from the source hash,
convert them to XML-RPC booleans 
and store them in target hash

=over 4

=item parameter

    - hash: target XML-RPC hash to be filled
    - hash: source
    - list: keys to be converted

=item return

    result is stored in first parameter

=back

=cut

sub xmlrpc_convertHashBooleans
{
    my $xmlrpc_hash = shift;
    my $hash        = shift;
    my $aref_keys   = shift;

    foreach my $key ( @{$aref_keys} ) {
        if ( defined( $hash->{$key} ) ) {
            $xmlrpc_hash->{$key} = RPC::XML::boolean->new( $hash->{$key} );
        }
    }
}

=head2 xmlrpc_convertResult( $;$ )

checks if a XML Rpc response marks errors.
If yes, log in error message and code

=over 4

=item parameter

    - href: XML-RPC response
    - string: error message prefix

=item return

    result is stored in first parameter

=back

=cut

sub xmlrpc_convertResult( $;$ )
{
    my $logger = Log::Log4perl->get_logger();

    my $response = shift;
    my $error_message = shift || "";

    my $resp_data;

    #$logger->debug($response);

    # if response is not a hash (object),
    # an error is occured
    if ( !ref($response) ) {
        my $code = XMLRPC_INVALID_RESPONSE;
        $logger->error("$error_message: $response ($code)");
        return ($code);
    }

    # convert to regular hash
    $resp_data = $response->value();

    #$logger->debug( Dumper($resp_data) );

    # check for default XML RPC errors
    # eg. calling with wrong parameter
    if ( $response->is_fault() ) {

        my $code = XMLRPC_FAULT;
        $logger->error( "$error_message: "
              . $resp_data->{'faultString'}
              . " ($code, "
              . $resp_data->{'faultCode'}
              . ")" );
        return ($code);

    }

    my $response_code = $resp_data->{'code'};
    if ( !defined($response_code) ) {
        $response_code = XMLRPC_INCOMPLETE;
    }
    my $response_message = $resp_data->{'message'} || "";

    if ( $response_code != XMLRPC_OK ) {
        $logger->error("$error_message: $response_message  ($response_code)");
    }
    return ( $response_code, $resp_data );
}

=head2 xmlrpc_convertRefToRpcXmlStructure

converts input parameter recursively to RPC::XML::* objects

=head3 Example

  my rpc_xml_structure = xmlrpc_convertRefToRpcXmlStructure([{'test' => 1}, 'String', [0.23, "whatever"]]);

=over 4

=item Parameter

a scalar or a reference to hash or array

=item Return

RPC::XML::* object

=back

=cut

sub xmlrpc_convertRefToRpcXmlStructure
{
    my $ref = shift;
    my $level = shift || 0;
    return if ( $level > 20 );
    my $type = ref $ref;
    if ( $type eq "ARRAY" ) {
        my @new_array = ();
        foreach my $sub_ref (@$ref) {
            push @new_array,
              xmlrpc_convertRefToRpcXmlStructure( $sub_ref, $level + 1 );
        }
        return RPC::XML::array->new( \@new_array );
    } elsif ( $type eq "HASH" ) {
        my $href_new_hash = {};
        foreach my $sub_key ( keys %$ref ) {
            $href_new_hash->{$sub_key} =
              xmlrpc_convertRefToRpcXmlStructure( $ref->{$sub_key},
                $level + 1 );
        }
        return RPC::XML::struct->new($href_new_hash);
    } elsif ( length( ref $ref ) > 0 ) {

        #we are a perl object, but we treat it as a hash (problems ahead)
        my $href_new_hash = {};
        foreach my $sub_key ( keys %$ref ) {
            $href_new_hash->{$sub_key} =
              xmlrpc_convertRefToRpcXmlStructure( $ref->{$sub_key},
                $level + 1 );
        }
        return RPC::XML::struct->new($href_new_hash);
    } elsif ( defined $ref ) {
        if ( $ref =~ m/[<>&]/o ) {
            return RPC::XML::base64->new($ref);
        } elsif ( $ref =~ m/^[+-]?\d+$/o ) {
            return RPC::XML::int->new($ref);
        } elsif ( $ref =~ m/^[+-]?\d*\.\d*$/o ) {
            return RPC::XML::double->new($ref);
        } else {
            return RPC::XML::string->new($ref);
        }
    }
    return;
}

=head2 isFinalResponseCode( $ )

checks the XML-RPC response code 
and returns, if a retry can useful

=over 4

=item Parameter

    int: XML-RPC response code

=item Return

    undef: if response code is unknown
    0: RETRY (not final)
    1: FINAL_GOOD
    2: FINAL_BAD

=back

=cut

sub isFinalResponseCode( $ )
{
    my $code = shift;

    use constant RETRY      => 0;
    use constant FINAL_GOOD => 1;
    use constant FINAL_BAD  => 2;

    if ( !defined($code) ) {
        return RETRY;
    } elsif ( $code == XMLRPC_OK ) {
        return FINAL_GOOD;
    } elsif (
        ( $code == XMLRPC_NO_SUCH_ID )
        or ( $code == XMLRPC_NO_PERMISSION )
        or ( $code == XMLRPC_ERROR_MANUAL )
        or ( $code == XMLRPC_UNKNOWN_ACTION )
        ## no retry on LDAP error,
        ## because some LDAP action may already be performed
        or ( $code == XMLRPC_LDAP_FAILURE )
        or ( $code == XMLRPC_INCOMPLETE )
        or ( $code == SHARED_FUNCTION_DOES_NOT_EXIST )
        or ( $code == SHARED_FUNCTION_STORE_RESULT_ERROR )
      )
    {
        return FINAL_BAD;
    } elsif ( ( $code == XMLRPC_SHARED_FUNCTION_TIMEOUT )
        or ( $code == XMLRPC_INVALID_RESPONSE )
        or ( $code == XMLRPC_TIMEOUT )
        or ( $code == XMLRPC_FAULT ) )
    {
        return RETRY;
    } else {

        # unknown response code
        return;
    }
}

=head2 xmlrpc_sendRequest( % )

takes care about a XML-RPC request:
sends it and checks the result.
If result is a temporary error (eg. HTTP timeout)
the request is send again, until own timeout is expired.
Is server side shared function is pending 
(detected from returned 'id')
a getResult is tried first, 
before sending a new shared funtion request

=over 4

=item Parameter

    'rpc_xml_daemon'     
    'xmlrpc_function'    
    'href_xmlrpc_params' 
    'timeout'            

=item Return

    result from 'xmlrpc_function', 
    already converted to Perl variables

=back

=cut

sub xmlrpc_sendRequest( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon'  => undef,
        'xmlrpc_function' => undef,
        ## default: empty struct
        'href_xmlrpc_params' => RPC::XML::struct->new(),
        'timeout'            => 0,
        @_,
    );

    my $rpc_xml_daemon     = $params{'rpc_xml_daemon'};
    my $xmlrpc_function    = $params{'xmlrpc_function'};
    my $href_xmlrpc_params = $params{'href_xmlrpc_params'};
    if (   blessed($href_xmlrpc_params)
        && $href_xmlrpc_params->isa("RPC::XML::struct")
        && defined $href_xmlrpc_params->{'timeout'}
        && ( $href_xmlrpc_params->{'timeout'} > RPCXML_TIMEOUT ) )
    {
        $logger->warn(
            "Timout for XMLRPC function is higher than allowed. Will set it to "
              . RPCXML_TIMEOUT
              . ". Params for method $xmlrpc_function were: "
              . Dumper($href_xmlrpc_params) );
        $href_xmlrpc_params->{'timeout'} = RPC::XML::int->new(RPCXML_TIMEOUT);
    }
    my $timeout = $params{'timeout'} || 0;

    # store start time
    my $t0 = [gettimeofday];

    my $code;
    my $resp_data;
    my $elapsed = 0;
    my $try     = 0;
    ## TODO: get it from parameter?
    my $max_getResult     = 1;
    my $getResult_counter = 0;
    my $id;

    # because $timeout may be larger than HTTP timeout
    # request is handled by a loop
    do {

        sleep $try;
        $try++;
        if ( $try > 1 ) {
            $logger->debug(
                "$xmlrpc_function: $try. try (elapsed time: $elapsed). Result so far: "
                  . Dumper($resp_data) );
        }

        my $resp;

        #  if second try and id from former call are known,
        #  this can be used instead of
        #  creating a new shared function request
        if (   ( not isFinalResponseCode($code) )
            && $id
            && ( $getResult_counter < $max_getResult ) )
        {

            # request has already been send,
            # but result is still pending
            # => trying again, to get result
            $logger->debug(
                "$xmlrpc_function [id: $id] try to get pending response");
            $resp =
              $rpc_xml_daemon->send_request( 'getResult',
                RPC::XML::struct->new( 'id' => RPC::XML::string->new($id) ) );
            $getResult_counter++;
        } else {

            # if the 'href_xmlrpc_params'-struct is default (empty RPC::XML::struct)
            # send_request must call with the name of the remote routine as only argument
            if ( ref $href_xmlrpc_params->value eq 'HASH'
                and !keys %{ $href_xmlrpc_params->value } )
            {
                $resp = $rpc_xml_daemon->send_request($xmlrpc_function);
            } else {
                $resp =
                  $rpc_xml_daemon->send_request( $xmlrpc_function,
                    $href_xmlrpc_params );
            }
            $getResult_counter = 0;
        }

        ( $code, $resp_data ) = xmlrpc_convertResult( $resp,
                "$xmlrpc_function "
              . Dumper( $href_xmlrpc_params->value )
              . " failed" );

        if ( ref($resp_data) ) {
            $id = $resp_data->{'id'};
        }

        $elapsed = tv_interval( $t0, [gettimeofday] );

      } while ( ( not isFinalResponseCode($code) )
        and ( $elapsed < $timeout ) );

    # default log level for result message
    my $log = $DEBUG;

    # create result message string
    my $stat_string = "$xmlrpc_function ";
    if ($id) {
        $stat_string .= "[id: $id] ";
    }
    if ( defined $code ) {
        $stat_string .= "result code: $code";
    }
    $stat_string .= "; duration: $elapsed";

    # no timeout, therefore valid result
    if ( $elapsed < $timeout ) {
        if ( $try > 1 ) {

            # if try > 1 then error messages has been displayed,
            # therefore we should also display the success code
            $log = $INFO;
        }
    }
    $logger->log( $log, "$stat_string" );

    return $resp_data;
}

=head2 getResult(%)

=over 4

=item Parameter

        'rpc_xml_daemon' =>
        'device_dn' =>
        'timeout'       => 0 (default: no timeout => single try)


=item Return

    hash (already converted from XML-RPC)
    or undef on error

=back

=cut

sub getResult( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon' => undef,
        'id'             => undef,
        'timeout'        => 0,
        @_,
    );

    my $rpc_xml_daemon = $params{'rpc_xml_daemon'};
    my $id             = $params{'id'};
    my $timeout        = $params{'timeout'};

    $logger->debug("getResult($id) (timeout: $timeout)");

    return xmlrpc_sendRequest(
        'rpc_xml_daemon'  => $rpc_xml_daemon,
        'xmlrpc_function' => 'getResult',
        'href_xmlrpc_params' =>
          RPC::XML::struct->new( 'id' => RPC::XML::string->new($id) ),
        'timeout' => $timeout,
    );
}

=head2 notifyOfModification(%)

=over 4

=item Parameter

        'rpc_xml_daemon'    =>
        'aref_modified_dns' => reference to an array of strings with modified dns
        'timeout'           => 0 (default: no timeout => single try)

=item Return

    hash (already converted from XML-RPC)
    or undef on error

=back

=cut

sub notifyOfModification ( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon'    => undef,
        'aref_modified_dns' => undef,
        'timeout'           => 0,
        @_
    );

    my $rpc_xml_daemon    = $params{'rpc_xml_daemon'};
    my $aref_modified_dns = $params{'aref_modified_dns'};
    my $timeout           = $params{'timeout'};

    my $dn_list =
      "* parameter \'aref_modified_dns\' is not a array-reference *";
    if ( ref $aref_modified_dns eq 'ARRAY' ) {
        $dn_list = join( ', ', @{$aref_modified_dns} );
    }

    $logger->debug("notifyOfModification($dn_list) (timeout: $timeout)");

    return xmlrpc_sendRequest(
        'rpc_xml_daemon'  => $rpc_xml_daemon,
        'xmlrpc_function' => 'notifyOfModification',
        'href_xmlrpc_params' =>
          xmlrpc_convertRefToRpcXmlStructure($aref_modified_dns),
        'timeout' => $timeout,
    );

}

=head2 getDeviceStatus(%)

=over 4

=item Parameter

        'rpc_xml_daemon' =>
        'device_dn' =>
        'timeout'       => 0 (default: no timeout => single try)


=item Return

    hash (already converted from XML-RPC)
    or undef on error

=back

=cut

sub getDeviceStatus( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon' => undef,
        'device_dn'      => undef,
        'timeout'        => 0,
        @_,
    );

    my $rpc_xml_daemon = $params{'rpc_xml_daemon'};
    my $device_dn      = $params{'device_dn'};
    my $timeout        = $params{'timeout'};

    $logger->debug("getDeviceStatus($device_dn) (timeout: $timeout)");

    return xmlrpc_sendRequest(
        'rpc_xml_daemon'     => $rpc_xml_daemon,
        'xmlrpc_function'    => 'getDeviceStatus',
        'href_xmlrpc_params' => RPC::XML::struct->new(
            'device_dn' => RPC::XML::string->new($device_dn)
        ),
        'timeout' => $timeout
    );
}

=head2 waitForDeviceStatus(%)


=over 4

=item Parameter

        'rpc_xml_daemon' =>
        'device_dn' =>
        'netboot'   => undef | 0 | 1 
        'dbkImageVersion' => undef | version
        'timeout'       => 

=item Return

    hash (already converted from XML-RPC)
    or undef on error

=back

=cut

sub waitForDeviceStatus(%)
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon'  => undef,
        'device_dn'       => undef,
        'netboot'         => undef,
        'dbkImageVersion' => undef,
        'timeout'         => 60,
        @_,
    );

    my $rpc_xml_daemon  = $params{'rpc_xml_daemon'};
    my $device_dn       = $params{'device_dn'};
    my $netboot         = $params{'netboot'};
    my $dbkImageVersion = $params{'dbkImageVersion'};
    my $timeout         = $params{'timeout'};

    my $t0 = [gettimeofday];

    my $href_parameter = {
        'device_dn'       => $device_dn,
        'netboot'         => $netboot,
        'dbkImageVersion' => $dbkImageVersion,
        'timeout'         => $timeout,
    };

    $logger->debug( "sending request for (" . Dumper($href_parameter) . ")" );

    my $href_xmlrpc_parameter = {};
    xmlrpc_convertHashStrings( $href_xmlrpc_parameter, $href_parameter,
        [ 'device_dn', 'dbkImageVersion' ] );
    xmlrpc_convertHashInts( $href_xmlrpc_parameter, $href_parameter,
        ['timeout'] );
    xmlrpc_convertHashBooleans( $href_xmlrpc_parameter, $href_parameter,
        ['netboot'] );

    #     my $resp_data;
    #     my $elapsed = 0;
    #     my $try     = 0;
    #
    #     # because $timeout may be large than HTTP timeout
    #     # request is handled by a loop
    #     do {
    #
    #         $try++;
    #         if ( $try > 1 ) {
    #             $logger->debug("$try. try (elapsed time: $elapsed)");
    #             # sleep required, to prevent fast loop,
    #             # if there is no connection
    #             sleep $try;
    #         }
    #
    #         my $resp =
    #           $rpc_xml_daemon->send_request( 'waitForDeviceStatus',
    #             RPC::XML::struct->new($href_xmlrpc_parameter) );
    #
    #         ( my $code, $resp_data ) =
    #           xmlrpc_convertResult( $resp,
    #             "waitForDeviceStatus(" . Dumper($href_parameter) . ") failed" );
    #
    #         $elapsed = tv_interval( $t0, [gettimeofday] );
    #
    #     } while ( ( !$resp_data ) and ( $elapsed < $timeout ) );
    #
    #     $logger->debug("duration: $elapsed");
    #
    #     return $resp_data;

    return xmlrpc_sendRequest(
        'rpc_xml_daemon'     => $rpc_xml_daemon,
        'xmlrpc_function'    => 'waitForDeviceStatus',
        'href_xmlrpc_params' => RPC::XML::struct->new($href_xmlrpc_parameter),
        'timeout'            => $timeout
    );
}

=head2 getScheduleUpdateWorkstations( % )


=over 4

=item Parameter


=item Return


=back

=cut

sub getScheduleUpdateWorkstations( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon' => undef,
        'location_dn'    => undef,
        'refimage_dn'    => undef,
        'devicetype_dn'  => undef,
        'timeout'        => undef,
        @_,
    );
    my $rpc_xml_daemon = $params{'rpc_xml_daemon'};
    my $location_dn    = lc( $params{'location_dn'} );
    my $refimage_dn    = lc( $params{'refimage_dn'} );
    my $devicetype_dn  = lc( $params{'devicetype_dn'} );
    my $timeout        = $params{'timeout'};

    $logger->debug(
        "sending request for ( 
        location_dn => $location_dn,
        refimage_dn => $refimage_dn,
        devicetype_dn => $devicetype_dn )"
    );
    return xmlrpc_sendRequest(
        'rpc_xml_daemon'     => $rpc_xml_daemon,
        'xmlrpc_function'    => 'getScheduleUpdateWorkstations',
        'href_xmlrpc_params' => RPC::XML::struct->new(
            'location_dn'   => RPC::XML::string->new($location_dn),
            'refimage_dn'   => RPC::XML::string->new($refimage_dn),
            'devicetype_dn' => RPC::XML::string->new($devicetype_dn),
        ),
        'timeout' => $timeout
    );
}

=head2 setUpdate( % )


=over 4

=item Parameter


=item Return


=back

=cut

sub setUpdate( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon' => undef,
        'location_dn'    => undef,
        'device_type_dn' => undef,
        'workstations'   => undef,
        'tftpboot'       => undef,
        'rsync'          => undef,
        'timeout'        => undef,
        @_,
    );

    my $rpc_xml_daemon = $params{'rpc_xml_daemon'};
    ## delete param for easier debug output
    delete $params{'rpc_xml_daemon'};

    my $timeout = $params{'timeout'};

    $logger->debug( "sending request for (" . Dumper(%params) . ")" );

    my $href_xmlrpc_parameter = {};
    xmlrpc_convertHashStrings( $href_xmlrpc_parameter, \%params,
        [ 'location_dn', 'device_type_dn', 'tftpboot', 'rsync' ] );
    $href_xmlrpc_parameter->{'workstation'} =
      RPC::XML::smart_encode( $params{'workstations'} );

    return xmlrpc_sendRequest(
        'rpc_xml_daemon'     => $rpc_xml_daemon,
        'xmlrpc_function'    => 'setUpdate',
        'href_xmlrpc_params' => RPC::XML::struct->new($href_xmlrpc_parameter),
        'timeout'            => $timeout
    );
}

=head2 createSecondaryConfig(%)

=over 4

=item Parameter

        'rpc_xml_daemon' =>
        'timeout'       => 0 (default: no timeout => single try)


=item Return

    hash (already converted from XML-RPC)
    or undef on error

=back

=cut

sub createSecondaryConfig( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon' => undef,
        'timeout'        => 0,
        @_,
    );

    my $rpc_xml_daemon = $params{'rpc_xml_daemon'};
    my $timeout        = $params{'timeout'};

    $logger->debug("createSecondaryConfig() (timeout: $timeout)");

    return xmlrpc_sendRequest(
        'rpc_xml_daemon'  => $rpc_xml_daemon,
        'xmlrpc_function' => 'createConfig',
        'timeout'         => $timeout
    );
}

=head1 functions for communicating with Knut Cache XML_RPC server



=head2 setKnutStatus(%)

set the status message of a Knut system via XML-RPC.
Also logs this message

=over 4

=item Parameter

    'rpc_xml_daemon' => XML-RPC connection
    'message'        => Knut Status message. Date will be added automatically

=item Return

    'true'  on success
    'false' otherwise

=back

=cut

sub setKnutStatus( % )
{
    my $logger = Log::Log4perl->get_logger();

    my %params = (
        'rpc_xml_daemon' => undef,
        'message'        => "",
        @_,
    );

    my $rpc_xml_daemon = $params{'rpc_xml_daemon'};
    my $message        = $params{'message'};

    if ( !$rpc_xml_daemon ) {

        #$logger->warn( "failed to set XML-RPC Knut-Status ($message): no XML-RPC connection" );
        return;
    }

    # set XMLRPC status to "...: localtime".
    my $time = POSIX::strftime( "%d.%m.%Y %H:%M:%S", localtime );
    chomp($time);
    my $resp =
      $rpc_xml_daemon->send_request( "status.setKnutStatus",
        RPC::XML::string->new("$message:  $time") )->value;

    if ($resp) {
        $logger->info($message);
    } else {
        $logger->error("failed to set XML-RPC Knut-Status to $message");
    }

    return $resp;
}

1;
