# Class SmartClient::ScCache
# inherits from SmartClient::LdapCache

# $Id: ScCache.pm 10050 2014-11-05 17:02:43Z HZuchel $

package SmartClient::ScCache;
use strict;
use warnings;

use base qw(SmartClient::LdapCache);
our $VERSION = "0.01";

use Data::Dumper;
use SmartClient::Util;
use SmartClient::LdapUtil;
require SmartClient::NameGenerator;
use SmartClient::DeviceStatusCache;
use SmartClient::IP;
use RPC::XML;
use RPC::XML::Client;

my $LOGGER = Log::Log4perl->get_logger();

=head2 new

constructor for this class

=over 4

=item Parameter

    see LdapCache
    plus
    mmap_device_status_filename

=item Return

=back

=cut

sub new
{
    my $classname = shift;
    my $self      = $classname->SUPER::new(@_);

    my $logger = $LOGGER;

    # defaults for class specific parameters
    #$self->{'ip_device_objectclasses'} = [];
    if ( !$self->{'ip_device_objectclasses'} ) {
        $self->{'ip_device_objectclasses'} =
          [ 'dbkWorkstation', 'dbkPrinter', 'dbkServer', 'dbkService' ];
        $logger->warn( "$classname invoked without parameter "
              . "ip_device_objectclasses, using defaults" );
    }

    # defaults for reference attributes,
    # this can be changed later
    # to be configurable
    $self->{'reference_attributes'} = {};
    my $ref_attrs = $self->{'reference_attributes'};
    $ref_attrs->{'dbkWorkstation'} = [
        'dbkRefImageDn', 'scRefBootOptionsDn',
        'dbkDoTftpBoot', 'dbkDoRsync',
        'dbkUpdateTime', 'dbkDeviceTypeDn'
    ];
    $ref_attrs->{'dbkPrinter'} = [ 'scRefBootOptionsDn', 'dbkDoTftpBoot' ];
    $ref_attrs->{'dbkServer'}  = [ 'scRefBootOptionsDn', 'dbkDoTftpBoot' ];
    $ref_attrs->{'dbkService'} = [];
    $ref_attrs->{'scLocation'} =
      [ 'dbkDoTftpBoot', 'dhcpOption', 'scRefBootOptionsDn',
        'dbkDeviceTypeDn' ];

    $self->checkReferenceAttributes();

    $self->{'ip_addresses'} = {};

    $self->{'name_generator'} =
      SmartClient::NameGenerator->new( 'sc_cache' => $self );

    # create a FastMmap file
    # for storing device status information
    # FS: I think the parameter is mmap_device_status_filename, so TODO should be done, right?
    # TODO: get device status file as parameter,
    #   because otherwise,
    #   using ScCache from 2 different programms
    #   will currupt data
    if ( defined $self->{'mmap_device_status_filename'} ) {
        my $status_cache_filename = $self->{'mmap_device_status_filename'};
        my $status_cache          = SmartClient::DeviceStatusCache->new(
            'filename' => $status_cache_filename );
        if ( ref $status_cache ) {
            $self->{'status_cache'} = $status_cache;
        } else {
            $logger->warn(
                "could not instantiate SmartClient::DeviceStatusCache for $status_cache_filename"
            );
        }
    } else {
        $logger->warn(
            "mmap_device_status_filename not defined. Device status will not be stored in file"
        );
    }

    return $self;
}

=head2 generateSerialsForLocations

Reads the DNS serial number from the SOA Record of the zone files
and stores them in the locations object data structure in the object.

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<refreshed_locations_aref>

Optional C<parameter refreshed_locations_aref>: if passed a reference
to a list of location DNs only these are refreshed.

=item C<config>

defaults to the C<SmartClient::Config> object of the instance of this class

=back

=item Return

none

=back

=cut

sub generateSerialsForLocations
{
    my $self  = shift;
    my %param = (
        'refreshed_locations_aref' => undef,
        'config'                   => $self->{'config'},
        @_,
    );

    my $process_location_dn_list = [];

    # check if a list of refreshed DNs was passed
    if ( $param{'refreshed_locations_aref'} ) {
        $process_location_dn_list = $param{'refreshed_locations_aref'};
    } else {
        $process_location_dn_list = [ $self->getAllLocationDns() ];
    }

    foreach my $location_dn ( @{$process_location_dn_list} ) {
        my $serial = $self->getSerialForLocation( location_dn => $location_dn );
        next
          if (
            !$self->isLocation( 'dn' => $location_dn, 'with_network' => 1 ) );
        my @rev_zones =
          $self->getRevZonesForLocation( location_dn => $location_dn );
        foreach my $rev_zone (@rev_zones) {
            $self->getSerialForRevZone( rev_zone_name => $rev_zone );
        }
    }
}

=head2 setSerialForLocation

Set the DNS serial number for a given location DN in the cache

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<location_dn>

DN of the location for which to store the serial

=item C<serial>

DNS serial number

=back

=item Return

1 (true) if the serial could be set in the cache,
0 (false) if the given dn does not exist in the cache.

=back

=cut

sub setSerialForLocation
{
    my $self  = shift;
    my %param = (
        'location_dn' => undef,
        'serial'      => undef,
        @_,
    );

    my $location_dn = $param{'location_dn'};
    my $serial      = $param{'serial'};

    my $href_location = $self->getInternalEntryHash($location_dn);

    my $logger = $LOGGER;

    if ( !$href_location ) {
        $logger->error(
            "[DNS,LDAP-DATA] setSerialForLocation: DN $location_dn not found");
        return 0;
    }
    $href_location->{'serial'} = $serial;
    return 1;
}

=head2 setSerialForRevZone

Set the DNS serial number for a given reverse zone in the cache

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<rev_zone_name>

name of the reverse zone for which to store the serial

=item C<serial>

DNS serial number

=back

=item Return

1 (true) if the serial could be set in the cache,
0 (false) if the given dn does not exist in the cache.

=back

=cut

sub setSerialForRevZone
{
    my $self  = shift;
    my %param = (
        'rev_zone_name' => undef,
        'serial'        => undef,
        @_,
    );

    my $rev_zone_name = $param{'rev_zone_name'};
    my $serial        = $param{'serial'};

    my $rev_zones_hash = $self->{'rev_zones'};

    my $logger = $LOGGER;

    if ( !$rev_zones_hash->{$rev_zone_name} ) {
        $logger->error("[DNS] $rev_zone_name not found");
        return 0;
    }
    $rev_zones_hash->{$rev_zone_name}->{'serial'} = $serial;
    return 1;
}

=head2 addRevZone

Add the reverse zone for a given location DN in the cache,
also stores the relationship between reverse zone and location, so
that the location for a given reverse zone can be found.

=head3 Example

  $sc_cache->addRevZone( 
       location_dn => 'cn=gs123,ou=lgs,o=company,c=de', 
       rev_zone => '10.1.10.in-addr.arpa');

=over 4

=item Parameter

=over 4

=item C<location_dn>

DN of the location for which to store the reverse zone

=item C<rev_zone>

reverse zone names

=back

=item Return

1 (true) if the reverse zone could be set in the cache,
0 (false) if the given dn does not exist in the cache.

=back

=cut

sub addRevZone
{
    my $self  = shift;
    my %param = (
        'location_dn' => undef,
        'rev_zone'    => undef,
        @_,
    );

    my $location_dn = $param{'location_dn'};
    my $rev_zone    = $param{'rev_zone'};

    my $href_location  = $self->getInternalEntryHash($location_dn);
    my $rev_zones_hash = $self->{'rev_zones'};

    my $logger = $LOGGER;

    if ( !$href_location ) {
        $logger->error("[DNS,LDAP-DATA] addRevZone: DN $location_dn not found");
        return 0;
    }

    $href_location->{'rev_zones'}->{$rev_zone}++;
    $rev_zones_hash->{$rev_zone}->{'locations'}->{$location_dn}++;
    return 1;
}

=head2 getRevZonesForLocation

get all reverse zones for a given location

=head3 Example

  my @rev_zones = $sc_cache->getRevZonesForLocation(
      'location_dn' => 'ou=testgs,ou=test,o=company,c=de'
  );

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<location_dn> dn of the location

=back

=item Return

array with all reverse zone names which can be found in the location

=back

=cut

sub getRevZonesForLocation
{
    my $self  = shift;
    my %param = (
        'location_dn' => undef,
        @_,
    );

    my $location_dn   = $param{'location_dn'};
    my $href_location = $self->getInternalEntryHash($location_dn);

    my $logger = $LOGGER;

    if ( !$href_location ) {
        $logger->error(
            "[DNS,LDAP-DATA] getRevZonesForLocation: DN $location_dn not found"
        );
        return;
    }
    if ( !$href_location->{'rev_zones'} ) {
        if ( $self->isLocation( 'dn' => $location_dn, 'with_network' => 1 ) ) {
            $logger->error( "[DNS,LDAP-DATA] no reverse zones for "
                  . " locationDN $location_dn found" );
        } else {
            $logger->debug(
                "[DNS,LDAP-DATA] no reverse zones for $location_dn found. This is ok, since it has no network"
            );
        }
        return;
    }
    my @rev_zones = keys %{ $href_location->{'rev_zones'} };
    return @rev_zones;
}

=head2 assignIpToRevZone

assign an ip address for a given entry and location. This means, that the internal structures will
be updated, so that the ip address and locations can be found by searching in the reverse zones 
data structures.

=head3 Example

  my $rev_zone = $sc_cache->assignIpToRevZone(
    'ip_addr' => '10.1.2.3',
    'entry_dn' => 'cn=pc00259,ou=testgs,o=test,o=company,c=de',
    'location_dn' => 'ou=testgs,o=test,o=company,c=de'
  );
  # $rev_zone should now be '2.1.10.in-addr.arpa'

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<ip_addr>

=item C<location_dn>

=item C<entry_dn>

=back

=item Return

reverse zone for the ip

=back

=cut

sub assignIpToRevZone
{
    my $self  = shift;
    my %param = (
        'ip_addr'     => undef,
        'location_dn' => undef,
        'entry_dn'    => undef,
        @_,
    );
    my $ip_addr     = $param{'ip_addr'};
    my $location_dn = $param{'location_dn'};
    my $entry_dn    = $param{'entry_dn'};

    my $rev_zones_hash = $self->{'rev_zones'};

    my $logger = $LOGGER;

    if ( !$ip_addr ) {
        $logger->debug("[DNS] no ip_addr given, so we have nothing todo.");
        return;
    }

    my $rev_zone = $self->getRevZoneForIp($ip_addr);
    if ( !$rev_zone ) {
        $logger->info("[DNS] no rev zone found for $ip_addr");
        return;
    }
    $self->addRevZone(
        location_dn => $location_dn,
        rev_zone    => $rev_zone
    );

    $rev_zones_hash->{$rev_zone}->{'ip_addresses'}->{$ip_addr}++;
    $self->{'ip_addresses'}->{$ip_addr}->{'dn_hash'}->{$entry_dn}++;
    return $rev_zone;
}

=head2 removeIpFromRevZone

removes a given up address from the internal data structures for reverse zones.

=head3 Example

  $sc_cache->removeIpFromRevZone('ip_addr' => '10.1.2.3', 'collect_rev_zones' => $href_collect_rev_zones);

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<ip_addr>

=item C<collect_rev_zones>

reference to hash in which rev_zones of the ip gets collected (optional)

If C<collect_rev_zones> is given, the caller has to call C<updateLocationsForRevZones> on his own.

=back

=item Return

always true

=back

=cut

#TODO: should we mark changed reverse zones, for getAssignedLocations?
sub removeIpFromRevZone
{
    my $self  = shift;
    my %param = (
        'ip_addr'           => undef,
        'collect_rev_zones' => undef,
        @_,
    );
    my $ip_addr                = $param{'ip_addr'};
    my $href_collect_rev_zones = $param{'collect_rev_zones'};
    my $logger                 = $LOGGER;

    if ( $logger->is_debug() ) {
        $logger->debug("call removeIpFromRevZone for $ip_addr");
    }

    if ( !$ip_addr ) {
        $logger->info(
            "[DNS] no ip_addr was given, so we will not remove anything from rev_zones"
        );
        return;
    }

    my $rev_zone = $self->getRevZoneForIp($ip_addr);
    if ( !$rev_zone ) {
        $logger->error(
            "[DNS,LDAP-DATA] no rev_zone could be found for $ip_addr");
        return;
    }

    my $rev_zones_hash   = $self->{'rev_zones'};
    my $href_rev_ip_addr = $rev_zones_hash->{$rev_zone}->{'ip_addresses'};
    if ( !exists $href_rev_ip_addr->{$ip_addr} ) {
        $logger->error(
            "[DNS,LDAP-DATA] $ip_addr does not exist in rev_zone_hash");
        return;
    }

    delete( $href_rev_ip_addr->{$ip_addr} );

    #update the locations for the ip addresses left in rev_zone_hash
    if ( ref $href_collect_rev_zones ) {
        $href_collect_rev_zones->{$rev_zone}++;
    } else {
        $self->updateLocationsForRevZones($rev_zone);
    }
    return 1;
}

=head2 updateLocationsForRevZones

reconstructs the locations hash in the C<rev_zones> hash entries 
with the C<dn_hash> entries in the C<ip_addresses> hash

=head3 Example

  $sc_cache->updateLocationsForRevZones('29.5.10.in-addr.arpa', '28.5.10.in-addr.arpa');

=over 4

=item Parameter

list of reverse zones to sync locations hash

=item Return

always true

=back

=cut

sub updateLocationsForRevZones
{
    my ( $self, @rev_zones ) = @_;

    my $href_ip_addresses = $self->{'ip_addresses'};
    my $href_rev_zones    = $self->{'rev_zones'};
    my $logger            = $LOGGER;

    if ( $logger->is_debug() ) {
        $logger->debug( "call updateLocationsForRevZones for("
              . join( ", ", keys %{$href_rev_zones} )
              . ")" );
    }

    my @all_location_dns = $self->getAllLocationDns();

    #iterate over all given reverse zones
    foreach my $rev_zone (@rev_zones) {
        my $href_rev_zone = $href_rev_zones->{$rev_zone};

        #reinitialize locations hash in reverse zone entry
        $href_rev_zone->{'locations'} = {};
        my $href_rev_zone_locations = $href_rev_zone->{'locations'};

        #iterate over all ip addresses in reverse zone entry
        foreach my $ip_addr ( keys %{ $href_rev_zone->{'ip_addresses'} } ) {

            # and find all dn's which have that ip
            foreach
              my $dn ( keys %{ $href_ip_addresses->{$ip_addr}->{'dn_hash'} } )
            {

                # to get all locations which span this reverse zone
                $href_rev_zone_locations->{ $self->getLocationForDn($dn) }++;
            }
        }

        # there could be locations, where no ip addresses can be found (yet)
        for my $location_dn (@all_location_dns) {
            for my $location_rev_zone ( grep { $_ eq $rev_zone }
                $self->getRevZonesForLocation( location_dn => $location_dn ) )
            {
                $href_rev_zone_locations->{$location_dn}++;
            }
        }
    }
    return 1;
}

=head2 removeIp

remove ip for a specific dn from the internal cache structures.
This method will try to remove the ip from the rev_zones by calling removeIpFromRevZone, too.

=head3 Example

 $sc_cache->removeIp(
   'ip_addr' => '10.1.2.3', 
   'dn'      => 'cn=test,ou=server,o=company,c=de'
 );

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<ip_addr>

ip address of the entry to be removed

=item C<dn>

dn for this ip

=item C<collect_rev_zones>

reference to hash in which rev_zones of the ip gets collected (optional)

If C<collect_rev_zones> is given, the caller has to call C<updateLocationsForRevZones> on his own.

=back

=item Return

C<true> if ip could be removed correctly

=back

=cut

#TODO: should we mark the changed locations for getAssignedLocations?
sub removeIp
{
    my $self  = shift;
    my %param = (
        'ip_addr'           => undef,
        'dn'                => undef,
        'collect_rev_zones' => undef,
        @_,
    );
    my $ip_addr                = $param{'ip_addr'};
    my $dn                     = $param{'dn'};
    my $href_collect_rev_zones = $param{'collect_rev_zones'};

    # TODO pruefe href_collect_rev_zones

    my $logger = $LOGGER;

    if ( $logger->is_debug() ) {
        $logger->debug("call removeIp for $ip_addr ($dn)");
    }

    if ( !$ip_addr ) {
        $logger->error("No ip address was given. Doing nothing.");
        return;
    }
    my $ip_hash = $self->{'ip_addresses'};

    my $href_ip_hash_for_addr = $ip_hash->{$ip_addr};
    if ( !$href_ip_hash_for_addr ) {
        $logger->error("[LDAP-DATA] $ip_addr does not exist in cache");
        return;
    }
    if ( !exists $href_ip_hash_for_addr->{'dn_hash'}->{$dn} ) {
        $logger->warn(
            "[LDAP-DATA] dn $dn was not assigned to ip $ip_addr, so we cant remove it."
        );
        return;
    }

    #delete the dn for the ip address
    delete( $href_ip_hash_for_addr->{'dn_hash'}->{$dn} );

    #delete complete entry for ip address, if we deleted the last entry in it
    if ( !keys %{ $href_ip_hash_for_addr->{'dn_hash'} } ) {
        delete $ip_hash->{$ip_addr};

        #remove the ip from rev zone (this method will update the stored locations in rev_zone entry)
        $self->removeIpFromRevZone(
            'ip_addr'           => $ip_addr,
            'collect_rev_zones' => $href_collect_rev_zones
        );
    } else {

        #or we have to update the other entries in this structure
        #TODO zone_name is not nessecarily for every entry the same, which one to take?
        foreach
          my $dn_for_zone_name ( keys %{ $href_ip_hash_for_addr->{'dn_hash'} } )
        {
            $href_ip_hash_for_addr->{'zone_name'} =
              $self->getLocationZoneName(
                $self->getLocationForDn($dn_for_zone_name) );
            my $entry = $self->getEntry($dn);
            if ( ref $entry ) {
                $href_ip_hash_for_addr->{'cn'} = scalar $entry->get_value('cn');
            } else {
                $href_ip_hash_for_addr->{'cn'} = '';
            }
            last;
        }
        my $rev_zone = $self->getRevZoneForIp($ip_addr);
        if ($rev_zone) {
            if ( ref $href_collect_rev_zones ) {
                $href_collect_rev_zones->{$rev_zone}++;
            } else {
                $self->updateLocationsForRevZones($rev_zone);
            }
        }
    }

    #remove ip address from internal entry, if necessary
    my $href_internal_entry = $self->getInternalEntryHash($dn);
    if ( ref $href_internal_entry ) {
        if ( $href_internal_entry->{'ip_address'} eq $ip_addr ) {
            delete $href_internal_entry->{'ip_address'};
        }
    }
    return 1;
}

=head2 getIpAddressesForRevZone

get all ip addresses for a given reverse zone

=head3 Example

  foreach my $ip ( $sc_cache->getIpAddressesForRevZone('2.1.10.in-addr.arpa') ) {
     # do something with the ip
  }

=over 4

=item Parameter

name of the reverse zone

=item Return

array with all ip addresses for the reverse zone

=back

=cut

sub getIpAddressesForRevZone
{
    my $self          = shift;
    my $rev_zone_name = shift;

    my $rev_zones_hash = $self->{'rev_zones'};

    my @ip_addresses =
      keys %{ $rev_zones_hash->{$rev_zone_name}->{'ip_addresses'} };
    return @ip_addresses;
}

=head2 setIpAddress

store the IP address assigned to an entry in the cache specified
by its DN

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<entry_dn>

DN of the entry

=item C<ip_address>

IP address

=back

=item Return

1 (true) on success of 0 (false) if the DN does not exist in the cache

=back

=cut

sub setIpAddress
{
    my $self  = shift;
    my %param = (
        'entry_dn'   => undef,
        'ip_address' => undef,
        @_,
    );

    my $entry_dn = $param{'entry_dn'};
    my $ip_addr  = $param{'ip_address'};

    my $entry_cache = $self->{'entrycache'};

    my $logger = $LOGGER;

    if ( !$entry_cache->{$entry_dn} ) {
        $logger->error(
            "[LDAP-DATA] setIpAddress: $entry_dn not found in entry cache");
        return 0;
    }

    # check if IP address changed and refresh ip information
    # if applicable
    #TODO: if this information should be used by index,
    #ldap attributes should be used
    if ( $entry_cache->{$entry_dn}->{'ip_address'} ) {
        my $old_ip_addr = $entry_cache->{$entry_dn}->{'ip_address'};
        if ( $old_ip_addr ne $ip_addr ) {
            $self->removeIp( 'ip_addr' => $old_ip_addr, 'dn' => $entry_dn );
        }
    }

    # store the IP address
    $entry_cache->{$entry_dn}->{'ip_address'} = $ip_addr;
    return 1;
}

=head2 getIpAddress

get the IP address assigned to an entry in the cache specified
by its DN. 

Since an entry can have more than one ip address, the 
behaviour of this method can be switched to give back all ip addresses
by specifying the C<all_ip_addresses> flag.

TODO: This should probably be splitted in two functions.

=head3 Example

  my $ip_address = $sc_cache->getIpAddress(
    'entry_dn' => 'cn=www,ou=services,o=company,c=de'
  );
  my @ip_addresses = $sc_cache->getIpAddress(
    'entry_dn'         => 'cn=www,ou=services,o=company,c=de',
    'all_ip_addresses' => 1
  );

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<entry_dn>

DN of the entry

=item C<all_ip_addresses>

boolean to decide if all ip addresses for a given entry should be returned

=back

=item Return

IP address or undef if the DN does not exist in the cache ( if C<all_ip_addresses> are asked for, an array will be returned)

=back

=cut

sub getIpAddress
{
    my $self  = shift;
    my %param = (
        'entry_dn'         => undef,
        'all_ip_addresses' => undef,
        @_,
    );

    my $entry_dn = $param{'entry_dn'};

    my $href_internal_entry = $self->getInternalEntryHash($entry_dn);

    my $logger = $LOGGER;

    if ( !$href_internal_entry ) {
        $logger->error("[LDAP-DATA] $entry_dn not found in entry cache");
        return;
    }
    if ( !exists $href_internal_entry->{'ip_address'} ) {
        my $entry        = $self->getEntry($entry_dn);
        my $dbkNoDNSName = ( scalar $entry->get_value('dbkNoDNSName') )
          || 'false';
        if ( lc($dbkNoDNSName) eq 'true' ) {

            #ok, we have no ip address
            $logger->debug(
                "[LDAP-DATA] IP address for $entry_dn not found in entry cache (ok: dbkNoDNSName is true)"
            );
            return;
        } elsif ( !$self->isDevice($entry) ) {
            $logger->debug(
                "[LDAP-DATA] $entry_dn is not a device, so it has no ip address. OK"
            );
            return;
        }
        if ( !$param{'all_ip_addresses'} ) {

            # we were asked only for this computed ip address, and that was not found
            $logger->error(
                "[LDAP-DATA] IP address for $entry_dn not found in entry cache"
            );
            return;
        }
    }

    #if all ip addresses are wanted, look them up in the ldap entry object
    #and add them to the calculated one
    if ( $param{'all_ip_addresses'} ) {
        my $entry        = $self->getEntry($entry_dn);
        my %ip_addresses = ();
        if ( exists $href_internal_entry->{'ip_address'} ) {
            my $ip = $href_internal_entry->{'ip_address'};
            $ip_addresses{$ip} = 1 if isValidIp( $ip, $entry_dn );
        }
        if ( ref $entry ) {
            my @ip_addresses = $entry->get_value('ipHostNumber');
            foreach my $ip (@ip_addresses) {
                $ip_addresses{$ip} = 1 if isValidIp( $ip, $entry_dn );
            }
        }
        return keys %ip_addresses;
    }
    my $ip = $href_internal_entry->{'ip_address'};
    if ( isValidIp( $ip, $entry_dn ) ) {
        return $ip;
    }
    return;
}

# checks if an ip is valid (meaning contains other characters than space)
sub isValidIp
{
    my ( $ip, $dn ) = @_;
    my $logger = $LOGGER;
    if ( !defined($ip) || $ip =~ /^\s*$/ ) {
        $logger->error("[LDAP-DATA] IP address for $dn not valid");
        return 0;
    } else {
        return 1;
    }
}

=head2 getDnsNameForIp

Obsolete: use getDnsNamesForIp

=over 4

=item Parameter

ip address 

=item Return

dns name for the given ip address
 
=back

=cut

sub getDnsNameForIp
{

    # obsoleted by getDnsNamesForIp
    my $self    = shift;
    my $ip_addr = shift;

    my $ip_hash = $self->{'ip_addresses'};
    my $logger  = $LOGGER;

    if ( !$ip_hash->{$ip_addr} ) {
        $logger->error(
            "[DNS,LDAP-DATA] getDnsNameForIp: $ip_addr does not exist in cache"
        );
        return;
    }
    my $dns_name =
      $ip_hash->{$ip_addr}->{'cn'} . '.' . $ip_hash->{$ip_addr}->{'zone_name'};
    return $dns_name;
}

=head2 getDnsNamesForIp

Get the DNS names for a given IP address from the cache.
This is used for reverse Zone generation, for this purpose
multiple names will only returned for server/service entries.

=head3 Example

  foreach my $dns_name ( $sc_cache->getDnsNamesForIp('10.1.2.3') ) {
    #do something with the dns name
  }

=over 4

=item Parameter

- IP Address

=item Return

List of DNS names

=back

=cut

sub getDnsNamesForIp
{
    my $self    = shift;
    my $ip_addr = shift;

    my $ip_hash = $self->{'ip_addresses'};
    my $logger  = $LOGGER;

    if ( !ref $ip_hash or !$ip_hash->{$ip_addr} ) {
        $logger->error("[DNS,LDAP-DATA] $ip_addr does not exist in cache");
        return;
    }
    my %dns_names;

    #$ip_hash->{$ip_addr}->{'cn'} . '.' . $ip_hash->{$ip_addr}->{'zone_name'};
    my @dn_list = $self->getDnListForIp($ip_addr);
    if ( !@dn_list ) {
        $logger->error("[LDAP-DATA] error getting DN List for $ip_addr");
        return;
    }

    #TODO: why does this loop break after the first found entry?
    #if it is not neccesary we can leave it commented out and use the inner loop further down
    #DNLOOP: foreach my $dn (@dn_list) {
    #      my $entry = $self->getEntry($dn);
    #      next if ( !ref $entry );
    #      if ( $self->isServer($entry) || $self->isService($entry) ) {
    #          foreach
    #            my $dns_name ( keys %{ $ip_hash->{$ip_addr}->{'dns_names'} } )
    #          {
    #              $dns_names{ $dns_name . '.'
    #                    . $ip_hash->{$ip_addr}->{'zone_name'} }++;
    #          }
    #          last DNLOOP;
    #      }
    #  }

    #build dns names for every entry in dn_hash
    foreach my $dn (@dn_list) {
        my $entry = $self->getEntry($dn);
        next if ( !ref $entry );

        #get zonename for location of dn (this can be different from just looking at dn)
        my $location_zone_name =
          $self->getLocationZoneName( $self->getLocationForDn($dn) );
        if ($location_zone_name) {
            if ( $self->isServer($entry) || $self->isService($entry) ) {
                my @dns_names = $entry->get_value('dbkDnsName');
                foreach my $dns_name (@dns_names) {
                    $dns_names{ $dns_name . '.' . $location_zone_name }++;
                }
            }

            #try to build dns name by cn and zonename of location for dn
            #zonenames can be different for each dn
            my $cn = $entry->get_value('cn');
            next if ( !$cn );
            next if ( lc( scalar $entry->get_value('dbkNoDNSRev') ) eq 'true' );
            $dns_names{ $cn . '.' . $location_zone_name }++;
        } else {

            #try to build dns name by dn
            my $dns_name_by_dn = ldap_dn2dnsname($dn);
            next if ( !$dns_name_by_dn );
            $dns_names{$dns_name_by_dn}++;
        }
    }

    #TODO: I don't think this is needed anymore, since we have that logic as part of the loop above
    #    if ($ip_hash->{$ip_addr}->{'zone_name'}) {
    #        $dns_names{ $ip_hash->{$ip_addr}->{'cn'} . '.'
    #          . $ip_hash->{$ip_addr}->{'zone_name'} }++;
    #    }
    my @valid_dns_names =
      grep { defined $_ } map { convertToValidDnsName($_) } keys %dns_names;
    return @valid_dns_names;
}

=head2 checkLocationIgnore

checks whether a location can safely be ignored for ip configurations in respect to smartclients.
This means, that there will be no dynamically updated clients in this location. 

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<location_dn>

name of the location to check

=item C<config>

config object; defaults to $self->{'config'}

=back

=item Return

boolean, C<true> if location can be ignored regarding ip configuration, otherwise C<false>

=back

=cut

sub checkLocationIgnore
{
    my $self  = shift;
    my %param = (
        'location_dn' => undef,
        'config'      => $self->{'config'},
        @_,
    );

    my $location_dn = $param{'location_dn'};
    my $config      = $param{'config'};

    my $logger = $LOGGER;

    my $ignore_location = 0;
    if (   $config->get( 'LDAP2DNS_LOCATION_TYPE_ATTR', 1 )
        && $config->get( 'LDAP2DNS_LOCATION_TYPE_IGNORE_REX', 1 ) )
    {
        my $location_entry = $self->getEntry($location_dn);
        if ( !ref $location_entry ) {
            $logger->warn(
                "[LDAP-DATA] Could not find a location entry for $location_dn");
            return;
        }
        my $location_ignore_attr = $config->get('LDAP2DNS_LOCATION_TYPE_ATTR');
        my $location_type_ignore_string =
          $config->getWithDefault( 'LDAP2DNS_LOCATION_TYPE_IGNORE_REX',
            'SVB|SERVER|RL' );

        my $location_ignore_rex = qr/$location_type_ignore_string/io;

        if ( $location_entry->exists($location_ignore_attr) ) {
            my $location_ignore_value =
              $location_entry->get_value($location_ignore_attr);
            if ( uc($location_ignore_value) =~ /$location_ignore_rex/ ) {
                $ignore_location = 1;
            }
        }
    } else {
        $logger->warn(
            "checkLocationIgnore: configuration parameters LDAP2DNS_LOCATION_TYPE_ATTR and LDAP2DNS_LOCATION_TYPE_IGNORE_REX are not defined"
        );
    }
    return $ignore_location;
}

=head2 setZoneNameForLocation

Set the DNS zone name for a given location DN in the cache.
If the optional zone_name parameter is not given, the zone name
will be set from the attribute specified in the configuration
parameter LDAP2DNS_LOCATION_DNS_DOMAIN_ATTR if it is present
in the locations LDAP entry. Otherwise it is derived from the
DN of the location entry.

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<location_dn>

DN of the location for which to store the serial

=item C<zone_name>

DNS zone name (optional)

=item C<config>

C<SmartClient::Config> object, defaults to the config of
this object instance

=back

=item Return

1 (true) if the zone name could be set in the cache,
0 (false) if the given dn does not exist in the cache.

=back

=cut

sub setZoneNameForLocation
{
    my $self  = shift;
    my %param = (
        'location_dn' => undef,
        'zone_name'   => undef,
        'config'      => $self->{'config'},
        @_,
    );

    my $location_dn = $param{'location_dn'};
    my $zone_name   = $param{'zone_name'};
    my $config      = $param{'config'};

    my $href_location = $self->getInternalEntryHash($location_dn);

    my $logger = $LOGGER;

    if ( !$href_location ) {
        $logger->error(
            "[DNS,LDAP-DATA] setZoneNameForLocation: DN $location_dn not found in cache"
        );
        return;
    }

    if ( !$zone_name ) {
        my $location_entry = $self->getEntry($location_dn);
        my $location_domain_attr =
          $config->get('LDAP2DNS_LOCATION_DNS_DOMAIN_ATTR');
        $logger->debug(
            "[DNS,LDAP-DATA] checking $location_dn for $location_domain_attr");
        if ( $location_entry->exists($location_domain_attr) ) {
            ($zone_name) = $location_entry->get_value($location_domain_attr);
            $logger->debug(
                "[DNS,LDAP-DATA] got $zone_name for $location_dn from $location_domain_attr"
            );
        } else {
            $zone_name = get_zonename( $location_dn, 0 );
            $logger->debug(
                "[DNS,LDAP-DATA] derived $zone_name for $location_dn by DN");
        }
    }

    $logger->debug(
        "[DNS,LDAP-DATA] storing zone name $zone_name for $location_dn");
    $href_location->{'zone_name'} = $zone_name;
    return 1;
}

#TODO: this seems to implement the same functionality as getLocationZoneName
sub getZoneNameForLocation
{
    my $self        = shift;
    my $location_dn = shift;

    return $self->getLocationZoneName($location_dn);
}

sub assignNetworksToLocations
{
    my $self  = shift;
    my %param = (
        'refreshed_locations_aref' => undef,
        'config'                   => $self->{'config'},
        @_,
    );

    my $config                   = $param{'config'};
    my $process_location_dn_list = [];
    my $networks                 = {};
    my $rev_zones                = {};

    my $logger = $LOGGER;

    # check if a list of refreshed DNs was passed
    if ( $param{'refreshed_locations_aref'} ) {
        $process_location_dn_list = $param{'refreshed_locations_aref'};

        # use networks hash from object
        $networks = $self->{'networks'};

        # use reverse networks from object
        $rev_zones = $self->{'rev_zones'};
    } else {
        $process_location_dn_list = [ $self->getAllLocationDns() ];

        # initialize empty networks and reverse zones
        $self->{'networks'}  = $networks;
        $self->{'rev_zones'} = $rev_zones;
    }

    foreach my $location_dn ( @{$process_location_dn_list} ) {
        if ( !$self->hasEntry($location_dn) ) {
            $logger->warn(
                "[LDAP-DATA] no location entry for $location_dn found");
            next;
        }
        my $location_entry = $self->getEntry($location_dn);

        my $network_obj = $self->getLocationIpObject($location_dn);

        # location is maybe special location
        # without network attributes
        if ($network_obj) {
            my $network_addr = $network_obj->network();
            $self->setZoneNameForLocation( location_dn => $location_dn );
            my @rev_zones = $network_obj->getReverseZones();
            $networks->{$network_addr}->{'location_dn'} = $location_dn;
            for my $rev_zone (@rev_zones) {
                $self->addRevZone(
                    location_dn => $location_dn,
                    rev_zone    => $rev_zone
                );
            }

            # assign reference attributes for location
            my $ref_attr_aref = $self->{'reference_attributes'}->{'scLocation'};
            $self->assignReferencesByTreeSearch(
                entry_dn                  => $location_dn,
                reference_attributes_aref => $ref_attr_aref
            );
        } else {
            if ( !$self->checkLocationIgnore( 'location_dn' => $location_dn ) )
            {
                $logger->info(
                    "[LDAP-DATA] Can't determine network address for $location_dn"
                );
            }
        }
    }
    return $networks;
}

=head2 assignNameGenerators

Gets all name generators by their C<objectClass dbkDeviceType> and registers them in the 
name generator which is stored in $self->{'name_generator'}

Obsolete, since name generators can be used dynamically now.

=over 4

=item Parameter

None

=item Return

boolean always true

=back

=cut

sub assignNameGenerators
{
    my $self   = shift;
    my $logger = $LOGGER;

    my @devicetypes = $self->getObjectClassIndex(
        'base'             => $self->getBase(),
        'scope'            => 'sub',
        'objectclass_name' => 'dbkDeviceType'
    );
    foreach my $devicetype_dn (@devicetypes) {
        $logger->debug(
            "[LDAP-DATA] register name generator for " . $devicetype_dn );
        $self->{'name_generator'}
          ->addGeneratorByEntry( $self->getEntry($devicetype_dn) );
    }
    return 1;
}

=head2 getSubLocations

determines the sub-locations to a given base/sub-location.
 
Currently not used

=over 4

=item Parameter

    dn: 

=item Return

    array of DNs (sub-locations)

=back

=cut

sub getSubLocations
{
    ##my $logger = $LOGGER;

    my $self   = shift;
    my %params = (
        'dn' => undef,
        @_,
    );

    my $dn = $params{'dn'};

    my @locations = $self->getObjectClassIndex(
        'base'             => $dn,
        'objectclass_name' => 'scLocation'
    );

    # remove location dn from list of sub_locations
    my @sub_locations = grep { !/^$dn$/ } @locations;

    return @sub_locations;
}

=head2 getLocationEntries

determines the sub-locations to a given base/sub-location.
 
Currently not used

=over 4

=item Parameter

        'base'            
        'objectclass_name'

=item Return

    array of DNs from objects of type 'objectclass_name' 
        in location given by 'base'

=back

=cut

sub getLocationEntries
{
    ##my $logger = $LOGGER;

    my $self   = shift;
    my %params = (
        'base'             => undef,
        'objectclass_name' => undef,
        @_,
    );

    return $self->getObjectClassIndex( %params,
        'subs_exclude' => "scLocation" );
}

sub assignIpDevices
{
    my $logger = $LOGGER;

    my $self  = shift;
    my %param = (
        'refreshed_locations_aref' => undef,
        'config'                   => $self->{'config'},
        @_,
    );

    my $ip_device_objectclasses  = $self->{'ip_device_objectclasses'};
    my $ip_hash                  = $self->{'ip_addresses'};
    my $process_location_dn_list = [];

    # check if a list of refreshed DNs was passed
    if ( $param{'refreshed_locations_aref'} ) {
        $process_location_dn_list = $param{'refreshed_locations_aref'};
    } else {
        $process_location_dn_list = [ $self->getAllLocationDns() ];
    }

    # switch on or flush treesearch cache
    $self->{'treesearch_cache'} = {};

    foreach my $location_dn ( @{$process_location_dn_list} ) {
        if ( !$self->hasEntry($location_dn) ) {
            $logger->warn(
                "[LDAP-DATA] no location entry found for $location_dn");
            next;
        }
        my $location_data = $self->getInternalEntryHash($location_dn);
        my $zone_name     = $self->getZoneNameForLocation($location_dn);

        my $location_entry = $self->getEntry($location_dn);
        my $network_addr   = $location_entry->get_value('ipNetworkNumber');

        foreach my $ip_device_objectclass ( @{$ip_device_objectclasses} ) {

            my @assigned_dn_list = $self->getLocationEntries(
                'base'             => $location_dn,
                'objectclass_name' => $ip_device_objectclass
            );
            foreach my $dn (@assigned_dn_list) {
                my $ip_device_entry = $self->getEntry($dn);
                if ( !ref $ip_device_entry ) {
                    $logger->error(
                        "[LDAP-DATA] could not get entry $dn from cache, skipping"
                    );
                    next;
                }

                # ldap_entry2ip calculates the IP if not stored in LDAP
                my $ip_addr = ldap_entry2ip(
                    ENTRY   => $ip_device_entry,
                    NETWORK => $network_addr
                );

                # store calculated ip or ip from entry
                # and register the entry in the ip devices structure
                if ($ip_addr) {
                    $self->setIpAddress(
                        entry_dn   => $dn,
                        ip_address => $ip_addr
                    );
                    foreach my $curr_ip_addr (
                        $self->getIpAddress(
                            'entry_dn'         => $dn,
                            'all_ip_addresses' => 1
                        )
                      )
                    {

                        # gather DNS names, and store it in IP address hash,
                        # using a hash of the names and incrementing the value
                        # is done get a unique list of names
                        my $dns_names;
                        if ( !$ip_hash->{$curr_ip_addr} ) {
                            $ip_hash->{$curr_ip_addr} = {
                                'cn'        => '',
                                'dns_names' => {},
                                'zone_name' => $zone_name,
                                'dn_hash'   => {}
                            };
                        }
                        $ip_hash->{$curr_ip_addr}->{'dn_hash'}->{$dn}++;
                        $dns_names = $ip_hash->{$curr_ip_addr}->{'dns_names'};

                        #TODO: this cn is completely useless, why should we have it anymore?
                        if ( $ip_device_entry->exists('cn') ) {
                            my $cn_name = ldap_cn2dnsname(
                                $ip_device_entry->get_value('cn') );
                            $ip_hash->{$ip_addr}->{'cn'} = $cn_name;
                        }
                        if ( $ip_device_entry->exists('dbkDnsName') ) {
                            foreach my $ldap_dns_name (
                                $ip_device_entry->get_value('dbkDnsName') )
                            {
                                my $dns_name = ldap_cn2dnsname($ldap_dns_name);
                                $dns_names->{$dns_name}++;
                            }
                        }

                        # assign IP address to reverse zone
                        $self->assignIpToRevZone(
                            'ip_addr'     => $curr_ip_addr,
                            'location_dn' => $location_dn,
                            'entry_dn'    => $dn,
                        );

                    }
                } else {
                    if (
                        !(
                            lc(
                                scalar $ip_device_entry->get_value(
                                    'dbkNoDNSName')
                            ) eq 'true'
                        )
                      )
                    {
                        $logger->warn(
                            "[LDAP-DATA] can't determine IP address for $dn");
                    } else {
                        $logger->debug(
                            "[LDAP-DATA] skipping IP for $dn since it is marked as dbkNoDNSName"
                        );
                    }
                }

                if ( $self->{'reference_attributes'}->{$ip_device_objectclass} )
                {
                    my $ref_attr_aref =
                      $self->{'reference_attributes'}->{$ip_device_objectclass};
                    $self->assignReferencesByTreeSearch(
                        entry_dn                  => $dn,
                        reference_attributes_aref => $ref_attr_aref
                    );
                }

                # write device status of $dn to mmap-file
                # (to be used from controld)
                $self->updateDeviceStatusFile($dn);
            }
        }
    }

    # switch off treesearch_cache
    $self->{'treesearch_cache'} = undef;

    return 1;
}

=head2 changeDeviceTypeForEntry

=head3 Example

 my $changed_entry = $sc_cache->changeDeviceTypeForEntry(
    'dn' => 'cn=pc0001,ou=test,o=company,c=de',
    'device_type_dn' => 'cn=deviceTest,ou=global,o=debeka,c=de'
 );

=over 4

=item Parameter

hash containing of

=over 4

=item dn

=item device_type_dn

=back

=item Return

entry with chages or undef, if there were errors

=back

=cut

sub changeDeviceTypeForEntry
{
    my $self   = shift;
    my %params = (
        'dn'             => undef,
        'device_type_dn' => undef,
        @_
    );
    my $logger         = $LOGGER;
    my $dn             = $params{'dn'};
    my $device_type_dn = $params{'device_type_dn'};
    my $entry          = $self->getEntry($dn);
    my $location       = $self->getLocation( 'dn' => $dn );
    if ( !ref $entry ) {
        $logger->warn("[LDAP] no entry found for dn: $dn");
    } else {
        if ( $self->hasEntry($device_type_dn) ) {
            my $location_entry = $self->getEntry($location);
            my $ip             = $self->getIpAddress( 'entry_dn' => $dn );
            my $new_name       = $self->{'name_generator'}->generateCn(
                'new_location_entry' => $location_entry,
                'old_entry'          => $entry,
                'new_entry_infos'    => {
                    'devicetype_dn' => $device_type_dn,
                    'ip_addr'       => $ip
                }
            );
            if ( !$new_name ) {

                # no name could be generated
                return;
            }
            my @old_cn = $entry->get_value('cn');
            my $new_dn = "cn=$new_name," . ldap_shiftbase($dn);
            $logger->info(
                "new name $new_name ($new_dn) for dn: $dn with new devicetype: $device_type_dn "
            );
            my $ldap_con = $self->getLdapConnection( 'needcap' => 'rw' );

            # remove old entry from cache
            $self->removeEntry($dn);

            # change the cn of the updated entry and update those changes to ldap
            $ldap_con->modify(
                $dn,
                replace => {
                    'dbkDeviceTypeDn' => $device_type_dn,
                    'dbkAuthorName'   => 'ScCache change devicetype'
                }
            );

            # change dn and update it to ldap
            $ldap_con->moddn(
                'dn'     => $dn,
                'newrdn' => "cn=$new_name"
            );

            $ldap_con->modify( $new_dn, delete => { 'cn' => \@old_cn } );

            # refetch the new entry from ldap and insert it into the cache
            my $got_correct_entry;
            my $tries = 0;
            do {
                my $message = $ldap_con->search(
                    'base'   => $new_dn,
                    'filter' => "cn=$new_name",
                    'scope'  => 'base'
                );
                if ( ref $message and !$message->code ) {
                    $got_correct_entry = $message->entry(0);
                } else {
                    sleep($tries);
                    $logger->info("[LDAP-DATA] waiting for $new_dn");
                }
            } until ( ref $got_correct_entry or $tries++ > 10 );
            if ( ref $got_correct_entry ) {
                $self->setEntry($got_correct_entry);

                # correct all ip related stuff in changed location
                $self->assignLocations(
                    'refreshed_locations_aref' => [$new_dn] );
                $self->assignIpDevices(
                    'refreshed_locations_aref' => [$location] );
                return {
                    'dn'             => $got_correct_entry->dn,
                    'device_type_dn' =>
                      $got_correct_entry->get_value('dbkDeviceTypeDn'),
                    'ip' => $self->getIpAddress(
                        'entry_dn' => $got_correct_entry->dn
                    )
                };
            }
        } else {
            $logger->warn(
                "[LDAP] no deviceType found for $device_type_dn for dn $dn");
        }
    }
    return;
}

=head2 getFreeIpInLocation

Find first free ip address in location.

=over 4

=item Parameter

	- location_dn string representing location

=item Return

	- string containing ip or undef if no ip could be found

=back

=cut

sub getFreeIpInLocation
{
    my $self   = shift;
    my %params = (
        'location_dn' => undef,
        @_
    );
    my $logger      = $LOGGER;
    my $location_dn = $params{'location_dn'};

    #check if all paramaters were given
    if ( !$location_dn ) {
        $logger->error('method must be called with param location_dn');
        return;
    }

    my $location_entry = $self->getEntry($location_dn);
    if ( !ref $location_entry ) {
        $logger->error( '[LDAP-DATA] location_dn '
              . $location_dn
              . ' could not be found in entrycache' );
        return;
    }
    my $location_fixed_range = $location_entry->get_value('dbkdhcpfixedrange');
    if ( !$location_fixed_range ) {
        $logger->error( '[LDAP-DATA,DHCP] location '
              . $location_dn
              . ' has no dbkDhcpFixedRange. Strange.' );
        return;
    }

    #convert dbkDhcpFixedRange in a format SmartClient::IP can use
    $location_fixed_range =~ s/\s+//go;
    $location_fixed_range =~ s/,/ - /o;

    #find first free ip
    my $location_ip_block = SmartClient::IP->new($location_fixed_range);
    if ( !$location_ip_block ) {
        $logger->error( '[LDAP-DATA,DHCP] location '
              . $location_dn
              . ' has broken dbkDhcpFixedRange.' );
        return;
    }
    my $free_ip;
    my $ip_next           = $location_ip_block->getIpIterator();
    my $href_ip_addresses = $self->{'ip_addresses'};
    while ( my $test_ip = $ip_next->() ) {
        next if $test_ip =~ m/\.0$/o;
        if ( !exists $href_ip_addresses->{$test_ip} ) {

            #found free ip
            $free_ip = $test_ip;
            last;
        }
    }
    return $free_ip;
}

sub getSerialForLocation
{
    my $self  = shift;
    my %param = (
        'location_dn' => undef,
        'config'      => $self->{'config'},
        @_,
    );

    my $logger        = $LOGGER;
    my $location_dn   = $param{'location_dn'};
    my $config        = $param{'config'};
    my $href_location = $self->getInternalEntryHash($location_dn);
    my $zone_name     = $self->getZoneNameForLocation($location_dn);
    my $serial;
    my $zone_file_path;

    # if the serial number is already stored, return it,
    # other read it from the zone file
    if ( $href_location->{'serial'} ) {
        $serial = $href_location->{'serial'};
    } else {
        ( $serial, $zone_file_path ) =
          $self->getSerialFromFile( zone_name => $zone_name );
        $logger->debug(
            "[DNS] storing zonefile path $zone_file_path for $location_dn");
        $href_location->{'zonefile_path'} = $zone_file_path;
        $logger->debug("[DNS] storing serial $serial for $location_dn");
        $self->setSerialForLocation(
            location_dn => $location_dn,
            serial      => $serial
        );
    }
    return $serial;
}

sub getSerialForRevZone
{
    my $self  = shift;
    my %param = (
        'rev_zone_name' => undef,
        'config'        => $self->{'config'},
        @_,
    );

    my $logger = $LOGGER;

    #my $location_dn    = $param{'location_dn'};
    my $config        = $param{'config'};
    my $rev_zone_name = $param{'rev_zone_name'};

    my $rev_zones_hash = $self->{'rev_zones'};
    my $serial;
    my $zone_file_path;

    # if the serial number is already stored, return it,
    # otherwise read it from the zone file
    if ( $rev_zones_hash->{$rev_zone_name}->{'serial'} ) {
        $serial = $rev_zones_hash->{$rev_zone_name}->{'serial'};
    } else {
        my ( $serial, $zone_file_path ) =
          $self->getSerialFromFile( zone_name => $rev_zone_name );
        $logger->debug(
            "[DNS] storing zonefile path $zone_file_path for $rev_zone_name");
        $rev_zones_hash->{$rev_zone_name}->{'zonefile_path'} = $zone_file_path;
        $logger->debug("[DNS] storing serial $serial for $rev_zone_name");
        $self->setSerialForRevZone(
            rev_zone_name => $rev_zone_name,
            serial        => $serial
        );
    }
    return $serial;
}

=head2 getSerialFromFile

Read serial from zone file. If the zone file does not exist,
"0000000000" will be returned.

=over 4

=item Parameter

 - zonename (also works for reverse zones terminating with "arpa")

Needed configuration variables

 - 'LDAP2DNS_ZONEFILEPATH'

=item Rückgabewert

serial from SOA record or "0000000000" if the zone file does not exist.

=back

=cut

sub getSerialFromFile
{
    my $self  = shift;
    my %param = (
        'zone_name' => undef,
        'config'    => $self->{'config'},
        @_,
    );
    my $zone_name    = $param{'zone_name'};
    my $config       = $param{'config'};
    my $zonefilename = $zone_name;
    my @zone_lines;
    my $zonefile_path = $config->get('LDAP2DNS_ZONEFILEPATH');
    my $serial;

    my $logger = $LOGGER;

    # assemble path name
    my $base_zonename;
    my $path_name;
    if ( $zone_name =~ /arpa$/o ) {
        $base_zonename = 'rev';
        $path_name     = $zonefile_path . "/" . $base_zonename;
    } else {
        $base_zonename = lc( ldap_dn2dnsname( $config->get('ALL_LDAPBASE') ) );
        my $zone_name_path = zoneNameToPath( $zone_name, $base_zonename );
        if ($zone_name_path) {
            $path_name = $zonefile_path . "/" . $zone_name_path;
        } else {
            $path_name = $zonefile_path;
        }
    }

    my $input_filename = $path_name . "/" . $zonefilename;
    if ( !open( my $fh, '<', $input_filename ) ) {
        $logger->warn(
            "[DNS] opening $input_filename for reading serial failed: $!");
        $serial = "0000000000";
    } else {
        $serial = parseSerial($fh);
        close($fh);
        if ( !$serial ) {
            $logger->error("[DNS] error reading serial from $input_filename");
            $serial = "0000000000";
        }
    }
    $logger->debug("[DNS] got $serial for $input_filename");
    return ( $serial, $input_filename );
}

sub parseSerial
{
    my $fh    = shift;
    my $index = 0;
    my $serial_number;
    my $get_serial_rex = qr/.+  # may not be the @ character, but a string
                        \s+     # then spaces
                        IN      # 'IN'
                        \s+     # again only spaces
                        SOA     # 'SOA'
                        \s+     # spaces only
                        \S+     # something without spaces
                        \s+     # again some space
                        \S+     # something without spaces
                        \s*     # may be no whitespace before paren
                        \(      # '('
                        \s*     # may be no whitespace after paren
                        (\d+)   # this is the serial number, which we are interested in
                        /ox;

    my ($soa_sn_line) = getSoaLine($fh);
    if ( $soa_sn_line =~ /$get_serial_rex/ ) {
        $serial_number = $1;
    }
    return $serial_number;

}

sub getSoaLine
{
    my $fh       = shift;
    my $soa_line = '';

    my $soa_rex         = qr/\s+IN\s+SOA/ox;
    my $soa_closing_rex = qr/\)/ox;

    # find first and last line of SOA
    my $line;
    my $in_soa_lines = 0;
    while ( defined( $line = <$fh> ) ) {
        chomp($line);
        if ( $line =~ /$soa_rex/ ) {
            $in_soa_lines = 1;
        }
        if ($in_soa_lines) {
            $soa_line .= $line;
        }
        if ( $line =~ /$soa_closing_rex/ ) {
            last;
        }
    }

    return $soa_line;
}

=head2 zoneNameToPath

Builds a path from a zone name by bringing the parts of the zone
name separated by "." into reverse order while replacing the "."
by "/". By optionally passing a trailing part of the zone name,
this part will be excluded from conversion into the path name.

=head3 Example

an example without optional parameter
  my $zone_path = $sc_cache->zoneNameToPath(
    'gs345.koeln.smartclient.de'
  );
  #$zone_path should be 'de/smartclient/koeln/gs345'

and now with the optional parameter
  my $zone_path = $sc_cache->zoneNameToPath(
    'gs345.koeln.smartclient.de',
    'smartclient.de'
  );
  #$zone_path should now be 'koeln/gs345'

=over 4

=item Parameter

 - zonename (example: gs345.koeln.smartclient.de)
 - optional: trailing part of zone name to be excluded from conversion
   (example: smartclient.de)

=item Return

path name

=back

=cut

sub zoneNameToPath
{
    my ( $zone_name, $remove_part ) = @_;

    my $zone_path = $zone_name;
    my $new_zone_path;
    my @zone_path_parts;
    my @reverse_zone_path_parts;

    # replace . with \. for use in regex
    $remove_part =~ s/\./\\./g;

    # remove the unwanted part
    $zone_path =~ s/$remove_part//;

    # remove a trailing .
    $zone_path =~ s/\.$//;

    # replace . with /
    $zone_path =~ s#\.#/#g;

    # reverse order
    @zone_path_parts         = split( /\//, $zone_path );
    @reverse_zone_path_parts = reverse(@zone_path_parts);
    $new_zone_path           = join( '/', @reverse_zone_path_parts );

    return $new_zone_path;
}

=head2 getLocationZoneFileName

Returns a locations zone file name including path from the cache

=over 4

=item Parameter

location DN

=item Rückgabewert

zone file name including path or undef if the data could
not be found in the cache

=back

=cut

sub getLocationZoneFileName
{
    my $self        = shift;
    my $location_dn = shift;

    my $zonefile_path;

    my $href_location = $self->getInternalEntryHash($location_dn);

    my $logger = $LOGGER;

    if ( !$href_location ) {
        $logger->error(
            "[DNS,LDAP-DATA] getLocationZoneFileName: location $location_dn "
              . "does not exist in the cache" );
        return;
    }
    $zonefile_path = $href_location->{'zonefile_path'};
    if ( !$zonefile_path ) {
        $logger->error(
            "[DNS,LDAP-DATA] getLocationZoneFileName: zonefile path $location_dn "
              . "does not exist in the cache" );
        return;
    }
    return $zonefile_path;
}

=head2 getRevZoneFileName

Returns a reverse zone file name including path from the cache

=over 4

=item Parameter

reverse zone name

=item Rückgabewert

zone file name including path or undef if the data could
not be found in the cache

=back

=cut

sub getRevZoneFileName
{
    my $self          = shift;
    my $rev_zone_name = shift;

    my $zonefile_path;

    my $rev_zones_hash = $self->{'rev_zones'};

    my $logger = $LOGGER;

    if ( !exists $rev_zones_hash->{$rev_zone_name} ) {
        $logger->error( "[DNS] getRevZoneFileName: reverse zone $rev_zone_name "
              . "does not exist in the cache" );
        return;
    }
    $zonefile_path = $rev_zones_hash->{$rev_zone_name}->{'zonefile_path'};
    if ( !$zonefile_path ) {

        #TODO: is this the right way to get the zonefile_path?
        my ( $serial, $filepath ) =
          $self->getSerialFromFile( 'zone_name' => $rev_zone_name );
        if ($filepath) {
            $rev_zones_hash->{$rev_zone_name}->{'zonefile_path'} = $filepath;
            return $filepath;
        }
        $logger->error(
                "[DNS] getRevZoneFileName: zonefile path for $rev_zone_name "
              . "does not exist in the cache" );
        return;
    }
    return $zonefile_path;
}

=head2 getLocationZoneName

Returns a locations zone name from the cache

=over 4

=item Parameter

location DN

=item Rückgabewert

zone name or undef if the data could
not be found in the cache

=back

=cut

sub getLocationZoneName
{
    my $self        = shift;
    my $location_dn = shift;

    my $zone_name;

    my $href_location = $self->getInternalEntryHash($location_dn);

    my $logger = $LOGGER;

    if ( !$href_location ) {
        $logger->error(
                "[DNS,LDAP-DATA] getLocationZoneName: location $location_dn "
              . "does not exist in the cache" );
        return;
    }
    $zone_name = $href_location->{'zone_name'};
    if ( !$zone_name ) {
        if ( !$self->setZoneNameForLocation( 'location_dn' => $location_dn ) ) {
            $logger->error(
                "[DNS,LDAP-DATA] getLocationZoneName: zone name $location_dn "
                  . " could not be set in the cache" );
            return;
        }
        $zone_name = $href_location->{'zone_name'};
    }
    return $zone_name;
}

=head2 getLocationAliasZones

Returns the alias zone names for a location from the cache, the
values come from the LDAP attribute zoneName in an scLocation
object. As this is a multivalue attribute, an array will be
returned.

=over 4

=item Parameter

location DN

=item Rückgabewert

Array of zone names or undef if the data could
not be found in the cache. If no alias zone names
are stored in LDAP, an empty array will be returned.

=back

=cut

sub getLocationAliasZones
{
    my $self        = shift;
    my $location_dn = shift;

    my @zone_aliases = ();

    my $href_location = $self->getInternalEntryHash($location_dn);

    my $logger = $LOGGER;

    if ( !ref $href_location ) {
        $logger->error(
                "[DNS,LDAP-DATA] getLocationAliasZones: location $location_dn "
              . "does not exist in the cache" );
        return;
    }

    my $location_entry = $self->getEntry($location_dn);
    if ( $location_entry->exists('zoneName') ) {
        @zone_aliases = $location_entry->get_value('zoneName');
    }
    return @zone_aliases;
}

=head2 checkReferenceAttributes

checks if a Reference Attribute is part of the entry cache attributes.
If not, it logs a warning and removes the reference attribute 
from the list.

=over 4

=item Parameter

 - entry_dn: DN of a IP device entry (eg. Workstation, Printer, Server, Service)
 - reference_attributes_href: reference to a hash containing the reference attribute names
   and their values

=item Return

1 (true) if the data could be stored, 0 (false) on error, eg. if
the entry does not exist in the cache

=back

=cut

sub checkReferenceAttributes
{
    my $logger = $LOGGER;

    my $self = shift;

    my $aref_entrycache_attrs          = $self->{'entrycache_attrs'};
    my $href_aref_reference_attributes = $self->{'reference_attributes'};

    if ( !$aref_entrycache_attrs ) {
        $logger->error(
            "[LDAP-DATA] no attributes defined for this entrycache. This can not work"
        );
        return;
    }

    my @entrycache_attrs = @{$aref_entrycache_attrs};

    if ( grep( /^\*$/, @entrycache_attrs ) ) {
        $logger->debug(
            "[LDAP-DATA] all attributes are in the cache. No checks required");
        return 1;
    }

    # example:
    # $self->{'reference_attributes'}->{'dbkWorkstation'} = [
    #    'dbkRefImageDn', 'scRefBootOptionsDn',
    #    'dbkDoTftpBoot', 'dbkDoRsync',
    #    'dbkUpdateTime', 'dbkDeviceTypeDn' ];
    my $result = 1;
    foreach my $objectclass ( keys( %{$href_aref_reference_attributes} ) ) {
        my $aref_reference_attributes =
          $href_aref_reference_attributes->{$objectclass};
        my @confirmed_reference_attributes = ();

        foreach my $reference_attribute ( @{$aref_reference_attributes} ) {
            if ( grep( /^${reference_attribute}$/, @entrycache_attrs ) ) {
                push( @confirmed_reference_attributes, $reference_attribute );
            } else {
                $logger->warn(
                    "[LDAP-DATA] $reference_attribute is not a entrycache attribute and can therefore not be a reference attribute => ignored"
                );
                $result = 0;
            }
            $href_aref_reference_attributes->{$objectclass} =
              \@confirmed_reference_attributes;
        }

    }
    $logger->debug(
        "reference_attributes: " . Dumper( $self->{'reference_attributes'} ) );

    return $result;
}

=head2 setReferenceAttributes

store reference attributes to a given DN in the entry cache

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<entry_dn>

DN of a IP device entry (eg. Workstation, Printer, Server, Service)

=item C<reference_attributes_href>

reference to a hash containing the reference attribute names
and their values

=back

=item Return

1 (true) if the data could be stored, 0 (false) on error, eg. if
the entry does not exist in the cache

=back

=cut

sub setReferenceAttributes
{
    my $self  = shift;
    my %param = (
        'entry_dn'                  => undef,
        'reference_attributes_href' => undef,
        @_,
    );

    my $entry_dn       = $param{'entry_dn'};
    my $ref_attrs_href = $param{'reference_attributes_href'};

    my $entry_cache = $self->{'entrycache'};

    my $logger = $LOGGER;

    if ( !$entry_cache->{$entry_dn} ) {
        $logger->error(
            "[LDAP-DATA] setReferenceAttributes: $entry_dn does not exist in the cache"
        );
        return 0;
    }
    $entry_cache->{$entry_dn}->{'reference_attributes'} = $ref_attrs_href;
    return 1;
}

=head2 assignReferencesByTreeSearch

assign reference attributes to a given DN by doing a treesearch
for a given list of attribute names

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<entry_dn>

DN of a IP device entry (eg. Workstation, Printer, Server, Service)

=item C<ref_attrs>

List of reference attribute names

=back

=item Return

reference to hash of found reference attributes

=back

=cut

sub assignReferencesByTreeSearch
{
    my $self  = shift;
    my %param = (
        'entry_dn'                  => undef,
        'reference_attributes_aref' => undef,
        @_,
    );

    my $entry_dn      = $param{'entry_dn'};
    my $ref_attr_list = $param{'reference_attributes_aref'};

    my $logger = $LOGGER;

    my %ref_attrs;

    # TODO: this seems to be quite expensive and not the expected behaviour.
    # first treeSearch is really expensive
    # second I would think that only ONE standard should be used and not a standard for
    #        every attribute in $ref_attr_list
    # It would be probably better to determine the standards for a objectClass and deviceType
    # for each location once and use that one, but for this to happen we have to change the way
    # assignIpDevices, setEntry and this methods works.
    foreach my $ref_attr ( @{$ref_attr_list} ) {
        my $ref_entry = $self->treeSearch(
            'start_dn'        => $entry_dn,
            'attr'            => $ref_attr,
            'smartdevicetype' => 1
        );
        if ( !$ref_entry ) {
            $logger->warn(
                "[LDAP-DATA] could not assign " . "$ref_attr to $entry_dn" );
            next;
        }

        # do this multival safe
        my @values = $ref_entry->get_value($ref_attr);
        $ref_attrs{$ref_attr} = \@values;
    }
    $self->setReferenceAttributes(
        'entry_dn'                  => $entry_dn,
        'reference_attributes_href' => \%ref_attrs
    );
    return \%ref_attrs;
}

=head2 getReferenceAttributes

Get all reference attributes for a given entry DN from the cache

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<entry_dn>

DN of a IP device entry (eg. Workstation, Printer, Server, Service)

=back

=item Return

Reference to the hash containing the reference attribute name as key
and a reference to an array of values. Data should not be written
using this hash reference because it points to the entry cache.

If the entry DN does not exist in the cache, or no reference
attributes are stored for the entry, undef will be returned.

=back

=cut

sub getReferenceAttributes
{
    my $logger = $LOGGER;

    my $self  = shift;
    my %param = (
        'entry_dn' => undef,
        @_,
    );
    my $entry_dn    = $self->normalize( $param{'entry_dn'} );
    my $entry_cache = $self->{'entrycache'};

    if ( !$entry_cache->{$entry_dn} ) {
        $logger->error(
            "[LDAP-DATA] getReferenceAttributes: $entry_dn does not exist in the cache"
        );
        return;
    }

    my $reference_attributes =
      $entry_cache->{$entry_dn}->{'reference_attributes'};

    if ( !defined($reference_attributes) ) {
        $logger->warn(
            "[LDAP-DATA] getReferenceAttributes: no reference attributes exist "
              . " in the cache for $entry_dn" );
        return;
    }
    return $reference_attributes;
}

=head2 getReferenceAttribute

Get a specified reference attribute for a given entry DN from the cache

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<entry_dn>

DN of a IP device entry (eg. Workstation, Printer, Server, Service)

=item C<reference_attribute_name>

name of the reference attribute to retrieve

=back

=item Return

Array of values of the the reference attribute or an empty array if the reference
attribute is not stored or the given entry DN does not exist in the cache.
An empty is returned because an array which is assigned undef is true for perl.

=back

=cut

sub getReferenceAttribute
{
    my $self  = shift;
    my %param = (
        'entry_dn'                 => undef,
        'reference_attribute_name' => undef,
        @_,
    );
    my $entry_dn      = $param{'entry_dn'};
    my $ref_attr_name = $param{'reference_attribute_name'};

    my $entry_cache = $self->{'entrycache'};

    my $logger = $LOGGER;

    my $reference_attributes =
      $self->getReferenceAttributes( 'entry_dn' => $entry_dn );
    if ( !$reference_attributes ) {

        # error message already from getReferenceAttributes
        return;
    }

    if ( !$reference_attributes->{$ref_attr_name} ) {

        #TODO: check if objectclass should have ref_attr_name and log debug or error correspondingly
        $logger->error(
            "[LDAP-DATA] getReferenceAttribute: $ref_attr_name dn does not exist "
              . " in the cache for $entry_dn" );
        return;
    }

    # return array so that this is multivalue safe
    my @values = @{ $reference_attributes->{$ref_attr_name} };
    return @values;
}

=head2 getRefImageVersion

Get the reference image version for a given workstation DN

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<workstation_dn>

DN of Workstation entry

=back

=item Return

Version number of the reference image assigned to this workstation
or undef in case of error

=back

=cut

sub getRefImageVersion
{
    my $self  = shift;
    my %param = (
        'workstation_dn' => undef,
        @_,
    );

    my $workstation_dn = $param{'workstation_dn'};
    my $lc             = $self->{'entrycache_lowercase'};

    my $logger = $LOGGER;

    if ( !$self->isWorkstation($workstation_dn) ) {
        return;
    }

    my ($ref_image_dn) = $self->getReferenceAttribute(
        entry_dn                 => $workstation_dn,
        reference_attribute_name => 'dbkRefImageDn'
    );

    if ( !$ref_image_dn ) {

        # getReferenceAttribute already log error message
        return;
    }
    $ref_image_dn = lc($ref_image_dn) if $lc;

    # get the reference entry, and return the image version
    my $ref_image_entry = $self->getEntry($ref_image_dn);
    if ( !ref $ref_image_entry ) {
        $logger->error(
                "[LDAP-DATA] getRefImageVersion: could not get ref image "
              . "entry $ref_image_dn" );
        return;
    }
    my ($ref_image_version) = $ref_image_entry->get_value('dbkImageVersion');
    return $ref_image_version;
}

=head2 hasNewRefImageVersion

Check if the reference image version for a given workstation DN is
newer than the image version of the workstation

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<workstation_dn>

DN of Workstation entry

=back

=item Return

1 (true) if the reference image version differs, otherwise 0 (false)
or undef in case of error

=back

=cut

sub hasNewRefImageVersion
{
    my $self  = shift;
    my %param = (
        'workstation_dn' => undef,
        @_,
    );

    my $workstation_dn = $param{'workstation_dn'};

    my $logger = $LOGGER;

    my $ref_image_version =
      $self->getRefImageVersion( workstation_dn => $workstation_dn );

    # 0 is a valid image version for us.
    if ( !defined $ref_image_version ) {
        return;
    }

    # get workstation image version
    my $workstation_entry = $self->getEntry($workstation_dn);
    if ( !ref $workstation_entry ) {
        return;
    }
    my $workstation_image_version =
      $workstation_entry->get_value('dbkImageVersion');

    # compare ref image version with the workstation image version
    if ( $workstation_image_version eq $ref_image_version ) {
        return 0;
    }
    return 1;
}

=head2 getNetBootState

Get the netboot state for a given DN which may be a IP device (workstation,
server, printer) or location.

If the given DN points to a workstation, its image version will be compared
with the reference image version to which it is assigned. If the according
bkDoTftpBoot setting is set to LOCALBOOT, it has the precedence and image
version numbers are ignored.

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<dn>

DN of the entry for which to retrieve

=back 

=item Return

dbkDoTftpBoot value considering image version check for workstations,
which may be 'lan', 'wan', the name of the tftp server or 'LOCALBOOT'.

=back

=cut

sub getNetBootState
{
    my $self  = shift;
    my %param = (
        'dn' => undef,
        @_,
    );

    my $dn = $self->normalize( $param{'dn'} );

    my ($dbk_do_tftp_boot) = $self->getReferenceAttribute(
        entry_dn                 => $dn,
        reference_attribute_name => 'dbkDoTftpBoot'
    );

    if ( !$dbk_do_tftp_boot ) {

        # getReferenceAttribute already logs error message
        return;
    }

    if ( lc($dbk_do_tftp_boot) ne 'localboot' ) {
        my $entry = $self->getEntry($dn);
        if ( !ref $entry ) {
            return;
        }
        if ( has_objectclass( $entry, 'dbkWorkstation' ) ) {

            # check if the reference image version is newer
            if ( !$self->hasNewRefImageVersion( workstation_dn => $dn ) ) {
                $dbk_do_tftp_boot = 'LOCALBOOT';
            }
        }
    } else {
        $dbk_do_tftp_boot = uc($dbk_do_tftp_boot);
    }

    return $dbk_do_tftp_boot;
}

=head2 isNetboot

Get the netboot state for a given DN which may be a IP device (workstation,
server, printer) or location.

If the given DN points to a workstation, its image version will be compared
with the reference image version to which it is assigned. 
If the according dbkDoTftpBoot setting is set to LOCALBOOT, 
it has the precedence 
and image version numbers are ignored.

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<dn>

DN of the entry for which to retrieve

=back 

=item Return

    0: if device is set to localboot
    1: if device is set to netboot

=back

=cut

sub isNetboot
{
    my $self  = shift;
    my %param = (
        'dn' => undef,
        @_,
    );

    my $result = 1;

    my $netboot_state = $self->getNetBootState(%param);
    if ($netboot_state) {
        if ( $netboot_state eq 'LOCALBOOT' ) {
            $result = 0;
        }
    }

    return $result;
}

=head2 isCachingImageServer


=over 4

=item Parameter

hash consisting of:

=over 4

=item C<dn>

DN of the entry for which to retrieve

=back 

=item Return

    0: is not a caching image server
    1: is caching image server

=back

=cut

sub isCachingImageServer
{
    my $logger = $LOGGER;
    my $self   = shift;
    my %param  = (
        'dn' => undef,
        @_,
    );

    my $dn = $param{'dn'};

    if ( !$self->hasEntry($dn) ) {
        return 0;
    }
    my $cis = $self->getEntry($dn)->get_value('dbkCachingImageServer');

    if ( uc($cis) eq 'TRUE' ) {
        return 1;
    }

    return 0;
}

=head2 getNextAvailableDeviceInfos

generate a C<DN>, C<ip> and C<devicetype> for a new device in a given location. 
The location can be specified by passing a DN or a Net::LDAP::Entry object. If an entry
should be moved, the old entry can be given as C<old_entry>.

Caution:

The returned information has to be used immediately. It will be out of date as soon as the ScCache 
is altered by anyone.

=head3 Example

Get infos for a complete new device.

  my ($dn, $ip, $devicetype_dn) = $sc_cache->getNextAvailableDeviceInfos(
      'location_dn' => 'ou=location,o=test,c=de',
  );
 
Move an old entry into a new location. This will keep the devicetype of the old entry.

  my ($dn, $ip, $devicetype_dn) = $sc_cache->getNextAvailableDeviceInfos(
      'location_dn' => 'ou=location,o=test,c=de',
      'old_entry'   => $old_entry,
  );

Move an old entry into a new location while changing devicetype.

  my ($dn, $ip, $devicetype_dn) = $sc_cache->getNextAvailableDeviceInfos(
      'location_dn'     => 'ou=location,o=test,c=de',
      'old_entry'       => $old_entry,
      'new_entry_infos' => {
          'devicetype_dn' => 'cn=deviceRLPC,ou=global,o=test,c=de',
      },
  );

=over 4

=item Parameter

hash containing:

=over 4

=item C<location_dn>

dn of the location in which the new entry should be placed.
This string will be overriden by C<location_entry> if given.
[optional, must if no location_entry is given]

=item C<location_entry>

entry of the location in which the new entry should be placed. 
This entry will be taken instead of C<location_dn>, if given.
[optional, must if no location_dn is given]

=item C<old_entry>

will be used to determine the new C<devicetype_dn> and C<cn>.
[optional]

=item C<new_entry_infos>

hash containing:
[optional]

=over 4

=item C<ip_addr>

string representation of the ip address, the new entry should get. 
If none is given, first free ip in location will be used.
[optional]

=item C<devicetype_dn>

of the new entry. If none is given, the default for the location, or
(preferred) the devicetype of the C<old_entry> will be used.
[optional]

=item C<dhcp_hostname>

Name which was sent to DHCP from client. Will be used by NameGenerator
[optional]

=item C<mac_address>
Mac which was sent to DHCP from client. Will be used by NameGenerator
[optional]

=back

=back

=item Return

array containing:

C<dn> complete dn of the new entry

C<ip address> ip address used for calculating the name of the entry

C<devicetype> used for calculating the name of the entry

=back

=cut

sub getNextAvailableDeviceInfos
{
    my $self   = shift;
    my %params = (
        'location_dn'     => undef,
        'location_entry'  => undef,
        'old_entry'       => undef,
        'new_entry_infos' => undef,
        @_
    );

    my $logger = $LOGGER;

    #check params
    my $location_dn    = $params{'location_dn'};
    my $location_entry = $params{'location_entry'};
    if ( !$location_entry ) {
        if ( !$location_dn ) {
            $logger->error('[LDAP-DATA] no location_dn was given');
            return;
        }
        $location_entry = $self->getEntry($location_dn);
        if ( !$location_entry ) {
            $logger->error( '[LDAP-DATA] location_dn '
                  . $location_dn
                  . ' could not be found in entrycache' );
            return;
        }
    }
    $location_dn = $location_entry->dn;

    # if no old entry was given and no new_devicetype, try to guess the right devicetype
    my $old_entry         = $params{'old_entry'};
    my $new_devicetype_dn = $params{'new_entry_infos'}->{'devicetype_dn'};
    if ( !$old_entry and !$new_devicetype_dn ) {

        #TODO: referenceAttributes seem to be missing from scLocation Objects
        #        ($new_devicetype_dn) = $self->getReferenceAttribute(
        #            'entry_dn'                 => $location_dn,
        #            'reference_attribute_name' => 'dbkDeviceTypeDn'
        #        );
        my ($devicetype_entry) = $self->treeSearch(
            'start_dn' => $location_dn,
            'attr'     => 'dbkDeviceTypeDn'
        );
        if ($devicetype_entry) {
            $new_devicetype_dn =
              $devicetype_entry->get_value('dbkDeviceTypeDn');
        }
        if ( !$new_devicetype_dn ) {
            $logger->error(
                '[LDAP-DATA] no devicetype given, no old entry found and I could not figure it out by location_dn '
                  . $location_dn
                  . '. Whooa' );
            return;
        }
    } elsif ( $old_entry and !$new_devicetype_dn ) {

        #try to get devicetype from old entry
        if ( $old_entry->exists('dbkDeviceTypeDn') ) {
            $new_devicetype_dn = lc( $old_entry->get_value('dbkDeviceTypeDn') );
        }
    }
    $params{'new_entry_infos'}->{'devicetype_dn'} = $new_devicetype_dn;

    # if no ip_addr was given, try to guess one
    my $ip_addr = $params{'new_entry_infos'}->{'ip_addr'};
    if ( !$ip_addr ) {
        $ip_addr = $self->getFreeIpInLocation( 'location_dn' => $location_dn );
        $params{'new_entry_infos'}->{'ip_addr'} = $ip_addr;
    }

    #generate name
    my $cn = $self->{'name_generator'}->generateCn(
        'new_location_entry' => $location_entry,
        'old_entry'          => $old_entry,
        'new_entry_infos'    => $params{'new_entry_infos'},
    );
    if ( !$cn ) {
        $logger->error( 'No cn could be generated for devicetype '
              . $new_devicetype_dn
              . ' in location '
              . $location_dn );
        return;
    }
    my $new_dn = 'cn=' . $cn . ',' . $location_dn;
    return ( $new_dn, $ip_addr, $new_devicetype_dn );
}

=head2 getNextAvailableDn

generate a DN for a new device in a given location. The location can be specified by 
passing a DN or a Net::LDAP::Entry object.

This method is a simple wrapper for getNextAvailableDeviceInfos (look for parameter usage infos there)

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<location_dn>

=item C<location_entry>

=item C<old_entry>

=item C<new_entry_infos>

=over 4

=item C<ip_addr>

=item C<devicetype_dn>

=item C<dhcp_hostname>

=item C<mac_address>

=back

=back

=item Return

string representing dn of device

=back

=cut

sub getNextAvailableDn
{
    my $self   = shift;
    my %params = (
        'location_dn'     => undef,
        'location_entry'  => undef,
        'old_entry'       => undef,
        'new_entry_infos' => undef,
        @_
    );

    my ( $new_dn, $new_ip, $new_devicetype_dn ) =
      $self->getNextAvailableDeviceInfos(%params);
    if ($new_dn) {
        return $new_dn;
    } else {
        return;
    }
}

=head2 getDeviceStatus

This is called by sc_ldap2dnsdhcp-statusd.pl for the
XMLRPC method getDeviceStatus

=over 4

=item Parameter

device DN

=item Return

returns a hash for being processed into a XMLRPC struct.

=back

=cut

sub getDeviceStatus
{
    my $self = shift;

    my $logger = $LOGGER;

    my $device_dn = $self->normalize(shift);

    my %result = ();

    my $entry = $self->getEntry($device_dn);
    if ( !$entry ) {
        $result{'code'}    = 1;
        $result{'message'} = "$device_dn not found in cache";
        return %result;
    }

    if ( !$self->isDevice($entry) ) {
        $result{'code'} = 1;
        $result{'message'} =
          "$device_dn is not a scDevice, therefore device status is not available";
        return %result;
    }

    # are prior checks needed?
    #     if ( has_objectclass( $device_entry, 'dbkWorkstation' ) ) {
    #         my $ref_image_version =
    #           $self->getRefImageVersion( workstation_dn => $device_dn );
    #         if ($ref_image_version) {
    #             $result{'dbkRefImageVersion'} = $ref_image_version;
    #         }
    #         $result{'dbkImageVersion'} =
    #           $device_entry->get_value('dbkImageVersion');
    #     }

    # TODO: cn in ldap_cn2dnsname is probably not a good idea. fqdn is wanted
    my %result_special_attributes = ();
    my $ip_address = $self->getIpAddress( entry_dn => $device_dn );
    if ($ip_address) {
        $result_special_attributes{'ipHostNumber'} = $ip_address;
    }
    $result_special_attributes{'dnsName'} =
      ldap_cn2dnsname( $entry->get_value('cn') );

    if (   $self->isWorkstation($entry)
        or $self->isServer($entry)
        or $self->isPrinter($entry) )
    {
        %result_special_attributes = (
            %result_special_attributes,
            'macAddress' => scalar $entry->get_value('macAddress'),
            'netboot'    => $self->isNetboot( 'dn' => $device_dn ),
        );
    }

    if ( $self->isWorkstation($device_dn) ) {
        %result_special_attributes = (
            %result_special_attributes,
            'dbkRefImageVersion' =>
              $self->getRefImageVersion( workstation_dn => $device_dn ),
            'dbkImageVersion' => scalar $entry->get_value('dbkImageVersion'),
            'dbkCachingImageServer' =>
              $self->isCachingImageServer( 'dn' => $device_dn ),
        );
    }

    # add reference attributes
    %result = (
        $self->getReferenceAttributesValues($device_dn),
        %result_special_attributes
    );

    # remove all keys with undef value?

    return %result;
}

=head2 getReferenceAttributesValues

    
    # while the reference attributes are stored multivalue
    # compatible in the cache, this is not supported here


=over 4

=item Parameter

C<dn> of entry

=item Return

hash containing all current values for reference attributes of the objectclass

=back

=cut

sub getReferenceAttributesValues
{
    my $self = shift;
    my $dn   = shift;

    my %result = ();

    my $ref_attrs_href = $self->getReferenceAttributes( entry_dn => $dn );
    if ($ref_attrs_href) {
        foreach my $ref_attr_name ( keys %{$ref_attrs_href} ) {
            my ($ref_attr_value) = $self->getReferenceAttribute(
                entry_dn                 => $dn,
                reference_attribute_name => $ref_attr_name
            );
            $result{$ref_attr_name} = $ref_attr_value;
        }
    }

    if (%result) {
        return %result;
    }
    return;
}

=head2 getScheduleUpdateWorkstations

used by schedule_updates

get lists of workstations of a location,
that uses refImageDn 
and are not excluded from updates (dbkUpdateTime < 0)

    also evaluate dbkDoRsync != NONE ?

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<location_dn>

        'location_dn' =>    location

=item C<ref_image_dn>

        'ref_image_dn' =>   refImageDn

=item C<devicetype_dn>

        'devicetype_dn' =>  dbkDeviceTypeDn

=back

=item Return

    TODO: sort by ImageVersion

    hash ref:
        'code'          error code (0: okay, !=0: error)
        'workstations'  hash_ref
            <workstation_dn_1>  hash_ref from getDeviceStatus(workstation_dn_1)
                'macAddress'    ...
                'netboot'       ...
                ...
            <workstation_dn_2>  ...
                ...

=back

=cut

sub getScheduleUpdateWorkstations
{
    my $self   = shift;
    my %params = (
        'location_dn'   => undef,
        'devicetype_dn' => undef,
        'refimage_dn'   => undef,
        ##'netboot'      => undef,
        @_,
    );

    my $logger        = $LOGGER;
    my $location_dn   = $self->normalize( $params{'location_dn'} );
    my $devicetype_dn = $self->normalize( $params{'devicetype_dn'} );
    my $refimage_dn   = $self->normalize( $params{'refimage_dn'} );

    my %all_results;

    if ( $self->hasEntry($location_dn) ) {

        # get entries of a locations
        my @assigned_dn_list = $self->getLocationEntries(
            'objectclass_name' => 'dbkWorkstation',
            'base'             => $location_dn
        );
        $logger->debug( "[LDAP-DATA] location $location_dn entry DNs:\n"
              . join( "; ", @assigned_dn_list ) );

        my %workstations;

        # check all entries ...
        foreach my $entry_dn (@assigned_dn_list) {

            #$logger->debug($entry_dn);

            my $entry = $self->getEntry($entry_dn);
            if ( !ref $entry ) {
                $logger->warn("[LDAP-DATA] failed to get entry $entry_dn");
                next;
            }

            #$logger->debug(Dumper($entry));

            # is excluded from updates?
            my $dbkUpdateTime = $entry->get_value('dbkUpdateTime');
            if ( $dbkUpdateTime < 0 ) {

                # 'dbkUpdateTime < 0: ignore for update
                $logger->debug(
                    "[LDAP-DATA] $entry_dn excluded: dbkUpdateTime = $dbkUpdateTime ( < 0)"
                );
                next;
            }

            # reference_attributes contains array.
            # dbkRefImage is single value
            # use join to form a single string

            # is of type $devicetype_dn ?
            if ($devicetype_dn) {
                my $entry_devicetype_dn = join(
                    "",
                    $self->getReferenceAttribute(
                        entry_dn                 => $entry_dn,
                        reference_attribute_name => 'dbkDeviceTypeDn'
                    )
                );
                if ( !compare( $entry_devicetype_dn, $devicetype_dn, 0 ) ) {
                    $logger->debug(
                        "[LDAP-DATA] $entry_dn excluded: different device types: $entry_devicetype_dn != $devicetype_dn"
                    );
                    next;
                }
            }

            # uses $ref_image_dn ?
            if ($refimage_dn) {
                my $entry_ref_image = join(
                    "",
                    $self->getReferenceAttribute(
                        entry_dn                 => $entry_dn,
                        reference_attribute_name => 'dbkRefImageDn'
                    )
                );
                if ( !compare( $entry_ref_image, $refimage_dn, 0 ) ) {
                    $logger->debug(
                        "[LDAP-DATA] $entry_dn excluded: different ref image: $entry_ref_image != $refimage_dn"
                    );
                    next;
                }
            }

            my %result = $self->getDeviceStatus($entry_dn);

            #$logger->debug( "result: " . Dumper(%result) );
            $workstations{$entry_dn} = \%result;
        }

        #$logger->debug( "all: " . Dumper(%all_results) );

        %all_results = (
            'code'         => 0,
            'workstations' => \%workstations
        );

    } else {

        $logger->warn("[LDAP-DATA] unknown location $location_dn");
        %all_results = (
            'code'         => 0,
            'message'      => 'unknown location ' . $location_dn,
            'workstations' => {}
        );

    }

    return \%all_results;

}

=head2 setUpdate( % )

used by schedule_updates

    TODO: move to LdapCache?

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<ldap>

        'ldap'          => undef,

=item C<location_dn>

        'location_dn'   => undef,

=item C<workstation_dn>

        'workstation_dn' => undef,

=item C<device_type_dn>

        'device_type_dn' => undef,

=item C<tftpboot>

        'tftpboot'      => LOCALBOOT, WAN, DNS name

=item C<rsync>

        'rsync'         => LAN, WAN, NONE

=back

=item Return

    hash ref:
        'code'          error code.
                        max from all other codes
                        0: okay,
                        >0: error 
        <location_dn>   error code
        'workstations'  hash ref
            <workstation_dn_1>      error code
            <workstation_dn_2>      error code
            ...

=back

=cut

sub setUpdate
{
    my $self   = shift;
    my %params = (
        'ldap'           => undef,
        'location_dn'    => undef,
        'device_type_dn' => undef,
        'workstations'   => undef,
        'tftpboot'       => undef,
        'rsync'          => undef,
        @_,
    );

    my $logger            = $LOGGER;
    my $ldap              = $params{'ldap'};
    my $location_dn       = $self->normalize( $params{'location_dn'} );
    my $device_type_dn    = $self->normalize( $params{'device_type_dn'} );
    my $aref_workstations = $params{'workstations'};
    my $tftpboot          = $params{'tftpboot'};
    my $rsync             = $params{'rsync'};

    my %result   = ();
    my $code     = 0;
    my $code_max = 0;

    if ($location_dn) {
        $code = $self->setUpdateLocation(
            'ldap'           => $ldap,
            'location_dn'    => $location_dn,
            'device_type_dn' => $device_type_dn,
            'tftpboot'       => $tftpboot,
            'rsync'          => $rsync,
        );
        %result = ( $location_dn => $code );
        $code_max = $code;
    }

    if ( ref($aref_workstations) eq 'ARRAY' ) {

        my %result_workstations = ();
        foreach my $dn ( @{$aref_workstations} ) {

            #         # is 'dbkWorkstation's ?
            #         if ( !has_objectclass( $entry, 'dbkWorkstation' ) ) {
            #             $logger->debug("not a workstation");
            #             next;
            #         }

            my $dn    = $self->normalize($dn);
            my $entry = $self->getEntry($dn);
            if ($entry) {

                # check if entry has an tftpboot attribute
                # if not, modification is not required,
                # because standard is used
                if (   $entry->get_value('dbkDoTftpboot')
                    or $entry->get_value('dbkDoRsync') )
                {

                    # entry has dbkDoTftpboot value
                    # so entry must be changed
                    $code = $self->setUpdateEntry(
                        'ldap'     => $ldap,
                        'dn'       => $dn,
                        'tftpboot' => $tftpboot,
                        'rsync'    => $rsync,
                    );
                }
            } else {
                $code = 1024;
                $logger->warn("[LDAP-DATA] no entry found in cache for $dn");
            }

            $result_workstations{$dn} = $code;

            if ( $code > $code_max ) {
                $code_max = $code;
            }
        }
        $result{'workstations'} = \%result_workstations;
    }

    $result{'code'} = $code_max;
    return \%result;
}

=head2 setUpdateLocation

set boot option for location to localboot

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<ldap>

=item C<location_dn>

=item C<device_type_dn>

=item C<tftpboot>

=item C<rsync>

=back

=item Return

integer: 0: success

=back

=cut

sub setUpdateLocation
{

    my $self   = shift;
    my %params = (
        'ldap'           => undef,
        'location_dn'    => undef,
        'device_type_dn' => undef,
        'tftpboot'       => undef,
        'rsync'          => undef,
        @_
    );

    my $logger         = $LOGGER;
    my $ldap           = $params{'ldap'};
    my $location_dn    = $params{'location_dn'};
    my $device_type_dn = $params{'device_type_dn'};
    my $tftpboot       = $params{'tftpboot'};
    my $rsync          = $params{'rsync'};

    #$logger->debug( Dumper($ldap) );

    # TODO:
    #to support the old smartclient way and the new devicetype based approach
    #we should pass both devicetype_dn and standard_dc to constructStandardsName.
    #Or at least we should test for devicetype_dn.
    #constructStandardsName will  give back the ldap_shiftbase($location_dn),
    # if no standards could be found for the given location. (THIS IST PROBABLY NOT WANTED!)
    #my $location_standards_dn = "cn=standards," . $location_dn;
    my $location_standards_dn = $self->constructStandardsName(
        'dn'            => $location_dn,
        'devicetype_dn' => $device_type_dn,
    );

    if ( !$location_standards_dn ) {
        $logger->error(
            "[LDAP-DATA] failed to get standards DN for location $location_dn and device type $device_type_dn"
        );
        return 1025;
    }

    $logger->debug(
        "[LDAP-DATA] standards object for location $location_dn (devicetype $device_type_dn): $location_standards_dn"
    );

    return $self->setUpdateEntry(
        'ldap'     => $ldap,
        'dn'       => $location_standards_dn,
        'tftpboot' => $tftpboot,
        'rsync'    => $rsync,
    );
}

=head2 setUpdateEntry

set boot option for location to localboot

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<ldap>

=item C<dn>

=item C<tftpboot>

=item C<rsync>

=back

=item Return

integer: 0: success

=back

=cut

sub setUpdateEntry
{

    my $self   = shift;
    my %params = (
        'ldap'     => undef,
        'dn'       => undef,
        'tftpboot' => undef,
        'rsync'    => undef,
        @_
    );

    my $logger = $LOGGER;
    my $ldap   = $params{'ldap'};
    my $dn     = $params{'dn'};

    # set to "" to prevent warnings
    my $tftpboot = $params{'tftpboot'} || "";
    my $rsync    = $params{'rsync'}    || "";

    my $code    = 0;
    my $message = "";

    #$logger->debug( Dumper( %params ) );

    my $entry = $self->getEntry($dn);

    if ( !$entry ) {
        $code    = 1024;
        $message = "no entry found in cache for $dn";
    } else {
        my $tftpboot_old = $entry->get_value('dbkDoTftpboot') || "";
        my $rsync_old    = $entry->get_value('dbkDoRsync')    || "";

        # check if modification is required
        if (   ( $tftpboot ne $tftpboot_old )
            or ( $rsync ne $rsync_old ) )
        {

            # TODO: schema check
            # use scDeviceOptions and may attribute is available
            my %values = (
                'objectClass'   => 'dbkUpdate',
                'dbkUpdateTime' => $entry->get_value('dbkUpdateTime') || 0,
            );

            $values{'dbkDoTftpBoot'} =
                 $tftpboot
              || $tftpboot_old
              || "LOCALHOST";

            $values{'dbkDoRsync'} = $rsync || $rsync_old || "LAN";

            ( $code, $message ) = $ldap->set_values(
                DN     => $dn,
                ENTRY  => $entry,
                VALUES => \%values
            );
        }
    }

    if ($code) {
        $logger->error(
            "[LDAP-DATA] failed to set $dn update ($tftpboot, $rsync): $message ($code)"
        );
    }

    # note: $code is 0 on success.
    return $code;
}

=head2 isWorkstation

=over 4

=item Parameter

dn or entry

=item Return

C<true> if entry is a workstation or dn points to an entry which is a workstation

=back

=cut

#TODO: should we support both styles? dn and entry, or only one like isPrinter, isServer, ...
sub isWorkstation
{
    my $self = shift;
    my $dn   = shift;
    my $entry;
    if ( ref $dn ) {
        $entry = $dn;
        $dn    = $entry->dn;
    } else {
        $entry = $self->getEntry($dn);
    }

    if ( ref $entry ) {
        return has_objectclass( $entry, 'dbkWorkstation' );
    }
    return;
}

=head2 getIpDeviceObjectClasses

return list of IP device objectclasses

=over 4

=item Parameter

none

=item Return

array of IP device objectclass names

=back

=cut

sub getIpDeviceObjectClasses
{
    my $self = shift;

    return @{ $self->{'ip_device_objectclasses'} };
}

=head2 isDevice

test if an entry is a device for which devicestatus can be calculated

=over 4

=item Parameter

C<entry> or C<dn> of the object to be tested

=item Return

C<boolean> true if devicestatus can be calculated

=back

=cut

sub isDevice
{
    my $self  = shift;
    my $entry = shift;

    #if no entry was given, maybe it was an dn?
    if ( !ref $entry ) {
        $entry = $self->getEntry($entry);
        if ( !ref $entry ) {
            return;
        }
    }

    foreach my $test_objectclass ( $self->getIpDeviceObjectClasses() ) {
        if ( has_objectclass( $entry, $test_objectclass ) ) {
            return 1;
        }
    }
    return;
}

=head2 isServer

test if an entry is a server entry

=over 4

=item Parameter

C<entry> or C<dn> of the object to be tested

=item Return

C<boolean> true if entry is a server entry

=back

=cut

sub isServer
{
    my $self  = shift;
    my $entry = shift;

    #if no entry was given, maybe it was an dn?
    if ( !ref $entry ) {
        $entry = $self->getEntry($entry);
        if ( !ref $entry ) {
            return;
        }
    }
    if ( has_objectclass( $entry, 'dbkServer' ) ) {
        return 1;
    }
    return;
}

=head2 isPrinter

test if entry is a printer

=over 4

=item Parameter

C<entry> or C<dn> of the object to be tested

=item Return

C<boolean> true if entry is a server entry

=back

=cut

sub isPrinter
{
    my $self  = shift;
    my $entry = shift;

    #if no entry was given, maybe it was an dn?
    if ( !ref $entry ) {
        $entry = $self->getEntry($entry);
        if ( !ref $entry ) {
            return;
        }
    }
    if ( has_objectclass( $entry, 'dbkPrinter' ) ) {
        return 1;
    }
    return;
}

=head2 isService

test if an entry is a service entry

=over 4

=item Parameter

C<entry> or C<dn> of the object to be tested

=item Return

C<boolean> true if entry is a service entry

=back

=cut

sub isService
{
    my $self  = shift;
    my $entry = shift;

    #if no entry was given, maybe it was an dn?
    if ( !ref $entry ) {
        $entry = $self->getEntry($entry);
        if ( !ref $entry ) {
            return;
        }
    }
    if ( has_objectclass( $entry, 'dbkService' ) ) {
        return 1;
    }
    return;
}

=head2 getDnListForIp

get list of entry DNs assigned to an IP address

=over 4

=item Parameter

C<IP address> for which to find according DNs

=item Return

C<Array> of DN

=back

=cut

sub getDnListForIp
{
    my $self    = shift;
    my $ip_addr = shift;

    my $ip_hash = $self->{'ip_addresses'};

    my $logger = $LOGGER;

    if ( !$ip_hash->{$ip_addr} ) {
        $logger->error(
            "[LDAP-DATA] getDnListForIp: $ip_addr does not exist in cache");
        return;
    }
    return keys %{ $ip_hash->{$ip_addr}->{'dn_hash'} };
}

=head2 setEntry

overloaded method from LdapCache

=over 4

=item Parameter

=over 4

=item C<Net::LDAP::Entry>

=item reference to hash, which collects rev_zones of the entry

=back

=item Return

C<true> if entry was successfully set in ScCache

=back

=cut

sub setEntry
{
    my ( $self, $entry, $href_collect_rev_zones ) = @_;

    my $logger = $LOGGER;

    if ( $logger->is_debug() ) {
        $logger->debug( "call setEntry for "
              . ( ref($entry) ? $entry->dn : "No entry: $entry" ) );
    }

    if ( !ref $entry ) {
        return;
    }
    my $dn = $entry->dn;
    if ( $self->hasEntry($dn) ) {

        #remove old attributes for ScCache internal values
        $self->removeDnFromIpAddresses( $dn, $href_collect_rev_zones );
        $self->removeDnFromNameGenerator($dn);
    }
    $self->SUPER::setEntry($entry);

    # update ScCache internal values
    return 1;
}

=head2 setEntries

Sets all entries in one mass operation (overwrites the functionality of L<SmartClient::LdapCache::setEntries> completely.)
    
=head3 Example
        
  $ldap_cache->setEntries({ a => $entryA, b => $entryB, c => $entryC });

=over 4

=item Parameter

reference to hash with the entries as values

=item Return

nothing

=back

=cut

sub setEntries
{
    my ( $self, $href_entries ) = @_;
    my %rev_zones = ();
    for my $entry ( values %{$href_entries} ) {
        $self->setEntry( $entry, \%rev_zones );
    }
    $self->updateLocationsForRevZones( keys %rev_zones );
}

=head2 removeEntry

overloaded method from LdapCache

=over 4

=item Parameter

C<dn> of entry to remove from ScCache

=item Return

C<true> if entry could be removed

=back

=cut

sub removeEntry
{
    my $self = shift;
    my $dn   = shift;
    if ( ref $dn ) {
        $dn = $self->normalize( $dn->dn() );
    } else {
        $dn = $self->normalize($dn);
    }

    if ( $self->hasEntry($dn) ) {

        # remove ScCache internal attributes
        $self->removeDnFromIpAddresses($dn);

        # remove from device status file
        # (it is set in assignIpDevices)
        $self->removeDnFromDeviceStatusFile($dn);

        # remove dn from name generator
        $self->removeDnFromNameGenerator($dn);

        # call SUPER method
        my $scall_res = $self->SUPER::removeEntry($dn);

        # notify secondary server (result is ignored)
        my $controld_secondary = $self->getXmlRpcConnection();
        if ( ref($controld_secondary) ) {
            my $logger = $LOGGER;
            $logger->debug( "sending notifyOfDelete for $dn to "
                  . $controld_secondary->uri() );
            $controld_secondary->send_request( 'notifyOfDelete',
                RPC::XML::array->new( RPC::XML::string->new($dn) ) );
        }

        return $scall_res;
    }
    return;
}

=head2 removeDnFromNameGenerator

=over 4

=item Parameter

=item Return

=back

=cut

sub removeDnFromNameGenerator
{
    my $self = shift;
    my $dn   = shift;
    if ( ref $self and $self->{'name_generator'} ) {
        $self->{'name_generator'}->remove($dn);
    }
    return $self;
}

=head2 removeDnFromIpAddresses

tries to update the internal C<ip_addresses> hash when an entry is removed or updated

=over 4

=item Parameter

=over 4

=item string representing the C<dn> of the entry which ip address shall be removed/updated

=item reference to hash, to collect rev_zones for the dn

=back

=item Return

true if successful

=back

=cut

sub removeDnFromIpAddresses
{
    my ( $self, $_dn, $href_collect_rev_zones ) = @_;
    my $dn     = $self->normalize($_dn);
    my $logger = $LOGGER;

    if ( $logger->is_debug() ) {
        $logger->debug("call removeDnFromIpAddresses for $dn");
    }

    if ( !$self->hasEntry($dn) ) {
        return;
    }

    my $href_internal_entry = $self->getInternalEntryHash($dn);
    if ( !ref $href_internal_entry ) {
        return;
    }

    #remove all ip addresses for this dn from the ip_addresses hash
    foreach my $ip (
        $self->getIpAddress( 'entry_dn' => $dn, 'all_ip_addresses' => 1 ) )
    {

        # remove ip from ip_addresse.
        # removeIp will remove the ip from rev_zone automatically
        $self->removeIp(
            'ip_addr'           => $ip,
            'dn'                => $dn,
            'collect_rev_zones' => $href_collect_rev_zones
        );
    }

    return 1;
}

=head2 updateDeviceStatusFile( $ )

stores result of getDeviceStatus into mmap device status file.
The device status file is used by the 
SmartClient Control Daemon

=over 4

=item Parameter

C<dn>

=item Return

C<true>, if successful

C<undef>, if result of $self->getStatusCache() is not defined

=back

=cut

sub updateDeviceStatusFile
{
    my $logger = $LOGGER;
    my ( $self, $device_dn ) = @_;

    if ( !defined $device_dn ) {
        $logger->debug("It is useless to call this method without a dn.");
        return;
    }

    my $result;

    $logger->debug("check if device status cache exists");

    my $status_cache = $self->getStatusCache();

    if ( ref $status_cache ) {

        $logger->debug("updating status for $device_dn");
        my %device_status = $self->getDeviceStatus($device_dn);
        if ( %device_status and ( !$device_status{'error'} ) ) {
            $logger->debug("set status for $device_dn");
            $result =
              $status_cache->setDeviceStatus( $device_dn, \%device_status );
            if ( !$result ) {
                $logger->error("failed to store device status of $device_dn");
            }
        } else {
            $logger->debug("remove status for $device_dn");
            if ( $status_cache->getDeviceStatus($device_dn) ) {
                $result = $status_cache->removeDeviceStatus($device_dn);
                if ( !$result ) {
                    $logger->error(
                        "failed to remove device status of $device_dn");
                }
            }
        }
    } else {
        $logger->debug("no status cache used");
    }

    return $result;
}

=head2 removeDnFromDeviceStatusFile( $ )

removes dn from mmap device status file.
The device status file is used by the 
SmartClient Control Daemon

=over 4

=item Parameter

C<dn>

=item Return

C<true>, if successful

C<undef>, if  $self->getStatusCache() is not defined

=back

=cut

sub removeDnFromDeviceStatusFile
{
    my $logger = $LOGGER;
    my ( $self, $dn ) = @_;

    my $result;
    my $status_cache = $self->getStatusCache();

    if ( ref $status_cache ) {

        $result = $status_cache->removeDeviceStatus($dn);

    }

    return $result;
}

=head2 getNetbootLeasetimeForDn

=over 4

=item Parameter

C<dn>

=item Return

leasetime for netbooting device in seconds

=back

=cut

sub getNetbootLeasetimeForDn
{
    my $self = shift;
    my $dn   = shift;

    my ($ref_boot_options_dn) = $self->getReferenceAttribute(
        'entry_dn'                 => $dn,
        'reference_attribute_name' => 'scRefBootOptionsDn'
    );
    if ($ref_boot_options_dn) {
        my $ref_boot_options_entry = $self->getEntry($ref_boot_options_dn);
        if ( ref $ref_boot_options_entry ) {
            if ( $ref_boot_options_entry->exists('dbkDhcpNetbootLeasetime') ) {
                return
                  scalar $ref_boot_options_entry->get_value(
                    'dbkDhcpNetbootLeasetime');
            }
        }
    }
    return;
}

=head2 getRevZoneForIp

get the reverse zone for an ip address

=head3 Example

 my $rev_zone = $sc_cache->getRevZoneForIp('10.1.2.3');

should return "2.1.10.in-addr.arpa"

=over 4

=item Parameter

string reprenstation of an ip address

=item Returns

string containing reverse zone name

=back

=cut

sub getRevZoneForIp
{
    my $self    = shift;
    my $ip_addr = shift;
    my $logger  = $LOGGER;

    if ( !$ip_addr ) {
        $logger->info("[DNS,DHCP] no ip_addr was given");
        return;
    }

    my $ip_obj = SmartClient::IP->new( $ip_addr, '255.255.255.255' );
    if ( !ref $ip_obj ) {
        $logger->warn(
            "[DNS,DHCP] Could not get an SmartClient::IP Object for ip: $ip_addr"
        );
        return;
    }
    my ($rev_zone) = $ip_obj->getReverseZones();
    return $rev_zone;
}

=head2 getLocationForIp

get a location which is responsible for this ip (there can be more locations, for server locations)

=head3 Example

  my $location_dn = $sc_cache->getLocationForIp('10.1.2.3');

=over 4

=item Parameter

string representation of ip address

=item Return

dn for location, or undef if none found

=back

=cut

sub getLocationForIp
{
    my $self    = shift;
    my $ip_addr = shift;

    my $logger = $LOGGER;

    if ( !$ip_addr ) {
        $logger->debug("[DNS,DHCP] no ip address given, so no location found");
        return;
    }
    my $rev_zone = $self->getRevZoneForIp($ip_addr);
    if ( !$rev_zone ) {
        $logger->warn("[DNS,LDAP-DATA] no rev zone found for ip $ip_addr");
        return;
    }
    if ( $self->{'rev_zones'}->{$rev_zone} ) {
        my @locations =
          keys %{ $self->{'rev_zones'}->{$rev_zone}->{'locations'} };
        if (@locations) {
            foreach my $location (@locations) {
                my $ip_object = $self->getLocationIpObject($location);
                if ( ref $ip_object and $ip_object->contains($ip_addr) ) {
                    $logger->debug(
                        "found $location for $ip_addr with reverse zone $rev_zone"
                    );
                    return $location;
                }
            }
            $logger->warn(
                '[LDAP-DATA] strange, we found only locations in which the ip '
                  . $ip_addr
                  . ' doesn\'t really belong. will return '
                  . $locations[0]
                  . ' from ('
                  . join( ', ', @locations )
                  . ')' );
            return $locations[0];
        } else {
            $logger->warn(
                "[DNS,LDAP-DATA] no locations found in reverse zone $rev_zone"
            );
        }
    } else {
        $logger->debug("[DNS,LDAP-DATA] no reverse zone found for $rev_zone");
    }
    return;
}

=head2 getXmlRpcConnection

get an XML-RPC connection to controld of secondary server

=head3 Example

    my $controld_secondary = $sc_cache->getXmlRpcConnection();

=over 4

=item Parameter

none

=item Return

Returns an B<RPC::XML::Client> object which is connected to secondary server.
Returns undef if ALL_CONTROLD_SECONDARY_URL not defined or if the attempt
to call the method C<system.listMethods> failed.

=back

=cut

sub getXmlRpcConnection
{
    my $self   = shift;
    my $config = $self->{'config'};

    my $logger = $LOGGER;

    my $controld_secondary = $self->{'controld_secondary'};

    if ( !ref($controld_secondary) ) {

        my $controld_secondary_url =
          $config->get( 'ALL_CONTROLD_SECONDARY_URL', 1 );
        if ($controld_secondary_url) {
            $logger->info(
                "creating XMLRPC-Connection to " . $controld_secondary_url );
            $controld_secondary =
              RPC::XML::Client->new($controld_secondary_url);
            my $init_resp =
              $controld_secondary->send_request('system.listMethods');
            if ( ref $init_resp ) {
                $logger->debug( "$controld_secondary_url provides: "
                      . join( ', ', @{ $init_resp->value } ) );
                $logger->info( "using "
                      . $controld_secondary_url
                      . " for notification of changes" );
                $self->{'controld_secondary'} = $controld_secondary;
            } else {
                $logger->warn( "failed to connect to $controld_secondary_url: "
                      . $init_resp );
                return;
            }
        }

    }

    return $controld_secondary;

}

=head2 notifyOfModification

overwritten method fromq L<SmartClient::LdapCache::notifyOfModification>. 
This method tries to correct the C<< $self->{ip_addresses} >> hash after
the locations were changed. 

It expects an array reference where changes have happened. The method will scan those areas
and resync the cache to it. As the resync happens all valid entries will be cummulated in an 
array. A reference to that array will be returned, if it holds at least one entry.

=over 4

=item Parameter

reference to array with dn's of modified entries or locations 

=item Return

reference to array with dn's

=back

=cut

sub notifyOfModification
{
    my $self               = shift;
    my $aref_new_entry_dns = $self->SUPER::notifyOfModification(@_);
    if ($aref_new_entry_dns) {

        # TODO: this would not be needed, if setEntry and removeEntry would do
        # the ip changes instead of relying on assignIpDevices
        my @changed_location_dns = keys %{ $self->{'changed_locations_hash'} };
        $self->assignIpDevices(
            'refreshed_locations_aref' => \@changed_location_dns );
        return $aref_new_entry_dns;
    }
    return;
}

=head2 getStatusCache

=over 4

=item Parameter

=item Return

=back

=cut

sub getStatusCache
{
    my $self = shift;

    my $result = $self->{'status_cache'};

    if ( ref $result ) {
        return $result;
    }

    return;

}

1;
