#!/usr/bin/perl -w

#
# $Id: SharedFunction.pm 7578 2011-11-07 13:50:50Z FSchumacher $
#

#
# package name
#
package SmartClient::SharedFunction;
use strict;
use warnings;

#
# version of the package
#
our $VERSION = 2.0;

=encoding utf8

=head1 NAME

C<SmartClient::SharedFunction> - handling functions between processes


=head1 SYNOPSIS

    use SmartClient::SharedFunction;
    my $functionQueue = SmartClient::SharedFunction->new( 'id1', 'id2' );

=head1 DESCRIPTION

=head2 Purpose

C<SmartClient::SharedFunction> is used from slave processes to queue a function to a master process.
The master process performs the function 
and stores the result in shared memory.
This result can be evaluated by the client.

Using this class, all the real data can be stored
in the master process 
and functions are performed with a defined schedule.



=head2 Realisation

=head2 Error handling

=head2 Context and Dependencies

=over 4

=item Context

=item Dependencies


=back



=head2 Interfaces

=begin comment

    Command line parameter are already descripied 
    at SYNOPSIS

=end comment

=over 4

=item Programm

=item User

=back



=head2 Status

=begin comment

    descripes the TODOs/Errors for this programm

=end comment

    TODO
        minor_id % MAX
        cleanup/destroy of unused entries. Maybe generated in previous runs

=head2 Installation and other Notes



=head1 AUTHOR

    smartclient@dass-it.de




=head1 SEE ALSO

L<SmartClient::Queue>

=cut

#
# module definition
#

#use Exporter;
use base qw( SmartClient::Queue );

# =head1 symbols exported by default
#
# - variable1
# - function1
# - function2
#
# =cut
#
# @EXPORT = qw(  );

# =head1 symbols exported on request
#
# =cut
#
# @EXPORT_OK = qw( );

#
# used modules
#

use Data::Dumper;
use Log::Log4perl;
use Params::Validate qw(:all);
use Time::HiRes qw( gettimeofday tv_interval );
use Cache::FastMmap;
use SmartClient::Constant;
use SmartClient::Util;
use SmartClient::Queue;

#
# declarations
#

=head1 Functions (public)

=head2 new

The C<new> function creates a new empty queue for functions
and a shared memory hash for the results.

=over 4

=item Parameter

    - see check_parameter command

=item Return

  instance reference

=back

=cut

sub new
{
    my $class = shift;

    my %params = check_parameter(
        @_,
        {
            'shm_id_queue'     => { 'type' => SCALAR },
            'mmap_result_file' => {
                'type'    => SCALAR,
                'default' => "/var/cache/smartclient/shared_function_results",
            },
            'init_mmap_result_file' => {
                'type'    => SCALAR,
                'default' => 0
            },
            'busy_waiting_period' => {
                'type'     => SCALAR,
                'optional' => 1
            },
            'mmap_page_size' => {
                ## FastMmap default is 64k,
                ## but this isn't enough for
                ## getScheduleUpdateWorkstations
                ## with more than 130 workstations
                'type'    => SCALAR,
                'default' => "128k"
            },
            'mmap_num_pages' => {
                ## FastMmap default is 89
                ## should be a prime number
                'type'    => SCALAR,
                'default' => "313"
            },
            'mmap_expire_time' => {
                ## FastMmap setting
                ## maximum time in seconds after which
                ## mmap cache entry expires
                ## With a value of 0,
                ## cache entry are only removed based on LRU
                'type'    => SCALAR,
                'default' => "3600"
            },
        }
    );

    my $self =
      $class->SUPER::new( $params{'shm_id_queue'},
        $params{'busy_waiting_period'} );

    $self = {
        %params,
        %$self,
        'id_counter' => 0,
        'results'    => undef,
    };

    my $result = bless( $self, $class );

    $self->connect_results();

    return $result;
}

=head2 enqueue( function, parameter1, ... )

The C<enqueueFunction> method adds a hash, that descripes a function call to SmartClient::Queue.

Example:
    enqueueFunction( "testfunction", 1, "a", \@array, \%hash )

=over 4

=item Parameter

 - function name
 - parameter


=item Return

  none

=back

=cut

sub enqueue
{
    my $self = shift;

    # id: process id and counter,
    # in case a process is reused or a thread is used
    my $id = $$ . "." . $self->{'id_counter'};
    $self->{'id_counter'}++;

    my $data = {
        'id'           => $id,
        'function'     => shift,
        'params'       => [@_],
        'time_enqueue' => time(),
    };

    if ( $self->SUPER::enqueue($data) ) {
        return $id;
    }

    return;
}

=head2 run

C<run> will perform functions from queue 
for the amount of specified seconds.
The result from the function calls 
is stored in the result hash.
The result will be written in the result hash.

=over 4

=item Parameter

    - integer: timeout in seconds
    - rounds: number of tries to perform calls (default 0, which means at max one round)

=item Return

  integer: number of successful function calls

=back

=cut

sub run
{
    my $logger = Log::Log4perl->get_logger();

    my $self = shift;

    # parameter is timeout
    my $timeout         = shift;
    my $rounds          = shift || 0;
    my $round           = 0;
    my $counter         = 0;
    my $counter_success = 0;
    my $old_counter;

    do {
        $old_counter = $counter;
        $logger->info("run shared function queue in round ${round}/${rounds}");
        my $start_time = [gettimeofday];
        my $elapsed    = 0;
        while ( $elapsed < $timeout ) {
            my $data = $self->dequeue( $timeout - $elapsed );
            if ($data) {

                # return 1 for success, 0 on failure
                # count the successful function calls
                $counter_success += $self->perform($data);
                $counter++;
            }
            $elapsed = tv_interval( $start_time, [gettimeofday] );
        }
    } until ( $counter == $old_counter or ++$round > $rounds );
    $logger->debug(
        "successful performed $counter_success (of $counter) in ${round} with ${timeout}s duration"
    );

    return $counter_success;
}

=head2 run_nb

The C<run_nb> method is identical to the C<run> method,
except that it doesn't block.

=over 4

=item Parameter

 none

=item Return

  boolean: success

=back

=cut

sub run_nb
{
    my $self = shift;
    return $self->run(0);
}

=head2 perform( HASHREF )

The C<perform> method gets an function descriptions
and performs it.


=over 4

=item Parameter

 - HASHREF


=item Return

  boolean

=back

=cut

sub perform
{
    my $logger = Log::Log4perl->get_logger();

    my $self = shift;
    my $call = shift;

    my $results = $self->{'results'};

    my $id            = $call->{'id'};
    my $function      = $call->{'function'};
    my @params        = @{ $call->{'params'} };
    my $time_enqueue  = $call->{'time_enqueue'};
    my $params_string = join( ", ", @params );

    # disable strict refs checking
    # to be able to call functions
    # by their name as string
    no strict 'refs'; ## no critic

    # check if function exists
    if ( not exists( &{$function} ) ) {
        my $message =
          "failed to perform function $function( $params_string ) [id: $id]. Function doesn't exist";
        $logger->error($message);
        $results->set(
            $id,
            {
                'id'      => $id,
                'code'    => SHARED_FUNCTION_DOES_NOT_EXIST,
                'message' => "SharedFunction::perform: " . $message,
                'time_shared_function_enqueue' => $time_enqueue,
            }
        );

        return 0;
    }

    $logger->debug("perform function: $function( $params_string ) [id: $id]");

    my $time_start_perform = time();
    my %returnhash         = &{$function}(@params);

    # FastMmap::set sometimes fails. Why?
    # therefore it is tried several times
    my $result_stored;
    my $counter = 1;
    do {
        if ( $counter > 1 ) {
            sleep 1;
            $logger->warn("try $counter to store result $id");
        }
        $result_stored = $results->set(
            $id,
            {
                'id'                           => $id,
                'time_shared_function_enqueue' => $time_enqueue,
                'time_shared_function_perform' => $time_start_perform,
                'time_shared_function_result'  => time(),

                # add error status?
                # errorcode => 0,
                # errormsg  => "",
                %returnhash,
            }
        );
        $counter++;
    } until ( $result_stored or ( $counter > 5 ) );

    if ( !$result_stored ) {
        my $message =
          "failed to store result of function $function( $params_string ) [id: $id]. Probaly page_size is to small";
        $logger->error($message);
        $logger->debug(
            "result of [id: $id] should have been " . Dumper(%returnhash) );
        $results->set(
            $id,
            {
                'id'      => $id,
                'code'    => SHARED_FUNCTION_STORE_RESULT_ERROR,
                'message' => "SharedFunction::perform: " . $message,
                'time_shared_function_enqueue' => $time_enqueue,
                'time_shared_function_perform' => $time_start_perform,
            }
        );
        return 0;
    }

    return 1;
}

=head2 checkForResult

The C<checkForResult> return true
if a result is available, otherwise false.


=over 4

=item Parameter

    - none


=item Return

    boolean

=back

=cut

sub checkForResult
{
    my $self = shift;
    my $id   = shift;

    my $results = $self->{'results'};

    return defined( $results->get($id) );
}

=head2 getResult( $;$ )

The C<getResult> method waits for a result (or timeout)
and returns it

=over 4

=item Parameter

    - string: id
    - [ integer: timeout in seconds (default: no timeout)]


=item Return

    result hash ref or undef on timeout

=back

=cut

sub getResult
{
    my $self    = shift;
    my $id      = shift;
    my $timeout = shift;

    my $logger = Log::Log4perl->get_logger();

    while ( !$self->checkForResult($id)
        and ( !defined($timeout) or ( $timeout > 0 ) ) )
    {
        sleep_log( $self->{'busy_waiting_period'} );
        $timeout -= $self->{'busy_waiting_period'} if defined($timeout);
    }

    my $result = $self->getResult_nb($id);
    if ( !$result ) {
        $logger->error("failed to get result for id $id. Timeout");
    }

    return $result;
}

=head2 getResult_nb()

The C<getResults> method waits for a result (or timeout)
and returns it

=over 4

=item Parameter

    none


=item Return

    result hash ref or undef

=back

=cut

sub getResult_nb
{
    my $self = shift;
    my $id   = shift;

    my $results = $self->{'results'};

    my $return = $results->get($id);
    ## TODO: remove or expire??? Add and test expiring
    #$results->remove($id);

    return $return;
}

=head1 Functions (internal)



=head2 connect_results()

The C<connect_results> method 
creates a shared memory hash for storing the results

=cut

sub connect_results
{
    my $self = shift;

    # create a filename that includes
    # the num pages and page size information
    $self->{'share_file'} =
        $self->{'mmap_result_file'} . "_"
      . $self->{'mmap_num_pages'} . "_"
      . $self->{'mmap_page_size'} . ".mmap";

    # create parameter for Cache::FastMmap
    my %params = (
        'share_file'  => $self->{'share_file'},
        'init_file'   => $self->{'init_mmap_result_file'},
        'page_size'   => $self->{'mmap_page_size'},
        'num_pages'   => $self->{'mmap_num_pages'},
        'expire_time' => $self->{'mmap_expire_time'},
    );

    my $results = Cache::FastMmap->new(%params);

    # TODO: find better solution. Function?
    chmod 0666, $self->{'share_file'};

    $self->{'results'} = $results;

    return 1;
}

1;
