#!/usr/bin/perl -w
package SmartClient::Workstation;
use strict;
use warnings;

use base 'Exporter';

use SmartClient::Config;
use SmartClient::LdapUtil;
use File::Basename;
require Net::LDAP;
require Net::LDAP::LDIF;
require Sys::Hostname;
use Data::Dumper;
use Log::Log4perl;

our $Ldap_Conn;
our @EXPORT =
  qw(get_pw_from_server get_dn_from_ldap write_dn_to_config create_config_file write_pw_to_config get_dn get_pw get_ldap_host get_base_of_dn check_ldap_connection check_ldap_connection_sync_wait check_ldap_err create_ldap_msg ldap_to_syslog);

sub get_pw_from_server
{

    #
    # If required fetch a password from a password server
    #

    # opens the SmartClient-Config file
    my $config         = SmartClient::Config->new();
    my $socket_program = $config->get('ALL_SOCKET_PROGRAM');
    my $server         = $config->get('ALL_PW_SERVER');
    my $port           = $config->get('ALL_PW_PORT');
    my @pw;
    my $password;
    my $ldap_base_host = get_dn();
    my $ldap_host      = get_ldap_host();
    my $status;

    #
    # Try to connect to the LDAP server with the actual password.
    # If it "works" use it and don't fetch a new one.
    #
    $pw[1] = get_pw();

    if ( $pw[1] ) {
        $Ldap_Conn = Net::LDAP->new( $ldap_host, version => 3 ) or die "$@";
        $status = $Ldap_Conn->bind( "$ldap_base_host", password => "$pw[1]" );
        $Ldap_Conn->unbind;
    }

    # if password is undefined or ldap bind has failed,
    # get a new password
    if ( !$pw[1] or $status->code ) {

        #
        # The old password didn't work, fetch a new password
        #
        @pw = `$socket_program $server $port`;
    }

    if ( !$pw[1] ) {
        die "can't get password from server $server:$port";
    } else {
        chomp $pw[1];
    }

    return $pw[1];
}

sub write_dn_to_config
{

    #
    # Writes the LDAP dn to file on the Workstation
    #

    # opens the SmartClient-Config file
    my $config      = SmartClient::Config->new();
    my $config_file = $config->get('ALL_WORKSTATION_CONFIG');
    my $dn;

    $dn = get_dn_from_ldap();

    open( my $cfg_fh, '>>', $config_file )
      or die "can't open $config_file for append: $!";

    print $cfg_fh "## The LDAP dn of this Workstation\n";
    print $cfg_fh "ALL_WORKSTATION_DN=\"$dn\"\n";

    close $cfg_fh;
}

sub create_config_file
{

    #
    # create an empty config file
    #

    # opens the SmartClient-Config file
    my $config      = SmartClient::Config->new();
    my $config_file = $config->get('ALL_WORKSTATION_CONFIG');
    my $config_dir  = $config_file;

    $config_dir =~ s/(\w)*$//;

    system( "mkdir", "-p", "$config_dir" );

    open( my $cfg_fh, '>', $config_file )
      or die "can't open $config_file for writing: $!";

    print $cfg_fh "## \n";
    print $cfg_fh "## Don't edit this file, its created by $0\n";
    print $cfg_fh "## \n\n";

    close $cfg_fh;
}

sub write_pw_to_config
{

    #
    # Writes the LDAP pw into the config file of the workstation
    # for workstation connects with write permissions
    #

    # opens the SmartClient-Config file
    my $config      = SmartClient::Config->new();
    my $config_file = $config->get('ALL_WORKSTATION_CONFIG');
    my $pw;

    $pw = get_pw_from_server;

    open( my $cfg_fh, '>>', $config_file )
      or die "can't open $config_file for append: $!";

    print $cfg_fh "## Secret\n";
    print $cfg_fh "ALL_WORKSTATION_PW=\"$pw\"\n";

    close $cfg_fh;
}

sub get_dn_from_ldap
{

    #
    # Returns the LDAP dn strings
    # Method: Ask the LDAP for the entries of the own MAC address,
    # and grep the dn from the result.
    #

    # Hardware address of the workstation
    my $mac_address;

    # opens the SmartClient-Config file
    my $config = SmartClient::Config->new();

    my $ldap_server = $config->get('ALL_LDAPHOST');
    my $ldap_base   = $config->get('ALL_LDAPBASE');
    my $result;
    my $dn;

    my $network_interfaces = $config->get('ALL_NETWORK_INTERFACES');

    my @network_interfaces = split( /\s/, $network_interfaces );

    #
    # Grep the hardware address fom the ifconfig output.
    # The output of /sbin/ifconfig differs
    # bwith different LANG settings, so at first unset LANG.
    #

    foreach my $network_interface (@network_interfaces) {
        ($mac_address) = `(unset LANG; /sbin/ifconfig $network_interface)`;
        last if $mac_address;
    }

    # extract MAC address from ifconfig output
    $mac_address =~ s/.*HWaddr\s*([[:alnum:]:]+)\s*$/$1/i;

    #print "mac_address{$mac_address}\n";
    $mac_address =~ tr/A-F/a-f/;

    #print "mac_address{$mac_address}\n";

    #
    # connect and bind to LDAP-Server
    #

    $Ldap_Conn = Net::LDAP->new( $ldap_server, version => 3 ) or die "$@";
    $Ldap_Conn->bind or die "Error: LDAP-bind to $ldap_server failed";

    $result = $Ldap_Conn->search(
        base   => $ldap_base,
        filter => "(macAddress=$mac_address)"
    );
    $result->code && die $result->error;
    foreach my $entry ( $result->all_entries ) {
        $dn = $entry->dn();
    }

    return $dn;

}

sub get_dn
{

    #
    # opens the SmartClient-Config file and "return the pointer to the file"
    #
    my $dbk_config_file = SmartClient::Config->new();

    #
    # fetch the the name of the workstation config file
    #
    my $local_config_file_name = $dbk_config_file->{'ALL_WORKSTATION_CONFIG'};

    #
    # opens the Workstation-Config file and "return the pointer to the file"
    #
    my $local_config_file = SmartClient::Config->new($local_config_file_name);
    my $dn                = $local_config_file->{'ALL_WORKSTATION_DN'};

    # print "dn{$dn}\n";
    return $dn;
}

#
# returns the LDAP password strings
#
sub get_pw
{

    #
    # opens the SmartClient-Config file
    # and "return the pointer to the file"
    #
    my $dbk_config_file = SmartClient::Config->new();

    #
    # fetch the the name of the workstation config file
    #
    my $local_config_file_name = $dbk_config_file->{'ALL_WORKSTATION_CONFIG'};

    #
    # opens the Workstation-Config file
    # and "return the pointer to the file"
    #
    my $local_config_file = SmartClient::Config->new($local_config_file_name);
    my $pw                = $local_config_file->{'ALL_WORKSTATION_PW'};

    return $pw;
}

sub get_ldap_host
{

    # read and return the name of the LDAP server
    # from the global SmartClient config fie
    my $config = SmartClient::Config->new();
    return $config->get('ALL_LDAPHOST');
}

sub get_base_of_dn
{

    #
    # return the dn without the first cn=*,
    #

    my $base_dn = get_dn();

    $base_dn =~ s/cn=(\w)*,//;

    return $base_dn;

}

sub check_ldap_connection
{
    my $dn       = $_[0];
    my $password = $_[1];

    my $result = 0;

    if ( !$dn ) {
        $dn = get_dn();
    }

    if ( !$password ) {
        $password = get_pw();
    }

    if ( $dn and $password ) {
        my $ldap_host = get_ldap_host();

        my $ldap_conn = Net::LDAP->new( $ldap_host, version => 3 );
        my $status = $ldap_conn->bind( $dn, password => $password );

        # if ldap bind has been successful,
        # set result to true
        if ( $status->code == 0 ) {
            $result = 1;
        }
        $ldap_conn->unbind;

    }

    return $result;
}

#
# waits until a ldap connection can be established
# This is used for Master/Slave LDAP environments
# where a changes need time to be propagated to slave servers
#
# Parameter:
# 0: timeout (in seconds)
# 1: dn
# 2: password
#
sub check_ldap_connection_sync_wait
{

    # get max waiting time (LDAP sync interval)
    my $config = SmartClient::Config->new();
    my $timeout = $_[0] || $config->getWithDefault('ALL_LDAP_SYNC_INTERVAL', 0);

    my $dn       = $_[1] || get_dn();
    my $password = $_[2] || get_pw();

    my $success = 0;

    #
    # if $dn and $password are available
    # check LDAP connection every few seconds
    # until a connection could be established
    # or the timeout occured
    #
    if ( $dn and $password ) {

        # make STDOUT unbuffered
        select(STDOUT);
        $| = 1;
        $success = check_ldap_connection( $dn, $password );
        while ( ( $timeout > 0 ) and !$success ) {
            print "$timeout ";
            sleep 1;
            $timeout--;
            if ( ( $timeout % 10 ) == 0 ) {
                $success = check_ldap_connection( $dn, $password );
            }
        }
    }

    return $success;
}

sub check_ldap_err
{

    #
    # Check the status of last LDAP call. If not ok return true
    #

    # parameter of the subroutine call
    my ($status) = @_;
    my $return_code = 1;
    if ( defined $status ) {
        if ( $status->code == 0 ) {
            $return_code = 0;
        }
    }

    return $return_code;
}

sub create_ldap_msg
{

    #
    # Prepare a complex message of the state from the last ldap
    # call (for the syslog).
    #

    # parameter of the subroutine call
    my ( $status, $message, $line ) = @_;

    # The prgramm name is used as the syslog IDENT
    my $ident = basename($0);

    if ( defined $status ) {
        $message = "at line " . --$line . ": " . $message;
        $message =
            $message
          . " LDAP id: "
          . $status->code
          . " message: "
          . $status->error;
    } else {
        $message =
            "at line "
          . ++$line . ": "
          . $message
          . ", the LDAP call returns the value UNDEF";
    }

    return $message;
}

sub ldap_to_syslog
{

    #
    # Make the handling of LDAP return codes as syslog entries more easy
    #

    # parameter of the subroutine call
    my ( $result, $message, $line ) = @_;

    my $logger = Log::Log4perl->get_logger();

    if ( check_ldap_err($result) ) {
        $logger->info( create_ldap_msg( $result, $message, $line ) );
        exit 1;
    }

    return;

}
